//
//  ServiceTableViewController.m
//  LEGODeviceDemo
//
//  Created by Søren Toft Odgaard on 16/04/14.
//  Copyright (c) 2014 LEGO. All rights reserved.
//

#import "DeviceDetailTableViewController.h"
#import "LEDevice.h"
#import "LEIO.h"
#import "LEDeviceManager.h"
#import "LogCollector.h"
#import "DeviceServiceCell.h"
#import "ServiceDetailTableViewController.h"
#import "ErrorLogViewController.h"
#import "EditDeviceNameViewController.h"
#import "NSArray+LEAdditional.h"
#import "Layout.h"
#import "WindmillTestViewController.h"

//Sections
static const int kDeviceInfoSectionIndex = 0;
static const int kServiceListSectionIndex = 1;
static const int kLogSectionIndex = 2;
static const int kWindmillTestSectionIndex = 3;
static const int kNumberOfSectionsInDeviceDetailTableView = 4;

//Rows in device info section
static const int kNumberOfRowsInDeviceInfoSection = 2;
static const int kDeviceNameRowIndex = 0;
static const int kBatteryLevelRowIndex = 1;

//Rows in log section
static const int kNumberOfRowsInLogSection = 2;
static const int kErrorLogRowIndex = 0;
static const int kDebugLogRowIndex = 1;


@interface DeviceDetailTableViewController () <LEDeviceDelegate, LEDeviceManagerDelegate, LEServiceDelegate>

@end

@implementation DeviceDetailTableViewController {
    NSArray *_serviceCells;
    UITableViewCell *_windmillButtonCell;
}

#pragma mark - LEDeviceManagerDelegate

- (void)deviceManager:(LEDeviceManager *)manager didDisconnectFromDevice:(LEDevice *)device willAttemptAutoReconnect:(BOOL)autoReconnect error:(NSError *)error
{
    if ([self.device isEqualToDevice:device]) {
        [self.navigationController popViewControllerAnimated:YES];
    }
    [[LogCollector sharedInstance] clearLog];
}

#pragma mark - LEDeviceDelegate
- (void)device:(LEDevice *)device didUpdateDeviceInfo:(LEDeviceInfo *)deviceInfo error:(NSError *)error
{
    [self.tableView reloadRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:kDeviceNameRowIndex inSection:kDeviceInfoSectionIndex] ] withRowAnimation:UITableViewRowAnimationNone];
    if (error) {
        UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:nil message:error.localizedDescription delegate:nil cancelButtonTitle:@"OK" otherButtonTitles:nil];
        [alertView show];
    }
}

- (void)device:(LEDevice *)device didAddService:(LEService *)service
{
    NSUInteger row = [self addDeviceServiceCellForService:service];
    [self.tableView insertRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:row inSection:kServiceListSectionIndex] ] withRowAnimation:UITableViewRowAnimationAutomatic];
}

- (void)device:(LEDevice *)device didRemoveService:(LEService *)service
{
    NSInteger row = [self removeDeviceServiceCellForService:service];
    [self.tableView deleteRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:row inSection:kServiceListSectionIndex] ]  withRowAnimation:UITableViewRowAnimationAutomatic];
}

- (void)device:(LEDevice *)device didChangeNameFrom:(NSString *)oldName to:(NSString *)newName
{
    [self.tableView reloadRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:kDeviceNameRowIndex inSection:kDeviceInfoSectionIndex] ] withRowAnimation:UITableViewRowAnimationNone];
}

- (void)device:(LEDevice *)device didUpdateBatteryLevel:(NSNumber *)newLevel
{
    [self.tableView reloadRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:kBatteryLevelRowIndex inSection:kDeviceInfoSectionIndex] ] withRowAnimation:UITableViewRowAnimationNone];
}


#pragma mark - View Lifecycle

- (void)viewDidLoad
{
    [super viewDidLoad];
    [self.device addDelegate:self];
    [[LEDeviceManager sharedDeviceManager] addDelegate:self];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(logUpdated) name:LogCollectorLoggedErrorNotification object:nil];

    _serviceCells = @[];
    for (LEService *service in self.device.services) {
        [self addDeviceServiceCellForService:service];
    }
    
    _windmillButtonCell = [Layout createButtonCellWithTitle:@"Windmill Test" target:self action:@selector(didTouchWindmillButton)];
}

- (NSUInteger)addDeviceServiceCellForService:(LEService *)service
{
    DeviceServiceCell *deviceServiceCell = [DeviceServiceCell deviceServiceCellWithService:service];
    _serviceCells = [_serviceCells arrayByAddingObject:deviceServiceCell];
    return _serviceCells.count - 1; //insert index
}

- (NSInteger)removeDeviceServiceCellForService:(LEService *)service
{
    for (NSUInteger i = 0; i < _serviceCells.count; i++) {
        DeviceServiceCell *cell = _serviceCells[i];
        if ([cell.service isEqualToService:service]) {
            _serviceCells = [_serviceCells arrayByRemovingObject:cell];
            return i;
        }
    }
    return -1;
}


- (void)logUpdated
{
    [self.tableView reloadRowsAtIndexPaths:
            @[ [NSIndexPath indexPathForRow:0 inSection:kLogSectionIndex], [NSIndexPath indexPathForRow:1 inSection:kLogSectionIndex]]
            withRowAnimation:UITableViewRowAnimationNone];
}

- (void)viewDidAppear:(BOOL)animated
{
    [self.tableView reloadData];
    self.title = self.device.name;
    [super viewDidAppear:animated];
}

- (IBAction)disconnectTouched:(id)sender
{
    [[LEDeviceManager sharedDeviceManager] cancelDeviceConnection:self.device];
}

- (void)dealloc
{
    [self.device removeDelegate:self];
    [[LEDeviceManager sharedDeviceManager] removeDelegate:self];
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    // Return the number of sections.
    return kNumberOfSectionsInDeviceDetailTableView;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    if (section == kDeviceInfoSectionIndex) {
        return kNumberOfRowsInDeviceInfoSection;
    } else if (section == kServiceListSectionIndex) {
        return _serviceCells.count;
    } else if (section == kLogSectionIndex) {
        return kNumberOfRowsInLogSection;
    } else if (section == kWindmillTestSectionIndex) {
        return 1;
    }
    return 0;
}


- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    UITableViewCell *cell;
    if (indexPath.section == kDeviceInfoSectionIndex) {
        if (indexPath.row == kDeviceNameRowIndex) {
            cell = [tableView dequeueReusableCellWithIdentifier:@"DeviceNameCellID" forIndexPath:indexPath];
            cell.textLabel.text = self.device.name;
            cell.detailTextLabel.text = [NSString stringWithFormat:@"FW: %@, SW: %@", self.device.deviceInfo.firmwareRevision, self.device.deviceInfo.softwareRevision];
        } else if (indexPath.row == kBatteryLevelRowIndex) {
            cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:nil];
            cell.textLabel.text = @"Battery Level";
            cell.detailTextLabel.text = [NSString stringWithFormat:@"%ld%%", (long) self.device.batteryLevel.intValue];
        }
    } else if (indexPath.section == kServiceListSectionIndex) {
        cell = _serviceCells[indexPath.row];
    } else if (indexPath.section == kLogSectionIndex) {
        cell = [tableView dequeueReusableCellWithIdentifier:@"LogViewCellID" forIndexPath:indexPath];
        if (indexPath.row == kErrorLogRowIndex) {
            cell.textLabel.text = @"Error / Warn Log";
            cell.detailTextLabel.text = [NSString stringWithFormat:@"%lu / %lu",
                            (unsigned long) [LogCollector sharedInstance].errorCount, (unsigned long) [LogCollector sharedInstance].warningCount];
        } else if (indexPath.row == kDebugLogRowIndex) {
            cell.textLabel.text = @"Info / Debug Log";
            cell.detailTextLabel.text = @"";
        }
    } else if (indexPath.section == kWindmillTestSectionIndex) {
        cell = _windmillButtonCell;
    }

    return cell;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    if (indexPath.section == kServiceListSectionIndex) {
        ServiceDetailTableViewController *vc = [self.storyboard instantiateViewControllerWithIdentifier:@"ServiceDetailTableViewController"];
        DeviceServiceCell *cell = _serviceCells[self.tableView.indexPathForSelectedRow.row];
        vc.service = cell.service;
        vc.device = self.device;
        [self.navigationController pushViewController:vc animated:YES];
    }
}

- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
{
    if ([segue.identifier isEqualToString:@"LogSegue"]) {
        ErrorLogViewController *vc = segue.destinationViewController;
        NSIndexPath *indexPath = self.tableView.indexPathForSelectedRow;
        vc.showDebugLog = (indexPath.row == 1);
    } else if ([segue.identifier isEqualToString:@"EditDeviceNameSegue"]) {
        EditDeviceNameViewController *vc = segue.destinationViewController;
        vc.device = self.device;
    }
}

- (void)didTouchWindmillButton
{
    WindmillTestViewController *vc = [[UIStoryboard storyboardWithName:@"WindmillTestStoryboard" bundle:nil] instantiateViewControllerWithIdentifier:@"WindmillTestViewController"];
    vc.device = self.device;
    [self.navigationController pushViewController:vc animated:YES];
}



@end
