//
//  DeviceDetailsViewController.m
//  LEGODeviceDemo-OSX
//
//  Created by Bartlomiej Hyzy on 12/02/2015.
//  Copyright (c) 2015 lego. All rights reserved.
//

#import "DeviceDetailsViewController.h"
#import "LEDevice+UI.h"
#import "LEInputFormat+Mutable.h"
#import "ServiceControlViewController.h"
#import "WindmillTestViewController.h"


@interface DeviceDetailsViewController () <LEDeviceDelegate, LEServiceDelegate>

@property (strong, nonatomic) LEDevice *device;
@property (strong, nonatomic) LEInputFormat *serviceInputFormat;

@property (weak, nonatomic) IBOutlet NSArrayController *serviceListController;

@property (weak, nonatomic) IBOutlet NSTextField *deviceNameTextField;
@property (weak, nonatomic) IBOutlet NSButton *assignDeviceNameButton;

@property (weak, nonatomic) ServiceControlViewController *serviceControlViewController;

@end


@implementation DeviceDetailsViewController

#pragma mark - Factory Methods

+ (NSWindowController *)instantiateWindowControllerForDevice:(LEDevice *)device delegate:(id <DeviceDetailsViewControllerDelegate>)delegate
{
    NSStoryboard *storyboard = [NSStoryboard storyboardWithName:@"Main" bundle:nil];
    NSWindowController *windowController = [storyboard instantiateControllerWithIdentifier:@"deviceDetailsController"];
    
    DeviceDetailsViewController *deviceViewController = (DeviceDetailsViewController *)windowController.contentViewController;
    deviceViewController.representedObject = device;
    deviceViewController.delegate = delegate;
    
    return windowController;
}

#pragma mark - View Appearance Methods

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(deviceNameTextFieldDidBeginEditing:) name:NSControlTextDidBeginEditingNotification object:self.deviceNameTextField];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(deviceNameTextFieldDidEndEditing:) name:NSControlTextDidEndEditingNotification object:self.deviceNameTextField];
    
    [self.serviceListController addObserver:self forKeyPath:@"selection" options:0 context:NULL];
}

- (void)viewWillAppear
{
    [super viewWillAppear];
    
    [self.device addDelegate:self];
}

- (void)viewDidDisappear
{
    [super viewDidDisappear];
    
    [self.device removeDelegate:self];
}

#pragma mark - Overriden

- (void)setRepresentedObject:(id)representedObject
{
    [super setRepresentedObject:representedObject];
    
    [self.device removeDelegate:self];
    self.device = (LEDevice *)representedObject;
    [self.device addDelegate:self];
    
    [self updateWindowTitle];
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    if (object == self.serviceListController && [keyPath isEqualToString:@"selection"]) {
        LEService *selectedService = (LEService *)self.serviceListController.selectedObjects.firstObject;

        // Update mutable input format to reflect selected service's format
        if (selectedService != nil) {
            self.serviceInputFormat = [selectedService.inputFormat copy];
        } else {
            self.serviceInputFormat = nil;
        }
        
        // Display the right sensor control view controller for the selected service
        self.serviceControlViewController.controlledService = selectedService;
    }
}

- (void)prepareForSegue:(NSStoryboardSegue *)segue sender:(id)sender
{
    // Grab the service control view controller instantiated from a storyboard - we'll need it to set the controlled service.
    if ([segue.identifier isEqualToString:@"serviceControl"]) {
        self.serviceControlViewController = segue.destinationController;
    }
}

#pragma mark - UI Actions

- (IBAction)assignDeviceNamePressed:(id)sender
{
    [self.view.window makeFirstResponder:nil];
}

- (IBAction)updateInputFormatPressed:(id)sender
{
    [self.view.window makeFirstResponder:nil];
    
    LEService *selectedService = (LEService *)self.serviceListController.selectedObjects.firstObject;
    [selectedService updateInputFormat:self.serviceInputFormat];
}

- (IBAction)readValuePressed:(id)sender
{
    LEService *selectedService = (LEService *)self.serviceListController.selectedObjects.firstObject;
    [selectedService sendReadValueRequest];
}

- (IBAction)resetSensorPressed:(id)sender
{
    LEService *selectedService = (LEService *)self.serviceListController.selectedObjects.firstObject;
    [selectedService sendResetStateRequest];
}

- (IBAction)openWindmillTestSheet:(id)sender
{
    NSWindowController *windmillWindowController = [WindmillTestViewController instantiateWindowControllerForDevice:self.device];
    [self.view.window beginSheet:windmillWindowController.window completionHandler:nil];
}

- (IBAction)disconnectFromDevice:(id)sender
{
    NSAssert(self.delegate != nil, @"");
    [self.delegate deviceDetailsViewControllerDidRequestDisconnect:self];
}

#pragma mark - UI Update Methods

- (void)updateWindowTitle
{
    self.view.window.title = self.device.name ?: @"Unknown Device";
}

#pragma mark - LEDeviceDelegate

- (void)device:(LEDevice *)device didChangeNameFrom:(NSString *)oldName to:(NSString *)newName
{
    [self updateWindowTitle];
}

- (void)device:(LEDevice *)device didAddService:(LEService *)service
{
    [service addDelegate:self];
}

- (void)device:(LEDevice *)device didRemoveService:(LEService *)service
{
    [service removeDelegate:self];
}

#pragma mark - LEServiceDelegate

- (void)service:(LEService *)service didUpdateInputFormatFrom:(LEInputFormat *)oldFormat to:(LEInputFormat *)newFormat
{
    // Update input format if it's currently selected
    LEService *selectedService = (LEService *)self.serviceListController.selectedObjects.firstObject;
    if (service == selectedService) {
        self.serviceInputFormat = [newFormat copy];
    }
}

#pragma mark - Device Name Text Field Notifications

- (void)deviceNameTextFieldDidBeginEditing:(NSNotification *)notification
{
    self.assignDeviceNameButton.hidden = NO;
}

- (void)deviceNameTextFieldDidEndEditing:(NSNotification *)notification
{
    NSString *newName = self.deviceNameTextField.stringValue;
    if (![newName isEqualToString:self.device.name]) {
        self.device.name = newName;
    }
    
    self.assignDeviceNameButton.hidden = YES;
}

@end
