//
//  ViewController.m
//  LEGODeviceDemo-OSX
//
//  Created by Bartlomiej Hyzy on 10/02/2015.
//  Copyright (c) 2015 lego. All rights reserved.
//

#import "DeviceListViewController.h"
#import <LEGODevice/LEGODevice.h>
#import <LEGODevice/LEBluetoothDevice.h>
#import "LEDevice+UI.h"
#import "AppSettings.h"
#import "MessageLogViewController.h"
#import "DeviceDetailsViewController.h"


@interface DeviceListViewController () <DeviceDetailsViewControllerDelegate, LEDeviceManagerDelegate, LEDeviceDelegate, NSTableViewDelegate>

@property (weak, nonatomic) IBOutlet NSTableView *tableView;

@property (strong, nonatomic) LEDeviceManager *connectionManager;
@property (strong, nonatomic) NSMutableArray *devices;

@property (strong, nonatomic) NSMutableDictionary *deviceDetailsWindowControllers;

@property (assign, nonatomic) BOOL autoReconnect;
@property (assign, nonatomic) NSInteger rssiThreshold;

@property (strong, nonatomic) NSWindowController *logViewController;

@property (weak, nonatomic) IBOutlet NSArrayController *deviceListController;
@property (readonly) NSPredicate *deviceListFilterPredicate;

@end


@implementation DeviceListViewController

#pragma mark - View Appearance Methods

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.devices = [NSMutableArray new];
    self.deviceDetailsWindowControllers = [NSMutableDictionary new];
    
    self.connectionManager = [LEDeviceManager sharedDeviceManager];
    self.connectionManager.automaticReconnectOnConnectionLostEnabled = self.autoReconnect;
    [self.connectionManager addDelegate:self];
    
    self.logViewController = [MessageLogViewController instantiateWindowController];
    [self.logViewController showWindow:self];
}

- (void)viewWillAppear
{
    [super viewWillAppear];
    [self.connectionManager scan];
}

- (void)viewWillDisappear
{
    [super viewDidDisappear];
    [self.connectionManager stopScanning];
}

#pragma mark - Table View Handling

- (void)refreshUIForDevice:(LEDevice *)device
{
    // reload table view row for the device
    NSInteger index = [self.deviceListController.arrangedObjects indexOfObjectIdenticalTo:device];
    if (index != NSNotFound) {
        [self.tableView reloadDataForRowIndexes:[NSIndexSet indexSetWithIndex:index]
                                  columnIndexes:[NSIndexSet indexSetWithIndexesInRange:NSMakeRange(0, self.tableView.numberOfColumns)]];
    }
}

#pragma mark - UI Actions

- (IBAction)refresh:(id)sender
{
    NSMutableArray *devices = [self mutableArrayValueForKey:@"devices"];
    [devices removeAllObjects];
    [devices addObjectsFromArray:self.connectionManager.allDevices];
}

#pragma mark - Helper Methods

- (void)toggleConnectionStateForDevice:(LEDevice *)device
{
    if (device.connectState == LEDeviceStateDisconnectedAdvertising) {
        [self.connectionManager connectToDevice:device];
    } else if (device.connectState == LEDeviceStateInterrogationFinished) {
        [self.connectionManager cancelDeviceConnection:device];
    }
}

- (void)showDetailsWindowForDevice:(LEDevice *)device
{
    NSWindowController *windowController = self.deviceDetailsWindowControllers[device.deviceId];
    if (windowController == nil) {
        windowController = [DeviceDetailsViewController instantiateWindowControllerForDevice:device delegate:self];
        self.deviceDetailsWindowControllers[device.deviceId] = windowController;
    }
    [windowController showWindow:self];
}

- (void)dismissDetailsWindowForDevice:(LEDevice *)device
{
    NSWindowController *windowController = self.deviceDetailsWindowControllers[device.deviceId];
    [self.deviceDetailsWindowControllers removeObjectForKey:device.deviceId];
    [windowController close];
}

- (void)showAlertWithTitle:(NSString *)title message:(NSString *)message
{
    NSAlert *alert = [[NSAlert alloc] init];
    [alert setMessageText:title];
    [alert setInformativeText:message];
    [alert addButtonWithTitle:@"OK"];
    [alert runModal];
}

#pragma mark - Settings

- (BOOL)autoReconnect
{
    return [AppSettings defaultSettings].autoReconnectEnabled;
}

- (void)setAutoReconnect:(BOOL)autoReconnect
{
    [AppSettings defaultSettings].autoReconnectEnabled = autoReconnect;
    self.connectionManager.automaticReconnectOnConnectionLostEnabled = autoReconnect;
}

- (NSInteger)rssiThreshold
{
    return [AppSettings defaultSettings].rssiThreshold;
}

- (void)setRssiThreshold:(NSInteger)rssiThreshold
{
    [AppSettings defaultSettings].rssiThreshold = rssiThreshold;
}

#pragma mark - Filtering

- (NSPredicate *)deviceListFilterPredicate
{
    return [NSPredicate predicateWithFormat:@"connectState != %@ OR deviceRSSI > %@", @(LEDeviceStateDisconnectedAdvertising), @(self.rssiThreshold)];;
}

+ (NSSet *)keyPathsForValuesAffectingDeviceListFilterPredicate
{
    return [NSSet setWithObject:@"rssiThreshold"];
}

#pragma mark - DeviceDetailsViewControllerDelegate

- (void)deviceDetailsViewControllerDidRequestDisconnect:(DeviceDetailsViewController *)controller
{
    LEDevice *device = (LEDevice *)controller.representedObject;
    [self.connectionManager cancelDeviceConnection:device];
    [self dismissDetailsWindowForDevice:device];
}

#pragma mark - NSTableViewDelegate

- (void)tableView:(NSTableView *)tableView willDisplayCell:(id)cell forTableColumn:(NSTableColumn *)tableColumn row:(NSInteger)row
{
    LEDevice *device = self.devices[row];
    if (device.buttonPressed) {
        // mark devices with their buttons pressed down using a yellow background
        [cell setBackgroundColor:[NSColor yellowColor]];
        [cell setDrawsBackground:YES];
    } else if (device.lowVoltage && [tableColumn.identifier isEqualToString:@"BatteryLevel"]) {
        // mark the battery level cell red if a low voltage alert has been received
        [cell setBackgroundColor:[NSColor redColor]];
        [cell setDrawsBackground:YES];
    } else {
        [cell setDrawsBackground:NO];
    }
}

#pragma mark - LEDeviceManagerDelegate

- (void)deviceManager:(LEDeviceManager *)manager deviceDidAppear:(LEDevice *)device
{
    [device addDelegate:self];
    
    NSMutableArray *devices = [self mutableArrayValueForKey:@"devices"];
    [devices removeObject:device]; // try to remove the device from the list in case it's already in there
    [devices addObject:device];
}

- (void)deviceManager:(LEDeviceManager *)manager deviceDidDisappear:(LEDevice *)device
{
    [device removeDelegate:self];
}

- (void)deviceManager:(LEDeviceManager *)manager didFinishInterrogatingDevice:(LEDevice *)device
{
    [self showDetailsWindowForDevice:device];
}

- (void)deviceManager:(LEDeviceManager *)manager didDisconnectFromDevice:(LEDevice *)device willAttemptAutoReconnect:(BOOL)autoReconnect error:(NSError *)error
{
    [self dismissDetailsWindowForDevice:device];
    
    if (autoReconnect) {
        NSLog(@"Attempting to reconnect to device %@", device.name);
    } else if (error != nil) {
        [self showAlertWithTitle:@"Connection lost"
                         message:[NSString stringWithFormat:@"Connection lost to device %@\n[%@]", device.name, error.localizedDescription]];
    }
}

- (void)deviceManager:(LEDeviceManager *)manager didFailToConnectToDevice:(LEDevice *)device willAttemptAutoReconnect:(BOOL)autoReconnect error:(NSError *)error
{
    if (autoReconnect) {
        NSLog(@"Attempting to reconnect to device %@", device.name);
    } else {
        [self showAlertWithTitle:@"Failed to connect"
                         message:[NSString stringWithFormat:@"Could not connect to device %@\n[%@]", device.name, error.localizedDescription]];
    }
}

#pragma mark - LEDeviceDelegate

- (void)device:(LEDevice *)device didAddService:(LEService *)service
{
    if (service.defaultInputFormat != nil) {
        [service updateInputFormat:service.defaultInputFormat];
    }
}

- (void)device:(LEDevice *)device didFailToAddServiceWithError:(NSError *)error
{
    [self showAlertWithTitle:@"Service Definition Failture" message:error.localizedDescription];
}

- (void)device:(LEDevice *)device didChangeButtonState:(BOOL)pressed
{
    [self refreshUIForDevice:device];
}

- (void)device:(LEDevice *)device didUpdateLowVoltageState:(BOOL)lowVoltage
{
    [self refreshUIForDevice:device];
}

@end
