//
//  DeviceTableViewController.m
//  LDPeripheralDemo
//
//  Created by Søren Toft Odgaard on 8/15/13.
//  Copyright (c) 2013 LEGO. All rights reserved.
//

#import "DeviceTableViewController.h"
#import "DeviceDetailTableViewController.h"
#import "DeviceTableViewCell.h"
#import "SettingsViewController.h"
#import "NSArray+LEAdditional.h"
#import "LEBluetoothDevice.h"

@interface DeviceTableViewController () <LEDeviceManagerDelegate, LEDeviceDelegate, UIAlertViewDelegate>
@end

@implementation DeviceTableViewController {


    LEDeviceManager *_connectionManager;
    NSMutableDictionary *_peripheralIdentifierToDetailViewController;

    NSMutableArray *_devices;
    UIAlertView *_connectionLostAlertView;

}

- (void)viewDidLoad
{
    [super viewDidLoad];
    [self enablePullToRefresh];

    _peripheralIdentifierToDetailViewController = [NSMutableDictionary new];
    _connectionManager = [LEDeviceManager sharedDeviceManager];
    [_connectionManager addDelegate:self];

}

- (void)enablePullToRefresh
{
    // Initialize the refresh control.
    self.refreshControl = [[UIRefreshControl alloc] init];
    [self.refreshControl addTarget:self
                    action:@selector(updateTableData)
                  forControlEvents:UIControlEventValueChanged];
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];
    _connectionManager.automaticReconnectOnConnectionLostEnabled = [SettingsViewController autoReconnectEnabled];

    [_connectionManager scan];
    [self updateTableData];

    [_devices enumerateObjectsUsingBlock:^(LEDevice *device, NSUInteger idx, BOOL *stop) {
        [device addDelegate:self];
    }];

}

- (void)updateTableData
{

    [_connectionManager addDelegate:self];
    
    _devices = [NSMutableArray arrayWithArray:_connectionManager.allDevices];

    NSMutableArray *filteredList = [NSMutableArray new];
    for (LEDevice *device in _connectionManager.allDevices) {
        if ([device isKindOfClass:[LEBluetoothDevice class]]) {
            LEBluetoothDevice *btDevice = (LEBluetoothDevice *) device;
            if (btDevice.connectState != LEDeviceStateDisconnectedAdvertising) {
                [filteredList addObject:btDevice];
            } else if (btDevice.connectState == LEDeviceStateDisconnectedAdvertising && btDevice.RSSI.integerValue > [SettingsViewController rssiThreshold]) {
                [filteredList addObject:btDevice];
            }
        }
    }
    _devices = filteredList;

    [self.tableView reloadData];
    [self.refreshControl endRefreshing];
}

- (void)viewDidDisappear:(BOOL)animated
{
    [_connectionManager stopScanning];
    [super viewDidAppear:animated];
}

#pragma mark - LEDeviceManagerDelegate

- (void)deviceManager:(LEDeviceManager *)manager deviceDidAppear:(LEDevice *)device
{
    if ([device isKindOfClass:[LEBluetoothDevice class]]) {
        LEBluetoothDevice *btDevice = (LEBluetoothDevice *) device;
        if (btDevice.RSSI.intValue < [SettingsViewController rssiThreshold]) {
            NSLog(@"Ignoring new device below RSSI threshold");
            return;
        }
    }

    [device addDelegate:self];

    NSUInteger index = [_devices indexOfObjectPassingTest:^BOOL(LEDevice *obj, NSUInteger idx, BOOL *stop) {
        return [obj.deviceId isEqualToString:device.deviceId];
    }];

    //If we have an 'old' version of same device that is no long advertising in the list, replace it with this one
    if (index != NSNotFound) {
        _devices[index] = device;
    } else {
        [_devices addObject:device];
    }

    [self.tableView reloadData];

}

- (void)deviceManager:(LEDeviceManager *)manager deviceDidDisappear:(LEDevice *)device
{
    [device removeDelegate:self];
    [self.tableView reloadData];
}

- (void)deviceManager:(LEDeviceManager *)manager didStartInterrogatingDevice:(LEDevice *)device
{
    [self.tableView reloadData];
    
}

- (void)deviceManager:(LEDeviceManager *)manager didFinishInterrogatingDevice:(LEDevice *)device {
    NSLog(@"Showing hub detail for device: %@", device.name);
    [self showHubDetailsTableViewController:device];
}

- (void)deviceManager:(LEDeviceManager *)manager didDisconnectFromDevice:(LEDevice *)device willAttemptAutoReconnect:(BOOL)autoReconnect error:(NSError *)error
{
    [_peripheralIdentifierToDetailViewController removeObjectForKey:device.deviceId];
    
    [self.tableView reloadData];

    if (autoReconnect) {
        NSLog(@"Attempting to reconnect to device %@", device.name);
    }

    if (error && !autoReconnect) {
        if (error.code == 6) {
            //Happens for instance if Hub looses power (battery removed) or if device is out of range
            _connectionLostAlertView = [[UIAlertView alloc] initWithTitle:@"Connection lost"
                    message:[NSString stringWithFormat:@"Connection lost to Device %@\n", device.name]
                    delegate:self cancelButtonTitle:@"OK" otherButtonTitles:nil];
            [_connectionLostAlertView show];
        } else {
            //Other (unknown) errors
            _connectionLostAlertView = [[UIAlertView alloc] initWithTitle:@"Connection lost"
                    message:[NSString stringWithFormat:@"Connection lost to Device %@\n[%@]", device.name, error.localizedDescription]
                    delegate:self cancelButtonTitle:@"OK" otherButtonTitles:nil];
            [_connectionLostAlertView show];
        }
    }
}

- (void)alertView:(UIAlertView *)alertView clickedButtonAtIndex:(NSInteger)buttonIndex
{
    [self.tableView reloadData];
    [self.navigationController popToViewController:self animated:YES];
}

- (void)deviceManager:(LEDeviceManager *)manager didFailToConnectToDevice:(LEDevice *)device willAttemptAutoReconnect:(BOOL)autoReconnect error:(NSError *)error
{
    [self.tableView reloadData];

    if (autoReconnect) {
        NSLog(@"Attempting to reconnect to device %@", device.name);
    }

    if (!autoReconnect) {
        [self showAlertViewWithTitle:@"Failed to connect"
                message:[NSString stringWithFormat:@"Could not connect to Device %@\n[%@]", device.name, error.localizedDescription]];
    }
}

#pragma mark - LDDeviceDelegate

- (void)device:(LEDevice *)device didAddService:(LEService *)service
{
    if (service.defaultInputFormat) {
        [service updateInputFormat:service.defaultInputFormat];
    }
}

- (void)device:(LEDevice *)device didFailToAddServiceWithError:(NSError *)error
{
    [self showAlertViewWithTitle:@"Service Definition Failture" message:error.localizedDescription];
}

- (void)showAlertViewWithTitle:(NSString *)title message:(NSString *)message
{
    UIAlertView *alertView = [[UIAlertView alloc] initWithTitle:title message:message delegate:self cancelButtonTitle:@"OK" otherButtonTitles:nil];
    [alertView show];
}

- (void)device:(LEDevice *)device didChangeNameFrom:(NSString *)oldName to:(NSString *)newName
{
    [self.tableView reloadRowsAtIndexPaths:[self indexPathsForDevice:device] withRowAnimation:UITableViewRowAnimationFade];
}

- (void)device:(LEDevice *)device didChangeButtonState:(BOOL)pushed
{
    [self.tableView reloadRowsAtIndexPaths:[self indexPathsForDevice:device] withRowAnimation:UITableViewRowAnimationNone];
}

- (void)device:(LEDevice *)device didUpdateLowVoltageState:(BOOL)lowVoltage
{
    [self.tableView reloadRowsAtIndexPaths:[self indexPathsForDevice:device] withRowAnimation:UITableViewRowAnimationNone];
}


- (void)showHubDetailsTableViewController:(LEDevice *)device
{
    if (self.navigationController.visibleViewController == self) {
        [self.navigationController pushViewController:[self detailViewControllerForDevice:device] animated:YES];
    }
}


- (DeviceDetailTableViewController *)detailViewControllerForDevice:(LEDevice *)device
{
    DeviceDetailTableViewController *detailTableViewController = _peripheralIdentifierToDetailViewController[device.deviceId];
    if (!detailTableViewController) {
        detailTableViewController = [self.storyboard instantiateViewControllerWithIdentifier:@"DeviceDetailTableViewController"];
        detailTableViewController.device = device;
        _peripheralIdentifierToDetailViewController[device.deviceId] = detailTableViewController;
    }
    return detailTableViewController;
}

#pragma mark - Table view data source

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    return 1;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    return [_devices count];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    static NSString *CellIdentifier = @"DeviceInfoCellID";
    DeviceTableViewCell *cell = [tableView dequeueReusableCellWithIdentifier:CellIdentifier forIndexPath:indexPath];

    LEDevice *device = _devices[indexPath.row];
    [cell configureWithDevice:device];

    return cell;
}

- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    LEDevice *device = _devices[indexPath.row];
    assert(device);
    if (device.connectState == LEDeviceStateDisconnectedAdvertising) {
        NSLog(@"Connecting to device %@", device.name);
        [_connectionManager connectToDevice:device];
        [self updateTableData];
    } else if (device.connectState == LEDeviceStateConnecting || device.connectState == LEDeviceStateInterrogating) {
        //Do nothing....
        NSLog(@"Connection already in progress");
    } else if (device.connectState == LEDeviceStateDisconnectedNotAdvertising) {
        //Do nothing
        NSLog(@"Device no longer advertising - will not attempt to connect");
    } else if (device.connectState == LEDeviceStateInterrogationFinished) {
        NSLog(@"Showing table view controller for device %@", device.name);
        [self showHubDetailsTableViewController:device];
    }
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
    if (section == 0) {
        return @"Devices";
    }
    return @"";
}

- (NSArray *)indexPathsForDevice:(LEDevice *)device {
    NSMutableArray *indexPaths = [NSMutableArray arrayWithCapacity:1];
    for (NSInteger i = 0; i < [_devices count]; i++) {
        LEDevice *aDevice = [_devices objectAtIndex:i];
        if ([aDevice isEqualToDevice:device]) {
            [indexPaths addObject:[NSIndexPath indexPathForRow:i inSection:0]];
        }
    }
    return indexPaths;
}

@end
