//
// Created by Søren Toft Odgaard on 08/05/14.
// Copyright (c) 2014 LEGO. All rights reserved.
//

#import "GenericServiceDetailDataSource.h"
#import "LEInputFormat.h"
#import "Layout.h"

@implementation GenericServiceDetailDataSource {

    NSArray *_cells;
    UITableViewCell *_inputFormatCell;
}

- (instancetype)initWithService:(LEService *)service tableView:(UITableView *)tableView
{
    self = [super init];
    if (self) {
        self.tableView = tableView;
        _service = service;
        [_service addDelegate:self];

        _cells = @[
                [self createServiceVersionInfoCell],
                [self createInputValueCell],
                [self createInputModeCell],
                [self createInputFormatCell],
                [self createReadValueNowButtonCell],
                [self createResetValueButtonCell]
        ];

        [self updateInputValue];
        [self updateInputFormat];
    }
    return self;
}


+ (instancetype)dataSourceWithService:(LEService *)service tableView:(UITableView *)tableView
{
    return [[self alloc] initWithService:service tableView:tableView];
}


- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    return 1;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    return [_cells count];
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    return _cells[indexPath.row];
}

- (NSString *)tableView:(UITableView *)tableView titleForHeaderInSection:(NSInteger)section
{
    if (section == 0) {
        return self.service.serviceName;
    }
    return @"";
}


- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    if (indexPath.row == [_cells indexOfObject:_inputFormatCell]) {
        if (self.showInputFormatViewController) self.showInputFormatViewController();
    }
}

#pragma mark - Updates UI when value or format changes
- (void)updateInputValue
{
    self.inputValueLabel.text = _service.valueData.description;
}

- (void)updateInputFormat {
    _inputModeLabel.text = [NSString stringWithFormat:@"%lu", (unsigned long) _service.inputFormat.mode];
    _inputFormatCell.detailTextLabel.text = [NSString stringWithFormat:@"Unit: %@, Delta: %lu",
                                                                       [self stringFromUnit:_service.inputFormat.unit],
                                                                       (unsigned long) _service.inputFormat.deltaInterval];
}

- (NSString *)stringFromUnit:(uint8_t)unit
{
    if (unit == 0) {
        return @"Raw";
    } else if (unit == 1) {
        return @"Pct";
    } else if (unit == 2) {
        return @"SI";
    }
    return @"?";
}


#pragma mark - LEServiceDelegate
- (void)service:(LEService *)service didUpdateValueDataFrom:(NSData *)oldValue to:(NSData *)newValue
{
    [self updateInputValue];
    [self.tableView reloadRowsAtIndexPaths:@[ [NSIndexPath indexPathForRow:1 inSection:0] ] withRowAnimation:UITableViewRowAnimationNone];
}

- (void)service:(LEService *)service didUpdateInputFormatFrom:(LEInputFormat *)oldFormat to:(LEInputFormat *)newFormat
{
    [self updateInputFormat];
    [self.tableView reloadData];
}

#pragma mark - Create Cells

- (UITableViewCell *)createServiceVersionInfoCell
{
    UITableViewCell *cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:nil];
    cell.textLabel.text = @"Hardware/Firmware";
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
    cell.detailTextLabel.text = [NSString stringWithFormat:@"%@/%@",
                    self.service.connectInfo.hardwareVersion, self.service.connectInfo.firmwareVersion];
    return cell;
}

- (UITableViewCell *)createInputValueCell
{
    UITableViewCell *cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:nil];
    cell.textLabel.text = @"Value";
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
    _inputValueLabel = cell.detailTextLabel;
    return cell;
}

- (UITableViewCell *)createReadValueNowButtonCell
{
    return [Layout createButtonCellWithTitle:@"Send read request" target:_service action:@selector(sendReadValueRequest)];
}

- (UITableViewCell *)createResetValueButtonCell
{
    return [Layout createButtonCellWithTitle:@"Reset sensor state" target:_service action:@selector(sendResetStateRequest)];
}


- (UITableViewCell *)createInputModeCell
{
    UITableViewCell *cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleValue1 reuseIdentifier:nil];
    cell.selectionStyle = UITableViewCellSelectionStyleNone;
    cell.textLabel.text = @"Mode";
    _inputModeLabel = cell.detailTextLabel;
    return cell;
}

- (UITableViewCell *)createInputFormatCell
{
    UITableViewCell *cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleSubtitle reuseIdentifier:nil];
    cell.accessoryType = UITableViewCellAccessoryDisclosureIndicator;
    cell.textLabel.text = @"Format";
    _inputFormatCell = cell;
    return cell;
}


@end