//
//  HubSimulatorTableViewController.m
//  LDPeripheralDemo
//
//  Created by Søren Toft Odgaard on 9/3/13.
//  Copyright (c) 2013 LEGO. All rights reserved.
//

#import "HubSimulatorTableViewController.h"
#import "HubSimulator.h"
#import "SensorSimulator.h"
#import "LEBluetoothIO.h"
#import "LEConnectInfo.h"
#import "LEBluetoothServiceDefinition.h"
#import "LEIOServiceDefinition.h"
#import "LEDeviceServiceDefinition.h"
#import "SimulatorServiceStore.h"
#import "SimulatorService.h"
#import "SimulatorServiceDetailViewController.h"

@interface HubSimulatorTableViewController () <CBPeripheralManagerDelegate>

@end

@implementation HubSimulatorTableViewController {


    CBPeripheralManager *_peripheralManager;
    NSMutableArray *_subscribedCentrals;

    HubSimulator *_hubSimulator;

    SensorSimulator *_sensorSimulator;

    UITableViewCell *_advertisingCell;
    UISwitch *_advertisingSwitch;

    LEIOServiceDefinition *_inputServiceDefinition;
    LEDeviceServiceDefinition *_hubServiceDefinition;

    SimulatorServiceStore *_store;
}

- (void)viewDidLoad
{
    _store = [SimulatorServiceStore sharedInstance];

    _inputServiceDefinition = [LEBluetoothServiceDefinition ioServiceDefinition];
    _hubServiceDefinition = [LEBluetoothServiceDefinition deviceServiceDefinition];

    _hubSimulator = [HubSimulator new];
    _sensorSimulator = [[SensorSimulator alloc] init];
    [super viewDidLoad];
    // Do any additional setup after loading the view.
}

- (void)viewWillAppear:(BOOL)animated
{
    [super viewWillAppear:animated];

    [self.tableView reloadData];

    if (!_peripheralManager) {
        _peripheralManager = [[CBPeripheralManager alloc] initWithDelegate:self queue:nil];
        _subscribedCentrals = [NSMutableArray new];
    }
}

- (IBAction)switchChanged:(id)sender
{
    if (_advertisingSwitch.on) {
        [self startAdvertising];
    }
    else {
        [self stopAdvertising];
    }
}

- (void)sendNotificationSensorAttached:(BOOL)attached type:(LEIOType)type atIndex:(NSUInteger)index
{
    [_hubSimulator attached:attached type:type index:index];
    [_sensorSimulator attached:attached index:index];
    BOOL success = [_peripheralManager updateValue:[_hubSimulator ioValues] forCharacteristic:_hubSimulator.ioCharacteristic onSubscribedCentrals:nil];
    if (!success) {
        NSLog(@"Failed to update Attached IO characteristic");
    } else {
        NSLog(@"Updated Attachedc IO characteristic: %@", [_hubSimulator ioValues]);
    }
}

- (void)startAdvertising
{
    // All we advertise is our service's UUID
    [_advertisingSwitch setOn:YES animated:YES];
    [_peripheralManager startAdvertising:@{
            CBAdvertisementDataServiceUUIDsKey : @[ [LEDeviceServiceDefinition sharedInstance].serviceUUID ],
            CBAdvertisementDataLocalNameKey : _hubSimulator.name }];
    [self debugLog:@"Started advertising"];
}

- (void)stopAdvertising
{
    [_advertisingSwitch setOn:NO animated:YES];
    [_peripheralManager stopAdvertising];
    [self debugLog:@"Stopped advertising"];
}

- (void)debugLog:(NSString *)string
{
    NSLog(string, nil);
}

#pragma mark - CBPeripheralManagerDelegate

- (void)peripheralManagerDidUpdateState:(CBPeripheralManager *)peripheral
{
    if (peripheral.state != CBPeripheralManagerStatePoweredOn) {
        [_peripheralManager removeAllServices];
        return;
    }

    [_peripheralManager addService:_hubSimulator.service];
    [_peripheralManager addService:_sensorSimulator.service];

    [self debugLog:@"Did add HubService to periphial"];
}

- (void)peripheralManager:(CBPeripheralManager *)peripheral
                     central:(CBCentral *)central
didSubscribeToCharacteristic:(CBCharacteristic *)characteristic
{
    NSLog(@"Central did subscribe to characteristic: %@", characteristic);
//    if ([characteristic isEqual:_sensorSimulator.sensorValueCharacteristic] && !_valueNotifyTimer.isValid) {
//        _valueNotifyTimer = [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(sendSensorValueNotification) userInfo:nil repeats:YES];
//    }
}

- (void)sendSensorValueNotification
{
    [_peripheralManager updateValue:[_sensorSimulator sensorValues] forCharacteristic:_sensorSimulator.sensorValueCharacteristic onSubscribedCentrals:nil];
}

- (void)peripheralManager:(CBPeripheralManager *)peripheral central:(CBCentral *)central didUnsubscribeFromCharacteristic:(CBCharacteristic *)characteristic
{
    NSLog(@"Central did un-subscribe from characteristic %@", characteristic);
//    if ([characteristic isEqual:_sensorSimulator.sensorValueCharacteristic] && _valueNotifyTimer.isValid) {
//        [_valueNotifyTimer invalidate];
//    }
}


- (void)peripheralManager:(CBPeripheralManager *)peripheral didReceiveWriteRequests:(NSArray *)requests
{
    NSLog(@"Did receive write request");
    for (CBATTRequest *request in requests) {
        if ([_hubServiceDefinition.deviceName matchesCharacteristic:request.characteristic]) {
            [_peripheralManager respondToRequest:request withResult:CBATTErrorSuccess];
            _hubSimulator.name = [[NSString alloc] initWithData:request.value encoding:NSUTF8StringEncoding];
            [self.tableView reloadData];
        } else {
            [_peripheralManager respondToRequest:request withResult:CBATTErrorAttributeNotFound];
        }
    }
}


- (void)peripheralManager:(CBPeripheralManager *)peripheral didReceiveReadRequest:(CBATTRequest *)request
{
    if ([_hubServiceDefinition.deviceName matchesCharacteristic:request.characteristic]) {
        NSLog(@"Did receive read request for hub name");
        request.value = [_hubSimulator.name dataUsingEncoding:NSUTF8StringEncoding];
        [_peripheralManager respondToRequest:request withResult:CBATTErrorSuccess];
    } else if ([_hubServiceDefinition.attachedIO matchesCharacteristic:request.characteristic]) {
        NSLog(@"Did receive read request for hub IO");
        request.value = _hubSimulator.ioValues;
        [_peripheralManager respondToRequest:request withResult:CBATTErrorSuccess];
    } else {

        if ([_inputServiceDefinition.inputValue matchesCharacteristic:request.characteristic]) {
            NSLog(@"Did receive read request sensor value");
            request.value = _sensorSimulator.sensorValues;
            [_peripheralManager respondToRequest:request withResult:CBATTErrorSuccess];
        } else {
            NSLog(@"Did receive read request sensor unsupported characteristic");
            [_peripheralManager respondToRequest:request withResult:CBATTErrorAttributeNotFound];
        }
    }
}


#pragma mark - UITableViewDelegate

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView
{
    return 2;
}

- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    if (section == 0) {
        return 1;
    } else if (section == 1) {
        return _store.services.count;
    }
    return 0;
}

- (UITableViewCell *)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    UITableViewCell *cell;
    if (indexPath.section == 0) {
        if (!_advertisingCell) {
            _advertisingCell = [self createAdvertisingCell];
        }
        _advertisingCell.textLabel.text = [NSString stringWithFormat:@"Advertising: %@", _hubSimulator.name];
        cell = _advertisingCell;
    } else if (indexPath.section == 1) {
        cell = [self.tableView dequeueReusableCellWithIdentifier:@"ServiceCellID" forIndexPath:indexPath];
        [self configureCell:cell withService:[_store.services objectAtIndex:indexPath.row]];
    }
    return cell;
}

- (void)configureCell:(UITableViewCell *)cell withService:(SimulatorService *)service
{
    cell.textLabel.text = service.name;
    cell.detailTextLabel.text = [NSString stringWithFormat:@"ConnectID: %lx, Hub: %lx, Type: %lx",
                    (long)service.connectID, (long)service.hubIndex, (long)service.typeID];
}

- (UITableViewCell *)createAdvertisingCell
{
    UITableViewCell *cell = [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault reuseIdentifier:nil];
    _advertisingSwitch = [[UISwitch alloc] init];
    [_advertisingSwitch setOn:NO animated:NO];
    [_advertisingSwitch addTarget:self action:@selector(switchChanged:) forControlEvents:UIControlEventValueChanged];
    cell.accessoryView = _advertisingSwitch;
    return cell;
}

- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender
{
    if ([segue.identifier isEqualToString:@"ServiceDetailSegue"]) {
        SimulatorServiceDetailViewController *vc = segue.destinationViewController;
        vc.service = _store.services[[self.tableView indexPathForSelectedRow].row];
    }
}

@end
