//
//  WindmillTestViewController.m
//  LEGODeviceDemo
//
//  Created by Bartlomiej Hyzy on 25/02/2015.
//  Copyright (c) 2015 lego. All rights reserved.
//

#import "WindmillTestViewController.h"
#import <LEGODevice/LEGODevice.h>


@interface WindmillTestViewController () <LEMotionSensorDelegate>

@property (strong, nonatomic) LEMotor *motor;
@property (strong, nonatomic) LEMotionSensor *motionSensor;
@property (strong, nonatomic) LERGBLight *rgbLight;

@property (assign, nonatomic) NSUInteger motorSpeed;

@property (assign, nonatomic) NSInteger motionSensorDistance;
@property (assign, nonatomic) NSInteger distanceTriggerLow;
@property (assign, nonatomic) NSInteger distanceTriggerHigh;
@property (assign, nonatomic) BOOL shiftDirectionEnabled;
@property (assign, nonatomic) NSUInteger shiftDirectionCount;
@property (assign, nonatomic) BOOL didShiftToHigh;

@end


@implementation WindmillTestViewController

#pragma mark - Factory Methods

+ (NSWindowController *)instantiateWindowControllerForDevice:(LEDevice *)device
{
    NSStoryboard *storyboard = [NSStoryboard storyboardWithName:@"Main" bundle:nil];
    NSWindowController *controller = [storyboard instantiateControllerWithIdentifier:@"windmillTest"];
    controller.window.contentViewController.representedObject = device;
    return controller;
}

#pragma mark - Initialization

- (void)setRepresentedObject:(id)representedObject
{
    NSParameterAssert([representedObject isKindOfClass:[LEDevice class]]);
    [super setRepresentedObject:representedObject];
    
    LEDevice *device = (LEDevice *)representedObject;
    
    for (LEService *service in device.services) {
        if ([service isKindOfClass:[LEMotor class]]) {
            self.motor = (LEMotor *)service;
        } else if ([service isKindOfClass:[LEMotionSensor class]]) {
            if (self.motionSensor != nil) {
                [self.motionSensor removeDelegate:self];
            }
            self.motionSensor = (LEMotionSensor *)service;
            self.motionSensorDistance = (NSInteger)self.motionSensor.distance;
            [self.motionSensor addDelegate:self];
        } else if ([service isKindOfClass:[LERGBLight class]]) {
            self.rgbLight = (LERGBLight *)service;
        }
    }
}

#pragma mark - UI Appearance Methods

- (void)viewWillAppear
{
    self.motorSpeed = 0;
    
    self.distanceTriggerLow = 3;
    self.distanceTriggerHigh = 8;
    self.shiftDirectionEnabled = YES;
}

#pragma mark - Sensor Manipulation

- (void)setMotorSpeed:(NSUInteger)motorSpeed
{
    NSParameterAssert(motorSpeed <= LEMotorMaxSpeed);
    
    if (motorSpeed < LEMotorMinSpeed) {
        _motorSpeed = 0;
        [self.motor brake];
    } else {
        _motorSpeed = motorSpeed;
        LEMotorDirection direction = (self.motor.direction == LEMotorDirectionRight ? LEMotorDirectionRight : LEMotorDirectionLeft);
        [self.motor runInDirection:direction power:motorSpeed];
    }
}

- (void)shiftDirection
{
    LEMotorDirection newDirection = LEMotorDirectionBraking;
    int newColorIndex = 0;
    if (self.motor.direction == LEMotorDirectionLeft) {
        newDirection = LEMotorDirectionRight;
        newColorIndex = 2;
    } else if (self.motor.direction == LEMotorDirectionRight) {
        newDirection = LEMotorDirectionLeft;
        newColorIndex = 6;
    }
    
    if (newDirection != LEMotorDirectionBraking) {
        [self.motor runInDirection:newDirection power:self.motor.power];
        self.rgbLight.colorIndex = newColorIndex;
        self.shiftDirectionCount++;
    }
}

#pragma mark - UI Actions

- (IBAction)dismissSheet:(id)sender
{
    [self.view.window.sheetParent endSheet:self.view.window];
}

- (IBAction)brake:(id)sender
{
    self.motorSpeed = 0;
}

- (IBAction)flatOut:(id)sender
{
    self.motorSpeed = LEMotorMaxSpeed;
}

- (IBAction)resetShiftCount:(id)sender
{
    self.shiftDirectionCount = 0;
}

#pragma mark - LEMotionSensorDelegate

- (void)motionSensor:(LEMotionSensor *)sensor didUpdateDistanceFrom:(CGFloat)oldDistance to:(CGFloat)newDistance
{
    self.motionSensorDistance = (NSInteger)newDistance;
    
    if (self.shiftDirectionEnabled) {
        if (self.motionSensorDistance >= self.distanceTriggerHigh && !self.didShiftToHigh) {
            [self shiftDirection];
            self.didShiftToHigh = YES;
        } else if (self.motionSensorDistance <= self.distanceTriggerLow && self.didShiftToHigh) {
            self.didShiftToHigh = NO;
        }
    }
}

@end
