﻿using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic.Data;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI;
using Newtonsoft.Json;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoDeviceSDK.Wrapper.Helpers
{
    public class WrapperSerialization
    {
        private const string Tag = "WrapperSerialization";

        // Device

        public virtual Dictionary<string, object> SerializeDevice(IDevice device)
        {
            var info = device.DeviceInfo;
            var deviceInfo = new Dictionary<string, object>()
            {
                { "FirmwareRevision", info != null && info.FirmwareRevision != null ? info.FirmwareRevision.StringRepresentation : "" },
                { "HardwareRevision", info != null && info.HardwareRevision != null ? info.HardwareRevision.StringRepresentation : "" },
                { "SoftwareRevision", info != null && info.SoftwareRevision != null ? info.SoftwareRevision.StringRepresentation : "" },
                { "ManufacturerName", info != null && info.ManufacturerName != null ? info.ManufacturerName : "" },
            };
             
            return new Dictionary<string, object>()
            {
                { "DeviceName", device.Name },
                { "DeviceID", device.DeviceId },
                { "DeviceInfo", deviceInfo },
                { "ButtonPressed", device.ButtonPressed },
                { "BatteryLevel", device.BatteryLevel },
                { "ConnectedState", (int)device.ConnectState },
                // "Category", "SupportedFunctions" and "LastConnectedNetworkId" are not serialized on purpose - they are not supported by the Windows SDK as of now.
            };
        }

        public virtual List<Dictionary<string, object>> SerializeDevices(List<IDevice> devices)
        {
            return devices.Select(d => SerializeDevice(d)).ToList();
        }

        // Device events

        public virtual Dictionary<string, object> SerializeDeviceNameChange(IDevice device, string oldName, string newName)
        {
            return new Dictionary<string, object>()
            {
                { "DeviceID", device.DeviceId },
                { "OldDeviceName", oldName },
                { "DeviceName", newName },
            };
        }

        public virtual Dictionary<string, object> SerializeDeviceButtonStateChange(IDevice device, bool pressed)
        {
            return new Dictionary<string, object>()
            {
                { "DeviceID", device.DeviceId },
                { "ButtonState", pressed },
            };
        }

        public virtual Dictionary<string, object> SerializeDeviceBatteryLevel(IDevice device, int level)
        {
            return new Dictionary<string, object>()
            {
                { "DeviceID", device.DeviceId },
                { "BatteryLevel", level },
            };
        }

        public virtual Dictionary<string, object> SerializeDeviceLowVoltageState(IDevice device, bool lowVoltage) {
            return new Dictionary<string, object>()
            {
                { "DeviceID", device.DeviceId },
                { "LowVoltageState", lowVoltage },
            };
        }

        public virtual Dictionary<string, object> SerializeDeviceError(IDevice device, Exception error)
        {
            return new Dictionary<string, object>()
            {
                { "DeviceID", device.DeviceId },
                { "Error", error.Message },
            };
        }

        // Input format

        public virtual Dictionary<string, object> SerializeInputFormat(InputFormat inputFormat)
        {
            if (inputFormat == null)
                return null;

            return new Dictionary<string, object>()
            {
                { "Revision", inputFormat.Revision },
                { "ConnectID", inputFormat.ConnectId },
                { "TypeID", inputFormat.TypeId },
                { "Mode", inputFormat.Mode },
                { "DeltaInterval", inputFormat.DeltaInterval },
                { "Unit", (int)inputFormat.Unit },
                { "NotificationsEnabled", inputFormat.NotificationsEnabled },
            };
        }

        // Service

        public virtual Dictionary<string, object> SerializeService(IService service, bool onlyBasicInfo = false)
        {
            if (onlyBasicInfo)
            {
                return new Dictionary<string, object>()
                {
                    { "DeviceID", service.Device.DeviceId },
                    { "ConnectID", service.ConnectInfo.ConnectId },
                };
            }
            else
            {
                return new Dictionary<string, object>()
                {
                    { "DeviceID", service.Device.DeviceId },
                    { "ServiceName", service.ServiceName },
                    { "DefaultInputFormat", SerializeInputFormat(service.DefaultInputFormat) },
                    { "InputFormat", SerializeInputFormat(service.InputFormat) },
                    { "IsInternalService", service.IsInternalService },
                    { "ConnectInfo", new Dictionary<string, object>() {
                        { "ConnectID", service.ConnectInfo.ConnectId },
                        { "HubIndex", service.ConnectInfo.HubIndex },
                        { "HardwareRevision", service.ConnectInfo.HardwareVersion.StringRepresentation },
                        { "SoftwareRevision", service.ConnectInfo.FirmwareVersion.StringRepresentation },
                        { "Type", service.ConnectInfo.Type } } },
                };
            }
        }

        public virtual Dictionary<string, object> SerializeServiceWithData(IService service)
        {
            Dictionary<string, object> serializedService = SerializeService(service);
            Dictionary<string, object> serviceData = SerializeServiceData(service);
            if (serviceData != null)
                serializedService.Add("ServiceData", serviceData);
            return serializedService;
        }

        // Specific services

        public virtual Dictionary<string, object> SerializeServiceData(IService service)
        {
            if (service is IMotor) {
                return SerializeServiceData(service as IMotor);
            }
            else if (service is IRgbLight) {
                return SerializeServiceData(service as IRgbLight);
            }
            else if (service is IMotionSensor) {
                return SerializeServiceData(service as IMotionSensor);
            }
            else if (service is ITiltSensor) {
                return SerializeServiceData(service as ITiltSensor);
            }
            else if (service is IVoltageSensor) {
                return SerializeServiceData(service as IVoltageSensor);
            }
            else if (service is ICurrentSensor) {
                return SerializeServiceData(service as ICurrentSensor);
            }
            else if (service is IPiezoTonePlayer) {
                return SerializeServiceData(service as IPiezoTonePlayer);
            }
            else {
                // return no service data for unknown services
                return null;
            }
        }

        public virtual Dictionary<string, object> SerializeServiceData(IMotor motor)
        {
            return new Dictionary<string, object>()
            {
                { "Power", motor.Power },
                { "MotorDirection", (int)motor.Direction },
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(IRgbLight rgbLight)
        {
            return new Dictionary<string, object>()
            {
                { "RGBLightMode", (int)rgbLight.RGBMode },
                { "Color", SerializeColor(rgbLight.Color) },
                { "ColorIndex", rgbLight.ColorIndex },
                { "DefaultColor", SerializeColor(rgbLight.DefaultColor) }, 
                { "DefaultColorIndex", rgbLight.DefaultColorIndex }, 
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(IMotionSensor motionSensor)
        {
            return new Dictionary<string, object>()
            {
                { "Count", motionSensor.Count },
                { "Distance", motionSensor.Distance },
                { "MotionSensorMode", (int)motionSensor.MotionSensorMode },
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(ITiltSensor tiltSensor)
        {
            return new Dictionary<string, object>()
            {
                { "TiltSensorDirection", (int)tiltSensor.Direction },
                { "TiltSensorMode", (int)tiltSensor.TiltSensorMode },
                { "Angle", new Dictionary<string, object>() {
                    { "X", tiltSensor.Angle.X },
                    { "Y", tiltSensor.Angle.Y } } },
                { "Crash", new Dictionary<string, object>() {
                    { "X", tiltSensor.Crash.X },
                    { "Y", tiltSensor.Crash.Y },
                    { "Z", tiltSensor.Crash.Z } } },
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(IVoltageSensor voltageSensor)
        {
            return new Dictionary<string, object>()
            {
                { "MilliVolts", voltageSensor.MilliVolts },
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(ICurrentSensor currentSensor)
        {
            return new Dictionary<string, object>()
            {
                { "MilliAmp", currentSensor.MilliAmp },
            };
        }

        public virtual Dictionary<string, object> SerializeServiceData(IPiezoTonePlayer piezoTonePlayer)
        {
            // intentionally return no data for the piezo player
            return null;
        }

        // Service events

        public virtual Dictionary<string, object> SerializeServiceValueDataChange(IService service, byte[] oldValue, byte[] newValue)
        {
            // TODO: Take size of data set into account

            var data = SerializeService(service, true);

            InputFormatUnit unit = InputFormatUnit.InputFormatUnitUnknown;
            if (service.InputFormat != null) {
                unit = service.InputFormat.Unit;
            }

            if (unit == InputFormatUnit.InputFormatUnitPercentage || 
                unit == InputFormatUnit.InputFormatUnitRaw) {
                if (oldValue != null) {
                    data.Add("OldValue", service.IntegerFromData(oldValue));
                }
                if (newValue != null) {
                    data.Add("NewValue", service.IntegerFromData(newValue));
                }
            }
            else if (unit == InputFormatUnit.InputFormatUnitSi) {
                if (oldValue != null) {
                    data.Add("OldValue", service.FloatFromData(oldValue));
                }
                if (newValue != null) {
                    data.Add("NewValue", service.FloatFromData(newValue));
                }
            }
            else {
                SdkLogger.E(Tag, "Cannot serialize value with unknown input format unit");
            }

            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceInputFormatChange(IService service, InputFormat oldFormat, InputFormat newFormat)
        {
            var data = SerializeService(service, true);
            if (oldFormat != null)
                data.Add("OldInputFormat", SerializeInputFormat(oldFormat));
            data.Add("NewInputFormat", SerializeInputFormat(newFormat));
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceCurrentChange(ICurrentSensor sensor, float milliAmps)
        {
            var data = SerializeService(sensor, true);
            data.Add("MilliAmp", milliAmps);
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceDistanceChange(IMotionSensor sensor, float oldDistance, float newDistance)
        {
            var data = SerializeService(sensor, true);
            data.Add("OldDistance", oldDistance);
            data.Add("NewDistance", newDistance);
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceCountChange(IMotionSensor sensor, int count)
        {
            var data = SerializeService(sensor, true);
            data.Add("Count", count);
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceColorChange(IRgbLight rgbLight, Color oldColor, Color newColor)
        {
            var data = SerializeService(rgbLight, true);
            if (oldColor != null)
                data.Add("OldColor", SerializeColor(oldColor));
            data.Add("NewColor", SerializeColor(newColor));
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceColorIndexChange(IRgbLight rgbLight, int oldIndex, int newIndex) {
            var data = SerializeService(rgbLight, true);
            data.Add("OldColorIndex", oldIndex);
            data.Add("NewColorIndex", newIndex);
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceDirectionChange(ITiltSensor sensor, TiltSensorDirection oldDirection, TiltSensorDirection newDirection)
        {
            var data = SerializeService(sensor, true);
            data.Add("OldDirection", (int)oldDirection);
            data.Add("NewDirection", (int)newDirection);
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceAngleChange(ITiltSensor sensor, TiltSensorAngle oldAngle, TiltSensorAngle newAngle)
        {
            var data = SerializeService(sensor, true);
            if (oldAngle != null)
            {
                data.Add("OldAngle", new Dictionary<string, object>() {
                    { "X", oldAngle.X },
                    { "Y", oldAngle.Y },
                });
            }
            data.Add("NewAngle", new Dictionary<string, object>() {
                { "X", newAngle.X },
                { "Y", newAngle.Y },
            });
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceCrashChange(ITiltSensor sensor, TiltSensorCrash oldCrash, TiltSensorCrash newCrash)
        {
            var data = SerializeService(sensor, true);
            if (oldCrash != null)
            {
                data.Add("OldCrash", new Dictionary<string, object>() {
                    { "X", oldCrash.X },
                    { "Y", oldCrash.Y },
                    { "Z", oldCrash.Z },
                });
            }
            data.Add("NewCrash", new Dictionary<string, object>() {
                { "X", newCrash.X },
                { "Y", newCrash.Y },
                { "Z", newCrash.Z },
            });
            return data;
        }

        public virtual Dictionary<string, object> SerializeServiceVoltageChange(IVoltageSensor sensor, float milliVolts)
        {
            var data = SerializeService(sensor, true);
            data.Add("MilliVolts", milliVolts);
            return data;
        }
                
        // Helpers

        public virtual Dictionary<string, object> SerializeColor(Color color)
        {
            return new Dictionary<string, object>()
            {
                { "R", color.R / 255.0 },
                { "G", color.G / 255.0 },
                { "B", color.B / 255.0 },
                { "A", color.A / 255.0 },
            };
        }

        public virtual Color DeserializeColor(Dictionary<string, object> color)
        {
            try
            {
                return Color.FromArgb((byte)(Convert.ToDouble(color["A"]) * 255), 
                                      (byte)(Convert.ToDouble(color["R"]) * 255), 
                                      (byte)(Convert.ToDouble(color["G"]) * 255), 
                                      (byte)(Convert.ToDouble(color["B"]) * 255));
            }
            catch (Exception e)
            {
                throw new ArgumentException("Failed to parse given color dictionary", e);
            }
        }

        public virtual string ConvertToJSONString(Dictionary<string, object> dictionary)
        {
            return JsonConvert.SerializeObject(dictionary);
        }

        public virtual string ConvertToJSONString(List<Dictionary<string, object>> list)
        {
            return JsonConvert.SerializeObject(list);
        }

        public virtual Dictionary<string, object> ParseDictionaryFromJSONString(string jsonString)
        {
            return JsonConvert.DeserializeObject<Dictionary<string, object>>(jsonString);
        }
    }
}
