﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using LEGODeviceUnitySDK;
using System.Diagnostics;
using System.Threading.Tasks;

namespace LegoDeviceSDK.Wrapper
{
    public class DeviceManagerWrapper : IDeviceManagerWrapperImpl
    {
        private readonly IDeviceManager wrappedManager;
        private readonly WrapperSerialization wrapperSerialization;

        private const string Tag = "DeviceManagerWrapper";

        public IDeviceWrapperImpl DeviceWrapper { get; set; }
        public ILEGODeviceManagerNativeDelegate UnityDeviceManagerDelegate { get; set; }

        public DeviceManagerWrapper(IDeviceManager manager, WrapperSerialization serialization)
        {
            wrappedManager = manager;
            wrapperSerialization = serialization;

            manager.AddDelegate(this);
        }

        internal DeviceManagerWrapper()
            : this(DeviceManager.SharedDeviceManager, new WrapperSerialization())
        { }

        public void Dispose()
        {
            wrappedManager.RemoveDelegate(this);
        }

        public void Scan()
        {
            // Start the scan task and continue immediately without waiting for it to finish
            Task task = wrappedManager.ScanAsync();
        }

        public void StopScanning()
        {
            // Intentionally left empty - the SDK does not allow stopping a scan.
        }

        public void ConnectToDevice(string deviceID)
        {
            IDevice device = FindDevice(deviceID);
            if (device != null) {
                // Start the connect task and continue immediately without waiting for it to finish
                Task task = wrappedManager.ConnectToDeviceAsync(device);
            }
        }

        public void DisconnectDevice(string deviceID)
        {
            IDevice device = FindDevice(deviceID);
            if (device != null)
                wrappedManager.CancelDeviceConnection(device);
        }

        public string AllDevices()
        {
            return wrapperSerialization.ConvertToJSONString(wrapperSerialization.SerializeDevices(wrappedManager.AllDevices));
        }

        public IDevice FindDevice(string deviceID)
        {
            IDevice device = wrappedManager.AllDevices.Find(d => d.DeviceId == deviceID);
            if (device == null)
                SdkLogger.W(Tag, string.Format("Device not found: {0}", deviceID));
            return device;
        }

        public IService FindService(string deviceID, string connectID)
        {
            IDevice device = FindDevice(deviceID);
            if (device == null)
                return null;

            int serviceConnectID;
            if (!int.TryParse(connectID, out serviceConnectID))
                throw new ArgumentException(string.Format("Invalid service connect ID: {0}", connectID));

            IService service = device.Services.Find(s => s.ConnectInfo.ConnectId == serviceConnectID);
            if (service == null)
                SdkLogger.W(Tag, string.Format("Service not found for device {0}: {1}", deviceID, connectID));
            return service;
        }

        #region IDeviceManagerDelegate

        private void InvokeUnityMethod(Action<string> method, IDevice device)
        {
            UnityInvoker.InvokeUnityMethod(method, wrapperSerialization.ConvertToJSONString(wrapperSerialization.SerializeDevice(device)));
        }

        public void DeviceDidAppear(IDeviceManager deviceManager, IDevice device)
        {
            Debug.Assert(DeviceWrapper != null, "Device wrapper should have been set by now!");
            device.AddDelegate(DeviceWrapper);
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDeviceDidAppear, device);
        }

        public void DeviceDidDisappear(IDeviceManager deviceManager, IDevice device)
        {
            Debug.Assert(DeviceWrapper != null, "Device wrapper should have been set by now!");
            device.RemoveDelegate(DeviceWrapper);
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDeviceDidDisappear, device);
        }

        public void DidDisconnectFromDevice(IDeviceManager deviceManager, IDevice device, bool autoReconnect)
        {
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDidDisconnectFromDevice, device);
        }

        public void DidFailToConnectToDevice(IDeviceManager deviceManager, IDevice device, bool autoReconnect)
        {
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDidFailToConnectToDevice, device);
        }

        public void DidFinishInterrogatingDevice(IDeviceManager deviceManager, IDevice device)
        {
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDidFinishInterrogatingDevice, device);
        }

        public void DidStartInterrogatingDevice(IDeviceManager deviceManager, IDevice device)
        {
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerDidStartInterrogatingDevice, device);
        }

        public void WillStartConnectingToDevice(IDeviceManager deviceManager, IDevice device)
        {
            InvokeUnityMethod(UnityDeviceManagerDelegate.LEDeviceManagerWillStartConnectingToDevice, device);
        }

        #endregion
    }
}
