﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Bluetooth;
using LEGODeviceUnitySDK;
using Newtonsoft.Json;
using Windows.UI;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoDeviceSDK.Wrapper
{
    public class ServiceWrapper : IServiceWrapperImpl, ICurrentSensorDelegate, IMotionSensorDelegate, IRgbLightDelegate, ITiltSensorDelegate, IVoltageSensorDelegate
    {
        private readonly IDeviceManagerWrapperImpl deviceManagerWrapper;
        private readonly WrapperSerialization wrapperSerialization;

        private const string Tag = "ServiceWrapper";

        public ILEGOServiceNativeDelegate UnityServiceDelegate { get; set; }

        public ServiceWrapper(IDeviceManagerWrapperImpl deviceManagerWrapper, WrapperSerialization wrapperSerialization)
        {
            this.deviceManagerWrapper = deviceManagerWrapper;
            this.wrapperSerialization = wrapperSerialization;
        }

        internal ServiceWrapper(DeviceManagerWrapper deviceManagerWrapper)
            : this(deviceManagerWrapper, new WrapperSerialization())
        { }

        #region Service API
        public void UpdateServiceData(string deviceID, string connectID)
        {
            IService service = deviceManagerWrapper.FindService(deviceID, connectID);
            if (service != null)
                InvokeUnityMethod(UnityServiceDelegate.LEServiceUpdateServiceData, wrapperSerialization.SerializeServiceWithData(service));
        }

        public void UpdateInputFormat(string deviceID, string connectID, string mode, string unit, string deltaInterval, bool notificationEnabled)
        {
            IService service = deviceManagerWrapper.FindService(deviceID, connectID);
            if (service == null)
                return;

            try
            {
                InputFormat inputFormat = InputFormat.InputFormatWithConnectId(
                    int.Parse(connectID), service.ConnectInfo.Type, int.Parse(mode), uint.Parse(deltaInterval), (InputFormatUnit)int.Parse(unit), notificationEnabled);
                service.UpdateInputFormat(inputFormat);
            }
            catch (Exception e)
            {
                if (e is ArgumentNullException || e is FormatException || e is OverflowException)
                    SdkLogger.E(Tag, string.Format("Failed to update input format of service {0} for device {1}: {2}", connectID, deviceID, e.Message));
                else
                    throw;
            }
        }
        #endregion

        #region Motor
        public void Run(string deviceID, string connectID, string motorDirection, string power)
        {
            IMotor motor = (IMotor)deviceManagerWrapper.FindService(deviceID, connectID);
            if (motor != null)
                motor.RunInDirection((MotorDirection)int.Parse(motorDirection), int.Parse(power));
        }

        public void Brake(string deviceID, string connectID)
        {
            IMotor motor = (IMotor)deviceManagerWrapper.FindService(deviceID, connectID);
            if (motor != null)
                motor.Brake();
        }

        public void Drift(string deviceID, string connectID)
        {
            IMotor motor = (IMotor)deviceManagerWrapper.FindService(deviceID, connectID);
            if (motor != null)
                motor.Drift();
        }
        #endregion

        #region RGB Light
        public void SetRGBLightMode(string deviceID, string connectID, string rgbLightMode) 
        {
            IRgbLight light = (IRgbLight)deviceManagerWrapper.FindService(deviceID, connectID);
            if (light != null)
                light.RGBMode = (RgbLightMode)int.Parse(rgbLightMode);
        }

        public void SwitchOff(string deviceID, string connectID)
        {
            IRgbLight light = (IRgbLight)deviceManagerWrapper.FindService(deviceID, connectID);
            if (light != null)
                light.SwitchOff();
        }

        public void ChangeColor(string deviceID, string connectID, string colorJSON)
        {
            IRgbLight light = (IRgbLight)deviceManagerWrapper.FindService(deviceID, connectID);
            if (light == null)
                return;

            try
            {
                light.Color = wrapperSerialization.DeserializeColor(wrapperSerialization.ParseDictionaryFromJSONString(colorJSON));
            }
            catch (JsonException)
            {
                SdkLogger.E(Tag, string.Format("Invalid color JSON format: {0}", colorJSON));
            }
            catch (ArgumentException)
            {
                SdkLogger.E(Tag, string.Format("Invalid color: {0}", colorJSON));
            }
        }

        public void ChangeColorIndex(string deviceID, string connectID, string colorIndex)
        {
            IRgbLight light = (IRgbLight)deviceManagerWrapper.FindService(deviceID, connectID);
            if (light != null)
                light.ColorIndex = int.Parse(colorIndex);
        }

        public void ChangeColorToDefault(string deviceID, string connectID)
        {
            IRgbLight light = (IRgbLight)deviceManagerWrapper.FindService(deviceID, connectID);
            if (light != null)
                light.SwitchToDefaultColor();
        }
        #endregion

        #region Tilt sensor
        public void SetTiltSensorMode(string deviceID, string connectID, string tiltSensorMode)
        {
            ITiltSensor tiltSensor = (ITiltSensor)deviceManagerWrapper.FindService(deviceID, connectID);
            if (tiltSensor != null)
                tiltSensor.TiltSensorMode = (TiltSensorMode)int.Parse(tiltSensorMode);
        }
        #endregion

        #region Motion sensor
        public void SetMotionSensorMode(string deviceID, string connectID, string motionSensorMode)
        {
            IMotionSensor motionSensor = (IMotionSensor)deviceManagerWrapper.FindService(deviceID, connectID);
            if (motionSensor != null)
                motionSensor.MotionSensorMode = (MotionSensorMode)int.Parse(motionSensorMode);
        }
        #endregion

        #region Piezo tone player
        public void PlayFrequency(string deviceID, string connectID, string frequency, string milliseconds)
        {
            IPiezoTonePlayer player = (IPiezoTonePlayer)deviceManagerWrapper.FindService(deviceID, connectID);
            if (player != null)
                player.PlayFrequency(int.Parse(frequency), int.Parse(milliseconds));
        }

        public void PlayNote(string deviceID, string connectID, string note, string octave, string milliseconds)
        {
            IPiezoTonePlayer player = (IPiezoTonePlayer)deviceManagerWrapper.FindService(deviceID, connectID);
            if (player != null)
                player.PlayNote((PiezoTonePlayerNote)int.Parse(note), int.Parse(octave), int.Parse(milliseconds));
        }

        public void StopPlaying(string deviceID, string connectID)
        {
            IPiezoTonePlayer player = (IPiezoTonePlayer)deviceManagerWrapper.FindService(deviceID, connectID);
            if (player != null)
                player.StopPlaying();
        }
        #endregion

        private void InvokeUnityMethod(Action<string> method, Dictionary<string, object> dataDictionary)
        {
            UnityInvoker.InvokeUnityMethod(method, wrapperSerialization.ConvertToJSONString(dataDictionary));
        }

        #region IServiceDelegate
        public void DidUpdateInputFormat(IService service, Generic.InputFormat oldFormat, Generic.InputFormat newFormat)
        {
            InvokeUnityMethod(UnityServiceDelegate.LEServiceDidUpdateInputFormat, wrapperSerialization.SerializeServiceInputFormatChange(service, oldFormat, newFormat));
        }

        public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData)
        {
            InvokeUnityMethod(UnityServiceDelegate.LEServiceDidUpdateValueData, wrapperSerialization.SerializeServiceValueDataChange(service, oldData, newData));
        }
        #endregion

        #region ICurrentSensorDelegate
        public void DidUpdateMilliAmp(ICurrentSensor service, float milliAmps)
        {
            InvokeUnityMethod(UnityServiceDelegate.LECurrentSensorDidUpdateMilliAmp, wrapperSerialization.SerializeServiceCurrentChange((ICurrentSensor)service, milliAmps));
        }
        #endregion

        #region IMotionSensorDelegate
        public void DidUpdateDistance(IMotionSensor motionSensor, float oldDistance, float newDistance)
        {
            InvokeUnityMethod(UnityServiceDelegate.LEMotionSensorDidUpdateDistance, wrapperSerialization.SerializeServiceDistanceChange((IMotionSensor)motionSensor, oldDistance, newDistance));
        }

        public void DidUpdateCount(IMotionSensor motionSensor, uint newCount)
        {
            InvokeUnityMethod(UnityServiceDelegate.LEMotionSensorDidUpdateCount, wrapperSerialization.SerializeServiceCountChange((IMotionSensor)motionSensor, (int)newCount));
        }
        #endregion

        #region IRgbLightDelegate
        public void DidUpdateColor(IRgbLight rgbLight, Windows.UI.Color oldColor, Windows.UI.Color newColor)
        {
            InvokeUnityMethod(UnityServiceDelegate.LERGBLightDidUpdateColor, wrapperSerialization.SerializeServiceColorChange((IRgbLight)rgbLight, oldColor, newColor));
        }

        public void DidUpdateColorIndex(IRgbLight rgbLight, int oldColorIndex, int newColorIndex) 
        {
            InvokeUnityMethod(UnityServiceDelegate.LERGBLightDidUpdateColorIndex, wrapperSerialization.SerializeServiceColorIndexChange((IRgbLight)rgbLight, oldColorIndex, newColorIndex));
        }
        #endregion

        #region ITiltSensorDelegate
        public void DidUpdateAngle(ITiltSensor tiltSensor, Generic.Data.TiltSensorAngle oldAngle, Generic.Data.TiltSensorAngle newAngle)
        {
            InvokeUnityMethod(UnityServiceDelegate.LETiltSensorDidUpdateAngle, wrapperSerialization.SerializeServiceAngleChange(tiltSensor, oldAngle, newAngle));
        }

        public void DidUpdateCrash(ITiltSensor tiltSensor, Generic.Data.TiltSensorCrash oldCrashValue, Generic.Data.TiltSensorCrash newCrashValue)
        {
            InvokeUnityMethod(UnityServiceDelegate.LETiltSensorDidUpdateCrash, wrapperSerialization.SerializeServiceCrashChange(tiltSensor, oldCrashValue, newCrashValue));
        }

        public void DidUpdateDirection(ITiltSensor tiltSensor, Generic.Data.TiltSensorDirection oldDirection, Generic.Data.TiltSensorDirection newDirection)
        {
            InvokeUnityMethod(UnityServiceDelegate.LETiltSensorDidUpdateDirection, wrapperSerialization.SerializeServiceDirectionChange(tiltSensor, oldDirection, newDirection));
        }
        #endregion

        #region IVoltageSensorDelegate
        public void DidUpdateMilliVolts(IVoltageSensor voltageSensor, float milliVolts)
        {
            InvokeUnityMethod(UnityServiceDelegate.LEVoltageSensorDidUpdateMilliVolts, wrapperSerialization.SerializeServiceVoltageChange(voltageSensor, milliVolts));
        }
        #endregion
    }
}
