﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Text;
using GalaSoft.MvvmLight.Threading;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;

namespace LegoBleTestApp.Helpers
{
	/// <summary>
	/// Handles viewing of lists with services on the GUI. This collection subscribes itself to the service callbacks and updates itself when changes happen
	/// </summary>
    public class ObservableServiceCollection : ObservableCollection<IService>, IServiceDelegate {

	    public ObservableServiceCollection(IEnumerable<IService> services) : base(services) {
				foreach (var service in Items) {
			    service.AddDelegate(this);
		    }
	    }

			/// <summary>
			/// Inserts a new item and subscribes to the callbacks
			/// </summary>
			/// <param name="index">The index the item is inserted to</param>
			/// <param name="item">The item that is added to the list</param>
			protected override void InsertItem(int index, IService item) {
		    base.InsertItem(index, item);
				item.AddDelegate(this);
	    }

			/// <summary>
			/// Removes an item from the list. Makes sure to unsubcribe from the removed item
			/// </summary>
			/// <param name="index">The index of the item to be removed</param>
	    protected override void RemoveItem(int index) {
				base[index].RemoveDelegate(this);
				base.RemoveItem(index);
	    }

			/// <summary>
			/// Clears the list and unsubscribes itself from all service callbacks
			/// </summary>
	    protected override void ClearItems() {
		    foreach (var service in Items) {
			    service.RemoveDelegate(this);
		    }
				base.ClearItems();
	    }

			/// <summary>
			/// Callback when the input format changes. Sends an update notification to the List, that it is updated
			/// </summary>
			/// <param name="service">The service with the changed input format</param>
			/// <param name="oldFormat">The old input format</param>
			/// <param name="newFormat">The new input format</param>
			public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
		    DispatcherHelper.CheckBeginInvokeOnUI(() => base.OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset)));
	    }

			/// <summary>
			/// Callback when the value changes. Sends an update notification to the List, that it is updated
			/// </summary>
			/// <param name="service">The service with the changed value</param>
			/// <param name="oldData">The old value</param>
			/// <param name="newData">The new value</param>
			public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
		    DispatcherHelper.CheckBeginInvokeOnUI(() => base.OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset)));
			}

			/// <summary>
			/// Subscribes to all services
			/// </summary>
	    public void BindDelegates() {
		    foreach (var service in Items) {
			    service.AddDelegate(this);
		    }
	    }

			/// <summary>
			/// Unsubscribes from all services
			/// </summary>
			public void UnbindDelegates() {
				foreach (var service in Items) {
					service.RemoveDelegate(this);
				}
			}


    }
}
