﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using Windows.UI.Core;
using Windows.UI.Popups;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Ioc;
using GalaSoft.MvvmLight.Threading;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// Lists the devices that can be connected to
	/// </summary>
	public class DeviceListingViewModel : ExtendedViewModelBase, ILocation, IDeviceManagerDelegate, IDeviceDelegate {
		readonly IViewService viewService;
		readonly DeviceManager deviceManager;

		/// <summary>
		/// List of devices that can be connected to
		/// </summary>
		List<IDevice> devices;
		public List<IDevice> Devices {
			get { return devices; }
			set { Set(ref devices, value); }
		}

		/// <summary>
		/// Is executed when scan button is pressed
		/// </summary>
		public RelayCommand ScanCommand {
			get { return new RelayCommand(async () => await Scan()); }
		}

		/// <summary>
		/// Is executed when trying to connect to a device
		/// </summary>
		RelayCommand<BluetoothDevice> connectToDeviceCommand;
		public RelayCommand<BluetoothDevice> ConnectToDeviceCommand {
			get { return connectToDeviceCommand; }
			set { Set(ref connectToDeviceCommand, value); }
		}

		public DeviceListingViewModel(IViewService viewService, DeviceManager deviceManager) {
			this.viewService = viewService;
			this.deviceManager = deviceManager;
			ConnectToDeviceCommand = new RelayCommand<BluetoothDevice>(Execute);
			deviceManager.AddDelegate(this);
		}

		/// <summary>
		/// Move to Device View with specified device
		/// </summary>
		/// <param name="device">The device to be shown in the Device View</param>
		void Execute(IDevice device) {
			viewService.NavigateTo<DeviceViewModel, int>(deviceManager.AllDevices.IndexOf(device));
		}

		/// <summary>
		/// Scan for devices
		/// </summary>
		/// <returns>Awaitable Task</returns>
		public async Task Scan() {
			await deviceManager.ScanAsync();
		}

		/// <summary>
		/// Called when page is navigated to. Scans and updates the device list
		/// </summary>
		public override async void OnNavigatedTo() {
			base.OnNavigatedTo();
			await Scan();
			UpdateDeviceList();
		}

		/// <summary>
		/// Updates the device list on the UI thread
		/// </summary>
		void UpdateDeviceList() {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				Devices = deviceManager.AllDevices;
			});
		}

		public void DeviceDidAppear(IDeviceManager deviceManager2, IDevice device) {
			UpdateDeviceList();
            device.AddDelegate(this);
		}

		public void DeviceDidDisappear(IDeviceManager deviceManager2, IDevice device) {
			UpdateDeviceList();
            device.RemoveDelegate(this);
		}

		/// <summary>
		/// Called when device disconnects
		/// </summary>
		/// <param name="deviceManager2">Devicemanager the device disconnects from</param>
		/// <param name="device">The device that disconnects</param>
		/// <param name="autoReconnect">If the device should automatically reconnect</param>
		public void DidDisconnectFromDevice(IDeviceManager deviceManager2, IDevice device, bool autoReconnect) {
			DispatcherHelper.CheckBeginInvokeOnUI(async () => {
				var msgDialog = new MessageDialog(string.Format("The device {0} was disconnected.", device.Name));
				await msgDialog.ShowAsync();
				// if on one of the pages below and the current device is the device that disconnects. Return to this View
				var type = viewService.CurrentPage;
				var currentViewModel = SimpleIoc.Default.GetInstance(type);
				var genericSensor = currentViewModel as GenericServiceViewModel;
				if (genericSensor != null && genericSensor.Service != null && genericSensor.Service.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var motionSensor = currentViewModel as MotionSensorViewModel;
				if (motionSensor != null && motionSensor.Sensor != null && motionSensor.Sensor.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var motor = currentViewModel as MotorViewModel;
				if (motor != null && motor.Motor != null && motor.Motor.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var piezo = currentViewModel as PiezoTonePlayerViewModel;
				if (piezo != null && piezo.PiezoTonePlayer != null && piezo.PiezoTonePlayer.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var tilt = currentViewModel as TiltSensorViewModel;
				if (tilt != null && tilt.TiltSensor != null && tilt.TiltSensor.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var rgb = currentViewModel as RgbLightViewModel;
				if (rgb != null && rgb.RgbLight != null && rgb.RgbLight.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
				var deviceViewModel = currentViewModel as DeviceViewModel;
				if (deviceViewModel != null && deviceViewModel.Device != null && deviceViewModel.Device == device)
					viewService.NavigateTo<DeviceListingViewModel>();
			});
			UpdateDeviceList();
		}

		public void DidFailToConnectToDevice(IDeviceManager deviceManager2, IDevice device, bool autoReconnect) {
			UpdateDeviceList();
		}

		public void DidFinishInterrogatingDevice(IDeviceManager deviceManager2, IDevice device) {
			UpdateDeviceList();
		}

		public void DidStartInterrogatingDevice(IDeviceManager deviceManager2, IDevice device) {
			UpdateDeviceList();
		}

		public void WillStartConnectingToDevice(IDeviceManager deviceManager2, IDevice device) {
			UpdateDeviceList();
		}

        public void DidUpdateDeviceInfo(IDevice device, DeviceInfo deviceInfo, Exception error) {
            // Do nothing
        }

        public void DidChangeButtonState(IDevice device, bool buttonState) {
            // Do nothing
        }

        public void DidChangeName(IDevice device, string fromName, string toName) {
            // Do nothing
        }

        public void DidUpdateBatteryLevel(IDevice device, uint newLevel) {
            // Do nothing
        }

        public void DidChangeLowVoltageState(IDevice device, bool lowVoltage) {
            UpdateDeviceList();
        }

        public void DidAddService(IDevice device, IService service) {
            // Do nothing
        }

        public void DidRemoveService(IDevice device, IService service) {
            // Do nothing
        }

        public void DidFailToAddServiceWithError(IDevice device, Exception error) {
            // Do nothing
        }
    }
}
