﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using Windows.UI.Xaml;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Threading;
using LegoBleTestApp.Common;
using LegoBleTestApp.Helpers;
using LegoBleTestApp.Services;
using LegoDeviceSDK;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using RelayCommand = GalaSoft.MvvmLight.Command.RelayCommand;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// View that shows all important aspects of a device i.e. services, button pressed, Name and all revisions
	/// </summary>
	public class DeviceViewModel : ExtendedViewModelBase, ILocation<int>, IDeviceDelegate, IDeviceManagerDelegate {
		readonly IViewService viewService;
		readonly DeviceManager deviceManager;

		/// <summary>
		/// The device to view
		/// </summary>
		IDevice iDevice;
		public IDevice Device {
			get { return iDevice; }
			set {
				Set(ref iDevice, value);
			}
		}

		/// <summary>
		/// Boolean that represents if the button on the device is pressed
		/// </summary>
		bool buttonPressed = false;
		public bool ButtonPressed {
			get { return buttonPressed; }
			set { Set(ref buttonPressed, value);
			}
		}

		/// <summary>
		/// The name of the device
		/// </summary>
		string name;
		public string Name {
			get { return name; }
			set {
				Set(ref name, value);
			}
		}

		ObservableServiceCollection externalServices = new ObservableServiceCollection(new List<IService>());
		/// <summary>
		/// List of external services
		/// </summary>
		public ObservableServiceCollection ExternalServices {
			get { return externalServices; }
			private set { Set(ref externalServices, value); }
		}
		ObservableServiceCollection internalServices = new ObservableServiceCollection(new List<IService>());
		/// <summary>
		/// List of internal services
		/// </summary>
		public ObservableServiceCollection InternalServices { 
			get { return internalServices; }
			private set { Set(ref internalServices, value); }
		}

		/// <summary>
		/// Command to go back to the Hub (DeviceListing)
		/// </summary>
		public RelayCommand GoToHubCommand {
			get { return new RelayCommand(() => viewService.NavigateTo<DeviceListingViewModel>()); }
		}

		/// <summary>
		/// Command to disconnect from device
		/// </summary>
		public RelayCommand DisconnectCommand {
			get { return new RelayCommand(() => deviceManager.CancelDeviceConnection(Device)); }
		}

		/// <summary>
		/// Command to go to the Windmill test
		/// </summary>
		public RelayCommand WindmillTestCommand {
			get { return new RelayCommand(() => viewService.NavigateTo<WindMillTestViewModel, int>(deviceManager.AllDevices.IndexOf(Device))); }
		}

		/// <summary>
		/// Command to change the name of the Device
		/// </summary>
		public RelayCommand UploadNameCommand {
			get { return new RelayCommand(() => Device.Name = Name); }
		}

		/// <summary>
		/// Command that directs the user to the appropiate view, when a service is clicked
		/// </summary>
		public RelayCommand<IService> ViewServiceCommand {
			get {
				return new RelayCommand<IService>((service) => {
					if (service is BluetoothMotor)
						viewService.NavigateTo<MotorViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
					else if (service is BluetoothMotionSensor)
						viewService.NavigateTo<MotionSensorViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
					else if (service is BluetoothPiezoTonePlayer)
						viewService.NavigateTo<PiezoTonePlayerViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
					else if (service is BluetoothRgbLight)
						viewService.NavigateTo<RgbLightViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
					else if (service is BluetoothTiltSensor)
						viewService.NavigateTo<TiltSensorViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
					else
						viewService.NavigateTo<GenericServiceViewModel, Tuple<int, int>>(new Tuple<int, int>(deviceManager.AllDevices.IndexOf(Device), Device.Services.IndexOf(service)));
				});
			}
		}

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device is loaded.
		/// </summary>
		/// <param name="parameter">integer that represents the index of the device in the device list of the device manager</param>
		public async void LoadParameter(int parameter) {
			var paramDevice = deviceManager.AllDevices[parameter];
			Device = paramDevice;
			Device.AddDelegate(this);
			if (Device.ConnectState == DeviceState.DeviceStateDisconnected) {
				await deviceManager.ConnectToDeviceAsync(paramDevice);
				DispatcherHelper.CheckBeginInvokeOnUI(() => RaisePropertyChanged("Device"));
			}
			Name = Device.Name;
		}

		public DeviceViewModel(IViewService viewService, DeviceManager deviceManager) {
			this.viewService = viewService;
			this.deviceManager = deviceManager;
			InternalServices = new ObservableServiceCollection(new List<IService>());
			ExternalServices = new ObservableServiceCollection(new List<IService>());
			DispatcherHelper.CheckBeginInvokeOnUI(() => RaisePropertyChanged("Device"));
		}

		public void DidUpdateDeviceInfo(IDevice device, DeviceInfo deviceInfo, Exception error) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => RaisePropertyChanged("Device"));
		}

		public void DidChangeButtonState(IDevice device, bool buttonState) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => ButtonPressed = buttonState);
		}

		public void DidChangeName(IDevice device, string fromName, string toName) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				RaisePropertyChanged("Device");
				Name = toName;
			});
		}

		public void DidUpdateBatteryLevel(IDevice device, uint newLevel) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => RaisePropertyChanged("Device"));
		}

        public void DidChangeLowVoltageState(IDevice device, bool lowVoltage) {
            // Do nothing
        }

		/// <summary>
		/// Called when a service is added
		/// </summary>
		/// <param name="device">The device the service was added to</param>
		/// <param name="service">The service that was added</param>
		public void DidAddService(IDevice device, IService service) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				lock (InternalServices) {
					InternalServices = new ObservableServiceCollection(device.InternalServices);
					ExternalServices = new ObservableServiceCollection(device.ExternalServices);
				}
			});
		}

		/// <summary>
		/// Called when a service is removed
		/// </summary>
		/// <param name="device">The device the service was removed from</param>
		/// <param name="service">The service that was removed</param>
		public void DidRemoveService(IDevice device, IService service) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				lock (InternalServices) {
					InternalServices = new ObservableServiceCollection(device.InternalServices);
					ExternalServices = new ObservableServiceCollection(device.ExternalServices);
				}
			});
		}

		public void DidFailToAddServiceWithError(IDevice device, Exception error) {
			throw new NotImplementedException();
		}

		/// <summary>
		/// When navigated to, add this to the device delegates and initialize the service lists
		/// </summary>
		public override void OnNavigatedTo() {
			base.OnNavigatedTo();
			DispatcherHelper.CheckBeginInvokeOnUI(() => RaisePropertyChanged("Device"));
			deviceManager.AddDelegate(this);
			lock (InternalServices) {
				InternalServices = new ObservableServiceCollection(Device.InternalServices);
				ExternalServices = new ObservableServiceCollection(Device.ExternalServices);
			}

		}

		/// <summary>
		/// When moving to another View, unbind all delegates from the service lists and remove itself from the Device
		/// </summary>
		public override void OnNavigatedFrom() {
			base.OnNavigatedFrom();
			Device.RemoveDelegate(this);
			deviceManager.RemoveDelegate(this);
			ExternalServices.UnbindDelegates();
			InternalServices.UnbindDelegates();
		}

		#region UpdateConnection color when device state changes

		/// <summary>
		/// Refreshes the device i.e. all values associated with the device
		/// </summary>
		void UpdateDeviceState() {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				RaisePropertyChanged("Device");
			});
		}

		public void DeviceDidAppear(IDeviceManager deviceManager2, IDevice device) {
			if(device == Device)
				UpdateDeviceState();
		}

		public void DeviceDidDisappear(IDeviceManager deviceManager2, IDevice device) {
			if (device == Device)
				UpdateDeviceState();
		}

		public void DidDisconnectFromDevice(IDeviceManager deviceManager2, IDevice device, bool autoReconnect) {
			if (device == Device)
				UpdateDeviceState();
		}

		public void DidFailToConnectToDevice(IDeviceManager deviceManager2, IDevice device, bool autoReconnect) {
			if (device == Device)
				UpdateDeviceState();
		}

		public void DidFinishInterrogatingDevice(IDeviceManager deviceManager2, IDevice device) {
			if (device == Device)
				UpdateDeviceState();
		}

		public void DidStartInterrogatingDevice(IDeviceManager deviceManager2, IDevice device) {
			if (device == Device)
				UpdateDeviceState();
		}

		public void WillStartConnectingToDevice(IDeviceManager deviceManager2, IDevice device) {
			if (device == Device)
				UpdateDeviceState();
		}

		#endregion

    }
}
