﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GalaSoft.MvvmLight.Command;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// The view for a motor service
	/// </summary>
	public class MotorViewModel : ExtendedViewModelBase, ILocation<Tuple<int, int>> {
		readonly DeviceManager deviceManager;
		readonly IViewService viewService;
		readonly StandardServiceControlViewModel standardServiceControl;
		readonly ModifyInputFormatControlViewModel modifyInputFormatControl;

		BluetoothMotor motor;
		/// <summary>
		/// The motor
		/// </summary>
		public BluetoothMotor Motor {
			get { return motor; }
			private set { Set(ref motor, value); }
		}

		int speed;

		MotorDirection direction;
		/// <summary>
		/// The motor direction
		/// </summary>
		public MotorDirection Direction {
			get { return direction; }
			set { Set(ref direction, value); }
		}

		string motorString;
		/// <summary>
		/// A string representing the motor direction and speed
		/// </summary>
		public string MotorString {
			get { return motorString; }
			set { Set(ref motorString, value); }
		}

		bool lastSelectedDirectionRight = true;
		/// <summary>
		/// The last selected Direction
		/// </summary>
		public bool LastSelectedDirectionRight {
			get { return lastSelectedDirectionRight; }
			set { Set(ref lastSelectedDirectionRight, value); }
		}

		// Commands for controling the Motor
		public RelayCommand AccelerateCommand { get; set; }
		public RelayCommand DecelerateCommand { get; set; }

		public RelayCommand LeftCommand { get; set; }
		public RelayCommand RightCommand { get; set; }

		public RelayCommand BrakeCommand { get; set; }
		public RelayCommand DriftCommand { get; set; }

		public RelayCommand GoBackCommand { get { return new RelayCommand(viewService.GoBack); } }

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device and service is loaded.
		/// </summary>
		/// <param name="parameter">Tuple of integers that represents the index of the device in the device list of the device manager, and the index of the service of that device</param>
		public void LoadParameter(Tuple<int, int> parameter) {
			var service = deviceManager.AllDevices[parameter.Item1].Services[parameter.Item2];
			Motor = service as BluetoothMotor;
			if (Motor == null) {
				viewService.NavigateTo<DeviceViewModel>();
				return;
			}
			speed = Motor.Power;
			Direction = Motor.Direction;
			standardServiceControl.Service = Motor;
			modifyInputFormatControl.Service = Motor;
			LastSelectedDirectionRight = Motor.Direction != MotorDirection.MotorDirectionLeft;
			GenerateMotorString();
		}



		public MotorViewModel(DeviceManager deviceManager, IViewService viewService, StandardServiceControlViewModel standardServiceControl, ModifyInputFormatControlViewModel modifyInputFormatControl) {
			this.deviceManager = deviceManager;
			this.viewService = viewService;
			this.standardServiceControl = standardServiceControl;
			this.modifyInputFormatControl = modifyInputFormatControl;
		
			AccelerateCommand = new RelayCommand(Accelerate);
			DecelerateCommand = new RelayCommand(Decelerate);

			LeftCommand = new RelayCommand(Left);
			RightCommand = new RelayCommand(Right);

			BrakeCommand = new RelayCommand(Brake);
			DriftCommand = new RelayCommand(Drift);
		}

		// Methods for controling the Motor
		void Right() {
			Direction = MotorDirection.MotorDirectionRight;
			LastSelectedDirectionRight = true;
			RunMotor();
		}

		void Left() {
			Direction = MotorDirection.MotorDirectionLeft;
			LastSelectedDirectionRight = false;
			RunMotor();
		}

		void Accelerate() {
			speed++;
			if (Direction == MotorDirection.MotorDirectionBraking || Direction == MotorDirection.MotorDirectionDrifting) {
				Direction = LastSelectedDirectionRight ? MotorDirection.MotorDirectionRight : MotorDirection.MotorDirectionLeft;
			}
			RunMotor();
		}
	
		void Decelerate() {
			speed--;
			if(speed <= 0)
				Direction = MotorDirection.MotorDirectionDrifting;
			RunMotor();
		}

		void Brake() {
			Direction = MotorDirection.MotorDirectionBraking;
			RunMotor();
		}

		void Drift() {
			Direction = MotorDirection.MotorDirectionDrifting;
			RunMotor();
		}

		/// <summary>
		/// Run Motor
		/// </summary>
		void RunMotor() {
			Motor.RunInDirection(Direction, speed);
			speed = Motor.Power;
			Direction = Motor.Direction;
			GenerateMotorString();
		}

		/// <summary>
		/// Generates the string shown in the View
		/// </summary>
		void GenerateMotorString() {
			switch (Direction) {
				case MotorDirection.MotorDirectionDrifting:
					MotorString = "Drifting";
					break;
				case MotorDirection.MotorDirectionLeft:
					MotorString = "Motor left " + speed;
					break;
				case MotorDirection.MotorDirectionRight:
					MotorString = "Motor right " + speed;
					break;
				case MotorDirection.MotorDirectionBraking:
					MotorString = "Braking";
					break;
				default:
					throw new ArgumentOutOfRangeException();
			}
		}


	}
}
