﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GalaSoft.MvvmLight.Command;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// The view for the Piezo tone player service
	/// </summary>
	public class PiezoTonePlayerViewModel : ExtendedViewModelBase, ILocation<Tuple<int, int>> {
		readonly DeviceManager deviceManager;
		readonly IViewService viewService;
		readonly StandardServiceControlViewModel standardServiceControl;
		readonly ModifyInputFormatControlViewModel modifyInputFormatControl;


		BluetoothPiezoTonePlayer piezoTonePlayer;
		/// <summary>
		/// The Piezo tone player
		/// </summary>
		public BluetoothPiezoTonePlayer PiezoTonePlayer {
			get { return piezoTonePlayer; }
			private set { Set(ref piezoTonePlayer, value); }
		}

		int frequency;
		/// <summary>
		/// The frequency to play
		/// </summary>
		public int Frequency {
			get { return frequency; }
			set {
				Set(ref frequency, value);
			}
		}

		string frequencyString;
		/// <summary>
		/// Frequency as string shown in view
		/// </summary>
		public string FrequencyString {
			get { return frequencyString; }
			set {
				Set(ref frequencyString, value);
				Frequency = Convert.ToInt32(value);
			}
		}

		int duration;
		/// <summary>
		/// The duration of the tone to be played
		/// </summary>
		public int Duration {
			get { return duration; }
			set {
				Set(ref duration, value);
			}
		}

		string durationString;
		/// <summary>
		/// The duration as string shown in the View
		/// </summary>
		public string DurationString {
			get { return durationString; }
			set {
				Set(ref durationString, value);
				Duration = Convert.ToInt32(value);
			}
		}

		bool playSequence = false;

		public RelayCommand GoBackCommand { get { return new RelayCommand(viewService.GoBack); } }

		// commands for starting and stop playing sounds
		public RelayCommand StartPlayingCommand{ get { return new RelayCommand(() => PiezoTonePlayer.PlayFrequency(Frequency, Duration)); } }
		public RelayCommand StartPlayingSequenceCommand{ get { return new RelayCommand(() => {
			playSequence = true;
			PlaySequence();
		}); } }
		public RelayCommand StopPlayingCommand{ get { return new RelayCommand(() => {
			playSequence = false;
			PiezoTonePlayer.StopPlaying();
		}); } }

		public RelayCommand<string> PlayToneCommand { get { return new RelayCommand<string>(i => PiezoTonePlayer.PlayNote((PiezoTonePlayerNote)Convert.ToInt32(i), 4, 10000)); } }

		public RelayCommand<string> OnPlayTonePressedCommand {
			get {
				return new RelayCommand<string>(i => PiezoTonePlayer.PlayNote((PiezoTonePlayerNote)Convert.ToInt32(i), 4, 1000));
			}
		}

		public RelayCommand<string> OnPlayToneReleasedCommand {
			get {
				return new RelayCommand<string>(i => PiezoTonePlayer.StopPlaying());
			}
		}

		/// <summary>
		/// Plays a sequence of sounds. starts at 10Hz and increases by <see cref="Frequency"/> every <see cref="Duration"/>
		/// </summary>
		async void PlaySequence() {
			var freq = 10;
			while (playSequence) {
				freq = (freq + Frequency) % BluetoothPiezoTonePlayer.PiezoToneMaxFrequency;
				PiezoTonePlayer.PlayFrequency(freq, 0);
				await Task.Delay(Duration);
			}
		}

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device and service is loaded.
		/// </summary>
		/// <param name="parameter">Tuple of integers that represents the index of the device in the device list of the device manager, and the index of the service of that device</param>
		public void LoadParameter(Tuple<int, int> parameter) {
			var service = deviceManager.AllDevices[parameter.Item1].Services[parameter.Item2];
			PiezoTonePlayer = service as BluetoothPiezoTonePlayer;
			if (PiezoTonePlayer == null) {
				viewService.NavigateTo<DeviceViewModel>();
				return;
			}
			standardServiceControl.Service = PiezoTonePlayer;
			modifyInputFormatControl.Service = PiezoTonePlayer;
		}

		public PiezoTonePlayerViewModel(DeviceManager deviceManager, IViewService viewService, StandardServiceControlViewModel standardServiceControl, ModifyInputFormatControlViewModel modifyInputFormatControl) {
			this.deviceManager = deviceManager;
			this.viewService = viewService;
			this.standardServiceControl = standardServiceControl;
			this.modifyInputFormatControl = modifyInputFormatControl;

			FrequencyString = "440";
			DurationString = "100";
		}


	}

}
