﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Media;
using GalaSoft.MvvmLight.Command;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using GalaSoft.MvvmLight.Threading;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// The view for the RGB light service
	/// </summary>
	public class RgbLightViewModel : ExtendedViewModelBase, ILocation<Tuple<int, int>>, IRgbLightDelegate {
		readonly DeviceManager deviceManager;
		readonly IViewService viewService;
		readonly StandardServiceControlViewModel standardServiceControl;
		readonly ModifyInputFormatControlViewModel modifyInputFormatControl;

		BluetoothRgbLight rgbLight;
		/// <summary>
		/// The RGB light that is displayed
		/// </summary>
		public BluetoothRgbLight RgbLight {
			get { return rgbLight; }
			private set { Set(ref rgbLight, value); }
		}

		/// <summary>
		/// Command to go back to the Device hub
		/// </summary>
		public RelayCommand GoBackCommand { get { return new RelayCommand(viewService.GoBack); } }

		string serviceName = "RGB Light";
		/// <summary>
		/// The service name
		/// </summary>
		public string ServiceName {
			get { return serviceName; }
			private set { Set(ref serviceName, value); }
		}

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device and service is loaded.
		/// </summary>
		/// <param name="parameter">Tuple of integers that represents the index of the device in the device list of the device manager, and the index of the service of that device</param>
		public void LoadParameter(Tuple<int, int> parameter) {
			var service = deviceManager.AllDevices[parameter.Item1].Services[parameter.Item2];
			RgbLight = service as BluetoothRgbLight;
			if (RgbLight == null) {
				viewService.NavigateTo<DeviceViewModel>();
				return;
			}
			ServiceName = RgbLight.ConnectInfo.TypeString;
			standardServiceControl.Service = RgbLight;
			modifyInputFormatControl.Service = RgbLight;
            colorIndex = RgbLight.ColorIndex;
            RgbLight.AddDelegate(this);
            UpdateModeUI();
		}

		SolidColorBrush colorBrush = new SolidColorBrush(Colors.Black);
		/// <summary>
		/// This Color is used to display the color the sensor should display
		/// </summary>
		public SolidColorBrush ColorBrush {
			get { return colorBrush; }
			set {
				Set(ref colorBrush, value);
			}
		}

		// all properties and slider colors for the color Red
		#region Red
		public LinearGradientBrush RedSliderBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Colors.Red, Offset = 1 } }, 0); }
		}
		public LinearGradientBrush RedSliderForegroundBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Color.FromArgb(255, (byte)red, 0, 0), Offset = 1 } }, 0); }
		}

		int red;
		public int Red {
			get { return red; }
			set {
				var changed = Set(ref red, value);
				if (changed) {
					var color = Color.FromArgb(255, (byte)value, (byte)green, (byte)blue);
					RgbLight.Color = color;
					ColorBrush = new SolidColorBrush(color);
					RaisePropertyChanged("RedSliderForegroundBrush");
				}
			}
		}
		#endregion

		// all properties and slider colors for the color Green
		#region Green
		public LinearGradientBrush GreenSliderBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Color.FromArgb(255, 0, 255, 0), Offset = 1 } }, 0); }
		}
		public LinearGradientBrush GreenSliderForegroundBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Color.FromArgb(255, 0, (byte)green, 0), Offset = 1 } }, 0); }
		}
		int green;
		public int Green {
			get { return green; }
			set {
				var changed = Set(ref green, value);
				if (changed) {
					var color = Color.FromArgb(255, (byte)red, (byte)value, (byte)blue);
					RgbLight.Color = color;
					ColorBrush = new SolidColorBrush(color);
					RaisePropertyChanged("GreenSliderForegroundBrush");
				}
			}
		}
		#endregion

		// all properties and slider colors for the color Blue
		#region Blue
		public LinearGradientBrush BlueSliderBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Colors.Blue, Offset = 1 } }, 0); }
		}
		public LinearGradientBrush BlueSliderForegroundBrush {
			get { return new LinearGradientBrush(new GradientStopCollection() { new GradientStop() { Color = Colors.Black, Offset = 0 }, new GradientStop() { Color = Color.FromArgb(255, 0, 0, (byte)blue), Offset = 1 } }, 0); }
		}
		int blue;
		public static readonly DependencyProperty SwitchOffCommandProperty = DependencyProperty.Register("SwitchOffCommand", typeof (RelayCommand), typeof (RgbLightViewModel), new PropertyMetadata(default(RelayCommand)));

		public int Blue {
			get { return blue; }
			set {
				var changed = Set(ref blue, value);
				if (changed) {
					var color = Color.FromArgb(255, (byte)red, (byte)green, (byte)value);
					RgbLight.Color = color;
					ColorBrush = new SolidColorBrush(color);
					RaisePropertyChanged("BlueSliderForegroundBrush");
				}
			}
		}
		#endregion

        #region Color Index
        int colorIndex;
        public int ColorIndex {
            get { return colorIndex; }
            set {
                bool changed = Set(ref colorIndex, value);
                if (changed) {
                    RgbLight.ColorIndex = value;
                }
            }
        }
        #endregion

        #region RGB Light Mode
        bool discreteMode;
        public bool InDiscreteMode {
            get { return discreteMode; }
            set { Set(ref discreteMode, value); }
        }

        bool absoluteMode;
        public bool InAbsoluteMode {
            get { return absoluteMode; }
            set { Set(ref absoluteMode, value); }
        }
        #endregion

        /// <summary>
		/// Switches of the Light
		/// </summary>
		public RelayCommand SwitchOffCommand {
			get { return new RelayCommand(() => {
				RgbLight.SwitchOff();
				Red = 0;
				Green = 0;
				Blue = 0;
			}); }
		}

        /// <summary>
        /// Switches the Light color to the default
        /// </summary>
        public RelayCommand SwitchToDefaultCommand {
            get {
                return new RelayCommand(() => {
                    RgbLight.SwitchToDefaultColor();
                    Red = 0;
                    Green = 0;
                    Blue = 0;
                });
            }
        }

        /// <summary>
        /// Decreases the RGB Light's color index
        /// </summary>
        public RelayCommand ColorIndexDecrease {
            get {
                return new RelayCommand(() => {
                    if (ColorIndex > 0) {
                        ColorIndex -= 1;
                    }
                });
            }
        }

        /// <summary>
        /// Increases the RGB Light's color index
        /// </summary>
        public RelayCommand ColorIndexIncrease {
            get {
                return new RelayCommand(() => {
                    ColorIndex += 1;
                });
            }
        }

		public RgbLightViewModel(DeviceManager deviceManager, IViewService viewService, StandardServiceControlViewModel standardServiceControl, ModifyInputFormatControlViewModel modifyInputFormatControl) {
			this.deviceManager = deviceManager;
			this.viewService = viewService;
			this.standardServiceControl = standardServiceControl;
			this.modifyInputFormatControl = modifyInputFormatControl;
		
		}

        private void UpdateModeUI() {
            DispatcherHelper.CheckBeginInvokeOnUI(() => {
                InAbsoluteMode = RgbLight.RGBMode == RgbLightMode.Absolute;
                InDiscreteMode = RgbLight.RGBMode == RgbLightMode.Discrete;
            });
        }

        public void DidUpdateColor(IRgbLight rgbLight, Color oldColor, Color newColor) {
            // Do nothing
        }

        public void DidUpdateColorIndex(IRgbLight rgbLight, int oldColorIndex, int newColorIndex) {
            // Do nothing
        }

        public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
            UpdateModeUI();
        }

        public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
            // Do nothing
        }
    }
}
