﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Threading;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// The view for the Tilt sensor service
	/// </summary>
	public class TiltSensorViewModel : ExtendedViewModelBase, ILocation<Tuple<int, int>>, ITiltSensorDelegate {
		readonly DeviceManager deviceManager;
		readonly IViewService viewService;
		readonly StandardServiceControlViewModel standardServiceControl;
		readonly ModifyInputFormatControlViewModel modifyInputFormatControl;


		BluetoothTiltSensor tiltSensor;
		/// <summary>
		/// The Tilt sensor to display
		/// </summary>
		public BluetoothTiltSensor TiltSensor {
			get { return tiltSensor; }
			private set {
				Set(ref tiltSensor, value);
				tiltSensor.AddDelegate(this);
				DidUpdateInputFormat(value, null, value.InputFormat);
			}
		}

		bool showAngle;
		/// <summary>
		/// Boolean if the angle should be shown
		/// </summary>
		public bool ShowAngle {
			get { return showAngle; }
			set { Set(ref showAngle, value); }
		}

		TiltSensorAngle angle;
		/// <summary>
		/// The angle of the Tilt sensor
		/// </summary>
		public TiltSensorAngle Angle {
			get { return angle; }
			set {
				Set(ref angle, value);
				AngleX = value.X;
				AngleY = value.Y;
			}
		}
		float angleX;
		/// <summary>
		/// The x value of the Angle
		/// </summary>
		public float AngleX {
			get { return angleX; }
			set { Set(ref angleX, value); }
		}
		float angleY;
		/// <summary>
		///  The y value of the Angle
		/// </summary>
		public float AngleY {
			get { return angleY; }
			set { Set(ref angleY, value); }
		}

		bool showCrash;
		/// <summary>
		/// Boolean if the crash values should be shown
		/// </summary>
		public bool ShowCrash {
			get { return showCrash; }
			set { Set(ref showCrash, value); }
		}
		TiltSensorCrash crash;
		/// <summary>
		/// The crash values
		/// </summary>
		public TiltSensorCrash Crash {
			get { return crash; }
			set {
				Set(ref crash, value);
				CrashX = value.X;
				CrashY = value.Y;
				CrashZ = value.Z;
			}
		}
		float crashX;
		/// <summary>
		/// The x value of the Crash
		/// </summary>
		public float CrashX {
			get { return crashX; }
			set { Set(ref crashX, value); }
		}
		float crashY;
		/// <summary>
		/// The y value of the Crash
		/// </summary>
		public float CrashY {
			get { return crashY; }
			set { Set(ref crashY, value); }
		}
		float crashZ;
		/// <summary>
		/// The z value of the Crash
		/// </summary>
		public float CrashZ {
			get { return crashZ; }
			set { Set(ref crashZ, value); }
		}

		bool showDirection;
		/// <summary>
		/// Boolean if the direction should be shown
		/// </summary>
		public bool ShowDirection {
			get { return showDirection; }
			set { Set(ref showDirection, value); }
		}
		TiltSensorDirection direction;
		/// <summary>
		/// The direction of the sensor
		/// </summary>
		public TiltSensorDirection Direction {
			get { return direction; }
			set { Set(ref direction, value); }
		}

		/// <summary>
		/// The mode the Tilt sensor is in
		/// </summary>
		TiltSensorMode Mode = TiltSensorMode.TiltSensorModeUnknown;

		/// <summary>
		/// Command that navigates to the Device view
		/// </summary>
		public RelayCommand GoBackCommand { get { return new RelayCommand(viewService.GoBack); } }

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device and service is loaded.
		/// </summary>
		/// <param name="parameter">Tuple of integers that represents the index of the device in the device list of the device manager, and the index of the service of that device</param>
		public void LoadParameter(Tuple<int, int> parameter) {
			var service = deviceManager.AllDevices[parameter.Item1].Services[parameter.Item2];
			TiltSensor = service as BluetoothTiltSensor;
			if (TiltSensor == null) {
				viewService.NavigateTo<DeviceViewModel>();
				return;
			}
			standardServiceControl.Service = service;
			modifyInputFormatControl.Service = service;
		}

		public TiltSensorViewModel(DeviceManager deviceManager, IViewService viewService, StandardServiceControlViewModel standardServiceControl, ModifyInputFormatControlViewModel modifyInputFormatControl) {
			this.deviceManager = deviceManager;
			this.viewService = viewService;
			this.standardServiceControl = standardServiceControl;
			this.modifyInputFormatControl = modifyInputFormatControl;
		
		}

		/// <summary>
		/// Callback when the input format changes. Updates the mode to show
		/// </summary>
		/// <param name="service">The service with the changed input format</param>
		/// <param name="oldFormat">The old input format</param>
		/// <param name="newFormat">The new input format</param>

		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
			Mode = (TiltSensorMode)newFormat.Mode;
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				ShowAngle = Mode == TiltSensorMode.TiltSensorModeAngle;
				ShowCrash = Mode == TiltSensorMode.LeTiltSensorModeCrash;
				ShowDirection = Mode == TiltSensorMode.TiltSensorModeTilt;
			});
		}

		/// <summary>
		/// Callback when the value changes. Not used since the callback from the Tilt sensor is used and not the generic service underneath
		/// </summary>
		/// <param name="service">The service with the changed value</param>
		/// <param name="oldData">The old value</param>
		/// <param name="newData">The new value</param>
		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {

		}

		/// <summary>
		/// Callback when angle is updated
		/// </summary>
		/// <param name="tiltSensor">The Tilt sensor that has an updated angle value</param>
		/// <param name="oldAngle">The old angle</param>
		/// <param name="newAngle">The new angle</param>
		public void DidUpdateAngle(ITiltSensor tiltSensor, TiltSensorAngle oldAngle, TiltSensorAngle newAngle) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				Angle = newAngle;
			});
		}

		/// <summary>
		/// Callback when crash is updated
		/// </summary>
		/// <param name="tiltSensor">The Tilt sensor that has an updated crash value</param>
		/// <param name="oldAngle">The old crash</param>
		/// <param name="newAngle">The new crash</param>
        public void DidUpdateCrash(ITiltSensor tiltSensor, TiltSensorCrash oldCrashValue, TiltSensorCrash newCrashValue) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				Crash = newCrashValue;
			});
		}

		/// <summary>
		/// Callback when direction is updated
		/// </summary>
		/// <param name="tiltSensor">The Tilt sensor that has an updated direction</param>
		/// <param name="oldAngle">The old direction</param>
		/// <param name="newAngle">The new direction</param>
        public void DidUpdateDirection(ITiltSensor tiltSensor, TiltSensorDirection oldDirection, TiltSensorDirection newDirection) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				Direction = newDirection;
			});
		}

	}
}