﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection.Context;
using System.Text;
using System.Threading.Tasks;
using Windows.System.Threading;
using Windows.UI.Popups;
using GalaSoft.MvvmLight.Command;
using GalaSoft.MvvmLight.Threading;
using LegoBleTestApp.Services;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoBleTestApp.ViewModel {
	/// <summary>
	/// View that performs the Windmill Test
	/// </summary>
	public class WindMillTestViewModel : ExtendedViewModelBase, ILocation<int>, IMotionSensorDelegate {
		readonly DeviceManager deviceManager;
		readonly IViewService viewService;

		/// <summary>
		/// Default high trigger
		/// </summary>
		const int DefaultMotionSensorTriggerHigh = 8;
		/// <summary>
		/// Default low trigger
		/// </summary>
		const int DefaultMotionSensorTriggerLow = 7;

		int speed = 0;
		/// <summary>
		/// The speed to perform the windmill test with
		/// </summary>
		public int Speed {
			get { return speed; }
			set { Set(ref speed, value); }
		}

		float currentValue = 0;
		/// <summary>
		/// The current value of the motion sensor
		/// </summary>
		public float CurrentValue {
			get { return currentValue; }
			set { Set(ref currentValue, value); }
		}

		/// <summary>
		/// The direction of the Motor
		/// </summary>
		MotorDirection direction;

		IDevice device;
		BluetoothMotor motor;
		/// <summary>
		/// The motor
		/// </summary>
		public BluetoothMotor Motor {
			get { return motor; }
			set { Set(ref motor, value); }
		}
		BluetoothMotionSensor sensor;
		/// <summary>
		/// The motion sensor
		/// </summary>
		public BluetoothMotionSensor Sensor {
			get { return sensor; }
			set { Set(ref sensor, value); }
		}
        BluetoothRgbLight rgbLight;
        /// <summary>
        /// The RGB light
        /// </summary>
        public BluetoothRgbLight RgbLight {
            get { return rgbLight; }
            set { Set(ref rgbLight, value); }
        }

		/// <summary>
		/// Command that returns to the device view
		/// </summary>
		public RelayCommand GoBackCommand { get { return new RelayCommand(() => viewService.NavigateTo<DeviceViewModel, int>(deviceManager.AllDevices.IndexOf(device))); } }
		/// <summary>
		/// Command that sets the motor to run at full speed
		/// </summary>
		public RelayCommand FullSpeedCommand { get {
			return new RelayCommand(() => {
				direction = MotorDirection.MotorDirectionRight;
				Speed = 100;
				RunMotor();
			});
			}
		}
		/// <summary>
		/// Command that stops the motor
		/// </summary>
		public RelayCommand StopCommand { get { return new RelayCommand(() => {
			direction = MotorDirection.MotorDirectionBraking;
			Speed = 0;
			RunMotor();
		});} }
		/// <summary>
		/// Command that resets the counter of the motion sensor
		/// </summary>
		public RelayCommand ResetCommand {
			get {
				return new RelayCommand(() => {
					Sensor.SendResetStateRequest();
					ShiftCount = 0;
				});
			}
		}

		// Commands that handle the button clicks from the GUI
		public RelayCommand AccelerateCommand { get { return new RelayCommand(Accelerate); } }
		public RelayCommand DecelerateCommand { get { return new RelayCommand(Decelerate); } }

		public RelayCommand TriggerHighUpCommand { get { return new RelayCommand(() => TriggerHigh = Math.Min(TriggerHigh+1, 10)); } }
		public RelayCommand TriggerHighDownCommand { get { return new RelayCommand(() => TriggerHigh = Math.Max(TriggerHigh-1, 0)); } }

		public RelayCommand TriggerLowUpCommand { get { return new RelayCommand(() => TriggerLow = Math.Min(TriggerLow+1, 10)); } }
		public RelayCommand TriggerLowDownCommand { get { return new RelayCommand(() => TriggerLow = Math.Max(TriggerLow - 1, 0)); } }

		int shiftCount;
		/// <summary>
		/// How many times has the windmill passed the motion sensor
		/// </summary>
		public int ShiftCount {
			get { return shiftCount; }
			set { Set(ref shiftCount, value); }
		}

		int triggerHigh;
		/// <summary>
		/// The high trigger value
		/// </summary>
		public int TriggerHigh {
			get { return triggerHigh; }
			set { Set(ref triggerHigh, value); }
		}

		bool lowTriggered = false;
		int triggerLow;
		/// <summary>
		/// The low trigger value
		/// </summary>
		public int TriggerLow {
			get { return triggerLow; }
			set { Set(ref triggerLow, value); }
		}

		bool shiftDirectionEnabled;
		/// <summary>
		/// if true the direction of the motor is reversed when the motion sensor is triggerd
		/// </summary>
		public bool ShiftDirectionEnabled {
			get { return shiftDirectionEnabled; }
			set { Set(ref shiftDirectionEnabled, value); }
		}

		/// <summary>
		/// LoadParameter is used in navigating between views. Here the correct device is loaded.
		/// </summary>
		/// <param name="parameter">integer that represents the index of the device in the device list of the device manager</param>
		public void LoadParameter(int parameter) {
			device = deviceManager.AllDevices[parameter];
			if (device == null) {
				viewService.NavigateTo<DeviceViewModel, int>(parameter);
				return;
			}
			// search for the motor and motion sensor on the device
			foreach (var service in device.Services) {
                if (service.ConnectInfo.TypeEnum == IoType.IoTypeMotor)
                    Motor = service as BluetoothMotor;
                else if (service.ConnectInfo.TypeEnum == IoType.IoTypeMotionSensor)
                    Sensor = service as BluetoothMotionSensor;
                else if (service.ConnectInfo.TypeEnum == IoType.IoTypeRgbLight)
                    RgbLight = service as BluetoothRgbLight;
			}
			// if motor or sensor == null return to device hub.
			if (Motor == null || Sensor == null) {
				DispatcherHelper.CheckBeginInvokeOnUI(async () => {
					var missingString = Motor == null ? " Motor is missing." : "";
					missingString += Sensor == null ? " Motion sensor is missing." : "";
					var msgDialog = new MessageDialog("The device " + device.Name + " does not have the required sensors to perform the Windmill Test." + missingString);
					await msgDialog.ShowAsync();
					viewService.NavigateTo<DeviceViewModel, int>(parameter);
				});
				return;
			}

			Sensor.AddDelegate(this);
	}


		public WindMillTestViewModel(DeviceManager deviceManager, IViewService viewService) {
			this.deviceManager = deviceManager;
			this.viewService = viewService;

			TriggerHigh = DefaultMotionSensorTriggerHigh;
			TriggerLow = DefaultMotionSensorTriggerLow;
		}

		/// <summary>
		/// Accelerate the motor
		/// </summary>
		void Accelerate() {
			Speed++;
			if(direction == MotorDirection.MotorDirectionBraking || direction == MotorDirection.MotorDirectionDrifting)
				direction = MotorDirection.MotorDirectionRight;

			RunMotor();
		}

		/// <summary>
		/// Decelerate the motor
		/// </summary>
		void Decelerate() {
			Speed--;
			if (Speed <= 0)
				direction = MotorDirection.MotorDirectionDrifting;
			RunMotor();
		}

		/// <summary>
		/// Run the motor
		/// </summary>
		void RunMotor() {
			Motor.RunInDirection(direction, speed);
			Speed = Motor.Power;
			direction = Motor.Direction;
		}

		// not used
		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {}
		// not used
		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {}

		/// <summary>
		/// Callback when the distance of the motion sensor is updated
		/// </summary>
		/// <param name="motionSensor">The motion sensor with the updated value</param>
		/// <param name="oldDistance">The old distance</param>
		/// <param name="newDistance">The new distance</param>
		public void DidUpdateDistance(IMotionSensor motionSensor, float oldDistance, float newDistance) {
			HandleNewValue((int)newDistance);
		}

		/// <summary>
		/// Handle the new updated motion sensor distance value
		/// </summary>
		/// <param name="newValue">The new distance reported back from the motion sensor</param>
		void HandleNewValue(int newValue) {
			UpdateMotionSensorValue(newValue);
			// if the value is below or equal to the low trigger value, low is triggered
			if (newValue <= TriggerLow) {
				lowTriggered = true;
				return;
			}
			// if low is not triggered ignore rest
			if (!lowTriggered)
				return;
			// if the new value is above the high trigger, trigger a shift count and reset lowTriggered
			if (newValue >= TriggerHigh) {
				TriggerShiftCount();
				lowTriggered = false;
			}
			
		}

		/// <summary>
		/// Updates the new motion sensor value
		/// </summary>
		/// <param name="value"></param>
		void UpdateMotionSensorValue(int value) {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				CurrentValue = value;
			});

		}

		/// <summary>
		/// Is called when the motion sensor detects that the wing of the windmill passes
		/// </summary>
		void TriggerShiftCount() {
			// if we want to change motor direction
			if (ShiftDirectionEnabled) {
				// switch on the direction and if it is left or right, reverse direction and run motor
				switch (direction) {
					case MotorDirection.MotorDirectionDrifting:
						break;
					case MotorDirection.MotorDirectionLeft:
						direction = MotorDirection.MotorDirectionRight;
						RunMotor();
                        RgbLight.ColorIndex = 2;
						break;
					case MotorDirection.MotorDirectionRight:
						direction = MotorDirection.MotorDirectionLeft;
						RunMotor();
                        RgbLight.ColorIndex = 6;
						break;
					case MotorDirection.MotorDirectionBraking:
						break;
					default:
						throw new ArgumentOutOfRangeException();
				}
			}
			// increases the shift count value
			IncrementShiftCountValue();
		}

		/// <summary>
		/// Increments the shift count value when the motion sensor is triggered
		/// </summary>
		void IncrementShiftCountValue() {
			DispatcherHelper.CheckBeginInvokeOnUI(() => {
				ShiftCount = ShiftCount + 1;
			});

		}

		// not used
		public void DidUpdateCount(IMotionSensor motionSensor, uint newCount) {}

	}
}
