﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Helpers;
using LegoDeviceSDK.Wrapper;

namespace LegoDeviceSDK.Bluetooth.Definitions {
	internal class BluetoothServiceDefinition {
		const string Tag = "BluetoothServiceDefinition";
		internal string ServiceName { get; set; }
		internal Guid ServiceUuid { get; set; }
		internal List<BluetoothCharacteristicDefinition> CharacteristicDefinitions { get; set; }
		internal List<Guid> CharacteristicUuids { get; set; }
		internal IGattDeviceService GattDeviceService { get; private set; }

		static IoServiceDefinition ioServiceDefinition;
		internal static IoServiceDefinition IoServiceDefinition { get { return ioServiceDefinition; } }
		static DeviceServiceDefinition deviceServiceDefinition;
		internal static DeviceServiceDefinition DeviceServiceDefinition { get { return deviceServiceDefinition; } }
		static DeviceInfoServiceDefinition deviceInfoServiceDefinition;
		internal static DeviceInfoServiceDefinition DeviceInfoServiceDefinition { get { return deviceInfoServiceDefinition; } }
		static BluetoothBatteryServiceDefinition batteryServiceDefinition;
		internal static BluetoothBatteryServiceDefinition BatteryServiceDefinition { get { return batteryServiceDefinition; } }

		/// <summary>
		/// Finds all services and their characteristics
		/// </summary>
		/// <param name="standartDeviceInformation">The standart device information used to find all other services and characteristics</param>
		/// <returns></returns>
		internal static async Task LoadDefinitions(IDeviceInformation standartDeviceInformation) {
			// find all Deviceinformation instances for the services
			var ioServices = await GattConfiguration.DeviceInformationWrapper.FindAllAsync(
				GattConfiguration.GattDeviceServiceWrapper.GetDeviceSelectorFromUuid(GuidHelper.GetFullGuid(IoServiceDefinition.InputServiceUuid)),
				new string[] { "System.Devices.ContainerId" });
			var deviceInfoServices = await GattConfiguration.DeviceInformationWrapper.FindAllAsync(
			  GattConfiguration.GattDeviceServiceWrapper.GetDeviceSelectorFromUuid(Windows.Devices.Bluetooth.GenericAttributeProfile.GattDeviceService.ConvertShortIdToUuid(DeviceInfoServiceDefinition.DeviceInfoServiceUuid)),
				new string[] {"System.Devices.ContainerId"});
			var batteryServices = await GattConfiguration.DeviceInformationWrapper.FindAllAsync(
			  GattConfiguration.GattDeviceServiceWrapper.GetDeviceSelectorFromUuid(Windows.Devices.Bluetooth.GenericAttributeProfile.GattDeviceService.ConvertShortIdToUuid(BluetoothBatteryServiceDefinition.BatteryServiceUuid)),
				new string[] {"System.Devices.ContainerId"});
			var deviceId = ExtractRealDeviceId(standartDeviceInformation.Id);

			// find all services for this device and create the corresponding DeviceServiceDefinitions
			var standartGattDeviceService = await GattConfiguration.GattDeviceServiceWrapper.FromIdAsync(standartDeviceInformation.Id);
			deviceServiceDefinition = new DeviceServiceDefinition(standartGattDeviceService);

			var ioDeviceInformation = ioServices.Find(information => ExtractRealDeviceId(information.Id) == deviceId);
			var deviceInfoDeviceInformation = deviceInfoServices.Find(information => ExtractRealDeviceId(information.Id) == deviceId);
			var batteryDeviceInformation = batteryServices.Find(information => ExtractRealDeviceId(information.Id) == deviceId);
			deviceServiceDefinition = new DeviceServiceDefinition(standartGattDeviceService);

			if (ioDeviceInformation != null) {
				var ioService = await GattConfiguration.GattDeviceServiceWrapper.FromIdAsync(ioDeviceInformation.Id);
				ioServiceDefinition = new IoServiceDefinition(ioService);
			}
			if (deviceInfoDeviceInformation != null) {
				var deviceInfoService = await GattConfiguration.GattDeviceServiceWrapper.FromIdAsync(deviceInfoDeviceInformation.Id);
				deviceInfoServiceDefinition = new DeviceInfoServiceDefinition(deviceInfoService);
			}
			if (batteryDeviceInformation != null) {
				var batteryService = await GattConfiguration.GattDeviceServiceWrapper.FromIdAsync(batteryDeviceInformation.Id);
				batteryServiceDefinition = new BluetoothBatteryServiceDefinition(batteryService);
			}
		}

		internal BluetoothServiceDefinition(IGattDeviceService gattDeviceService) {
			this.GattDeviceService = gattDeviceService;
			CharacteristicDefinitions = new List<BluetoothCharacteristicDefinition>();
			CharacteristicUuids = new List<Guid>();
		}

		/// <summary>
		/// Finds Characteristicts with the specified short uuid. This method works with Lego specific services
		/// </summary>
		/// <param name="uuid">the short uuid to search for</param>
		/// <returns>The BluetoothCharacteristic belonging to the short uuid</returns>
		internal BluetoothCharacteristicDefinition CharacteristicWithUuid(string uuid) {
			var longUuid = GuidHelper.GetFullGuid(uuid);
			return CharacteristicWithUuid(longUuid);
		}

		/// <summary>
		/// Finds Characteristicts with the specified short uuid. This method works with generic services
		/// </summary>
		/// <param name="uuid">The short uuid to search for</param>
		/// <returns>The BluetoothCharacteristic belonging to the short uuid</returns>
		internal BluetoothCharacteristicDefinition CharacteristicWithUuid(ushort uuid) {
			var longUuid = Windows.Devices.Bluetooth.GenericAttributeProfile.GattDeviceService.ConvertShortIdToUuid(uuid);
			return CharacteristicWithUuid(longUuid);
		}

		/// <summary>
		/// Finds Characteristicts with the specified long uuid.
		/// </summary>
		/// <param name="uuid">The uuid specifying the exact characteristic</param>
		/// <returns>The BluetoothCharacteristic belonging to the uuid</returns>
		internal BluetoothCharacteristicDefinition CharacteristicWithUuid(Guid uuid) {
			var characteristic = CharacteristicDefinitions.Find(definition => definition.Uuid.Equals(uuid));
			if(characteristic == null)
				SdkLogger.E(Tag, string.Format("Could not find Characteristic with Uuid: {0}", uuid));
			return characteristic;
		}

		private static string ExtractRealDeviceId(string longDeviceId) {
			var tokens = longDeviceId.Split(new string[] { "#" }, StringSplitOptions.None);
			if (tokens.Length != 4)
				return "";
			var tokens2 = tokens[1].Split(new string[] { "_" }, StringSplitOptions.None);
			if (tokens2.Length == 2)
				return tokens2[1];

			return "";
		}

		public override string ToString() {
			return String.Format("<{0}: ", GetType().ToString()) +
							String.Format("< {0}", ServiceName) +
							String.Format("({0})", ServiceUuid) +
							String.Format(", self.characteristicDefinitions={0}", CharacteristicDefinitions) + ">";
		}


	}
}
