﻿using System;
using LegoDeviceSDK.Interfaces;

namespace LegoDeviceSDK.Generic {
	/// <summary>
	/// Use this class to configure the log level that you want the LEGO Device SDK to use.
	/// You may provide your own log-writer that all log from the LEGO Device SDK will be written to, by setting the <see cref="LogWriter"/> property.
	/// </summary>
	public class SdkLogger {
		private static volatile SdkLogger instance;
		private static object syncRoot = new Object();
		/// <summary>
		/// The shared SdkLogger
		/// </summary>
		public static SdkLogger DefaultLogger {
			get {
				if (instance == null) {
					lock (syncRoot) {
						if (instance == null)
							instance = new SdkLogger();
					}
				}
				return instance;
			}
		}

        /// <summary>
        /// Public constructor for unit tests
        /// </summary>
		public SdkLogger() {
			LogLevel = LoggerLevel.LoggerLevelVerbose;
		}

		/// <summary>
		/// Convenience property for checking if the log level is set to debug (or below)
		/// </summary>
		public bool DebugLoggingEnabled { get { return LogLevel == LoggerLevel.LoggerLevelVerbose || LogLevel == LoggerLevel.LoggerLevelDebug; } }

		/// <summary>
		/// The log level. Default is Error.
		/// </summary>
		/// <remarks>For Windows Phone Store build you should set the log level to Error or Warn.</remarks>
		public LoggerLevel LogLevel { get; set; }

		ILogWriter logWriter;
		/// <summary>
		/// You may set a custom log writer that all log from the library will be written to. 
		/// </summary>
		// If set to null (the default) log will be from the library using the standard Debug.WriteLn
		public ILogWriter LogWriter {
			get {
				if (logWriter == null)
					logWriter = new DefaultConsoleLogWriter();
				return logWriter;
				
			}
			set {
				if(value == null)
					logWriter = new DefaultConsoleLogWriter();
				else
					logWriter = value;
			}
		}

		/// <summary>
		/// Write Debug message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
		public static void D(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelDebug)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelDebug);
		}

		/// <summary>
		/// Write Error message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
		public static void E(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelError)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelError);
		}

		/// <summary>
		/// Write Fatal message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
        public static void F(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelFatal)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelFatal);
		}

		/// <summary>
		/// Write Info message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
        public static void I(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelInfo)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelInfo);
		}

		/// <summary>
		/// Write Verbose message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
        public static void V(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelVerbose)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelVerbose);
		}

		/// <summary>
		/// Write Warning message
		/// </summary>
		/// <param name="tag">Class name where the message originated</param>
		/// <param name="message">The message to be logged</param>
        public static void W(string tag, string message) {
			if (DefaultLogger.LogLevel > LoggerLevel.LoggerLevelWarn)
				return;
			DefaultLogger.LogWriter.WriteMessage(FormatMessage(tag, message), LoggerLevel.LoggerLevelWarn);
		}

		static string FormatMessage(string tag, string message) {
			return string.Format("{0}: {1}", tag, message);
		}

	}
}
