The LEGO Device SDK is used communicate with various LEGO manufactured hardware such as sensors and motors.

Currently the library supports only communication with hardware connected through Bluetooth LE (BLE). The SDK will help
you scan for and connect to LEGO BLE devices. Once connected to a device the SDK will provide access to read and be notified about changes to sensor values,
and to send commands to various outputs (motors, piezo-tone, etc.)


Scanning for and Connecting to Advertising Devices
--------------------------------------------------

To scan for devices you use the DeviceManager as illustrated in the example below. Your class must implement the IDeviceManagerDelegate

	async void startScanningForDevices
	{
		DeviceManager.SharedInstance.AddDelegate(this);
	    await DeviceManager.SharedInstance.ScanAsync();
	}

	public void DeviceDidAppear(IDeviceManager deviceManager2, IDevice device) {
		// Do something with you device like putting it in a list
	}
		
When adding a delegate to the DeviceManager it will inform you whenever a new device did appear or disappear (devices are managed through the Bluetooth functionality of Windows. You have to pair a device with windows first before it is recognized in the SDK). After a device has been discovered you can connect to it.
Again the IDeviceManagerDelegate will notify you once the connection is established (or if the connection fails).
Before connecting to a device you should add a IDeviceDelegate to the device.
Through this delegate you will be notified on any services (motors, sensors, etc.) offered by the device.
Remember to unregister the listener again when no longer used!

	async void ConnectToDevice(IDevice device)
	{
		DeviceManager.SharedInstance.AddDelegate(this);
	    await DeviceManager.SharedInstance.ConnectToDeviceAsync(device);
	}

	public void DidFinishInterrogatingDevice(IDeviceManager deviceManager2, IDevice device) {
		// device is now connected.
	}

	public void DidAddService(IDevice device, IService service) {
		// the Service service was added to device.
	}

Read Values From a Known Input connected to the Device
-----------------------------------------------

Once connected to a BluetoothDevice, and having discovered one or more services you are ready to start read info from sensors and send commands to motors and other outputs.
For example, to start listening to updates from a tilt sensor:

	void StartUsingService(IService service)
	{
		var tiltSensor = service as BluetoothTiltSensor;
        if (tiltSensor != null) {
            tiltSensor.AddDelegate(this);
        }
	}

	public void DidUpdateAngle(BluetoothTiltSensor tiltSensor, TiltSensorAngle oldAngle, TiltSensorAngle newAngle) {
		// the new angle of the sensor is the newAngle variable
	}


Communicate with a Generic Input connected to the Device
---------------------------------------------------------

If a service is an instance of the class BluetoothGenericService this means that the IO type in unkown to the SDK.
As opposed to the known service types (tilt, motion, etc) a generic service does not have a predefined BluetoothService.DefaultInputFormat
and therefore an InputFormat is not automatically send to the Device. This means that per default no data is send from the
device to the SDK when the IO reads a new data value. To use a BluetoothGenericService to receive data updated from a sensor unknown to the SDK you
must create and send an InputFormat to the device. Find inspiration in one of the concrete service classes for this, like the BluetoothMotionSensor
or BluetoothVoltageSensor.

Also see the documentation of the BluetoothGenericService class for an example on how to configure and use an IO unkown to the SDK.


Getting Debug and Error log from the LEGO Device SDK
----------------------------------------------------

You should insert the following code snippet in e.g. the end of the OnLaunched method of your App.xaml.cs class:

	protected async override void OnLaunched(LaunchActivatedEventArgs e)

        ...

        if (BuildConfig.DEBUG) {
            SdkLogger.DefaultLogger.LogLevel = LoggerLevel.LoggerLevelDebug;
        } else {
            SdkLogger.DefaultLogger.LogLevel = LoggerLevel.LoggerLevelWarn;
        }

        ...
    }

It is also possible to redirect the Device log statement to another destination, for example an online logging
service or a (debug) text view in your app. You can do this by creating a class that implements the ILogWriter interface and assign it to SdkLogger.DefaultLogger.LogWriter.

public class MyLogWriter : ILogWriter {
	void WriteMessage(string message, LoggerLevel level) {
		Debug.WriteLine("{0}: {1}", level, message);
	}
}

// now assign on start the new LogWriter
SdkLogger.DefaultLogger.LogWriter = new MyLogWriter();
