﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Devices.Bluetooth;
using Windows.Devices.Bluetooth.GenericAttributeProfile;
using Windows.Foundation;
using Windows.Storage.Streams;
using Windows.UI.Core;

namespace LegoDeviceSDK.Wrapper {
	/// <summary>
	/// Internal wrapper for a GattCharacteristic
	/// </summary>
	internal class GattCharacteristic : IGattCharacteristic {
		/// <summary>
		/// The characteristic to wrap
		/// </summary>
		readonly Windows.Devices.Bluetooth.GenericAttributeProfile.GattCharacteristic characteristic;

		/// <summary>
		/// The protection level of the characteristic
		/// </summary>
		public GattProtectionLevel ProtectionLevel {
			get { return characteristic.ProtectionLevel; }
			set { characteristic.ProtectionLevel = value; }
		}

		/// <summary>
		/// The Uuid of the characteristic
		/// </summary>
		public Guid Uuid {
			get { return characteristic.Uuid; }
		}
		/// <summary>
		/// The value changed event handler
		/// </summary>
		public event TypedEventHandler<GattCharacteristic, GattValueChangedEventArgs> ValueChanged;

		/// <summary>
		/// Constructor with the characteristic to wrap
		/// </summary>
		/// <param name="characteristic">The characteristic to wrap</param>
		internal GattCharacteristic(Windows.Devices.Bluetooth.GenericAttributeProfile.GattCharacteristic characteristic) {
			this.characteristic = characteristic;
		}

		/// <summary>
		/// Register the Value changed callback on the characteristic
		/// </summary>
		public void RegisterValueChangedCallback() {
			characteristic.ValueChanged += CharacteristicOnValueChanged;
		}

		/// <summary>
		/// internal value changed event handler that forwards from the real characteristic
		/// </summary>
		/// <param name="sender">The sender of the event</param>
		/// <param name="args">The arguments of the event</param>
		void CharacteristicOnValueChanged(Windows.Devices.Bluetooth.GenericAttributeProfile.GattCharacteristic sender, GattValueChangedEventArgs args) {
			if(ValueChanged != null)
				ValueChanged(this, args);
		}

		/// <summary>
		/// Reads the client characteristic configuration.
		/// </summary>
		/// <returns>Succes or Unreachable and which mode the characteristic is configured on (None, Notify or Indicate)</returns>
		public async Task<GattReadClientCharacteristicConfigurationDescriptorResult> ReadClientCharacteristicConfigurationDescriptorAsync() {
			return await characteristic.ReadClientCharacteristicConfigurationDescriptorAsync();
		}

		/// <summary>
		/// Write the client characteristic configuration
		/// </summary>
		/// <param name="clientCharacteristicConfigurationDescriptorValue">The configuration to write</param>
		/// <returns>Success or Unreachable</returns>
		public async Task<GattCommunicationStatus> WriteClientCharacteristicConfigurationDescriptorAsync(GattClientCharacteristicConfigurationDescriptorValue clientCharacteristicConfigurationDescriptorValue) {
			return await characteristic.WriteClientCharacteristicConfigurationDescriptorAsync(clientCharacteristicConfigurationDescriptorValue);
		}

		/// <summary>
		/// Reads a value from the characteristic
		/// </summary>
		/// <returns>The value read</returns>
		public async Task<GattReadResult> ReadValueAsync() {
			if(characteristic != null)
				return await characteristic.ReadValueAsync(BluetoothCacheMode.Uncached);

			return null;
		}

		/// <summary>
		/// Write Value to characteristic
		/// </summary>
		/// <param name="buffer">The value to write</param>
		/// <returns>Success or Unreachable</returns>
		public async Task<GattCommunicationStatus> WriteValueAsync(IBuffer buffer) {
			if (characteristic != null)
				return await characteristic.WriteValueAsync(buffer);

			return GattCommunicationStatus.Unreachable;
		}

		/// <summary>
		/// Write Value to characteristic with specific write option.
		/// </summary>
		/// <param name="buffer">The value to write</param>
		/// <param name="writeOption">The write option used to write the value. (with or without response)</param>
		/// <returns>Success or Unreachable</returns>
		public async Task<GattCommunicationStatus> WriteValueAsync(IBuffer buffer, GattWriteOption writeOption) {
			if (characteristic != null)
				return await characteristic.WriteValueAsync(buffer, writeOption);

			return GattCommunicationStatus.Unreachable;
		}
	}
}
