﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using LegoBleTestApp.Wrapper;
using LegoDeviceSDK;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Core;
using NUnit.Framework;

namespace LegoSDKNunitTest.SDK {
	[TestFixture]
	public class BluetoothDeviceManagerTest : IDeviceManagerDelegate {
		AutoResetEvent testTrigger;
		List<IDevice> devicesAdded = new List<IDevice>();
		List<IDevice> devicesRemoved = new List<IDevice>();

		bool willStartConnectingCalled = false;
		bool didDisconnectFromDeviceCalled = false;
		bool didFinishInterrogatingDeviceCalled = false;
		bool didStartInterrogatingDeviceCalled = false;


		void Reset() {
			this.testTrigger = new AutoResetEvent(false);
			devicesAdded.Clear();
			devicesRemoved.Clear();

			willStartConnectingCalled = false;
			didDisconnectFromDeviceCalled = false;
			didFinishInterrogatingDeviceCalled = false;
			didStartInterrogatingDeviceCalled = false;

		}

		[Test]
		public void TestInstance() {
			// ARRANGE
			Reset();
			BluetoothDeviceManager manager = null;
			manager = BluetoothDeviceManager.SharedDeviceManager;

			// ACT

			// ASSERT
			Assert.IsNotNull(manager);
			Assert.IsInstanceOf<BluetoothDeviceManager>(manager);

		}

		[Test]
        [Ignore("Fails on the master branch")]
		public async void TestScan() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() {Count = 3});
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.AreEqual(3, devicesAdded.Count);


			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 2 });

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.AreEqual(1, devicesRemoved.Count);

			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3 });

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.AreEqual(1, devicesAdded.Count);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);
			devicesAdded.Clear();
			devicesRemoved.Clear();
		}

		[Test]
		public async void TestScanTooFewhashTags() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.TooFewHashTags});
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();

			// ASSERT
			Assert.AreEqual(0, DeviceManager.SharedDeviceManager.AllDevices.Count);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);

		}

		[Test]
		public async void TestScanTooManyhashTags() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.TooManyHashTags });
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();

			// ASSERT
			Assert.AreEqual(0, BluetoothDeviceManager.SharedDeviceManager.AllDevices.Count);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);

		}

		[Test]
		public async void TestScanNoAmpersand() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.NoAmpersand });
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			devicesAdded.Clear();
			devicesRemoved.Clear();
			await DeviceManager.SharedDeviceManager.ScanAsync();

			// ASSERT
			Assert.AreEqual(0, BluetoothDeviceManager.SharedDeviceManager.AllDevices.Count);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);

		}

		[Test]
		public void TestWillStartConnectingCallback() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.NoAmpersand });
			var testDevice = new BluetoothDevice("", null, BluetoothDeviceManager.SharedDeviceManager);
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			testDevice.ConnectState = DeviceState.DeviceStateConnecting;
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.True(willStartConnectingCalled);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);
		}

		[Test]
		public void TestDidStartInterrogatingCallback() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.NoAmpersand });
			var testDevice = new BluetoothDevice("", null, BluetoothDeviceManager.SharedDeviceManager);
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			testDevice.ConnectState = DeviceState.DeviceStateInterrogating;
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.True(didStartInterrogatingDeviceCalled);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);
		}

		[Test]
        [Ignore("Fails on the master branch")]
		public void TestDidFinishInterrogatingCallback() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.NoAmpersand });
			var testDevice = new BluetoothDevice("", null, BluetoothDeviceManager.SharedDeviceManager);
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			testDevice.ConnectState = DeviceState.DeviceStateInterrogationFinished;
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.True(didFinishInterrogatingDeviceCalled);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);
		}
		[Test]
		public void TestDidDisconnectCallback() {
			// ARRANGE
			Reset();
			GattConfiguration.ChangeConfiguration(new DeviceManagerGattDeviceServiceWrapperMock(), new DeviceManagerDeviceinformationWrapperMock() { Count = 3, ErrorEnum = DeviceManagerDeviceInformationMockErrorEnum.NoAmpersand });
			var testDevice = new BluetoothDevice("", null, BluetoothDeviceManager.SharedDeviceManager);
			DeviceManager.SharedDeviceManager.AddDelegate(this);

			// ACT
			testDevice.ConnectState = DeviceState.DeviceStateConnecting;
			testDevice.ConnectState = DeviceState.DeviceStateDisconnected;
			testTrigger.WaitOne(1000);

			// ASSERT
			Assert.True(didDisconnectFromDeviceCalled);

			// CLEANUP
			DeviceManager.SharedDeviceManager.RemoveDelegate(this);
		}

		public void DeviceDidAppear(IDeviceManager deviceManager, IDevice device) {
			devicesAdded.Add(device);
			testTrigger.Set();
		}

		public void DeviceDidDisappear(IDeviceManager deviceManager, IDevice device) {
			devicesRemoved.Add(device);
			testTrigger.Set();
		}

		public void DidDisconnectFromDevice(IDeviceManager deviceManager, IDevice device, bool autoReconnect) {
			didDisconnectFromDeviceCalled = true;
			testTrigger.Set();
		}

		public void DidFailToConnectToDevice(IDeviceManager deviceManager, IDevice device, bool autoReconnect) {
			testTrigger.Set();
		}

		public void DidFinishInterrogatingDevice(IDeviceManager deviceManager, IDevice device) {
			didFinishInterrogatingDeviceCalled = true;
			testTrigger.Set();
		}

		public void DidStartInterrogatingDevice(IDeviceManager deviceManager, IDevice device) {
			didStartInterrogatingDeviceCalled = true;
			testTrigger.Set();
		}

		public void WillStartConnectingToDevice(IDeviceManager deviceManager, IDevice device) {
			willStartConnectingCalled = true;
			testTrigger.Set();
		}
	}
}
