﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Windows.UI.Xaml.Controls;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Helpers;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Framework;

namespace LegoSDKNunitTest.SDK.Bluetooth {
	[TestFixture]
	public class BluetoothDeviceTest : IDeviceDelegate {
		AutoResetEvent testTrigger;
		BluetoothDevice bluetoothDevice;

		bool buttonState = false;
		int buttonPressedCount = 0;

		string fromName = "";
		string toName = "";

        bool lowVoltage = false;

		Dictionary<byte, IoType> ioTypeDict = new Dictionary<byte, IoType>();
		IoType AddedOrRemovedIoType;
		Dictionary<byte, byte> hubIndexDict = new Dictionary<byte, byte>();

		private void Reset() {
			this.testTrigger = new AutoResetEvent(false);
			buttonState = false;
			buttonPressedCount = 0;
		
			fromName = "";
			toName = "";

			bluetoothDevice = new BluetoothDevice("", null, BluetoothDeviceManager.SharedDeviceManager);
			bluetoothDevice.AddDelegate(this);

			ioTypeDict.Clear();
			ioTypeDict.Add(1, IoType.IoTypeMotor);
			ioTypeDict.Add(20, IoType.IoTypeVoltage);
			ioTypeDict.Add(21, IoType.IoTypeCurrent);
			ioTypeDict.Add(22, IoType.IoTypePiezoTone);
			ioTypeDict.Add(23, IoType.IoTypeRgbLight);
			ioTypeDict.Add(34, IoType.IoTypeTiltSensor);
			ioTypeDict.Add(35, IoType.IoTypeMotionSensor);
			ioTypeDict.Add(0, IoType.IoTypeGeneric);
			ioTypeDict.Add(2, IoType.IoTypeGeneric);
			ioTypeDict.Add(19, IoType.IoTypeGeneric);
			ioTypeDict.Add(24, IoType.IoTypeGeneric);
			ioTypeDict.Add(33, IoType.IoTypeGeneric);
			ioTypeDict.Add(36, IoType.IoTypeGeneric);

			hubIndexDict.Clear();
			hubIndexDict.Add(1, 1);
			hubIndexDict.Add(20, 51);
			hubIndexDict.Add(21, 52);
			hubIndexDict.Add(22, 53);
			hubIndexDict.Add(23, 54);
			hubIndexDict.Add(34, 2);
			hubIndexDict.Add(35, 3);
			hubIndexDict.Add(0, 4);
			hubIndexDict.Add(2, 5);
			hubIndexDict.Add(19, 6);
			hubIndexDict.Add(24, 7);
			hubIndexDict.Add(33, 8);
			hubIndexDict.Add(36, 9);

		}

		[Test]
		public void TestRemoveDelegate() {
			Reset();
			bluetoothDevice.AddDelegate(this);
			ProcessDataHelper.GetButtonPressedEvent(new byte[1] { 1 }, bluetoothDevice);
			testTrigger.WaitOne(1000);
			Assert.AreEqual(2, buttonPressedCount);

			buttonPressedCount = 0;
			bluetoothDevice.RemoveDelegate(this);
			ProcessDataHelper.GetButtonPressedEvent(new byte[1] { 1 }, bluetoothDevice);
			testTrigger.WaitOne(1000);
			Assert.AreEqual(1, buttonPressedCount);
		}

		[Test]
		public void TestButtonPressedTrue() {
			Reset();
			ProcessDataHelper.GetButtonPressedEvent(new byte[1] { 1 }, bluetoothDevice);
			testTrigger.WaitOne(1000);
			Assert.AreEqual(true, buttonState);
		}
		[Test]
		public void TestButtonPressedFalse() {
			Reset();
			ProcessDataHelper.GetButtonPressedEvent(new byte[1] { 0 }, bluetoothDevice);
			testTrigger.WaitOne(1000);
			Assert.AreEqual(false, buttonState);
		}

		[Test]
		public void TestButtonPressedWrongInput() {
			Reset();
			Assert.Catch<Exception>(() => ProcessDataHelper.GetButtonPressedEvent(new byte[2] { 0, 1 }, bluetoothDevice));
		}

		[Test]
		public async void TestAttachedIoDetachedIo() {
			Reset();
			await bluetoothDevice.Connect();
			foreach (var ioType in ioTypeDict) {
				var data = new byte[12] { ioType.Key, 1, hubIndexDict[ioType.Key], ioType.Key, 0, 0, 0, 0, 0, 0, 0, 0 };
				ProcessDataHelper.GetAttachedIoEvent(data, bluetoothDevice, new BluetoothServiceIoMock());
				testTrigger.WaitOne(1000);
				Assert.AreEqual(ioType.Value, AddedOrRemovedIoType); 
			}

			Assert.AreEqual(13, bluetoothDevice.Services.Count);
			Assert.AreEqual(4, bluetoothDevice.InternalServices.Count);
			Assert.AreEqual(9, bluetoothDevice.ExternalServices.Count);

			foreach (var ioType in ioTypeDict) {
				var data = new byte[2] { ioType.Key, 0 };
				ProcessDataHelper.GetAttachedIoEvent(data, bluetoothDevice);
				testTrigger.WaitOne(1000);
				Assert.AreEqual(ioType.Value, AddedOrRemovedIoType);
			}

			Assert.AreEqual(0, bluetoothDevice.Services.Count);
			Assert.AreEqual(0, bluetoothDevice.InternalServices.Count);
			Assert.AreEqual(0, bluetoothDevice.ExternalServices.Count);

			bluetoothDevice.StopProcessingIo();
		}

        [Test]
        public void TestLowVoltageAlertTrue() {
            Reset();
            ProcessDataHelper.GetLowVoltageEvent(new byte[1] { 1 }, bluetoothDevice);
            testTrigger.WaitOne(1000);
            Assert.AreEqual(true, lowVoltage);
        }
        [Test]
        public void TestLowVoltageAlertFalse() {
            Reset();
            ProcessDataHelper.GetLowVoltageEvent(new byte[1] { 0 }, bluetoothDevice);
            testTrigger.WaitOne(1000);
            Assert.AreEqual(false, lowVoltage);
        }


		public void DidUpdateDeviceInfo(IDevice device, DeviceInfo deviceInfo, Exception error) {
			throw new NotImplementedException();
		}

		public void DidChangeButtonState(IDevice device, bool buttonState) {
			this.buttonState = buttonState;
			buttonPressedCount++;
			testTrigger.Set();
		}

		public void DidChangeName(IDevice device, string fromName, string toName) {
			this.fromName = fromName;
			this.toName = toName;
			testTrigger.Set();
		}

		public void DidUpdateBatteryLevel(IDevice device, uint newLevel) {
			throw new NotImplementedException();
		}

        public void DidChangeLowVoltageState(IDevice device, bool lowVoltage) {
            this.lowVoltage = lowVoltage;
            testTrigger.Set();
        }

		public void DidAddService(IDevice device, IService service) {
			AddedOrRemovedIoType = service.ConnectInfo.TypeEnum;
			testTrigger.Set();
		}

		public void DidRemoveService(IDevice device, IService service) {
			AddedOrRemovedIoType = service.ConnectInfo.TypeEnum;
			testTrigger.Set();
		}

		public void DidFailToAddServiceWithError(IDevice device, Exception error) {
			throw new NotImplementedException();
		}
    }
}
