﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using LegoDeviceSDK.Annotations;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Bluetooth.Definitions;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Framework;

namespace LegoSDKNunitTest.SDK.Bluetooth {
	[TestFixture]
	public class BluetoothIoTest : BluetoothService {
		AutoResetEvent testTrigger;
		InputFormat receivedInputFormat;
		byte[] receivedData;

		public BluetoothIoTest()
			: base(null, null, new ConnectInfo() { ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeMotor }) {
		}

		[Test]
		public async void TestHandleUpdatedInputFormatSuccess() {
			testTrigger = new AutoResetEvent(false);
			var connectInfo = new ConnectInfo() {ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeMotor};
			var gattServiceMock = new GattDeviceServiceMock();
			var ioServiceDefinition = new IoServiceDefinition(gattServiceMock);
			var bluetoothIo = BluetoothIo.GetBluetoothIo(ioServiceDefinition);
			bluetoothIo.AddDelegate(this);
			await bluetoothIo.ConfigureNotifications();
			var version = 1;
			var mode = 0;
			uint deltaInterval = 1234;
			var unit = 1;
			var notificationsEnabled = true;
			var numberOfBytes = 2;
			var formatWriteData = BluetoothTestStupFactory.InputFormatWriteDataWithRevision(version, connectInfo.ConnectId, IoType.IoTypeMotionSensor, mode, deltaInterval, unit, notificationsEnabled, numberOfBytes);

			bluetoothIo.HandleUpdatedInputServiceCharacteristic(formatWriteData, ioServiceDefinition.CharacteristicWithUuid(IoServiceDefinition.CharacteristicInputFormatUuid));
			testTrigger.WaitOne(1000);

			Assert.That(receivedInputFormat.Mode == mode);
			Assert.That(receivedInputFormat.ConnectId == connectInfo.ConnectId);
			Assert.That(receivedInputFormat.TypeId == (int)IoType.IoTypeMotionSensor);
			Assert.That(receivedInputFormat.Mode == mode);
			Assert.That(receivedInputFormat.DeltaInterval == deltaInterval);
			Assert.That(receivedInputFormat.Unit == (InputFormatUnit)unit);
			Assert.True(receivedInputFormat.NotificationsEnabled == notificationsEnabled);
			Assert.That(receivedInputFormat.NumberOfBytes == numberOfBytes);

			bluetoothIo.RemoveDelegate(this);
			bluetoothIo.Dispose();
		}

		[Test]
		public async void TestHandleUpdatedInputValueSuccess() {
			testTrigger = new AutoResetEvent(false);
			var gattServiceMock = new GattDeviceServiceMock();
			var ioServiceDefinition = new IoServiceDefinition(gattServiceMock);
			var bluetoothIo = BluetoothIo.GetBluetoothIo(ioServiceDefinition);
			bluetoothIo.AddDelegate(this);
			await bluetoothIo.ConfigureNotifications();

			var formatRevision = 2;

			var connectId1 = 1;
			var connectId2 = 2;

			var connectInfo1 = new ConnectInfo() { ConnectId = connectId1, HubIndex = 1, TypeEnum = IoType.IoTypeMotor };
			var connectInfo2 = new ConnectInfo() { ConnectId = connectId2, HubIndex = 2, TypeEnum = IoType.IoTypeVoltage };
			SimulateReceivingInputFormatForConnectInfo(ref bluetoothIo, ioServiceDefinition, connectInfo1, formatRevision, 2);
			SimulateReceivingInputFormatForConnectInfo(ref bluetoothIo, ioServiceDefinition, connectInfo2, formatRevision, 1);

			short value1 = 1000;
			var value2 = 120;
			var value1Bytes = BitConverter.GetBytes(value1);
			var valueData = new byte[] {
				(byte) formatRevision,
				(byte) connectId1,
				value1Bytes[0],
				value1Bytes[1],
				(byte) connectId2,
				(byte) value2,
			};

			ConnectInfo = connectInfo1;
			bluetoothIo.HandleUpdatedInputServiceCharacteristic(valueData, ioServiceDefinition.CharacteristicWithUuid(IoServiceDefinition.CharacteristicInputValueUuid));
			testTrigger.WaitOne(1000);

			Assert.That(receivedData.Length == 2);
			Assert.That(receivedData[0] == value1Bytes[0]);
			Assert.That(receivedData[1] == value1Bytes[1]);

			ConnectInfo = connectInfo2;
			bluetoothIo.HandleUpdatedInputServiceCharacteristic(valueData, ioServiceDefinition.CharacteristicWithUuid(IoServiceDefinition.CharacteristicInputValueUuid));
			testTrigger.WaitOne(1000);

			Assert.That(receivedData.Length == 1);
			Assert.That(receivedData[0] == value2);

			bluetoothIo.RemoveDelegate(this);
			bluetoothIo.Dispose();
		}

		[Test]
		public async void TestHandleUpdatedInputValueIgnoreIFRevisionUnknown() {
			testTrigger = new AutoResetEvent(false);
			var gattServiceMock = new GattDeviceServiceMock();
			var ioServiceDefinition = new IoServiceDefinition(gattServiceMock);
			var bluetoothIo = BluetoothIo.GetBluetoothIo(ioServiceDefinition);
			bluetoothIo.AddDelegate(this);
			await bluetoothIo.ConfigureNotifications();

			var formatRevision = 2;
			var connectId = 1;
			var connectInfo = new ConnectInfo() { ConnectId = connectId, HubIndex = 1, TypeEnum = IoType.IoTypeMotor };
			SimulateReceivingInputFormatForConnectInfo(ref bluetoothIo, ioServiceDefinition, connectInfo, formatRevision, 2);

			var valueFormatRevision = 3;

			short value = 10;
			var valueData = new byte[] {
				(byte) valueFormatRevision,
				(byte) connectId,
				(byte) value,
			};

			receivedData = null;
			ConnectInfo = connectInfo;
			bluetoothIo.HandleUpdatedInputServiceCharacteristic(valueData, ioServiceDefinition.CharacteristicWithUuid(IoServiceDefinition.CharacteristicInputValueUuid));

			await Task.Delay(100);
			Assert.Null(receivedData);

			bluetoothIo.RemoveDelegate(this);
			bluetoothIo.Dispose();
		}

		private void SimulateReceivingInputFormatForConnectInfo(ref BluetoothIo io, IoServiceDefinition ioDef, ConnectInfo connectInfo, int revision, int numberOfBytes) {
			var mode = 0;
			uint deltaInterval = 1234;
			var unit = 1;
			var notificationsEnabled = true;
			var formatData = BluetoothTestStupFactory.InputFormatWriteDataWithRevision(revision, connectInfo.ConnectId, connectInfo.TypeEnum, mode, deltaInterval, unit, notificationsEnabled, numberOfBytes);
			io.HandleUpdatedInputServiceCharacteristic(formatData, ioDef.CharacteristicWithUuid(IoServiceDefinition.CharacteristicInputFormatUuid));
			testTrigger.WaitOne(1000);
		}


		internal override void HandleUpdatedInputFormat(InputFormat newInputFormat) {
			receivedInputFormat = newInputFormat;
			testTrigger.Set();

		}

		internal override bool HandleUpdatedValueData(byte[] data, out SdkError error) {
			error = null;
			receivedData = data;
			testTrigger.Set();
			return true;
		}
	}
}
