﻿using System;
using System.Threading;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Framework;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoSDKNunitTest.SDK.Bluetooth.Services {
	[TestFixture]
	public class BluetoothMotionSensorTest : IMotionSensorDelegate {
		AutoResetEvent testTrigger;

		BluetoothMotionSensor motionSensor;
		int notfiedCount;
		float notifiedDistance;
		byte[] notifiedValueData;
		ConnectInfo connectInfo = new ConnectInfo() {ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeMotionSensor};
		IIo ioStub = new BluetoothServiceIoMock();

		[SetUp]
		public void Setup() {
			testTrigger = new AutoResetEvent(false);
			ioStub = new BluetoothServiceIoMock();
			var service = BluetoothService.GetServiceFromConnectInfo(new BluetoothDeviceStub(), connectInfo, ioStub);
			motionSensor = service as BluetoothMotionSensor;
			if(motionSensor == null)
				Assert.Fail();
			motionSensor.AddDelegate(this);
		}

		[TearDown]
		public void TearDown() {
			motionSensor.RemoveDelegate(this);
			motionSensor = null;
		}

		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
		}

		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
			notifiedValueData = newData;
			testTrigger.Set();
		}

        public void DidUpdateDistance(IMotionSensor motionSensor, float oldDistance, float newDistance) {
			notifiedDistance = newDistance;
			testTrigger.Set();
		}

		public void DidUpdateCount(IMotionSensor motionSensor, uint newCount) {
			notfiedCount = (int)newCount;
			testTrigger.Set();
		}

		void UpdateValueWithData(byte[] data) {
			SdkError error;
			motionSensor.HandleUpdatedValueData(data, out error);
			Assert.Null(error);
		}

		void UpdateInputFormatMode(MotionSensorMode mode, InputFormatUnit unit, int numberOfBytes) {
			var data = BluetoothTestStupFactory.InputFormatWriteDataWithRevision(1, motionSensor.ConnectInfo.ConnectId, motionSensor.ConnectInfo.TypeEnum, (int)mode, 1, (int)unit, true, numberOfBytes);

			var inputFormat = InputFormat.InputFormatFromByteArray(data);
			motionSensor.HandleUpdatedInputFormat(inputFormat);
		}

		[Test]
		public void TestHandleUpdatedValue_DetectModeRaw() {
			UpdateInputFormatMode(MotionSensorMode.Detect, InputFormatUnit.InputFormatUnitRaw, 1);

			UpdateValueWithData(new byte[] {1});
			testTrigger.WaitOne(1000);
			Assert.That(1f == motionSensor.Distance);
			Assert.That(1f == notifiedDistance);

			UpdateValueWithData(new byte[] { 100 });
			testTrigger.WaitOne(1000);
			Assert.That(100f == motionSensor.Distance);
			Assert.That(100f == notifiedDistance);

		}

		[Test]
		public void TestHandleUpdatedValue_DetectModePercentage() {
			UpdateInputFormatMode(MotionSensorMode.Detect, InputFormatUnit.InputFormatUnitPercentage, 1);

			UpdateValueWithData(new byte[] { 1 });
			testTrigger.WaitOne(1000);
			Assert.That(1f == motionSensor.Distance);
			Assert.That(1f == notifiedDistance);

			UpdateValueWithData(new byte[] { 100 });
			testTrigger.WaitOne(1000);
			Assert.That(100f == motionSensor.Distance);
			Assert.That(100f == notifiedDistance);

		}

		[Test]
		public void TestHandleUpdatedValue_DetectModeSi() {
			UpdateInputFormatMode(MotionSensorMode.Detect, InputFormatUnit.InputFormatUnitSi, 4);

			float value = 123.45f;
			UpdateValueWithData(BitConverter.GetBytes(value));
			testTrigger.WaitOne(1000);
			Assert.That(value == motionSensor.Distance);
			Assert.That(value == notifiedDistance);

			value = -123.45f;
			UpdateValueWithData(BitConverter.GetBytes(value));
			testTrigger.WaitOne(1000);
			Assert.That(value == motionSensor.Distance);
			Assert.That(value == notifiedDistance);
		}

		[Test]
		public void TestHandleUpdatedValue_CountModeRaw() {
			UpdateInputFormatMode(MotionSensorMode.Count, InputFormatUnit.InputFormatUnitRaw, 4);

			var count = 12345;
			UpdateValueWithData(BitConverter.GetBytes(count));
			testTrigger.WaitOne(1000);
			Assert.That(count == motionSensor.Count);
			Assert.That(count == notfiedCount);

			Assert.That(0f == motionSensor.Distance);

		}

		[Test]
		public void TestHandleUpdatedValue_CountModePercentage() {
			UpdateInputFormatMode(MotionSensorMode.Count, InputFormatUnit.InputFormatUnitPercentage, 1);

			var count = 99;
			UpdateValueWithData(new byte[] {(byte)count});
			testTrigger.WaitOne(1000);
			Assert.That(count == motionSensor.Count);
			Assert.That(count == notfiedCount);

			Assert.That(0f == motionSensor.Distance);

		}

		[Test]
		public void TestHandleUpdatedValue_CountModeSi() {
			UpdateInputFormatMode(MotionSensorMode.Count, InputFormatUnit.InputFormatUnitSi, 4);

			var count = 12345;
			UpdateValueWithData(BitConverter.GetBytes(count));
			testTrigger.WaitOne(1000);
			Assert.That(((int)count) == motionSensor.Count);
			Assert.That(((int)count) == notfiedCount);

			Assert.That(0f == motionSensor.Distance);
		}

	
	}
}
