﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Windows.UI;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Framework;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoSDKNunitTest.SDK.Bluetooth.Services {
	[TestFixture]
	public class BluetoothRgbLightTest : IRgbLightDelegate {

		AutoResetEvent testTrigger;

		BluetoothRgbLight rgbLight;
		readonly ConnectInfo connectInfo = new ConnectInfo() { ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeRgbLight };
		BluetoothServiceIoMock ioStub = new BluetoothServiceIoMock();

		Color oldReceivedColor;
		Color newReceivedColor;

        byte[] receivedRgbColorData;

        int receivedColorIndex;

		[SetUp]
		public void Setup() {
			testTrigger = new AutoResetEvent(false);
			ioStub = new BluetoothServiceIoMock();
			var service = BluetoothService.GetServiceFromConnectInfo(new BluetoothDeviceStub(), connectInfo, ioStub);
			rgbLight = service as BluetoothRgbLight;
			rgbLight.AddDelegate(this);
            UpdateInputFormat(RgbLightMode.Absolute, InputFormatUnit.InputFormatUnitRaw, 3);
		}

		[TearDown]
		public void TearDown() {
			rgbLight.RemoveDelegate(this);
			rgbLight = null;
		}

		[Test]
		public void TestSetAbsoluteColor_success() {
			var expectedColor = Color.FromArgb(255, 0, 100, 255);
			
            rgbLight.Color = expectedColor;
			testTrigger.WaitOne(1000);

			Assert.AreEqual(expectedColor, newReceivedColor);
            Assert.AreEqual(expectedColor, ioStub.LastColorWritten);
            Assert.AreEqual(expectedColor, rgbLight.Color);
		}

        [Test]
        public void TestSetAbsoluteColor_ignored_when_in_discrete_mode() {
            UpdateInputFormat(RgbLightMode.Discrete, InputFormatUnit.InputFormatUnitRaw, 1);

            Color initialColor = rgbLight.Color;
            Color newColor = Color.FromArgb(255, 0, 10, 255);
            rgbLight.Color = newColor;

            Assert.AreNotEqual(newColor, initialColor);
            Assert.AreEqual(initialColor, rgbLight.Color);
        }

        [Test]
        public void TestSwitchOff_absolute_mode() {
            Color expectedColor = Colors.Black;

            rgbLight.SwitchOff();
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColor, newReceivedColor);
            Assert.AreEqual(expectedColor, ioStub.LastColorWritten);
            Assert.AreEqual(expectedColor, rgbLight.Color);
        }

        [Test]
        public void TestSwitchToDefaultColor_absolute_mode() {
            Color expectedColor = Color.FromArgb(0xFF, 0x00, 0x00, 0xFF);

            rgbLight.SwitchToDefaultColor();
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColor, newReceivedColor);
            Assert.AreEqual(expectedColor, ioStub.LastColorWritten);
            Assert.AreEqual(expectedColor, rgbLight.Color);
        }

        [Test]
        public void TestHandleUpdatedValue_absolute_raw() {
            var expectedColor = Color.FromArgb(255, 0, 100, 255);

            SdkError error;
            rgbLight.HandleUpdatedValueData(new byte[] { expectedColor.R, expectedColor.G, expectedColor.B }, out error);
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColor, newReceivedColor);
            Assert.AreEqual(expectedColor, rgbLight.Color);
            Assert.AreEqual(new byte[] { 0x00, 0x64, 0xFF }, receivedRgbColorData);
        }

        [Test]
        public void TestSetDiscreteColor_success() {
            UpdateInputFormat(RgbLightMode.Discrete, InputFormatUnit.InputFormatUnitRaw, 1);
            int expectedColorIndex = 1;

            rgbLight.ColorIndex = expectedColorIndex;
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColorIndex, receivedColorIndex);
            Assert.AreEqual(expectedColorIndex, ioStub.LastColorIndexWritten);
            Assert.AreEqual(expectedColorIndex, rgbLight.ColorIndex);
        }

        [Test]
        public void TestSwitchOff_discrete_mode() {
            UpdateInputFormat(RgbLightMode.Discrete, InputFormatUnit.InputFormatUnitRaw, 1);
            int expectedColorIndex = 0;

            rgbLight.SwitchOff();
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColorIndex, receivedColorIndex);
            Assert.AreEqual(expectedColorIndex, ioStub.LastColorIndexWritten);
            Assert.AreEqual(expectedColorIndex, rgbLight.ColorIndex);
        }

        [Test]
        public void TestSwitchToDefaultColor_discrete_mode() {
            UpdateInputFormat(RgbLightMode.Discrete, InputFormatUnit.InputFormatUnitRaw, 1);
            // Color index 3 happens to be the default color :-)
            int expectedColorIndex = 3;

            rgbLight.SwitchToDefaultColor();
            testTrigger.WaitOne(1000);

            Assert.AreEqual(expectedColorIndex, receivedColorIndex);
            Assert.AreEqual(expectedColorIndex, ioStub.LastColorIndexWritten);
            Assert.AreEqual(expectedColorIndex, rgbLight.ColorIndex);
        }

        private void UpdateInputFormat(RgbLightMode mode, InputFormatUnit unit, int numberOfBytes) {
            byte[] inputFormatData = BluetoothTestStupFactory.InputFormatWriteDataWithRevision(1, connectInfo.ConnectId, connectInfo.TypeEnum, (int)mode, (uint)1, (int)unit, true, numberOfBytes);
            InputFormat inputFormat = InputFormat.InputFormatFromByteArray(inputFormatData);
            rgbLight.HandleUpdatedInputFormat(inputFormat);
        }

		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
		}

		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
            receivedRgbColorData = newData;
		}

		public void DidUpdateColor(IRgbLight rgbLight, Color oldColor, Color newColor) {
			oldReceivedColor = oldColor;
			newReceivedColor = newColor;
			testTrigger.Set();
		}

        public void DidUpdateColorIndex(IRgbLight rgbLight, int oldColorIndex, int newColorIndex) {
            receivedColorIndex = newColorIndex;
            testTrigger.Set();
        }
    }
}
