﻿using System;
using System.Collections.Generic;
using System.Threading;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Framework;
using IService = LegoDeviceSDK.Interfaces.IService;

namespace LegoSDKNunitTest.SDK.Bluetooth.Services {
	[TestFixture]
	public class BluetoothServiceTest : IServiceDelegate{
		AutoResetEvent testTrigger;
		ConnectInfo testConnectInfo = new ConnectInfo() { ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeMotor, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
		ConnectInfo testConnectInfo2 = new ConnectInfo() { ConnectId = 2, HubIndex = 2, TypeEnum = IoType.IoTypeMotor, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };

		InputFormat testInputFormat1 = new InputFormat() { Revision = 0, ConnectId = 1, DeltaInterval = 1, Mode = 1, NotificationsEnabled = false, NumberOfBytes = 1, TypeId = 1, Unit = new InputFormatUnit() };
		InputFormat testInputFormat2 = new InputFormat() { Revision = 0, ConnectId = 2, DeltaInterval = 2, Mode = 2, NotificationsEnabled = true, NumberOfBytes = 2, TypeId = 2, Unit = new InputFormatUnit() };

		InputFormat receivedOldInputFormat;
		InputFormat receivedInputFormat;

		byte[] testValue1 = new byte[1] { 0 };
		byte[] testValue2 = new byte[1] { 1 };

		byte[] receivedOldData;
		byte[] receivedNewData;

		[Test]
		public void TestConnectInfo() {
			var testService = new BluetoothMotor(null, null, testConnectInfo);
			Assert.True(testService.ConnectInfo.IsEqualToConnectInfo(testConnectInfo));
			testService.ConnectInfo = testConnectInfo2;
			Assert.True(testService.ConnectInfo.IsEqualToConnectInfo(testConnectInfo2));
		}

		[Test]
		public void TestUpdateInputFormat() {
			this.testTrigger = new AutoResetEvent(false);
			var testService = (BluetoothService)BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), testConnectInfo);
			testService.InputFormat = testInputFormat1;
			testService.AddDelegate(this);
			testService.InputFormat = testInputFormat2;
			testTrigger.WaitOne(1000);
			Assert.True(receivedInputFormat.Equals(testInputFormat2));
			Assert.True(receivedOldInputFormat.Equals(testInputFormat1));
			Assert.True(testService.InputFormat.Equals(testInputFormat2));
			testService.RemoveDelegate(this);

		}

		[Test]
		public void TestUpdateValueData() {
			this.testTrigger = new AutoResetEvent(false);
			var testService = (BluetoothService)BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), testConnectInfo);
			testService.ValueData = testValue1;
			testService.AddDelegate(this);
			testService.ValueData = testValue2;
			testTrigger.WaitOne(1000);
			Assert.AreEqual(receivedOldData, testValue1);
			Assert.AreEqual(receivedNewData, testValue2);
			Assert.AreEqual(testService.ValueData, testValue2);
			testService.RemoveDelegate(this);
		}

		[Test]
		public void TestGetServiceFromConnectInfo() {
			var connectInfo = new ConnectInfo() {ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeCurrent, FirmwareVersion = new Revision(), HardwareVersion = new Revision()};
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothCurrentSensor>(testService);

			connectInfo.TypeEnum = IoType.IoTypeGeneric;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothGenericService>(testService);

			connectInfo.TypeEnum = IoType.IoTypeMotionSensor;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothMotionSensor>(testService);

			connectInfo.TypeEnum = IoType.IoTypeMotor;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothMotor>(testService);

			connectInfo.TypeEnum = IoType.IoTypePiezoTone;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothPiezoTonePlayer>(testService);

			connectInfo.TypeEnum = IoType.IoTypeRgbLight;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothRgbLight>(testService);

			connectInfo.TypeEnum = IoType.IoTypeTiltSensor;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothTiltSensor>(testService);

			connectInfo.TypeEnum = IoType.IoTypeVoltage;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, new BluetoothServiceIoMock());
			Assert.IsInstanceOf<BluetoothVoltageSensor>(testService);
		}

		[Test]
		public void TestValueAsNumbersRaw() {
			this.testTrigger = new AutoResetEvent(false);
			var connectInfo = new ConnectInfo() { ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitRaw, true);
			inputFormat.NumberOfBytes = 2;
			testService.UpdateInputFormat(inputFormat);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitRaw, 1, 2);
			testService.AddValidDataFormat(dataFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			var success = genericService.HandleUpdatedValueData(new byte[] {1, 2}, out error);
		
			Assert.True(success);
			Assert.AreEqual(genericService.NumbersFromValueData.Count, dataFormat.DataSetCount);
			Assert.AreEqual((int)genericService.NumbersFromValueData[0][0], 1);
			Assert.AreEqual((int)genericService.NumbersFromValueData[1][0], 2);

			testService.RemoveDelegate(this);

		}

		[Test]
		public void TestValueAsNumbersSi() {
			this.testTrigger = new AutoResetEvent(false);
			var numberOfBytes = 4;
			var numberOfDatasets = 2;
			var connectInfo = new ConnectInfo() { ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			var float1 = 12345.6f;
			var float2 = -12345.6f;
			var floatList = new List<byte>();
			floatList.AddRange(BitConverter.GetBytes(float1));
			floatList.AddRange(BitConverter.GetBytes(float2));
			var success = genericService.HandleUpdatedValueData(floatList.ToArray(), out error);

			Assert.True(success);
			Assert.AreEqual(genericService.NumbersFromValueData.Count, dataFormat.DataSetCount);
			Assert.AreEqual(BitConverter.ToSingle(genericService.NumbersFromValueData[0], 0), float1);
			Assert.AreEqual(BitConverter.ToSingle(genericService.NumbersFromValueData[1], 0), float2);
			testService.RemoveDelegate(this);

		}

		[Test]
		public void TestValueAsSignedInt() {
			testTrigger = new AutoResetEvent(false);
			var numberOfBytes = 1;
			var numberOfDatasets = 1;
			var connectInfo = new ConnectInfo() { ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			var int8 = -127;
			var success = genericService.HandleUpdatedValueData(new byte[] {(byte)int8}, out error);
			Assert.True(success);
			Assert.AreEqual(testService.ValueAsInteger, int8);

			numberOfBytes = 2;
			numberOfDatasets = 1;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			testTrigger.WaitOne(1000);

			short int16 = -12725;
			success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(int16), out error);
			Assert.True(success);
			Assert.AreEqual(testService.ValueAsInteger, int16);

			numberOfBytes = 4;
			numberOfDatasets = 1;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			testTrigger.WaitOne(1000);

			int int32 = -12725456;
			success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(int32), out error);
			Assert.True(success);
			Assert.AreEqual(testService.ValueAsInteger, int32);
			
			testService.RemoveDelegate(this);

		}

		[Test]
		public void TestValueAsSignedIntValueZeroIfDataTooBig() {
			testTrigger = new AutoResetEvent(false);
			var numberOfBytes = 4;
			var numberOfDatasets = 1;
			var connectInfo = new ConnectInfo() {ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision()};
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint) numberOfBytes, (uint) numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes*numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			long longInt = -2552525612345644;
			var success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(longInt), out error);
			Assert.AreEqual(testService.ValueAsInteger, 0);
			Assert.NotNull(error);
		}

		[Test]
		public void TestValueAsSignedFloat() {
			testTrigger = new AutoResetEvent(false);
			var numberOfBytes = 4;
			var numberOfDatasets = 1;
			var connectInfo = new ConnectInfo() { ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			var positiveFloat = 12345.12f;
			var success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(positiveFloat), out error);
			Assert.True(success);
			Assert.AreEqual(testService.ValueAsFloat, positiveFloat);

			numberOfBytes = 4;
			numberOfDatasets = 1;
			testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			testTrigger.WaitOne(1000);

			var negativeFloat = -12345.12f;
			success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(negativeFloat), out error);
			Assert.True(success);
			Assert.AreEqual(testService.ValueAsFloat, negativeFloat);


			testService.RemoveDelegate(this);

		}

		[Test]
		public void TestValueAsFloatValueZeroIfDataTooBig() {
			testTrigger = new AutoResetEvent(false);
			var numberOfBytes = 4;
			var numberOfDatasets = 1;
			var connectInfo = new ConnectInfo() { ConnectId = 0, HubIndex = 0, TypeEnum = IoType.IoTypeGeneric, FirmwareVersion = new Revision(), HardwareVersion = new Revision() };
			var io = new BluetoothServiceIoMock();
			var testService = BluetoothService.GetServiceFromConnectInfo(new BluetoothDevice("0", null, BluetoothDeviceManager.SharedDeviceManager), connectInfo, io);
			var dataFormat = DataFormat.FormatWithModeName("Count", 0, InputFormatUnit.InputFormatUnitSi, (uint)numberOfBytes, (uint)numberOfDatasets);
			testService.AddValidDataFormat(dataFormat);
			var inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 0, 1, InputFormatUnit.InputFormatUnitSi, true);
			inputFormat.NumberOfBytes = numberOfBytes * numberOfDatasets;
			testService.UpdateInputFormat(inputFormat);
			testService.AddDelegate(this);
			var genericService = testService as BluetoothGenericService;
			Assert.NotNull(genericService);
			SdkError error;
			double doubleFloat = -123451654564.12;
			var success = genericService.HandleUpdatedValueData(BitConverter.GetBytes(doubleFloat), out error);
			Assert.AreEqual(testService.ValueAsFloat, 0);
			Assert.NotNull(error);
		}


		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
			receivedOldInputFormat = oldFormat;
			receivedInputFormat = newFormat;
			testTrigger.Set();
		}

		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
			receivedOldData = oldData;
			receivedNewData = newData;
			testTrigger.Set();
		}



	}
}
