﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Generic.Data;
using LegoDeviceSDK.Interfaces;
using LegoSDKNunitTest.SDK.Bluetooth.Mocks;
using NUnit.Core;
using NUnit.Framework;
using IService = LegoDeviceSDK.Interfaces.IService;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoSDKNunitTest.SDK.Bluetooth.Services {
	[TestFixture]
	public class BluetoothTiltSensorTest : ITiltSensorDelegate {

		AutoResetEvent testTrigger;

		BluetoothTiltSensor tiltSensor;
		TiltSensorAngle oldReceivedAngle;
		TiltSensorAngle newReceivedAngle;

		TiltSensorCrash oldReceivedCrash;
		TiltSensorCrash newReceivedCrash;

		TiltSensorDirection oldReceivedDirection;
		TiltSensorDirection newReceivedDirection;

		ConnectInfo connectInfo = new ConnectInfo() { ConnectId = 1, HubIndex = 1, TypeEnum = IoType.IoTypeTiltSensor };
		IIo ioStub = new BluetoothServiceIoMock();

		[SetUp]
		public void Setup() {
			testTrigger = new AutoResetEvent(false);
			ioStub = new BluetoothServiceIoMock();
			var service = BluetoothService.GetServiceFromConnectInfo(new BluetoothDeviceStub(), connectInfo, ioStub);
			tiltSensor = service as BluetoothTiltSensor;
			if (tiltSensor == null)
				Assert.Fail();
			tiltSensor.AddDelegate(this);
		}

		[TearDown]
		public void TearDown() {
			tiltSensor.RemoveDelegate(this);
			tiltSensor = null;
		}

		void UpdateInputFormatMode(TiltSensorMode mode, InputFormatUnit unit, int numberOfBytes) {
			var data = BluetoothTestStupFactory.InputFormatWriteDataWithRevision(1, tiltSensor.ConnectInfo.ConnectId, tiltSensor.ConnectInfo.TypeEnum, (int)mode, 1, (int)unit, true, numberOfBytes);

			var inputFormat = InputFormat.InputFormatFromByteArray(data);
			tiltSensor.HandleUpdatedInputFormat(inputFormat);
		}

		void UpdateValueWithData(byte[] data) {
			SdkError error;
			tiltSensor.HandleUpdatedValueData(data, out error);
			Assert.Null(error);
		}


		#region Angle Tests
		[Test]
		public void TestHandleUpdatedValueAngleModeRawSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.TiltSensorModeAngle, InputFormatUnit.InputFormatUnitRaw, 2);

			UpdateValueWithData(new byte[] {1,2});
			testTrigger.WaitOne(1000);
			var expectedResult = new TiltSensorAngle() {X = 1, Y = 2};
			Assert.True(expectedResult.Equals(tiltSensor.Angle));
			Assert.True(expectedResult.Equals(newReceivedAngle));
		}

		[Test]
		public void TestHandleUpdatedValueAngleModeSiSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.TiltSensorModeAngle, InputFormatUnit.InputFormatUnitSi, 8);

			var expectedResult = new TiltSensorAngle() { X = 10.10f, Y = 20.20f };
			var data = new List<byte>();
			data.AddRange(BitConverter.GetBytes(expectedResult.X));
			data.AddRange(BitConverter.GetBytes(expectedResult.Y));
			UpdateValueWithData(data.ToArray());
		
			testTrigger.WaitOne(1000);

			Assert.True(expectedResult.Equals(tiltSensor.Angle));
			Assert.True(expectedResult.Equals(newReceivedAngle));
		}

		[Test]
		public void TestHandleUpdatedValueAngleModePercentSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.TiltSensorModeAngle, InputFormatUnit.InputFormatUnitPercentage, 2);

			var expectedResult = new TiltSensorAngle() { X = 32, Y = 64 };
			UpdateValueWithData(new byte[] { 32, 64 });

			testTrigger.WaitOne(1000);

			Assert.True(expectedResult.Equals(tiltSensor.Angle));
			Assert.True(expectedResult.Equals(newReceivedAngle));
		}

		#endregion

		#region Direction Tests
		[Test]
		public void TestHandleUpdatedValueTiltModeRawSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.TiltSensorModeTilt, InputFormatUnit.InputFormatUnitRaw, 1);

			UpdateValueWithData(new byte[] { 9 });
			testTrigger.WaitOne(1000);
			Assert.True(TiltSensorDirection.TiltSensorDirectionForward.Equals(tiltSensor.Direction));
			Assert.True(TiltSensorDirection.TiltSensorDirectionForward.Equals(newReceivedDirection));
		}

		[Test]
		public void TestHandleUpdatedValueAllTiltDirectionsRaw() {
			UpdateInputFormatMode(TiltSensorMode.TiltSensorModeTilt, InputFormatUnit.InputFormatUnitRaw, 1);

			TestHandleUpdatedValue(new byte[] { 0 }, TiltSensorDirection.TiltSensorDirectionNeutral);
			TestHandleUpdatedValue(new byte[] { 3 }, TiltSensorDirection.TiltSensorDirectionBackward);
            TestHandleUpdatedValue(new byte[] { 5 }, TiltSensorDirection.TiltSensorDirectionRight);
            TestHandleUpdatedValue(new byte[] { 7 }, TiltSensorDirection.TiltSensorDirectionLeft);
			TestHandleUpdatedValue(new byte[] { 9 }, TiltSensorDirection.TiltSensorDirectionForward);
			TestHandleUpdatedValue(new byte[] { 10 }, TiltSensorDirection.TiltSensorDirectionUnknown);
		}

		void TestHandleUpdatedValue(byte[] data, TiltSensorDirection direction) {
			UpdateValueWithData(data);
			testTrigger.WaitOne(1000);
			Assert.True(direction.Equals(tiltSensor.Direction));
			Assert.True(direction.Equals(newReceivedDirection));
			
		}
		#endregion

		#region Crash Tests

		[Test]
		public void TestHandleUpdatedValueCrashModeRawSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.LeTiltSensorModeCrash, InputFormatUnit.InputFormatUnitRaw, 3);

			var expectedResult = new TiltSensorCrash() { X = 10, Y = 20, Z = 30};
			UpdateValueWithData(new byte[] {10, 20, 30});

			testTrigger.WaitOne(1000);

			Assert.True(expectedResult.Equals(tiltSensor.Crash));
			Assert.True(expectedResult.Equals(newReceivedCrash));
		}

		[Test]
		public void TestHandleUpdatedValueCrashModeSiSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.LeTiltSensorModeCrash, InputFormatUnit.InputFormatUnitSi, 12);

			var expectedResult = new TiltSensorCrash() { X = 10.10f, Y = 20.20f, Z = 30.30f };
			var data = new List<byte>();
			data.AddRange(BitConverter.GetBytes(expectedResult.X));
			data.AddRange(BitConverter.GetBytes(expectedResult.Y));
			data.AddRange(BitConverter.GetBytes(expectedResult.Z));
			UpdateValueWithData(data.ToArray());

			testTrigger.WaitOne(1000);

			Assert.True(expectedResult.Equals(tiltSensor.Crash));
			Assert.True(expectedResult.Equals(newReceivedCrash));
		}

		[Test]
		public void TestHandleUpdatedValueCrashModePercentSuccessScenarion() {
			UpdateInputFormatMode(TiltSensorMode.LeTiltSensorModeCrash, InputFormatUnit.InputFormatUnitPercentage, 3);

			var expectedResult = new TiltSensorCrash() { X = 32, Y = 64, Z = 96 };
			UpdateValueWithData(new byte[] { 32, 64, 96 });

			testTrigger.WaitOne(1000);

			Assert.True(expectedResult.Equals(tiltSensor.Crash));
			Assert.True(expectedResult.Equals(newReceivedCrash));
		}
		#endregion

		#region Callbacks

		public void DidUpdateInputFormat(IService service, InputFormat oldFormat, InputFormat newFormat) {
		}

		public void DidUpdateValueData(IService service, byte[] oldData, byte[] newData) {
		}

		public void DidUpdateAngle(ITiltSensor tiltSensor1, TiltSensorAngle oldAngle, TiltSensorAngle newAngle) {
			oldReceivedAngle = oldAngle;
			newReceivedAngle = newAngle;
			testTrigger.Set();
		}

        public void DidUpdateCrash(ITiltSensor tiltSensor1, TiltSensorCrash oldCrashValue, TiltSensorCrash newCrashValue) {
			oldReceivedCrash = oldCrashValue;
			newReceivedCrash = newCrashValue;
			testTrigger.Set();
		}

        public void DidUpdateDirection(ITiltSensor tiltSensor1, TiltSensorDirection oldDirection, TiltSensorDirection newDirection) {
			oldReceivedDirection = oldDirection;
			newReceivedDirection = newDirection;
			testTrigger.Set();
		}

		#endregion
	}
}
