﻿using NUnit.Framework;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LegoDeviceSDK.Wrapper;
using Moq;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using LEGODeviceUnitySDK;

namespace LegoSDKNunitTest.SDK.Wrapper
{
    [TestFixture]
    public class DeviceManagerWrapperTest 
    {
        private DeviceManagerWrapper deviceManagerWrapper;
        private Mock<IDeviceManager> deviceManagerMock = new Mock<IDeviceManager>();

        private Mock<IDevice> deviceMock = new Mock<IDevice>();
        private string deviceID = "TestDevice-123";

        private Mock<IDevice> deviceMock2 = new Mock<IDevice>();

        private Mock<WrapperSerialization> serializationMock = new Mock<WrapperSerialization>();
        private Mock<ILEGODeviceManagerNativeDelegate> nativeDelegateMock = new Mock<ILEGODeviceManagerNativeDelegate>();
        private Mock<IDeviceWrapperImpl> deviceWrapperMock = new Mock<IDeviceWrapperImpl>();

        [SetUp]
        public void SetUp()
        {
            deviceManagerWrapper = new DeviceManagerWrapper(deviceManagerMock.Object, serializationMock.Object);
            deviceManagerWrapper.UnityDeviceManagerDelegate = nativeDelegateMock.Object;
            deviceManagerWrapper.DeviceWrapper = deviceWrapperMock.Object;

            deviceMock.Setup(d => d.DeviceId).Returns(deviceID);
            deviceMock2.Setup(d => d.DeviceId).Returns("AnotherDevice-123456789");

            // we need a partial mock - call the original implementation unless explicitly overriden
            serializationMock.CallBase = true;
            // we need to return some data when serializing devices since WrapperSerialization expects that internally
            serializationMock.Setup(s => s.SerializeDevice(It.IsAny<IDevice>())).Returns(new Dictionary<string, object>());
        }

        #region API Methods
        [Test]
        public void TestStartScanning() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.Scan();

            // ASSERT
            deviceManagerMock.Verify(manager => manager.ScanAsync());
        }

        [Test]
        public void TestConnectToDevice() 
        {
            // ARRANGE
            deviceManagerMock.Setup(m => m.AllDevices).Returns(new List<IDevice>() { deviceMock.Object });

            // ACT
            deviceManagerWrapper.ConnectToDevice(deviceMock.Object.DeviceId);

            // ASSERT
            deviceManagerMock.Verify(m => m.ConnectToDeviceAsync(deviceMock.Object));
        }

        [Test]
        public void TestDisconnectDevice()
        {
            // ARRANGE
            deviceManagerMock.Setup(m => m.AllDevices).Returns(new List<IDevice>() { deviceMock.Object });

            // ACT
            deviceManagerWrapper.DisconnectDevice(deviceMock.Object.DeviceId);

            // ASSERT
            deviceManagerMock.Verify(m => m.CancelDeviceConnection(deviceMock.Object));
        }

        [Test]
        public void TestAllDevices()
        {
            // ARRANGE
            deviceManagerMock.Setup(m => m.AllDevices).Returns(new List<IDevice>() { deviceMock.Object, deviceMock2.Object });

            // ACT
            string allDevices = deviceManagerWrapper.AllDevices();

            // ASSERT
            Assert.IsTrue(allDevices != null && allDevices.Length > 0);
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            serializationMock.Verify(s => s.SerializeDevice(deviceMock2.Object));
        }
        #endregion

        #region Callbacks/events
        [Test]
        public void TestOnDeviceAppeared() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DeviceDidAppear(deviceManagerMock.Object, deviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDeviceDidAppear(It.IsAny<string>()));
        }

        [Test]
        public void TestOnDeviceDisappeared() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DeviceDidDisappear(deviceManagerMock.Object, deviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDeviceDidDisappear(It.IsAny<string>()));
        }

        [Test]
        public void TestOnDidDisconnectFromDevice() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DidDisconnectFromDevice(deviceManagerMock.Object, deviceMock.Object, false);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDidDisconnectFromDevice(It.IsAny<string>()));
        }

        [Test]
        public void TestOnDidFailToConnectToDevice() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DidFailToConnectToDevice(deviceManagerMock.Object, deviceMock.Object, false);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDidFailToConnectToDevice(It.IsAny<string>()));
        }

        [Test]
        public void TestOnDidStartInterrogatingDevice() 
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DidStartInterrogatingDevice(deviceManagerMock.Object, deviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDidStartInterrogatingDevice(It.IsAny<string>()));
        }

        [Test]
        public void TestOnDidFinishInterrogatingDevice()
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.DidFinishInterrogatingDevice(deviceManagerMock.Object, deviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerDidFinishInterrogatingDevice(It.IsAny<string>()));
        }


        [Test]
        public void TestOnWillStartConnectingToDevice()
        {
            // ARRANGE

            // ACT
            deviceManagerWrapper.WillStartConnectingToDevice(deviceManagerMock.Object, deviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceManagerWillStartConnectingToDevice(It.IsAny<string>()));
        }
        #endregion
    }

}
