﻿using NUnit.Framework;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LegoDeviceSDK.Wrapper;
using Moq;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using LEGODeviceUnitySDK;
using LegoDeviceSDK.Generic;

namespace LegoSDKNunitTest.SDK.Wrapper
{
    [TestFixture]
    public class DeviceWrapperTest
    {
        private DeviceWrapper deviceWrapper;

        private Mock<IDevice> deviceMock = new Mock<IDevice>();
        private string deviceID = "TestDevice-123";

        private Mock<IService> serviceMock = new Mock<IService>();

        private Mock<WrapperSerialization> serializationMock = new Mock<WrapperSerialization>();
        private Mock<ILEGODeviceNativeDelegate> nativeDelegateMock = new Mock<ILEGODeviceNativeDelegate>();

        private Mock<IDeviceManagerWrapperImpl> deviceManagerWrapperMock = new Mock<IDeviceManagerWrapperImpl>();
        private Mock<IServiceWrapperImpl> serviceWrapperMock = new Mock<IServiceWrapperImpl>();

        [SetUp]
        public void SetUp() 
        {
            deviceWrapper = new DeviceWrapper(deviceManagerWrapperMock.Object, serializationMock.Object);
            deviceWrapper.UnityDeviceDelegate = nativeDelegateMock.Object;
            deviceWrapper.ServiceWrapper = serviceWrapperMock.Object;

            deviceMock.Setup(d => d.DeviceId).Returns(deviceID);
            serializationMock.Setup(s => s.SerializeDevice(It.IsAny<IDevice>())).Returns(new Dictionary<string, object>());
            serializationMock.Setup(s => s.SerializeService(serviceMock.Object, It.IsAny<bool>())).Returns(new Dictionary<string, object>());
        }

        #region API Methods
        [Test]
        public void TestChangeDeviceName()
        {
            // ARRANGE
            string newName = "FooBar-1";
            deviceManagerWrapperMock.Setup(m => m.FindDevice(deviceID)).Returns(deviceMock.Object);

            // ACT
            deviceWrapper.ChangeDeviceName(deviceMock.Object.DeviceId, newName);

            // ASSERT
            deviceMock.VerifySet(d => d.Name = newName);
        }
        #endregion

        #region Callbacks/events
        [Test]
        public void TestDidAddService() 
        {
            // ARRANGE

            // ACT
            deviceWrapper.DidAddService(deviceMock.Object, serviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeService(serviceMock.Object, false));
            nativeDelegateMock.Verify(d => d.LEDeviceDidAddService(It.IsAny<string>()));
        }

        [Test]
        public void TestDidRemoveService() 
        {
            // ARRANGE

            // ACT
            deviceWrapper.DidRemoveService(deviceMock.Object, serviceMock.Object);

            // ASSERT
            serializationMock.Verify(s => s.SerializeService(serviceMock.Object, true));
            nativeDelegateMock.Verify(d => d.LEDeviceDidRemoveService(It.IsAny<string>()));
        }

        [Test]
        public void TestDidChangeButtonState() 
        {
            // ARRANGE
            bool buttonPressed = true;

            // ACT
            deviceWrapper.DidChangeButtonState(deviceMock.Object, buttonPressed);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDeviceButtonStateChange(deviceMock.Object, buttonPressed));
            nativeDelegateMock.Verify(d => d.LEDeviceDidChangeButtonState(It.IsAny<string>()));
        }

        [Test]
        public void TestDidChangeName()
        {
            // ARRANGE
            string oldName = "OldName";
            string newName = "NewName";

            // ACT
            deviceWrapper.DidChangeName(deviceMock.Object, oldName, newName);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDeviceNameChange(deviceMock.Object, oldName, newName));
            nativeDelegateMock.Verify(d => d.LEDeviceDidChangeName(It.IsAny<string>()));
        }

        [Test]
        public void TestDidFailToAddServiceWithError()
        {
            // ARRANGE
            Exception error = new Exception("some error");

            // ACT
            deviceWrapper.DidFailToAddServiceWithError(deviceMock.Object, error);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDeviceError(deviceMock.Object, error));
            nativeDelegateMock.Verify(d => d.LEDeviceDidFailToAddServiceWithError(It.IsAny<string>()));
        }

        [Test]
        public void TestDidUpdateBatteryLevel()
        {
            // ARRANGE
            int batteryLevel = 72;

            // ACT
            deviceWrapper.DidUpdateBatteryLevel(deviceMock.Object, (uint)batteryLevel);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDeviceBatteryLevel(deviceMock.Object, batteryLevel));
            nativeDelegateMock.Verify(d => d.LEDeviceDidUpdateBatteryLevel(It.IsAny<string>()));
        }

        [Test]
        public void TestDidUpdateDeviceInfo()
        {
            // ARRANGE
            DeviceInfo deviceInfo = new DeviceInfo();

            // ACT
            deviceWrapper.DidUpdateDeviceInfo(deviceMock.Object, deviceInfo, null);

            // ASSERT
            serializationMock.Verify(s => s.SerializeDevice(deviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEDeviceDidUpdateDeviceInfo(It.IsAny<string>()));
        }
        #endregion
    }

}
