﻿using NUnit.Framework;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LegoDeviceSDK.Wrapper;
using Moq;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using LEGODeviceUnitySDK;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic.Data;
using Windows.UI;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoSDKNunitTest.SDK.Wrapper 
{
    [TestFixture]
    public class ServiceWrapperTest
    {
        private ServiceWrapper serviceWrapper;

        private Mock<IDevice> deviceMock = new Mock<IDevice>();
        private string deviceID = "TestDevice-123";

        private string serviceConnectID = "1";

        private Mock<WrapperSerialization> serializationMock = new Mock<WrapperSerialization>();
        private Mock<ILEGOServiceNativeDelegate> nativeDelegateMock = new Mock<ILEGOServiceNativeDelegate>();
        private Mock<IDeviceManagerWrapperImpl> deviceManagerWrapperMock = new Mock<IDeviceManagerWrapperImpl>();

        [SetUp]
        public void SetUp()
        {
            serviceWrapper = new ServiceWrapper(deviceManagerWrapperMock.Object, serializationMock.Object);
            serviceWrapper.UnityServiceDelegate = nativeDelegateMock.Object;

            serializationMock.CallBase = true;
        }

        private Mock<T> PrepareServiceMock<T>() where T : class, IService
        {
            ConnectInfo connectInfo = new ConnectInfo() 
            {
                ConnectId = int.Parse(this.serviceConnectID),
                HubIndex = 0,
                Type = (int)IoType.IoTypeGeneric
            };

            Mock<IDevice> deviceMock = new Mock<IDevice>();
            deviceMock.Setup(d => d.DeviceId).Returns(deviceID);

            Mock<T> serviceMock = new Mock<T>();
            serviceMock.Setup(s => s.ConnectInfo).Returns(connectInfo);
            serviceMock.Setup(s => s.Device).Returns(deviceMock.Object);
    
            deviceMock.Setup(d => d.Services).Returns(new List<IService>() { serviceMock.Object });

            deviceManagerWrapperMock.Setup(m => m.FindDevice(deviceID)).Returns(deviceMock.Object);
            deviceManagerWrapperMock.Setup(m => m.FindService(deviceID, serviceConnectID)).Returns(serviceMock.Object);

            serializationMock.Setup(s => s.SerializeServiceWithData(serviceMock.Object)).Returns(new Dictionary<string, object>());

            return serviceMock;
        } 

        #region API Methods
        [Test]
        public void TestUpdateServiceData()
        {
            // ARRANGE
            Mock<IService> serviceMock = PrepareServiceMock<IService>();

            // ACT
            serviceWrapper.UpdateServiceData(deviceID, serviceConnectID);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceWithData(serviceMock.Object));
            nativeDelegateMock.Verify(d => d.LEServiceUpdateServiceData(It.IsAny<string>()));
        }

        [Test]
        public void TestUpdateInputFormat()
        {
            // ARRANGE
            Mock<IService> serviceMock = PrepareServiceMock<IService>();

            int mode = 2;
            int deltaInterval = 15;
            InputFormatUnit unit = InputFormatUnit.InputFormatUnitRaw;
            bool notificationsEnabled = true;

            InputFormat inputFormat = null;
            serviceMock.Setup(s => s.UpdateInputFormat(It.IsAny<InputFormat>())).Callback<InputFormat>(f => inputFormat = f);

            // ACT
            serviceWrapper.UpdateInputFormat(deviceID, serviceConnectID, mode.ToString(), ((int)unit).ToString(), deltaInterval.ToString(), notificationsEnabled);

            // ASSERT
            serviceMock.Verify(s => s.UpdateInputFormat(It.IsAny<InputFormat>()));
            Assert.AreEqual(mode, inputFormat.Mode);
            Assert.AreEqual(deltaInterval, inputFormat.DeltaInterval);
            Assert.AreEqual(unit, inputFormat.Unit);
            Assert.AreEqual(notificationsEnabled, inputFormat.NotificationsEnabled);
        }

        [Test]
        public void TestMotorRun()
        {
            // ARRANGE
            Mock<IMotor> serviceMock = PrepareServiceMock<IMotor>();
            int power = 77;
            MotorDirection direction = MotorDirection.MotorDirectionLeft;

            // ACT
            serviceWrapper.Run(deviceID, serviceConnectID, ((int)direction).ToString(), power.ToString());

            // ASSERT
            serviceMock.Verify(s => ((IMotor)s).RunInDirection(direction, power));
        }

        [Test]
        public void TestMotorBrake()
        {
            // ARRANGE
            Mock<IMotor> serviceMock = PrepareServiceMock<IMotor>();

            // ACT
            serviceWrapper.Brake(deviceID, serviceConnectID);

            // ASSERT
            serviceMock.Verify(s => ((IMotor)s).Brake());
        }

        [Test]
        public void TestMotorDrift()
        {
            // ARRANGE
            Mock<IMotor> serviceMock = PrepareServiceMock<IMotor>();

            // ACT
            serviceWrapper.Drift(deviceID, serviceConnectID);

            // ASSERT
            serviceMock.Verify(s => ((IMotor)s).Drift());
        }

        [Test]
        public void TestLightSwitchOff()
        {
            // ARRANGE
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();

            // ACT
            serviceWrapper.SwitchOff(deviceID, serviceConnectID);

            // ASSERT
            serviceMock.Verify(s => ((IRgbLight)s).SwitchOff());
        }

        [Test]
        public void TestLightChangeColorToDefault() {
            // ARRANGE
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();

            // ACT
            serviceWrapper.ChangeColorToDefault(deviceID, serviceConnectID);

            // ASSERT
            serviceMock.Verify(s => ((IRgbLight)s).SwitchToDefaultColor());
        }


        [Test]
        public void TestLightChangeColor()
        {
            // ARRANGE
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();
            Color color = Color.FromArgb(255, 255, 120, 50);
            string colorJSONString = serializationMock.Object.ConvertToJSONString(serializationMock.Object.SerializeColor(color));

            // ACT
            serviceWrapper.ChangeColor(deviceID, serviceConnectID, colorJSONString);

            // ASSERT
            serviceMock.VerifySet(s => ((IRgbLight)s).Color = color);
        }

        [Test]
        public void TestLightChangeColorIndex() 
        {
            // ARRANGE
            int colorIndex = 7;
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();

            // ACT
            serviceWrapper.ChangeColorIndex(deviceID, serviceConnectID, colorIndex.ToString());

            // ASSERT
            serviceMock.VerifySet(s => ((IRgbLight)s).ColorIndex = colorIndex);
        }

        [Test]
        public void TestLightSetRGBMode()
        {
            // ARRANGE
            RgbLightMode mode = RgbLightMode.Discrete;
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();

            // ACT
            serviceWrapper.SetRGBLightMode(deviceID, serviceConnectID, ((int)mode).ToString());

            // ASSERT
            serviceMock.VerifySet(s => ((IRgbLight)s).RGBMode = mode);
        }

        [Test]
        public void TestMotionSensorSetMode()
        {
            // ARRANGE
            Mock<IMotionSensor> serviceMock = PrepareServiceMock<IMotionSensor>();
            MotionSensorMode mode = MotionSensorMode.Count;

            // ACT
            serviceWrapper.SetMotionSensorMode(deviceID, serviceConnectID, ((int)mode).ToString());

            // ASSERT
            serviceMock.VerifySet(s => ((IMotionSensor)s).MotionSensorMode = mode);
        }

        [Test]
        public void TestTiltSensorSetMode()
        {
            // ARRANGE
            Mock<ITiltSensor> serviceMock = PrepareServiceMock<ITiltSensor>();
            TiltSensorMode mode = TiltSensorMode.LeTiltSensorModeCrash;

            // ACT
            serviceWrapper.SetTiltSensorMode(deviceID, serviceConnectID, ((int)mode).ToString());

            // ASSERT
            serviceMock.VerifySet(s => s.TiltSensorMode = mode);
        }

        [Test]
        public void TestPiezoPlayerPlayFrequency()
        {
            // ARRANGE
            Mock<IPiezoTonePlayer> serviceMock = PrepareServiceMock<IPiezoTonePlayer>();
            int frequency = 5000;
            int milliseconds = 350;

            // ACT
            serviceWrapper.PlayFrequency(deviceID, serviceConnectID, frequency.ToString(), milliseconds.ToString());

            // ASSERT
            serviceMock.Verify(s => ((IPiezoTonePlayer)s).PlayFrequency(frequency, milliseconds));
        }

        [Test]
        public void TestPiezoPlayerPlayNote()
        {
            // ARRANGE
            Mock<IPiezoTonePlayer> serviceMock = PrepareServiceMock<IPiezoTonePlayer>();
            PiezoTonePlayerNote note = PiezoTonePlayerNote.PiezoNoteB;
            int octave = 2;
            int milliseconds = 350;

            // ACT
            serviceWrapper.PlayNote(deviceID, serviceConnectID, ((int)note).ToString(), octave.ToString(), milliseconds.ToString());

            // ASSERT
            serviceMock.Verify(s => ((IPiezoTonePlayer)s).PlayNote(note, octave, milliseconds));
        }

        [Test]
        public void TestPiezoPlayerStopPlaying()
        {
            // ARRANGE
            Mock<IPiezoTonePlayer> serviceMock = PrepareServiceMock<IPiezoTonePlayer>();

            // ACT
            serviceWrapper.StopPlaying(deviceID, serviceConnectID);

            // ASSERT
            serviceMock.Verify(s => ((IPiezoTonePlayer)s).StopPlaying());
        }
        #endregion

        #region Callbacks/events
        [Test]
        public void TestServiceDidUpdateInputFormat()
        {
            // ARRANGE
            Mock<IService> serviceMock = PrepareServiceMock<IService>();
            InputFormat oldInputFormat = InputFormat.InputFormatWithConnectId(int.Parse(serviceConnectID), (int)IoType.IoTypeGeneric, 3, 10, InputFormatUnit.InputFormatUnitSi, true);
            InputFormat newInputFormat = InputFormat.InputFormatWithConnectId(int.Parse(serviceConnectID), (int)IoType.IoTypeMotor, 4, 30, InputFormatUnit.InputFormatUnitRaw, false);

            // ACT
            serviceWrapper.DidUpdateInputFormat(serviceMock.Object, oldInputFormat, newInputFormat);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceInputFormatChange(serviceMock.Object, oldInputFormat, newInputFormat));
            nativeDelegateMock.Verify(d => d.LEServiceDidUpdateInputFormat(It.IsAny<string>()));
        }

        [Test]
        public void TestServiceDidUpdateValueData()
        {
            // ARRANGE
            Mock<IService> serviceMock = PrepareServiceMock<IService>();
            byte[] oldData = { (byte)'a' };
            byte[] newData = { (byte)'a' };

            // ACT
            serviceWrapper.DidUpdateValueData(serviceMock.Object, oldData, newData);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceValueDataChange(serviceMock.Object, oldData, newData));
            nativeDelegateMock.Verify(d => d.LEServiceDidUpdateValueData(It.IsAny<string>()));
        }
                
        [Test]
        public void TestCurrentSensorDidUpdateMilliAmps()
        {
            // ARRANGE
            Mock<ICurrentSensor> serviceMock = PrepareServiceMock<ICurrentSensor>();
            float current = 3.14f;

            // ACT
            serviceWrapper.DidUpdateMilliAmp(serviceMock.Object, current);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceCurrentChange((ICurrentSensor)serviceMock.Object, current));
            nativeDelegateMock.Verify(d => d.LECurrentSensorDidUpdateMilliAmp(It.IsAny<string>()));
        }

        [Test]
        public void TestVoltageSensorDidUpdateMilliVolts()
        {
            // ARRANGE
            Mock<IVoltageSensor> serviceMock = PrepareServiceMock<IVoltageSensor>();
            float voltage = 1234.0f;

            // ACT
            serviceWrapper.DidUpdateMilliVolts(serviceMock.Object, voltage);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceVoltageChange((IVoltageSensor)serviceMock.Object, voltage));
            nativeDelegateMock.Verify(d => d.LEVoltageSensorDidUpdateMilliVolts(It.IsAny<string>()));
        }

        [Test]
        public void TestMotionSensorDidUpdateCount()
        {
            // ARRANGE
            Mock<IMotionSensor> serviceMock = PrepareServiceMock<IMotionSensor>();
            int count = 35;

            // ACT
            serviceWrapper.DidUpdateCount(serviceMock.Object, (uint)count);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceCountChange((IMotionSensor)serviceMock.Object, count));
            nativeDelegateMock.Verify(d => d.LEMotionSensorDidUpdateCount(It.IsAny<string>()));
        }

        [Test]
        public void TestMotionSensorDidUpdateDistance()
        {
            // ARRANGE
            Mock<IMotionSensor> serviceMock = PrepareServiceMock<IMotionSensor>();
            float oldDistance = 6;
            float newDistance = 8;

            // ACT
            serviceWrapper.DidUpdateDistance((IMotionSensor)serviceMock.Object, oldDistance, newDistance);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceDistanceChange((IMotionSensor)serviceMock.Object, oldDistance, newDistance));
            nativeDelegateMock.Verify(d => d.LEMotionSensorDidUpdateDistance(It.IsAny<string>()));
        }

        [Test]
        public void TestRGBLightDidUpdateColor()
        {
            // ARRANGE
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();
            Color oldColor = Color.FromArgb(255, 255, 255, 255);
            Color newColor = Color.FromArgb(255, 200, 100, 50);

            // ACT
            serviceWrapper.DidUpdateColor((IRgbLight)serviceMock.Object, oldColor, newColor);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceColorChange((IRgbLight)serviceMock.Object, oldColor, newColor));
            nativeDelegateMock.Verify(d => d.LERGBLightDidUpdateColor(It.IsAny<string>()));
        }

        [Test]
        public void TestRGBLightDidUpdateColorIndex() {
            // ARRANGE
            Mock<IRgbLight> serviceMock = PrepareServiceMock<IRgbLight>();
            int oldColorIndex = 1;
            int newColorIndex = 2;

            // ACT
            serviceWrapper.DidUpdateColorIndex((IRgbLight)serviceMock.Object, oldColorIndex, newColorIndex);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceColorIndexChange((IRgbLight)serviceMock.Object, oldColorIndex, newColorIndex));
            nativeDelegateMock.Verify(d => d.LERGBLightDidUpdateColorIndex(It.IsAny<string>()));
        }

        [Test]
        public void TestTiltSensorDidUpdateAngle()
        {
            // ARRANGE
            Mock<ITiltSensor> serviceMock = PrepareServiceMock<ITiltSensor>();
            TiltSensorAngle oldAngle = new TiltSensorAngle() { X = 10, Y = 10 };
            TiltSensorAngle newAngle = new TiltSensorAngle() { X = 20, Y = 20 };

            // ACT
            serviceWrapper.DidUpdateAngle((ITiltSensor)serviceMock.Object, oldAngle, newAngle);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceAngleChange((ITiltSensor)serviceMock.Object, oldAngle, newAngle));
            nativeDelegateMock.Verify(d => d.LETiltSensorDidUpdateAngle(It.IsAny<string>()));
        }

        [Test]
        public void TestTiltSensorDidUpdateCrash()
        {
            // ARRANGE
            Mock<ITiltSensor> serviceMock = PrepareServiceMock<ITiltSensor>();
            TiltSensorCrash oldCrash = new TiltSensorCrash() { X = 1, Y = 2, Z = 3 };
            TiltSensorCrash newCrash = new TiltSensorCrash() { X = 7, Y = 8, Z = 9 };

            // ACT
            serviceWrapper.DidUpdateCrash((ITiltSensor)serviceMock.Object, oldCrash, newCrash);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceCrashChange((ITiltSensor)serviceMock.Object, oldCrash, newCrash));
            nativeDelegateMock.Verify(d => d.LETiltSensorDidUpdateCrash(It.IsAny<string>()));
        }

        [Test]
        public void TestTiltSensorDidUpdateDirection()
        {
            // ARRANGE
            Mock<ITiltSensor> serviceMock = PrepareServiceMock<ITiltSensor>();
            TiltSensorDirection oldDirection = TiltSensorDirection.TiltSensorDirectionLeft;
            TiltSensorDirection newDirection = TiltSensorDirection.TiltSensorDirectionRight;

            // ACT
            serviceWrapper.DidUpdateDirection((ITiltSensor)serviceMock.Object, oldDirection, newDirection);

            // ASSERT
            serializationMock.Verify(s => s.SerializeServiceDirectionChange((ITiltSensor)serviceMock.Object, oldDirection, newDirection));
            nativeDelegateMock.Verify(d => d.LETiltSensorDidUpdateDirection(It.IsAny<string>()));
        }
        #endregion
    }

}
