﻿using NUnit.Framework;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using LegoDeviceSDK.Generic;
using LegoDeviceSDK.Interfaces;
using LegoDeviceSDK.Wrapper.Helpers;
using Moq;
using LegoDeviceSDK.Bluetooth;
using LegoDeviceSDK.Generic.Data;
using Windows.UI;
using LegoDeviceSDK.Interfaces.Services;

namespace LegoSDKNunitTest.SDK.Wrapper 
{
    [TestFixture]
    public class WrapperSerializationTest
    {      
        private string deviceID = "DeviceID";
        private Mock<IDevice> deviceMock;

        private string deviceID2 = "AnotherDevice-123";
        private Mock<IDevice> deviceMock2;

        private int serviceConnectID = 2;
        private Mock<IService> serviceMock;

        private WrapperSerialization wrapperSerialization = new WrapperSerialization();
        
        [SetUp]
        public void SetUp()
        {
            deviceMock = PrepareDeviceMock(deviceID);
            deviceMock2 = PrepareDeviceMock(deviceID2);
            serviceMock = PrepareServiceMock<IService>();
        }

        private Mock<IDevice> PrepareDeviceMock(String deviceID) 
        {
            DeviceInfo deviceInfo = new DeviceInfo();
            deviceInfo.FirmwareRevision = Revision.FromString(Encoding.UTF8.GetBytes("1.1.1.1"));
            deviceInfo.SoftwareRevision = Revision.FromString(Encoding.UTF8.GetBytes("2.2.2.2"));
            deviceInfo.HardwareRevision = Revision.FromString(Encoding.UTF8.GetBytes("3.3.3.3"));
            deviceInfo.ManufacturerName = "LEGO";

            Mock<IDevice> deviceMock = new Mock<IDevice>();
            deviceMock.Setup(d => d.DeviceInfo).Returns(deviceInfo);
            deviceMock.Setup(d => d.Name).Returns("DeviceName");
            deviceMock.Setup(d => d.DeviceId).Returns(deviceID);
            deviceMock.Setup(d => d.ButtonPressed).Returns(true);
            deviceMock.Setup(d => d.BatteryLevel).Returns(90);
            deviceMock.Setup(d => d.ConnectState).Returns(DeviceState.DeviceStateInterrogating);

            return deviceMock;
        }

        private Mock<T> PrepareServiceMock<T>() where T : class, IService
        {
            InputFormat inputFormat = InputFormat.InputFormatWithConnectId(1, 1, 3, 10, InputFormatUnit.InputFormatUnitSi, true);
            InputFormat defaultInputFormat = InputFormat.InputFormatWithConnectId(5, 20, 8, 30, InputFormatUnit.InputFormatUnitRaw, false);
            ConnectInfo connectInfo = new ConnectInfo() 
            { 
                ConnectId = this.serviceConnectID, 
                HubIndex = 1, 
                Type = (int)IoType.IoTypeGeneric, 
                HardwareVersion = Revision.FromString(Encoding.UTF8.GetBytes("1.1.1.1")), 
                FirmwareVersion = Revision.FromString(Encoding.UTF8.GetBytes("2.2.2.2"))
            };

            Mock<T> serviceMock = new Mock<T>();
            serviceMock.Setup(s => s.ServiceName).Returns("Test service");
            serviceMock.Setup(s => s.InputFormat).Returns(inputFormat);
            serviceMock.Setup(s => s.DefaultInputFormat).Returns(defaultInputFormat);
            serviceMock.Setup(s => s.IsInternalService).Returns(true);
            serviceMock.Setup(s => s.ConnectInfo).Returns(connectInfo);
            serviceMock.Setup(s => s.Device).Returns(this.deviceMock.Object);

            return serviceMock;
        }

        private void VerifySerializedDevice(Dictionary<string, object> serializedDevice, Mock<IDevice> mock, bool onlyBasic) 
        {
            if (mock != deviceMock && mock != deviceMock2)
                throw new ArgumentException("Unrecognized device mock given");

            Assert.AreEqual(mock == deviceMock ? deviceID : deviceID2, serializedDevice["DeviceID"]);
            if (onlyBasic)
                return;

            Assert.AreEqual("DeviceName", serializedDevice["DeviceName"]);
            Assert.AreEqual(true, serializedDevice["ButtonPressed"]);
            Assert.AreEqual(90, serializedDevice["BatteryLevel"]);
            Assert.AreEqual(3, serializedDevice["ConnectedState"]);

            Dictionary<string, object> serializedDeviceInfo = (Dictionary<string, object>)serializedDevice["DeviceInfo"];
            Assert.AreEqual("1.1.1.1", serializedDeviceInfo["FirmwareRevision"]);
            Assert.AreEqual("2.2.2.2", serializedDeviceInfo["SoftwareRevision"]);
            Assert.AreEqual("3.3.3.3", serializedDeviceInfo["HardwareRevision"]);
            Assert.AreEqual("LEGO", serializedDeviceInfo["ManufacturerName"]);
        }

        private void VerifySerializedService(Dictionary<string, object> serializedService, bool onlyBasic) 
        {
            if (onlyBasic) 
            {
                Assert.AreEqual(deviceID, serializedService["DeviceID"]);
                Assert.AreEqual(serviceConnectID, serializedService["ConnectID"]);
                return;
            }

            Assert.AreEqual("Test service", serializedService["ServiceName"]);
            Assert.AreEqual(true, serializedService["IsInternalService"]);
            Assert.AreEqual(deviceID, serializedService["DeviceID"]);

            Dictionary<string, object> connectInfo = (Dictionary<string, object>)serializedService["ConnectInfo"];
            Assert.AreEqual(serviceConnectID, connectInfo["ConnectID"]);
            Assert.AreEqual(1, connectInfo["HubIndex"]);
            Assert.AreEqual("1.1.1.1", connectInfo["HardwareRevision"]);
            Assert.AreEqual("2.2.2.2", connectInfo["SoftwareRevision"]);
            Assert.AreEqual((int)IoType.IoTypeGeneric, connectInfo["Type"]);

            Dictionary<string, object> inputFormat = (Dictionary<string, object>)serializedService["InputFormat"];
            Assert.AreEqual(0, inputFormat["Revision"]);
            Assert.AreEqual(1, inputFormat["ConnectID"]);
            Assert.AreEqual(1, inputFormat["TypeID"]);
            Assert.AreEqual(3, inputFormat["Mode"]);
            Assert.AreEqual(10, inputFormat["DeltaInterval"]);
            Assert.AreEqual((int)InputFormatUnit.InputFormatUnitSi, inputFormat["Unit"]);
            Assert.AreEqual(true, inputFormat["NotificationsEnabled"]);

            Dictionary<string, object> defaultInputFormat = (Dictionary<string, object>)serializedService["DefaultInputFormat"];
            Assert.AreEqual(0, defaultInputFormat["Revision"]);
            Assert.AreEqual(5, defaultInputFormat["ConnectID"]);
            Assert.AreEqual(20, defaultInputFormat["TypeID"]);
            Assert.AreEqual(8, defaultInputFormat["Mode"]);
            Assert.AreEqual(30, defaultInputFormat["DeltaInterval"]);
            Assert.AreEqual((int)InputFormatUnit.InputFormatUnitRaw, defaultInputFormat["Unit"]);
            Assert.AreEqual(false, defaultInputFormat["NotificationsEnabled"]);
        }

        [Test]
        public void TestDevice()
        {
            // ARRANGE
            Mock<IDevice> deviceMock = this.deviceMock;

            // ACT
            Dictionary<string, object> serializedDevice = wrapperSerialization.SerializeDevice(deviceMock.Object);

            // ASSERT
            VerifySerializedDevice(serializedDevice, deviceMock, false);
        }

        [Test]
        public void TestDevices()
        {
            // ARRANGE
            Mock<IDevice>[] deviceMocks = {this.deviceMock, this.deviceMock2};

            // ACT
            List<Dictionary<string, object>> serializedDevices = wrapperSerialization.SerializeDevices(deviceMocks.Select(d => d.Object).ToList());

            // ASSERT
            Assert.AreEqual(2, serializedDevices.Count);
            VerifySerializedDevice(serializedDevices[0], deviceMocks[0], false);
            VerifySerializedDevice(serializedDevices[1], deviceMocks[1], false);
        }

        [Test]
        public void TestDeviceNameChange()
        {
            // ARRANGE
            Mock<IDevice> deviceMock = this.deviceMock;
            string deviceOldName = "OldName";
            string deviceNewName = "NewName";

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeDeviceNameChange(deviceMock.Object, deviceOldName, deviceNewName);

            // ASSERT
            VerifySerializedDevice(serialized, deviceMock, true);
            Assert.AreEqual(deviceOldName, serialized["OldDeviceName"]);
            Assert.AreEqual(deviceNewName, serialized["DeviceName"]);
        }

        [Test]
        public void TestDeviceButtonStateChange()
        {
            // ARRANGE
            Mock<IDevice> deviceMock = this.deviceMock;
            bool buttonState = true;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeDeviceButtonStateChange(deviceMock.Object, buttonState);

            // ASSERT
            VerifySerializedDevice(serialized, deviceMock, true);
            Assert.AreEqual(buttonState, serialized["ButtonState"]);
        }

        [Test]
        public void TestDeviceBatteryLevel() 
        {
            // ARRANGE
            Mock<IDevice> deviceMock = this.deviceMock;
            int batteryLevel = 42;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeDeviceBatteryLevel(deviceMock.Object, batteryLevel);

            // ASSERT
            VerifySerializedDevice(serialized, deviceMock, true);
            Assert.AreEqual(batteryLevel, serialized["BatteryLevel"]);
        }

        [Test]
        public void TestDeviceError() 
        {
            // ARRANGE
            Mock<IDevice> deviceMock = this.deviceMock;
            string errorDescription = "some error";
            Exception error = new Exception(errorDescription);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeDeviceError(deviceMock.Object, error);

            // ASSERT
            VerifySerializedDevice(serialized, deviceMock, true);
            Assert.AreEqual(errorDescription, serialized["Error"]);
        }

        [Test]
        public void TestInputFormat()
        {
            // ARRANGE
            InputFormat inputFormat = InputFormat.InputFormatWithConnectId(1, (int)IoType.IoTypeGeneric, 3, 10, InputFormatUnit.InputFormatUnitSi, true);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeInputFormat(inputFormat);

            // ASSERT
            Assert.AreEqual(0, serialized["Revision"]);
            Assert.AreEqual(1, serialized["ConnectID"]);
            Assert.AreEqual((int)IoType.IoTypeGeneric, serialized["TypeID"]);
            Assert.AreEqual(3, serialized["Mode"]);
            Assert.AreEqual(10, serialized["DeltaInterval"]);
            Assert.AreEqual((int)InputFormatUnit.InputFormatUnitSi, serialized["Unit"]);
            Assert.AreEqual(true, serialized["NotificationsEnabled"]);
        }

        [Test]
        public void TestServiceFull()
        {
            // ARRANGE
            Mock<IService> serviceMock = this.serviceMock;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeService(serviceMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
        }

        [Test]
        public void TestServiceBasic()
        {
            // ARRANGE
            Mock<IService> serviceMock = this.serviceMock;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeService(serviceMock.Object, true);

            // ASSERT
            VerifySerializedService(serialized, true);
        }

        [Test]
        public void TestMotorServiceData()
        {
            // ARRANGE
            Mock<IMotor> motorMock = PrepareServiceMock<IMotor>();
            motorMock.Setup(motor => motor.Power).Returns(24);
            motorMock.Setup(motor => motor.Direction).Returns(MotorDirection.MotorDirectionLeft);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(motorMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];
            Assert.AreEqual(24, serviceData["Power"]);
            Assert.AreEqual(1, serviceData["MotorDirection"]);
        }

        [Test]
        public void TestRGBLightServiceData()
        {
            // ARRANGE
            Mock<IRgbLight> lightMock = PrepareServiceMock<IRgbLight>();
            lightMock.Setup(light => light.RGBMode).Returns(RgbLightMode.Absolute);
            lightMock.Setup(light => light.Color).Returns(Color.FromArgb(200, 120, 130, 140));
            lightMock.Setup(light => light.DefaultColor).Returns(Color.FromArgb(255, 200, 210, 220));
            lightMock.Setup(light => light.ColorIndex).Returns(13);
            lightMock.Setup(light => light.DefaultColorIndex).Returns(7);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(lightMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];

            Assert.AreEqual(1, serviceData["RGBLightMode"]);
            Assert.AreEqual(13, serviceData["ColorIndex"]);
            Assert.AreEqual(7, serviceData["DefaultColorIndex"]);

            Dictionary<string, object> color = (Dictionary<string, object>)serviceData["Color"];
            Assert.AreEqual(120 / 255.0, color["R"]);
            Assert.AreEqual(130 / 255.0, color["G"]);
            Assert.AreEqual(140 / 255.0, color["B"]);
            Assert.AreEqual(200 / 255.0, color["A"]);

            Dictionary<string, object> defaultColor = (Dictionary<string, object>)serviceData["DefaultColor"];
            Assert.AreEqual(200 / 255.0, defaultColor["R"]);
            Assert.AreEqual(210 / 255.0, defaultColor["G"]);
            Assert.AreEqual(220 / 255.0, defaultColor["B"]);
            Assert.AreEqual(255 / 255.0, defaultColor["A"]);
        }

        [Test]
        public void TestMotionSensorServiceData()
        {
            // ARRANGE
            Mock<IMotionSensor> motionSensorMock = PrepareServiceMock<IMotionSensor>();
            motionSensorMock.Setup(sensor => sensor.Count).Returns(2);
            motionSensorMock.Setup(sensor => sensor.Distance).Returns(3.14f);
            motionSensorMock.Setup(sensor => sensor.MotionSensorMode).Returns(MotionSensorMode.Count);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(motionSensorMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];
            Assert.AreEqual(2, serviceData["Count"]);
            Assert.AreEqual(3.14f, serviceData["Distance"]);
            Assert.AreEqual(1, serviceData["MotionSensorMode"]);
        }

        [Test]
        public void TestTiltSensorServiceData()
        {
            // ARRANGE
            Mock<ITiltSensor> tiltSensorMock = PrepareServiceMock<ITiltSensor>();
            tiltSensorMock.Setup(sensor => sensor.Direction).Returns(TiltSensorDirection.TiltSensorDirectionRight);
            tiltSensorMock.Setup(sensor => sensor.TiltSensorMode).Returns(TiltSensorMode.LeTiltSensorModeCrash);
            tiltSensorMock.Setup(sensor => sensor.Crash).Returns(new TiltSensorCrash() { X = 1, Y = 2, Z = 3 });
            tiltSensorMock.Setup(sensor => sensor.Angle).Returns(new TiltSensorAngle() { X = 20, Y = 30 });

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(tiltSensorMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];

            Assert.AreEqual(5, serviceData["TiltSensorDirection"]);
            Assert.AreEqual(2, serviceData["TiltSensorMode"]);

            Dictionary<string, object> crash = (Dictionary<string, object>)serviceData["Crash"];
            Assert.AreEqual(1, crash["X"]);
            Assert.AreEqual(2, crash["Y"]);
            Assert.AreEqual(3, crash["Z"]);

            Dictionary<string, object> angle = (Dictionary<string, object>)serviceData["Angle"];
            Assert.AreEqual(20.0f, angle["X"]);
            Assert.AreEqual(30.0f, angle["Y"]);
        }

        [Test]
        public void TestVoltageSensorServiceData()
        {
            // ARRANGE
            Mock<IVoltageSensor> voltageSensorMock = PrepareServiceMock<IVoltageSensor>();
            voltageSensorMock.Setup(sensor => sensor.MilliVolts).Returns(1234.25f);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(voltageSensorMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];
            Assert.AreEqual(1234.25f, serviceData["MilliVolts"]);
        }

        [Test]
        public void TestCurrentSensorServiceData()
        {
            // ARRANGE
            Mock<ICurrentSensor> currentSensorMock = PrepareServiceMock<ICurrentSensor>();
            currentSensorMock.Setup(sensor => sensor.MilliAmp).Returns(1234.25f);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(currentSensorMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Dictionary<string, object> serviceData = (Dictionary<string, object>)serialized["ServiceData"];
            Assert.AreEqual(1234.25f, serviceData["MilliAmp"]);
        }

        [Test]
        public void TestPiezoPlayerServiceData()
        {
            // ARRANGE
            Mock<IPiezoTonePlayer> piezoPlayerMock = PrepareServiceMock<IPiezoTonePlayer>();

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(piezoPlayerMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Assert.IsFalse(serialized.ContainsKey("ServiceData")); // No service data should be generated for piezo tone player, at least for now
        }

        [Test]
        public void TestUnknownServiceData() 
        {
            // ARRANGE
            Mock<IService> unknownServiceMock = PrepareServiceMock<IService>();

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceWithData(unknownServiceMock.Object);

            // ASSERT
            VerifySerializedService(serialized, false);
            Assert.IsFalse(serialized.ContainsKey("ServiceData")); // No service data should be generated for unknown services
        }

        [Test]
        public void TestServiceValueDataDidChange() 
        {
            // ARRANGE
            // the exact data doesn't really matter, we're only interested in different data pointers
            byte[] oldData = { (byte)'a' };
            byte[] newData = { (byte)'b' };
            Mock<IService> serviceMock = this.serviceMock;
            serviceMock.Setup(service => service.FloatFromData(oldData)).Returns(1.0f);
            serviceMock.Setup(service => service.FloatFromData(newData)).Returns(2.0f);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceValueDataChange(serviceMock.Object, oldData, newData);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(1.0f, serialized["OldValue"]);
            Assert.AreEqual(2.0f, serialized["NewValue"]);
        }

        [Test]
        public void TestServiceInputFormatDidChange()
        {
            // ARRANGE
            Mock<IService> serviceMock = this.serviceMock;
            InputFormat oldInputFormat = InputFormat.InputFormatWithConnectId(1, (int)IoType.IoTypeMotor, 3, 10, InputFormatUnit.InputFormatUnitSi, true);
            InputFormat newInputFormat = InputFormat.InputFormatWithConnectId(1, (int)IoType.IoTypeMotionSensor, 4, 30, InputFormatUnit.InputFormatUnitRaw, false);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceInputFormatChange(serviceMock.Object, oldInputFormat, newInputFormat);

            // ASSERT
            VerifySerializedService(serialized, true);

            Dictionary<string, object> serializedOldFormat = (Dictionary<string, object>)serialized["OldInputFormat"];
            Assert.AreEqual(0, serializedOldFormat["Revision"]);
            Assert.AreEqual(1, serializedOldFormat["ConnectID"]);
            Assert.AreEqual(1, serializedOldFormat["TypeID"]);
            Assert.AreEqual(3, serializedOldFormat["Mode"]);
            Assert.AreEqual(10, serializedOldFormat["DeltaInterval"]);
            Assert.AreEqual(2, serializedOldFormat["Unit"]);
            Assert.AreEqual(true, serializedOldFormat["NotificationsEnabled"]);

            Dictionary<string, object> serializedNewFormat = (Dictionary<string, object>)serialized["NewInputFormat"];
            Assert.AreEqual(0, serializedNewFormat["Revision"]);
            Assert.AreEqual(1, serializedNewFormat["ConnectID"]);
            Assert.AreEqual(35, serializedNewFormat["TypeID"]);
            Assert.AreEqual(4, serializedNewFormat["Mode"]);
            Assert.AreEqual(30, serializedNewFormat["DeltaInterval"]);
            Assert.AreEqual(0, serializedNewFormat["Unit"]);
            Assert.AreEqual(false, serializedNewFormat["NotificationsEnabled"]);
        }


        [Test]
        public void TestMotionSensorDistanceDidChange()
        {
            // ARRANGE
            Mock<IMotionSensor> motionSensorMock = PrepareServiceMock<IMotionSensor>();
            float oldDistance = 42.0f;
            float newDistance = 3.14f;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceDistanceChange(motionSensorMock.Object, oldDistance, newDistance);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(oldDistance, serialized["OldDistance"]);
            Assert.AreEqual(newDistance, serialized["NewDistance"]);
        }

        [Test]
        public void TestMotionSensorCountDidChange()
        {
            // ARRANGE
            Mock<IMotionSensor> motionSensorMock = PrepareServiceMock<IMotionSensor>();
            int count = 4;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceCountChange(motionSensorMock.Object, count);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(count, serialized["Count"]);
        }

        [Test]
        public void TestRGBLightColorDidChange() 
        {
            // ARRANGE
            Mock<IRgbLight> lightMock = PrepareServiceMock<IRgbLight>();
            Color oldColor = Color.FromArgb(255, 255, 255, 255);
            Color newColor = Color.FromArgb(255, 200, 100, 150);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceColorChange(lightMock.Object, oldColor, newColor);

            // ASSERT
            VerifySerializedService(serialized, true);

            Dictionary<string, object> serializeOldColor = (Dictionary<string, object>)serialized["OldColor"];
            Assert.AreEqual(oldColor.R / 255.0, serializeOldColor["R"]);
            Assert.AreEqual(oldColor.G / 255.0, serializeOldColor["G"]);
            Assert.AreEqual(oldColor.B / 255.0, serializeOldColor["B"]);
            Assert.AreEqual(oldColor.A / 255.0, serializeOldColor["A"]);

            Dictionary<string, object> serializeNewColor = (Dictionary<string, object>)serialized["NewColor"];
            Assert.AreEqual(newColor.R / 255.0, serializeNewColor["R"]);
            Assert.AreEqual(newColor.G / 255.0, serializeNewColor["G"]);
            Assert.AreEqual(newColor.B / 255.0, serializeNewColor["B"]);
            Assert.AreEqual(newColor.A / 255.0, serializeNewColor["A"]);
        }

        [Test]
        public void TestTiltSensorDirectionDidChange()
        {
            // ARRANGE
            Mock<ITiltSensor> tiltSensorMock = PrepareServiceMock<ITiltSensor>();

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceDirectionChange(tiltSensorMock.Object, TiltSensorDirection.TiltSensorDirectionLeft, TiltSensorDirection.TiltSensorDirectionForward);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(7, serialized["OldDirection"]);
            Assert.AreEqual(9, serialized["NewDirection"]);
        }

        [Test]
        public void TestTiltSensorAngleDidChange() 
        {
            // ARRANGE
            Mock<ITiltSensor> tiltSensorMock = PrepareServiceMock<ITiltSensor>();

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceAngleChange(tiltSensorMock.Object, new TiltSensorAngle() { X = 10, Y = 20 }, new TiltSensorAngle() { X = 15, Y = 15 });

            // ASSERT
            VerifySerializedService(serialized, true);

            Dictionary<string, object> serializedOldAngle = (Dictionary<string, object>)serialized["OldAngle"];
            Assert.AreEqual(10.0f, serializedOldAngle["X"]);
            Assert.AreEqual(20.0f, serializedOldAngle["Y"]);

            Dictionary<string, object> serializedNewAngle = (Dictionary<string, object>)serialized["NewAngle"];
            Assert.AreEqual(15.0f, serializedNewAngle["X"]);
            Assert.AreEqual(15.0f, serializedNewAngle["Y"]);
        }

        [Test]
        public void TestTiltSensorCrashDidChange() 
        {
            // ARRANGE
            Mock<ITiltSensor> tiltSensorMock = PrepareServiceMock<ITiltSensor>();

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceCrashChange(tiltSensorMock.Object, new TiltSensorCrash() { X = 1, Y = 2, Z = 3 }, new TiltSensorCrash() { X = 3, Y = 5, Z = 7 });

            // ASSERT
            VerifySerializedService(serialized, true);

            Dictionary<string, object> serializedOldCrash = (Dictionary<string, object>)serialized["OldCrash"];
            Assert.AreEqual(1, serializedOldCrash["X"]);
            Assert.AreEqual(2, serializedOldCrash["Y"]);
            Assert.AreEqual(3, serializedOldCrash["Z"]);

            Dictionary<string, object> serializedNewCrash = (Dictionary<string, object>)serialized["NewCrash"];
            Assert.AreEqual(3, serializedNewCrash["X"]);
            Assert.AreEqual(5, serializedNewCrash["Y"]);
            Assert.AreEqual(7, serializedNewCrash["Z"]);
        }

        [Test]
        public void TestVoltageSensorValueDidChange()
        {
            // ARRANGE
            Mock<IVoltageSensor> voltageSensorMock = PrepareServiceMock<IVoltageSensor>();
            float voltage = 2015.25f;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceVoltageChange(voltageSensorMock.Object, voltage);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(voltage, serialized["MilliVolts"]);
        }

        [Test]
        public void TestCurrentSensorValueDidChange()
        {
            // ARRANGE
            Mock<ICurrentSensor> currentSensorMock = PrepareServiceMock<ICurrentSensor>();
            float current = 314.5f;

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeServiceCurrentChange(currentSensorMock.Object, current);

            // ASSERT
            VerifySerializedService(serialized, true);
            Assert.AreEqual(current, serialized["MilliAmp"]);
        }

        [Test]
        public void TestSerializeColor()
        {
            // ARRANGE
            Color color = Color.FromArgb(200, 50, 100, 170);

            // ACT
            Dictionary<string, object> serialized = wrapperSerialization.SerializeColor(color);

            // ASSERT
            Assert.AreEqual(color.R / 255.0, serialized["R"]);
            Assert.AreEqual(color.G / 255.0, serialized["G"]);
            Assert.AreEqual(color.B / 255.0, serialized["B"]);
            Assert.AreEqual(color.A / 255.0, serialized["A"]);
        }

        [Test]
        public void TestDeserializeColor()
        {
            // ARRANGE
            Dictionary<string, object> color = new Dictionary<string,object>(){ { "R", 0.5 }, { "G", 0.6 }, { "B", 0.7 }, { "A", 0.9 } };

            // ACT
            Color deserialized = wrapperSerialization.DeserializeColor(color);

            // ASSERT
            Assert.AreEqual((int)(0.5 * 255), deserialized.R);
            Assert.AreEqual((int)(0.6 * 255), deserialized.G);
            Assert.AreEqual((int)(0.7 * 255), deserialized.B);
            Assert.AreEqual((int)(0.9 * 255), deserialized.A);
        }
    }

}
