package bluetooth;

import android.bluetooth.BluetoothGattCharacteristic;
import android.os.Handler;
import android.os.Looper;
import android.test.InstrumentationTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.UUID;
import java.util.concurrent.CountDownLatch;

import bluetooth.mocks_and_stubs.AndroidBluetoothDeviceWrapperMock;
import bluetooth.mocks_and_stubs.CoreBluetoothMockFactory;
import bluetooth.mocks_and_stubs.TestStubFactory;
import dk.lego.devicesdk.bluetooth.BluetoothIO;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDevice;
import dk.lego.devicesdk.bluetooth.service_definitions.IOServiceDefinition;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.IOCallbackListener;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.input_output.OutputCommand;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class BluetoothIOTest extends InstrumentationTestCase implements IOCallbackListener {

    private ConnectInfo connectInfo;
    private BluetoothIO bluetoothIO;
    private ConnectInfo activeConnectInfo;
    private InputFormat notifiedInputFormat;
    private byte[] notifiedValueData;

    private CountDownLatch didReceiveInputFormatLatch;
    private CountDownLatch didReceiveValueDataLatchFirstTime;
    private CountDownLatch didReceiveValueDataLatchSecondTime;

    private LegoBluetoothDevice legoBluetoothDeviceMock;

    @Override
    public void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new Handler(Looper.getMainLooper()));

        didReceiveInputFormatLatch = new CountDownLatch(1);
        didReceiveValueDataLatchFirstTime = new CountDownLatch(1);
        didReceiveValueDataLatchSecondTime = new CountDownLatch(1);

        connectInfo = new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR);
        activeConnectInfo = connectInfo;

        AndroidBluetoothDeviceWrapperMock androidBluetoothDeviceMock = new AndroidBluetoothDeviceWrapperMock("TestName", "TestDeviceId");
        legoBluetoothDeviceMock = mock(LegoBluetoothDevice.class);
        when(legoBluetoothDeviceMock.getAndroidBluetoothDevice()).thenReturn(androidBluetoothDeviceMock.getDevice());

        bluetoothIO = BluetoothIO.bluetoothIO(CoreBluetoothMockFactory.inputService(), legoBluetoothDeviceMock);
        bluetoothIO.registerCallbackListener(this);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();

        bluetoothIO.unregisterCallbackListener(this);
        HandlerHelper.getInstance().setMainHandler(null);
    }

    @SmallTest
    public void testHandleUpdatedInputFormat_SuccessScenario() throws Exception {
        int version = 1;
        int mode = 0;
        long deltaInterval = 1234;
        InputFormat.InputFormatUnit unit = InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE;
        boolean notificationsEnabled = true;
        int numberOfBytes = 2;
        byte[] formatData = TestStubFactory.inputFormatWriteData(
                version,
                connectInfo.getConnectId(),
                connectInfo.getType(),
                mode,
                deltaInterval,
                unit,
                notificationsEnabled,
                numberOfBytes);
        BluetoothGattCharacteristic inputFormatCharacteristic = CoreBluetoothMockFactory.inputFormatCharacteristicWithData(formatData);

        // Parse the data to the method under test
        bluetoothIO.handleUpdatedInputServiceCharacteristic(inputFormatCharacteristic);

        didReceiveInputFormatLatch.await();

        // Test that an InputFormat is correctly created
        assertNotNull(notifiedInputFormat);
        assertEquals(connectInfo.getConnectId(), notifiedInputFormat.getConnectId());
        assertEquals(connectInfo.getType(), notifiedInputFormat.getType());
        assertEquals(mode, notifiedInputFormat.getMode());
        assertEquals(deltaInterval, notifiedInputFormat.getDeltaInterval());
        assertEquals(unit, notifiedInputFormat.getUnit());
        assertTrue(notifiedInputFormat.isNotificationsEnabled());
        assertEquals(numberOfBytes, notifiedInputFormat.getNumberOfBytes());
    }

    @SmallTest
    public void testHandleUpdatedValue_SuccessScenario() throws Exception {
        //vFirst make sure an input format is set up for the connectionID so received values can be parsed
        int formatRevision = 2;

        int connectID1 = 1;
        int connectID2 = 2;

        // We add two InputFormats
        ConnectInfo connectInfo1 = new ConnectInfo(connectID1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR);
        ConnectInfo connectInfo2 = new ConnectInfo(connectID2, 2, ConnectInfo.IOType.IO_TYPE_VOLTAGE);
        simulateReceivingInputFormat(connectInfo1, formatRevision, 2);
        simulateReceivingInputFormat(connectInfo2, formatRevision, 1);

        // Then set up the Input Value bytes to be received
        int value1 = 1000;
        int value2 = 120;

        ByteBuffer valueData = ByteBuffer.allocate(6).order(ByteOrder.LITTLE_ENDIAN);
        ByteUtils.putUnsignedByte(valueData, formatRevision);
        ByteUtils.putUnsignedByte(valueData, connectID1);
        ByteUtils.putUnsignedShort(valueData, value1);
        ByteUtils.putUnsignedByte(valueData, connectID2);
        ByteUtils.putUnsignedByte(valueData, value2);

        // We make sure that the delegate will return the connectInfo1, meaning that the delegate should receive value1
        activeConnectInfo = connectInfo1;

        // Parse the data to the method under test
        BluetoothGattCharacteristic characteristic = CoreBluetoothMockFactory.inputValueCharacteristicWithData(valueData.array());
        bluetoothIO.handleUpdatedInputServiceCharacteristic(characteristic);

        didReceiveValueDataLatchFirstTime.await();

        // Verify the result
        byte[] expectedResult = ByteUtils.byteArrayFromShortLE(value1);
        assertTrue(Arrays.equals(expectedResult, notifiedValueData));


        //Now, we repeat the test, only we make the delegate return connectInfo2, meaning that the delegate should receive value2
        activeConnectInfo = connectInfo2;

        // Parse the data to the method under test
        bluetoothIO.handleUpdatedInputServiceCharacteristic(characteristic);

        didReceiveValueDataLatchSecondTime.await();

        //verify the result
        expectedResult = ByteUtils.byteArrayFromByte(value2);
        assertTrue(Arrays.equals(expectedResult, notifiedValueData));
    }

    @SmallTest
    public void testHandleUpdatedValue_ignored_if_format_revision_unknown()
    {
        int formatRevision = 2;
        simulateReceivingInputFormat(connectInfo, formatRevision, 1);

        int valueFormatRevision = 3;    //NOTE: Value format is different from format above
        int connectID = connectInfo.getConnectId();
        int value = 10;

        ByteBuffer valueData = ByteBuffer.allocate(3).order(ByteOrder.LITTLE_ENDIAN);
        ByteUtils.putUnsignedByte(valueData, valueFormatRevision);
        ByteUtils.putUnsignedByte(valueData, connectID);
        ByteUtils.putUnsignedByte(valueData, value);

        // Execute method under test
        BluetoothGattCharacteristic characteristic = CoreBluetoothMockFactory.inputValueCharacteristicWithData(valueData.array());
        bluetoothIO.handleUpdatedInputServiceCharacteristic(characteristic);

        // Verify that the value is not updated (as the format version in the value data does not match the last received format)
        assertNull(notifiedValueData);
    }

    @SmallTest
    public void testWriteMotorPower_offset_zero_value_zero_is_correctly_translated()
    {
        int offset = 0;
        int motorPower = 0;
        int expectedPowerWritten = 0;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_zero_value_100_is_correctly_translated()
    {
        int offset = 0;
        int motorPower = 100;
        int expectedPowerWritten = 100;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_zero_value_minus_100_is_correctly_translated()
    {
        int offset = 0;
        int motorPower = -100;
        int expectedPowerWritten = -100;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_30_value_zero_is_correctly_translated()
    {
        int offset = 30;
        int motorPower = 0;
        int expectedPowerWritten = 30;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_30_value_50_is_correctly_translated()
    {
        int offset = 30;
        int motorPower = 50;
        int expectedPowerWritten = 65; // The range goes from 30-100. 50% of that places the value in the middle of that range: 65
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_30_value_minus_50_is_correctly_translated()
    {
        int offset = 30;
        int motorPower = -50;
        int expectedPowerWritten = -65; // The range goes from 30-100. -50% of that places the value in the middle of that range: -65
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_30_value_minus_100_is_correctly_translated()
    {
        int offset = 30;
        int motorPower = -100;
        int expectedPowerWritten = -100;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    @SmallTest
    public void testWriteMotorPower_offset_30_value_100_is_correctly_translated()
    {
        int offset = 30;
        int motorPower = 100;
        int expectedPowerWritten = 100;
        verifyWritingMotorPowerWithOffset(motorPower, offset, expectedPowerWritten);
    }

    private void verifyWritingMotorPowerWithOffset(int motorPower, int offset, int expectedPowerWritten) {
        int connectID = 0;
        OutputCommand outputCommand = OutputCommand.commandWriteMotorPower(expectedPowerWritten, connectID);

        bluetoothIO.writeMotorPower(motorPower, offset, connectID);
        verify(legoBluetoothDeviceMock).addWriteOfCharacteristicToQueue(any(BluetoothGattCharacteristic.class), eq(outputCommand.getData()), eq(BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE));
    }

    private void simulateReceivingInputFormat(ConnectInfo connectInfo, int version, int numberOfBytes) {
        int mode = 0;
        long deltaInterval = 1234;
        InputFormat.InputFormatUnit unit = InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE;
        boolean notificationsEnabled = true;
        byte[] formatData = TestStubFactory.inputFormatWriteData(version,
                connectInfo.getConnectId(),
                connectInfo.getType(),
                mode,
                deltaInterval,
                unit,
                notificationsEnabled,
                numberOfBytes);
        BluetoothGattCharacteristic inputFormatCharacteristic = CoreBluetoothMockFactory.inputFormatCharacteristicWithData(formatData);
        bluetoothIO.handleUpdatedInputServiceCharacteristic(inputFormatCharacteristic);

    }

    @Override
    public void didReceiveInputFormat(IO io, InputFormat inputFormat) {
        didReceiveInputFormatLatch.countDown();
        notifiedInputFormat = inputFormat;
    }

    @Override
    public void didReceiveValueData(IO io, byte[] valueData) {
        if (didReceiveValueDataLatchFirstTime.getCount() > 0) {
            didReceiveValueDataLatchFirstTime.countDown();
        } else {
            didReceiveValueDataLatchSecondTime.countDown();
        }
        notifiedValueData = valueData;
    }

    @Override
    public ConnectInfo didRequestConnectInfo(IO io) {
        return activeConnectInfo;
    }

}
