package bluetooth;

import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothProfile;
import android.test.AndroidTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CountDownLatch;

import bluetooth.mocks_and_stubs.AndroidBluetoothDeviceWrapperMock;
import bluetooth.mocks_and_stubs.CoreBluetoothMockFactory;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.android.bluetooth.le.ScanRecord;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDevice;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDeviceImpl;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothDeviceWrapper;
import dk.lego.devicesdk.device.DeviceCallbackListener;
import dk.lego.devicesdk.device.DeviceInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class LegoBluetoothDeviceImplTest extends AndroidTestCase implements DeviceCallbackListener {
    public static final String EXPECTED_NAME = "NAME";
    public static final String EXPECTED_DEVICE_ID = "DEVICE_ID";
    public static final byte[] EMPTY_ADVERTISEMENT_DATA = new byte[]{};

    private AndroidBluetoothDeviceWrapperMock androidBluetoothDeviceMock;
    private LegoBluetoothDeviceImpl legoBluetoothDevice;
    private Boolean notifiedButtonStatePressed = false;
    private Boolean notifiedLowVoltageAlert = false;

    private CountDownLatch valueChangeFirstTimeCountDownLatch;
    private CountDownLatch valueChangeSecondTimeCountDownLatch;

    private byte[] expectedScanRecord = new byte[] {
            0x02, 0x01, 0x1a, // advertising flags
            0x05, 0x02, 0x0b, 0x11, 0x0a, 0x11, // 16 bit service uuids
            0x04, 0x09, 0x50, 0x65, 0x64, // name
            0x02, 0x0A, (byte) 0xec, // tx power level
            0x05, 0x16, 0x0b, 0x11, 0x50, 0x64, // service data
            0x05, (byte) 0xff, (byte) 0xe0, 0x00, 0x02, 0x15, // manufacturer specific data
            0x03, 0x50, 0x01, 0x02, // an unknown data type won't cause trouble
    };

    @Override
    protected void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new android.os.Handler(getContext().getMainLooper()));

        androidBluetoothDeviceMock = new AndroidBluetoothDeviceWrapperMock(EXPECTED_NAME, EXPECTED_DEVICE_ID);
        legoBluetoothDevice = (LegoBluetoothDeviceImpl) LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        legoBluetoothDevice.registerCallbackListener(this);

        valueChangeFirstTimeCountDownLatch = new CountDownLatch(1);
        valueChangeSecondTimeCountDownLatch = new CountDownLatch(1);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();

        HandlerHelper.getInstance().setMainHandler(null);

        legoBluetoothDevice.unregisterCallbackListener(this);
    }

    @SmallTest
    public void testName() throws Exception {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);

        assertEquals(EXPECTED_NAME, legoBluetoothDevice.getName());
    }

    @SmallTest
    public void testDeviceId() throws Exception {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);

        assertEquals(EXPECTED_DEVICE_ID, legoBluetoothDevice.getDeviceId());
    }

    @SmallTest
    public void testAdvertisementData() throws Exception {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, expectedScanRecord, 0);

        assertTrue(Arrays.equals(expectedScanRecord, legoBluetoothDevice.getScanRecord().getBytes()));
    }

    @SmallTest
    public void testDidChangeButtonState() throws Exception
    {
        assertFalse(legoBluetoothDevice.isButtonPressed());

        // Pressed
        byte[] buttonData = ByteUtils.fromHexString("01");
        BluetoothGattCharacteristic characteristic = CoreBluetoothMockFactory.deviceButtonCharacteristic(buttonData);

        legoBluetoothDevice.handleReadOrUpdatedCharacteristic(characteristic);

        valueChangeFirstTimeCountDownLatch.await();

        assertTrue(legoBluetoothDevice.isButtonPressed());
        assertTrue(notifiedButtonStatePressed);

        // Not pressed
        buttonData = ByteUtils.fromHexString("00");
        characteristic = CoreBluetoothMockFactory.deviceButtonCharacteristic(buttonData);

        legoBluetoothDevice.handleReadOrUpdatedCharacteristic(characteristic);

        valueChangeSecondTimeCountDownLatch.await();

        assertFalse(legoBluetoothDevice.isButtonPressed());
        assertFalse(notifiedButtonStatePressed);
    }

    @SmallTest
    public void testDidUpdateLowVoltage() throws Exception
    {
        assertFalse(legoBluetoothDevice.isLowVoltage());

        // Alert
        byte[] lowVoltageData = ByteUtils.fromHexString("01");
        BluetoothGattCharacteristic characteristic = CoreBluetoothMockFactory.deviceLowVoltageAlertCharacteristic(lowVoltageData);

        legoBluetoothDevice.handleReadOrUpdatedCharacteristic(characteristic);

        valueChangeFirstTimeCountDownLatch.await();

        assertTrue(legoBluetoothDevice.isLowVoltage());
        assertTrue(notifiedLowVoltageAlert);

        // Not alert
        lowVoltageData = ByteUtils.fromHexString("00");
        characteristic = CoreBluetoothMockFactory.deviceLowVoltageAlertCharacteristic(lowVoltageData);

        legoBluetoothDevice.handleReadOrUpdatedCharacteristic(characteristic);

        valueChangeSecondTimeCountDownLatch.await();

        assertFalse(legoBluetoothDevice.isLowVoltage());
        assertFalse(notifiedLowVoltageAlert);
    }

    @SmallTest
    public void testDidUpdateRSSIValues() throws Exception
    {
        int rssiValue = -10;

        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(new byte[]{}, rssiValue);
        assertEquals(rssiValue, legoBluetoothDevice.getRSSIValue());
    }

    @SmallTest
    public void testDidUpdateRSSIValues_averaging_always_picks_the_middle_value_in_a_sorted_list() throws Exception
    {
        byte[] emptyScanRecord = {};

        int rssi = -1;
        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(emptyScanRecord, rssi);
        assertEquals(-1, legoBluetoothDevice.getRSSIValue());

        rssi = -5;
        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(emptyScanRecord, rssi);
        assertEquals(-1, legoBluetoothDevice.getRSSIValue());

        rssi = -3;
        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(emptyScanRecord, rssi);
        assertEquals(-3, legoBluetoothDevice.getRSSIValue());

        rssi = -4;
        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(emptyScanRecord, rssi);
        assertEquals(-3, legoBluetoothDevice.getRSSIValue());

        rssi = -4;
        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(emptyScanRecord, rssi);
        assertEquals(-4, legoBluetoothDevice.getRSSIValue());
    }

    @SmallTest
    // After having received more that only the last 10 values received are taken into account
    public void testDidUpdateRSSIValues_averaging_only_last_10_values_count() throws Exception
    {
        int numberOfValuesToAverage = 10;
        List<Integer> rssiValues = new ArrayList<>(LegoBluetoothDeviceImpl.NUMBER_OF_RSSI_VALUES_TO_AVERAGE);
        for (int i = 1; i <= numberOfValuesToAverage; ++i) {
            rssiValues.add(-i * 2);
        }

        // After this, the array now contains all numbers from -2..-4...-20 in a random order
        Collections.shuffle(rssiValues, new Random(10));

        int expectedAverage = -10;
        // Execute method under test
        for (Integer rssiValue : rssiValues) {
            legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(new byte[] {}, rssiValue);
        }

        assertEquals(expectedAverage, legoBluetoothDevice.getRSSIValue());

        // Now add the new numbers and make sure only the new number goes into the average
        // as the should have replaced all of the ten previously received numbers
        rssiValues.clear();

        for (int i = 1; i <= numberOfValuesToAverage; ++i) {
            rssiValues.add(-i * 2 - 40);
        }

        // After this, the array now contains all numbers from -42..-44...-60 in a random order
        Collections.shuffle(rssiValues, new Random(20));

        expectedAverage = -50;
        // Execute method under test
        for (Integer rssiValue : rssiValues) {
            legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(new byte[] {}, rssiValue);
        }

        assertEquals(expectedAverage, legoBluetoothDevice.getRSSIValue());
    }

    @SmallTest
    public void testUpdate() throws Exception {
        ScanRecord expectedAdvertisementData = ScanRecord.parseFromBytes(expectedScanRecord);
        int expectedRssiValue = 47;

        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);

        assertFalse(Arrays.equals(expectedAdvertisementData.getBytes(), legoBluetoothDevice.getScanRecord().getBytes()));
        assertFalse(Integer.valueOf(legoBluetoothDevice.getRSSIValue()).equals(expectedRssiValue));

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateWithAdvertisementDataAndRSSI(expectedScanRecord, expectedRssiValue);

        assertTrue(Arrays.equals(expectedAdvertisementData.getBytes(), legoBluetoothDevice.getScanRecord().getBytes()));
        assertEquals(legoBluetoothDevice.getRSSIValue(), expectedRssiValue);
    }

    @SmallTest
    public void testEquals() throws Exception {
        AndroidBluetoothDeviceWrapper androidBluetoothDeviceMockAlt = mock(AndroidBluetoothDeviceWrapper.class);

        //If two devices refer to holds the same peripheral, they are considered equals
        LegoBluetoothDevice legoBluetoothDevice1 = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, -10);
        LegoBluetoothDevice legoBluetoothDevice2 = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, expectedScanRecord, -20);
        LegoBluetoothDevice legoBluetoothDevice3 = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMockAlt, EMPTY_ADVERTISEMENT_DATA, -30);

        when(androidBluetoothDeviceMockAlt.getAddress()).thenReturn("3");

        assertTrue(legoBluetoothDevice1.equals(legoBluetoothDevice1));
        assertTrue(legoBluetoothDevice1.equals(legoBluetoothDevice2));
        assertTrue(legoBluetoothDevice2.equals(legoBluetoothDevice1));
        assertFalse(legoBluetoothDevice1.equals(legoBluetoothDevice3));
        assertFalse(legoBluetoothDevice3.equals(legoBluetoothDevice1));
    }

    @SmallTest
    public void testConnectState_disconnecting() throws Exception
    {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateConnectionStateUponConnectionStateChange(BluetoothGatt.GATT_FAILURE, BluetoothProfile.STATE_CONNECTED);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateConnectionStateUponConnectionStateChange(BluetoothGatt.GATT_SUCCESS, BluetoothProfile.STATE_DISCONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING, legoBluetoothDevice.getConnectState());
    }

    @SmallTest
    public void testConnectState_disconnected() throws Exception
    {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).setAdvertising(false);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateConnectionStateUponConnectionStateChange(BluetoothGatt.GATT_SUCCESS, BluetoothProfile.STATE_DISCONNECTING);

        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING, legoBluetoothDevice.getConnectState());
    }

    @SmallTest
    public void testConnectState_disconnected_not_advertising() throws Exception
    {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).setAdvertising(true);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateConnectionStateUponConnectionStateChange(BluetoothGatt.GATT_SUCCESS, BluetoothProfile.STATE_DISCONNECTING);

        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING, legoBluetoothDevice.getConnectState());
    }

    @SmallTest
    public void testConnectState_connecting() throws Exception
    {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateConnectionStateUponConnectionStateChange(BluetoothGatt.GATT_SUCCESS, BluetoothProfile.STATE_CONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING, legoBluetoothDevice.getConnectState());
    }

    @SmallTest
    public void testConnectState_connected_interrogation_finished() throws Exception
    {
        LegoBluetoothDevice legoBluetoothDevice = LegoBluetoothDeviceImpl.deviceWithWrapper(androidBluetoothDeviceMock, EMPTY_ADVERTISEMENT_DATA, 0);
        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING, legoBluetoothDevice.getConnectState());

        ((LegoBluetoothDeviceImpl) legoBluetoothDevice).updateDeviceState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
        assertEquals(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING, legoBluetoothDevice.getConnectState());
    }

    @Override
    public void didUpdateDeviceInfo(LegoDevice device, DeviceInfo deviceInfo, LDSDKError error) {

    }

    @Override
    public void didChangeNameFrom(LegoDevice device, String oldName, String newName) {

    }

    @Override
    public void didChangeButtonState(LegoDevice device, boolean pressed) {
        if (valueChangeFirstTimeCountDownLatch.getCount() > 0) {
            valueChangeFirstTimeCountDownLatch.countDown();
        } else {
            valueChangeSecondTimeCountDownLatch.countDown();
        }
        notifiedButtonStatePressed = pressed;
    }

    @Override
    public void didUpdateLowVoltageState(LegoDevice device, boolean lowVoltage) {
        if (valueChangeFirstTimeCountDownLatch.getCount() > 0) {
            valueChangeFirstTimeCountDownLatch.countDown();
        } else {
            valueChangeSecondTimeCountDownLatch.countDown();
        }
        notifiedLowVoltageAlert = lowVoltage;
    }

    @Override
    public void didUpdateBatteryLevel(LegoDevice device, int newLevel) {

    }

    @Override
    public void didAddService(LegoDevice device, LegoService service) {

    }

    @Override
    public void didRemoveService(LegoDevice device, LegoService service) {

    }

    @Override
    public void didFailToAddServiceWithError(LegoDevice device, LDSDKError error) {

    }
}