package bluetooth;

import android.test.AndroidTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.util.concurrent.CountDownLatch;

import bluetooth.mocks_and_stubs.AndroidBluetoothAdapterWrapperMock;
import bluetooth.mocks_and_stubs.AndroidBluetoothDeviceWrapperMock;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDevice;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDeviceManagerCallbackListener;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDeviceManagerImpl;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothDeviceWrapper;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.utils.HandlerHelper;

public class LegoBluetoothDeviceManagerImplTest extends AndroidTestCase implements LegoBluetoothDeviceManagerCallbackListener {

    public static final int LEGO_DEVICE_RSSI_VALUE = -47;

    private static final byte[] LEGO_DEVICE_ADVERTISEMENT_DATA = {2,1,6,17,7,35,-47,-68,-22,95,120,35,21,-34,-17,18,18,35,21,0,0,7,22,35,21,0,-1,0,0,5,18,16,0,32,0,2,10,0,4,9,103,117,108,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    private AndroidBluetoothDeviceWrapper androidBluetoothDeviceWrapperMock1;
    private AndroidBluetoothDeviceWrapper androidBluetoothDeviceWrapperMock2;
    private AndroidBluetoothAdapterWrapperMock androidBluetoothAdapterWrapperMock;
    private LegoBluetoothDeviceManagerImpl legoBluetoothDeviceManager;
    private CountDownLatch deviceAppearedLatch;
    private CountDownLatch deviceDisappearedLatch;

    @Override
    public void setUp() throws Exception {
        super.setUp();
        deviceAppearedLatch = new CountDownLatch(1);
        deviceDisappearedLatch = new CountDownLatch(1);
        androidBluetoothDeviceWrapperMock1 = new AndroidBluetoothDeviceWrapperMock("1", "1");
        androidBluetoothDeviceWrapperMock2 = new AndroidBluetoothDeviceWrapperMock("2", "2");
        androidBluetoothAdapterWrapperMock = new AndroidBluetoothAdapterWrapperMock();
        legoBluetoothDeviceManager = new LegoBluetoothDeviceManagerImpl(androidBluetoothAdapterWrapperMock);
        legoBluetoothDeviceManager.registerCallbackListener(this);
    }

    @SmallTest
    public void testDeviceDidAppear_two_different_peripherals_adds_two_new_devices() throws Exception {
        androidBluetoothAdapterWrapperMock.performOnLeScanCallback(androidBluetoothDeviceWrapperMock1, LEGO_DEVICE_RSSI_VALUE, LEGO_DEVICE_ADVERTISEMENT_DATA);
        androidBluetoothAdapterWrapperMock.performOnLeScanCallback(androidBluetoothDeviceWrapperMock2, LEGO_DEVICE_RSSI_VALUE, LEGO_DEVICE_ADVERTISEMENT_DATA);

        assertEquals(2, legoBluetoothDeviceManager.devicesInState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING).size());
    }

    @SmallTest
    public void testDeviceDidAppear_two_identical_peripherals_adds_only_one_device() throws Exception {
        androidBluetoothAdapterWrapperMock.performOnLeScanCallback(androidBluetoothDeviceWrapperMock1, LEGO_DEVICE_RSSI_VALUE, LEGO_DEVICE_ADVERTISEMENT_DATA);
        androidBluetoothAdapterWrapperMock.performOnLeScanCallback(androidBluetoothDeviceWrapperMock1, LEGO_DEVICE_RSSI_VALUE, LEGO_DEVICE_ADVERTISEMENT_DATA);

        assertEquals(1, legoBluetoothDeviceManager.devicesInState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING).size());
    }

    @SmallTest
    public void testDeviceDidDisappear() throws Exception
    {
        //Milliseconds to wait before removing a device that has not been seen advertising
        int millisecondsToWait = 1000;
        legoBluetoothDeviceManager.updateAdvertisingDevicesInterval(millisecondsToWait);

        //We need to call scan, as this starts the timer that will remove devices
        //that has not been seen advertising for some time
        legoBluetoothDeviceManager.scan(getContext());

        //Make the BluetoothDeviceManager discover a new device
        androidBluetoothAdapterWrapperMock.performOnLeScanCallback(androidBluetoothDeviceWrapperMock1, LEGO_DEVICE_RSSI_VALUE, LEGO_DEVICE_ADVERTISEMENT_DATA);

        deviceAppearedLatch.await();

        assertEquals(1, legoBluetoothDeviceManager.devicesInState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING).size());

        //Now, if the device is not seen advertising again within 'millisecondsToWait' it should
        //cause a 'deviceDidDisappear' notification
        deviceDisappearedLatch.await();

        assertEquals(0, legoBluetoothDeviceManager.devicesInState(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING).size());
    }

    @Override
    public void tearDown() throws Exception {
        legoBluetoothDeviceManager.stopScanning();
        legoBluetoothDeviceManager.unregisterCallbackListener(this);

        HandlerHelper.getInstance().setMainHandler(null);
        super.tearDown();
    }

    @Override
    public void onDeviceDidAppear(LegoBluetoothDevice bluetoothDevice) {
        deviceAppearedLatch.countDown();

    }

    @Override
    public void onDeviceDidDisappear(LegoBluetoothDevice bluetoothDevice) {
        deviceDisappearedLatch.countDown();

    }

    @Override
    public void onWillStartConnectingToDevice(LegoBluetoothDevice device) {

    }

    @Override
    public void onDidFailToConnectToDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error) {

    }

    @Override
    public void onDidDisconnectFromDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error) {

    }

    @Override
    public void onDidStartInterrogatingDevice(LegoBluetoothDevice device) {

    }

    @Override
    public void onDidFinishInterrogatingDevice(LegoBluetoothDevice device) {

    }
}
