package bluetooth.services;

import android.os.Looper;
import android.test.AndroidTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CountDownLatch;

import bluetooth.helper.IOStub;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.CurrentSensor;
import dk.lego.devicesdk.services.CurrentSensorCallbackListener;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

public class CurrentSensorTest extends AndroidTestCase implements CurrentSensorCallbackListener {
    private CountDownLatch valueUpdateCountdownLatch;
    private float notifiedValue;
    private CurrentSensor sensor;

    @Override
    public void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new android.os.Handler(Looper.getMainLooper()));

        IOStub ioStub = new IOStub();
        ConnectInfo connectInfo = new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_VOLTAGE);
        sensor = CurrentSensor.createService(connectInfo, ioStub);
        sensor.registerCallbackListener(this);

        valueUpdateCountdownLatch = new CountDownLatch(1);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();

        HandlerHelper.getInstance().setMainHandler(null);

        sensor.unregisterCallbackListener(this);
    }

    @SmallTest
    public void testHandleUpdatedValue_SI() throws Exception
    {
        // Setup test
        updateSensorUnit(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI);
        float value = 0.10f;

        // Execute method under test
        updateValueWithData(ByteUtils.byteArrayFromFloatLE(value));

        valueUpdateCountdownLatch.await();

        // Verify
        assertEquals(value, sensor.getValueAsMilliAmps());
        assertEquals(notifiedValue, sensor.getValueAsMilliAmps());
    }

    @SmallTest
    public void testHandleUpdatedValue_Raw() throws Exception
    {
        //Setup test
        updateSensorUnit(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW);
        float value = 3000.10f;

        //Execute method under test
        updateValueWithData(ByteUtils.byteArrayFromFloatLE(value));

        valueUpdateCountdownLatch.await();

        //Verify
        assertEquals(0.0f, sensor.getValueAsMilliAmps());
        assertEquals(notifiedValue, 0.0f);
        assertEquals(value, sensor.getValueAsFloat()); //It should be possible to get the raw value from the 'generic'  method
    }

    @SmallTest
    public void testHandleUpdatedValue_Pct() throws Exception
    {
        updateSensorUnit(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE);
        int value = 99;

        //Execute method under test
        updateValueWithData(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(value).array());

        valueUpdateCountdownLatch.await();

        //Verify
        assertEquals(0.0f, sensor.getValueAsMilliAmps());
        assertEquals(notifiedValue, 0.0f);
        assertEquals(value, sensor.getValueAsInteger()); // It should be possible to get the raw value from the 'generic' method
    }

    private void updateValueWithData(byte[] data) {
        try {
            sensor.handleUpdatedValueData(data);
        } catch (LDSDKError e) {
            assertNotNull(String.format("Did not expect error from handleUpdatedValueData: %s", e.getMessage()), null); // Fail on error
        }
    }


    private void updateSensorUnit(InputFormat.InputFormatUnit unit) {
        InputFormat inputFormat = sensor.getDefaultInputFormat().inputFormatBySettingModeAndUnit(sensor.getDefaultInputFormat().getMode(), unit);
        sensor.handleUpdatedInputFormat(inputFormat);
    }

    @Override
    public void didUpdateMilliAmps(CurrentSensor sensor, float milliAmps) {
        valueUpdateCountdownLatch.countDown();
        notifiedValue = milliAmps;
    }

    @Override
    public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
        // Do nothing
    }

    @Override
    public void didUpdateInputFormat(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        // Do nothing
    }
}
