package bluetooth.services;

import android.os.Looper;
import android.test.InstrumentationTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CountDownLatch;

import bluetooth.helper.IOStub;
import bluetooth.mocks_and_stubs.TestStubFactory;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.MotionSensor;
import dk.lego.devicesdk.services.MotionSensorCallbackListener;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

public class MotionSensorTest extends InstrumentationTestCase implements MotionSensorCallbackListener {

    MotionSensor motionSensor;
    float notifiedDistance;
    int notifiedCount;
    private ConnectInfo connectInfo;
    private IOStub ioStub;

    CountDownLatch distanceUpdatedLatch;
    CountDownLatch countUpdatedLatch;

    @Override
    public void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new android.os.Handler(Looper.getMainLooper()));

        ioStub = new IOStub();
        connectInfo = new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTION_SENSOR);
        motionSensor = MotionSensor.createService(connectInfo, ioStub);
        motionSensor.registerCallbackListener(this);

        distanceUpdatedLatch = new CountDownLatch(1);
        countUpdatedLatch = new CountDownLatch(1);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();

        HandlerHelper.getInstance().setMainHandler(null);

        motionSensor.unregisterCallbackListener(this);
    }

    @Override
    public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
        /**
         * It is easy to 'fix' the code to make this test fail when trying to 'cache' values in the TiltSensor
         * You should make sure that at the time this delegate is invoked, the value of the "direction"
         * will also return the updated value
         */
        if (motionSensor.getMotionSensorMode() == MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT
                && motionSensor.getInputFormat().getUnit() == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW) {
            float value = ByteUtils.getUnsignedByte(ByteBuffer.wrap(newValue).order(ByteOrder.LITTLE_ENDIAN));
            assertEquals(motionSensor.getDistance(), value);
        }
    }

    @Override
    public void didUpdateInputFormat(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        // Do nothing
    }

    @Override
    public void didUpdateDistance(MotionSensor sensor, float oldDistance, float newDistance) {
        distanceUpdatedLatch.countDown();
        notifiedDistance = newDistance;
    }

    @Override
    public void didUpdateCount(MotionSensor sensor, int count) {
        countUpdatedLatch.countDown();
        notifiedCount = count;
    }

    @SmallTest
    public void testHandleUpdatedValue_detect_mode_raw_1() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        updateValueWithData(ByteUtils.fromHexString("01"));

        distanceUpdatedLatch.await();

        assertEquals(1.0f, motionSensor.getDistance());
        assertEquals(1.0f, notifiedDistance);


        assertEquals(0, notifiedCount);
    }

    @SmallTest
    public void testHandleUpdatedValue_detect_mode_raw_100() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        updateValueWithData(ByteUtils.fromHexString("64"));

        distanceUpdatedLatch.await();

        assertEquals(100.0f, motionSensor.getDistance());
        assertEquals(100.0f, notifiedDistance);

        assertEquals(0, notifiedCount);
    }

    @SmallTest
    public void testHandleUpdatedValue_detect_mode_percentage_1() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1);

        byte value = 1;
        updateValueWithData(ByteUtils.byteArrayFromByte(value));

        distanceUpdatedLatch.await();

        assertEquals(1.0f, motionSensor.getDistance());
        assertEquals(1.0f, notifiedDistance);
    }

    @SmallTest
    public void testHandleUpdatedValue_detect_mode_percentage_100() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1);

        byte value = 100;
        updateValueWithData(ByteUtils.byteArrayFromByte(value));

        distanceUpdatedLatch.await();


        assertEquals(100.0f, motionSensor.getDistance());
        assertEquals(100.0f, notifiedDistance);
    }


    @SmallTest
    public void testHandleUpdatedValue_detect_mode_si_pos() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4);

        float value = 123.45f;
        updateValueWithData(ByteUtils.byteArrayFromFloatLE(value));

        distanceUpdatedLatch.await();

        assertEquals(value, motionSensor.getDistance());
        assertEquals(value, notifiedDistance);
    }

    @SmallTest
    public void testHandleUpdatedValue_detect_mode_si_neg() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_DETECT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4);

        float value = -123.45f;
        updateValueWithData(ByteUtils.byteArrayFromFloatLE(value));

        distanceUpdatedLatch.await();

        assertEquals(value, motionSensor.getDistance());
        assertEquals(value, notifiedDistance);
    }

    @SmallTest
    public void testHandleUpdatedValue_count_mode_raw() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 4);

        int count = 12345;

        updateValueWithData(ByteUtils.byteArrayFromUnsignedIntLE(count));

        countUpdatedLatch.await();

        assertEquals(count, motionSensor.getCount());
        assertEquals(count, notifiedCount);

        assertEquals(0.0f, notifiedDistance);
    }

    @SmallTest
    public void testHandleUpdatedValue_count_mode_percentage() throws Exception {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1);

        byte count = 99;

        updateValueWithData(ByteUtils.byteArrayFromByte(count));

        countUpdatedLatch.await();

        assertEquals(count, motionSensor.getCount());
        assertEquals(count, notifiedCount);

        assertEquals(0.0f, notifiedDistance);
    }

    @SmallTest
    public void testHandleUpdatedValue_count_mode_si() throws Exception

    {
        updateInputFormatMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4);

        float count = 12345.12f;

        updateValueWithData(ByteUtils.byteArrayFromFloatLE(count));

        countUpdatedLatch.await();

        assertEquals((int) count, motionSensor.getCount());
        assertEquals((int) count, notifiedCount);

        assertEquals(0.0f, notifiedDistance);
    }

    @SmallTest
    public void testSetMotionSensorMode() throws Exception {
        // Setup expected moc invocations
        InputFormat expectedFormat = motionSensor.getDefaultInputFormat().inputFormatBySettingMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT.getValue());

        // Run MUC
        motionSensor.setMotionSensorMode(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT);

        // Verify mock
        assertEquals(connectInfo.getConnectId(), ioStub.lastConnectIdReceived.intValue());
        assertEquals(expectedFormat, ioStub.lastInputFormatReceived);
    }

    void updateValueWithData(byte[] data) {
        try {
            motionSensor.handleUpdatedValueData(data);
        } catch (LDSDKError e) {
            assertNotNull(null); // Fail on error
        }
    }

    private void updateInputFormatMode(MotionSensor.MotionSensorMode mode, InputFormat.InputFormatUnit unit, int numberOfBytes) {
        byte[] data = TestStubFactory.inputFormatWriteData(1, motionSensor.getConnectInfo().getConnectId(), motionSensor.getConnectInfo().getTypeEnum(), mode.getValue(), 1, unit, true, numberOfBytes);

        InputFormat format = InputFormat.inputFormat(data);
        motionSensor.handleUpdatedInputFormat(format);
    }
}
