package bluetooth.services;

import android.graphics.Color;
import android.os.Handler;
import android.os.Looper;
import android.test.suitebuilder.annotation.SmallTest;

import junit.framework.TestCase;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.concurrent.CountDownLatch;

import bluetooth.helper.IOStub;
import bluetooth.mocks_and_stubs.TestStubFactory;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.RGBLight;
import dk.lego.devicesdk.services.RGBLightCallbackListener;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

public class RGBLightTest extends TestCase implements RGBLightCallbackListener {

    private ConnectInfo connectInfo;
    private IOStub ioStub;
    private RGBLight rgbLight;
    private int notifiedRGBColor;
    private byte[] notifiedRGBColorData;

    private CountDownLatch notifiedRGBColorLatch;
    private CountDownLatch notifiedRGBColorDataLatch;

    @Override
    public void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new Handler(Looper.getMainLooper()));

        connectInfo = new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_RGB_LIGHT);
        ioStub = new IOStub();

        rgbLight = RGBLight.createService(connectInfo, ioStub);
        rgbLight.registerCallbackListener(this);

        notifiedRGBColorLatch = new CountDownLatch(1);
        notifiedRGBColorDataLatch = new CountDownLatch(1);

        updateInputFormat(RGBLight.RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 3);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();
        rgbLight.unregisterCallbackListener(this);
        HandlerHelper.getInstance().setMainHandler(null);
    }

    @SmallTest
    public void testSetAbsoluteColor_success()
    {
        int red = 0;
        int green = 100;
        int blue = 255;

        // Execute MUC
        Integer expectedRGBColor = Color.rgb(red, green, blue);
        rgbLight.setColor(expectedRGBColor);

        // Verify
        assertEquals(expectedRGBColor, ioStub.lastColorWritten);
        assertEquals(expectedRGBColor, rgbLight.getColor());
    }

    @SmallTest
    public void testSetAbsoluteColor_ignored_when_in_discrete_mode()
    {
        updateInputFormat(RGBLight.RGBLightMode.RGB_LIGHT_MODE_DISCRETE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        // Execute MUC
        rgbLight.setColor(Color.rgb(0, 100, 255));

        // Verify
        assertNull(rgbLight.getColor());
    }

    @SmallTest
    public void testSwitchOff_absolute_mode()
    {
        Integer expectedRGBColor = Color.rgb(0, 0, 0);

        // Execute MUT
        rgbLight.switchOff();

        // Verify
        assertEquals(expectedRGBColor, ioStub.lastColorWritten);
        assertEquals(expectedRGBColor, rgbLight.getColor());
    }

    @SmallTest
    public void testSwitchToDefaultColor_absolute_mode()
    {
        Integer expectedRGBColor = Color.rgb(0x00, 0x00, 0xFF);

        // Execute MUT
        rgbLight.switchToDefaultColor();

        // Verify
        assertEquals(expectedRGBColor, ioStub.lastColorWritten);
        assertEquals(expectedRGBColor, rgbLight.getColor());
    }

    @SmallTest
    public void testHandleUpdatedValue_absolute_raw() throws Exception
    {
        int red = 0;
        int green = 100;
        int blue = 255;

        try {
            rgbLight.handleUpdatedValueData(dataFromColor(red, green, blue));
        } catch (LDSDKError e) {
            assertNotNull(String.format("Did not expect error from handleUpdatedValueData: %s", e.getMessage()), null); // Fail on error
        }

        int expectedColor = Color.rgb(red, green, blue);

        assertEquals(expectedColor, rgbLight.getColor().intValue());

        notifiedRGBColorDataLatch.await();
        assertTrue(Arrays.equals(notifiedRGBColorData, ByteUtils.fromHexString("0064FF")));

        notifiedRGBColorLatch.await();
        assertEquals(expectedColor, notifiedRGBColor);
    }

    @SmallTest
    public void testSetDiscreteColor_success()
    {
        updateInputFormat(RGBLight.RGBLightMode.RGB_LIGHT_MODE_DISCRETE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        int colorIndex = 1;

        // Execute MUC
        rgbLight.setColorIndex(colorIndex);

        // Verify
        assertEquals(colorIndex, ioStub.lastColorIndexWritten);
        assertEquals(colorIndex, rgbLight.getColorIndex());
    }

    @SmallTest
    public void testSwitchOff_discrete_mode()
    {
        updateInputFormat(RGBLight.RGBLightMode.RGB_LIGHT_MODE_DISCRETE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        int expectedRGBColorIndex = 0;

        // Execute MUT
        rgbLight.switchOff();

        // Verify
        assertEquals(expectedRGBColorIndex, ioStub.lastColorIndexWritten);
        assertEquals(expectedRGBColorIndex, rgbLight.getColorIndex());
    }

    @SmallTest
    public void testSwitchToDefaultColor_discrete_mode()
    {
        updateInputFormat(RGBLight.RGBLightMode.RGB_LIGHT_MODE_DISCRETE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        // Color index 3 happens to be the default color :-)
        int expectedRGBColorIndex = 3;

        // Execute MUT
        rgbLight.switchToDefaultColor();

        // Verify
        assertEquals(expectedRGBColorIndex, ioStub.lastColorIndexWritten);
        assertEquals(expectedRGBColorIndex, rgbLight.getColorIndex());
    }

    private byte[] dataFromColor(int red, int green, int blue)
    {
        ByteBuffer data = ByteBuffer.allocate(3);

        ByteUtils.putUnsignedByte(data, red);
        ByteUtils.putUnsignedByte(data, green);
        ByteUtils.putUnsignedByte(data, (byte) blue);
        return data.array();
    }

    private void updateInputFormat(RGBLight.RGBLightMode mode, InputFormat.InputFormatUnit unit, int numberOfBytes)
    {
        byte[] inputFormatData = TestStubFactory.inputFormatWriteData(1, connectInfo.getConnectId(), connectInfo.getType(), mode.getValue(), 1, unit, true, numberOfBytes);
        InputFormat inputFormat = InputFormat.inputFormat(inputFormatData);
        rgbLight.handleUpdatedInputFormat(inputFormat);
    }

    @Override
    public void didReadInitialColor(RGBLight rgbLight) {
        // Do nothing
    }

    @Override
    public void didUpdateRGBValue(RGBLight rgbLight, Integer oldRGBColor, Integer newRGBColor) {
        notifiedRGBColorLatch.countDown();
        notifiedRGBColor = newRGBColor;
    }

    @Override
    public void didUpdateRGBColorIndex(RGBLight rgbLight, int oldColorIndex, int newColorIndex) {
        // Do nothing
    }

    @Override
    public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
        notifiedRGBColorDataLatch.countDown();
        notifiedRGBColorData = newValue;
    }

    @Override
    public void didUpdateInputFormat(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        // Do nothing
    }
}
