package bluetooth.services;

import junit.framework.TestCase;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.List;

import bluetooth.helper.IOStub;
import bluetooth.mocks_and_stubs.TestStubFactory;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.DataFormat;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.utils.ByteUtils;

public class ServiceTest extends TestCase {

    private IOStub ioStub;
    private LegoService service;

    private byte[] dummyInputFormatData = new byte[InputFormat.INPUT_FORMAT_PACKAGE_SIZE];

    @Override
    public void setUp() throws Exception {
        super.setUp();
        ioStub = new IOStub();
    }

    public void testValueAsNumbers_raw() throws Exception {
        service = LegoService.createService(new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR), ioStub);
        service.handleUpdatedInputFormat(InputFormat.inputFormat(dummyInputFormatData));

        int mode = 0;
        InputFormat.InputFormatUnit unit = InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW;
        int numberOfDataSets = 2;
        int dataSetSize = 1;

        //Make the service 'receive' a new Input Format from device
        updateInputFormatMode(mode, unit, (numberOfDataSets * dataSetSize));

        //Setup the service to know about the new Input Format (which is required to call the 'getNumbersFromValueDataSet')
        service.addValidDataFormat(DataFormat.create("Count", mode, unit, dataSetSize, numberOfDataSets));

        byte[] valueData = ByteUtils.fromHexString("0102");

        try {
            service.handleUpdatedValueData(valueData);
        } catch (LDSDKError e) {
            assertNotNull(String.format("Failed: %s", e.getMessage()), null); // Fail on error
        }

        List<Number> dataSetNumbers = service.getNumbersFromValueDataSet();
        assertEquals(numberOfDataSets, dataSetNumbers.size());
        assertEquals(1, dataSetNumbers.get(0));
        assertEquals(2, dataSetNumbers.get(1));
    }

    public void testValueAsNumbers_si() throws Exception {
        service = LegoService.createService(new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR), ioStub);
        service.handleUpdatedInputFormat(InputFormat.inputFormat(dummyInputFormatData));

        byte mode = 0;
        InputFormat.InputFormatUnit unit = InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI;
        byte numberOfDataSets = 2;
        byte dataSetSize = 4; // float for SI

        //Make the service 'receive' a new Input Format from device
        updateInputFormatMode(mode, unit, (numberOfDataSets * dataSetSize));

        //Setup the service to know about the new Input Format (which is required to call the 'getNumbersFromValueDataSet')
        service.addValidDataFormat(DataFormat.create("Count", mode, unit, dataSetSize, numberOfDataSets));

        float value1 = 12345;
        float value2 = -12345;

        ByteBuffer valueBuffer = ByteBuffer.allocate(8).order(ByteOrder.LITTLE_ENDIAN);

        valueBuffer.putFloat(value1);
        valueBuffer.putFloat(value2);
        valueBuffer.array();

        try {
            service.handleUpdatedValueData(valueBuffer.array());
        } catch (LDSDKError e) {
            assertNotNull(String.format("Failed: %s", e.getMessage()), null); // Fail on error
        }

        List<Number> dataSetNumbers = service.getNumbersFromValueDataSet();
        assertEquals(numberOfDataSets, dataSetNumbers.size());
        assertEquals(value1, dataSetNumbers.get(0));
        assertEquals(value2, dataSetNumbers.get(1));
    }


    public void testValueAsSignedInt() throws Exception
    {
        service = LegoService.createService(new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR), ioStub);
        service.handleUpdatedInputFormat(InputFormat.inputFormat(dummyInputFormatData));

        int aSignedInt8 = -127;
        service.didReceiveValueData(ioStub, ByteUtils.byteArrayFromByte(aSignedInt8));
        assertEquals(aSignedInt8, service.getValueAsInteger());

        int aSignedInt16 = -12725;
        service.didReceiveValueData(ioStub, ByteUtils.byteArrayFromShortLE(aSignedInt16));
        assertEquals(aSignedInt16, service.getValueAsInteger());

        int aSignedInt32 = -12725456;
        service.didReceiveValueData(ioStub, ByteUtils.byteArrayFromUnsignedIntLE(aSignedInt32));
        assertEquals(aSignedInt32, service.getValueAsInteger());

    }

    public void testValueAsFloat() throws Exception
    {
        service = LegoService.createService(new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_MOTOR), ioStub);
        service.handleUpdatedInputFormat(InputFormat.inputFormat(dummyInputFormatData));

        float aFloat32 = 12345.12f;
        service.didReceiveValueData(ioStub, ByteUtils.byteArrayFromFloatLE(aFloat32));
        assertEquals(aFloat32, service.getValueAsFloat());

        float aNegativeFloat32 = -12345.12f;
        service.didReceiveValueData(ioStub, ByteUtils.byteArrayFromFloatLE(aNegativeFloat32));
        assertEquals(aNegativeFloat32, service.getValueAsFloat());
    }

//    TODO public void testValueAsFloat_value_is_zero_if_data_too_big

    public void testEquals() throws Exception
    {
        LegoService service1 = LegoService.createService(new ConnectInfo(1, 2, ConnectInfo.IOType.IO_TYPE_GENERIC), ioStub);
        LegoService service2 = LegoService.createService(new ConnectInfo(1, 2, ConnectInfo.IOType.IO_TYPE_GENERIC), ioStub);
        LegoService service3 = LegoService.createService(new ConnectInfo(3, 2, ConnectInfo.IOType.IO_TYPE_MOTOR), ioStub);

        assertTrue(service1.equals(service1));
        assertTrue(service1.equals(service2));
        assertTrue(service2.equals(service1));

        assertFalse(service1.equals(service3));
        assertFalse(service3.equals(service1));
    }

    public void updateInputFormatMode(int mode, InputFormat.InputFormatUnit unit, int numberOfBytes) {
        byte[] data = TestStubFactory.inputFormatWriteData(1,
                service.getConnectInfo().getConnectId(),
                service.getConnectInfo().getTypeEnum(),
                mode,
                1,
                unit,
                true,
                numberOfBytes);
        InputFormat format = InputFormat.inputFormat(data);
        service.handleUpdatedInputFormat(format);
    }
}
