package bluetooth.services;

import android.os.Looper;
import android.test.AndroidTestCase;
import android.test.suitebuilder.annotation.SmallTest;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.concurrent.CountDownLatch;

import bluetooth.helper.IOStub;
import bluetooth.mocks_and_stubs.TestStubFactory;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.TiltSensor;
import dk.lego.devicesdk.services.TiltSensorCallbackListener;
import dk.lego.devicesdk.utils.ByteUtils;
import dk.lego.devicesdk.utils.HandlerHelper;

public class TiltSensorTest extends AndroidTestCase implements TiltSensorCallbackListener {
    private TiltSensor.TiltSensorDirection notifiedDirection;
    private TiltSensor.TiltSensorAngle notifiedAngle;
    private TiltSensor.TiltSensorCrash notifiedCrash;
    private TiltSensor tiltSensor;

    private CountDownLatch angleCountDownLatch;
    private CountDownLatch directionCountDownLatch;
    private CountDownLatch crashCountDownLatch;
    private IOStub ioStub;
    private ConnectInfo connectInfo;

    @Override
    public void setUp() throws Exception {
        super.setUp();

        // We need to provide a handler since we're relying on callbacks
        HandlerHelper.getInstance().setMainHandler(new android.os.Handler(Looper.getMainLooper()));

        ioStub = new IOStub();
        connectInfo = new ConnectInfo(1, 1, ConnectInfo.IOType.IO_TYPE_TILT_SENSOR);
        tiltSensor = TiltSensor.createService(connectInfo, ioStub);
        tiltSensor.registerCallbackListener(this);

        angleCountDownLatch = new CountDownLatch(1);
        directionCountDownLatch = new CountDownLatch(1);
        crashCountDownLatch = new CountDownLatch(1);
    }

    @Override
    public void tearDown() throws Exception {
        super.tearDown();

        HandlerHelper.getInstance().setMainHandler(null);

        tiltSensor.unregisterCallbackListener(this);
    }

    @SmallTest
    public void testHandleUpdatedValue_angle_mode_raw_success_scenario() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_ANGLE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 2);

        // Execute method under test
        updateValueWithData(ByteUtils.fromHexString("0102"));

        angleCountDownLatch.await();

        // Verify result

        TiltSensor.TiltSensorAngle expectedResult = TiltSensor.tiltSensorAngleMake(1.0f, 2.0f);
        assertTrue(TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, tiltSensor.getAngle()));
        assertTrue(TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, notifiedAngle));
    }

    @SmallTest
    public void testHandleUpdatedValue_angle_mode_SI_success_scenario() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_ANGLE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 2 * 4);

        // Execute method under test
        byte[] data = generateData(10.10f, 20.20f);
        updateValueWithData(data);

        angleCountDownLatch.await();

        // Verify result
        TiltSensor.TiltSensorAngle expectedResult = TiltSensor.tiltSensorAngleMake(10.10f, 20.20f);
        assertTrue(String.format("Actual result x: %f, y: %f", tiltSensor.getAngle().x, tiltSensor.getAngle().y), TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, tiltSensor.getAngle()));
        assertTrue(TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, notifiedAngle));
    }

    @SmallTest
    public void testHandleUpdatedValue_angle_mode_pct_success_scenario() throws Exception {
        //Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_ANGLE, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 2);

        //Execute method under test
        byte[] data = ByteUtils.fromHexString("3264");
        updateValueWithData(data);

        angleCountDownLatch.await();

        //Verify result
        TiltSensor.TiltSensorAngle expectedResult = TiltSensor.tiltSensorAngleMake(0x32, 0x64);
        assertTrue(TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, tiltSensor.getAngle()));
        assertTrue(TiltSensor.tiltSensorAngleEqualToAngle(expectedResult, notifiedAngle));
    }

    @SmallTest
    public void testHandleUpdatedValue_tilt_mode_raw_success_scenario() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_TILT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        // Execute method under test
        updateValueWithData(ByteUtils.fromHexString("09")); // 09 = Forward

        directionCountDownLatch.await();

        // Verify result
        assertEquals(TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_FORWARD, tiltSensor.getDirection());
        assertEquals(TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_FORWARD, notifiedDirection);
    }

    @SmallTest
    public void testHandleUpdatedValue_all_tilt_mode_raw_values() throws Exception {
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_TILT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1);

        testHandleUpdatedValue("00", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_NEUTRAL);
        testHandleUpdatedValue("03", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_BACKWARD);
        testHandleUpdatedValue("05", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_RIGHT);
        testHandleUpdatedValue("07", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_LEFT);
        testHandleUpdatedValue("09", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_FORWARD);
        testHandleUpdatedValue("10", TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_UNKNOWN);
    }

    @SmallTest
    //SI mode should just return the same values as in raw
    public void testHandleUpdatedValue_tilt_mode_SI_success_scenario() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_TILT, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4);

        // Execute method under test
        updateValueWithData(ByteUtils.byteArrayFromFloatLE(7)); // 07 = Left

        directionCountDownLatch.await();

        // Verify result
        assertEquals(TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_LEFT, tiltSensor.getDirection());
    }

    @SmallTest
    public void testHandleUpdatedValue_tilt_mode_crash_raw() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 3);

        // Execute method under test (did crash)
        updateValueWithData(ByteUtils.fromHexString("010203"));

        crashCountDownLatch.await();

        // Verify result
        TiltSensor.TiltSensorCrash expectedResult = TiltSensor.tiltSensorCrashMake(1, 2, 3);
        assertTrue(TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, tiltSensor.getCrash()));
        assertTrue(TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, notifiedCrash));
    }

    @SmallTest
    public void testHandleUpdatedValue_tilt_mode_crash_percentage() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 3);

        // Execute method under test (did crash)
        updateValueWithData(ByteUtils.fromHexString("010203"));

        crashCountDownLatch.await();

        // Verify result
        TiltSensor.TiltSensorCrash expectedResult = TiltSensor.tiltSensorCrashMake(1, 2, 3);
        assertTrue(TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, tiltSensor.getCrash()));
        assertTrue(TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, notifiedCrash));
    }

    public void testHandleUpdatedValue_tilt_mode_crash_SI() throws Exception {
        // Setup test
        updateInputFormatMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 3 * 4);

        // Execute method under test (did crash)
        updateValueWithData(generateData(10.10f, 0.0f, 100.0f));

        crashCountDownLatch.await();

        // Verify result
        TiltSensor.TiltSensorCrash expectedResult = TiltSensor.tiltSensorCrashMake(10, 0, 100);
        assertTrue(String.format("Actual result x: %d, y: %d, z: %d", tiltSensor.getCrash().x, tiltSensor.getCrash().y, tiltSensor.getCrash().z), TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, tiltSensor.getCrash()));
        assertTrue(TiltSensor.tiltSensorCrashEqualToCrash(expectedResult, notifiedCrash));
    }

    @SmallTest
    public void testSetTiltSensorMode() throws Exception {
        // Setup expected moc invocations
        InputFormat expectedFormat = tiltSensor.getDefaultInputFormat().inputFormatBySettingMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH.getValue());

        // Run MUC
        tiltSensor.setTiltSensorMode(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH);

        // Verify mock
        assertEquals(connectInfo.getConnectId(), ioStub.lastConnectIdReceived.intValue());
        assertEquals(expectedFormat, ioStub.lastInputFormatReceived);
    }

    private void testHandleUpdatedValue(String valueHex, TiltSensor.TiltSensorDirection direction) {
        updateValueWithData(ByteUtils.fromHexString(valueHex));
        assertEquals(direction, tiltSensor.getDirection());
    }

    private void updateValueWithData(byte[] data) {
        try {
            tiltSensor.handleUpdatedValueData(data);
        } catch (LDSDKError e) {
            assertNotNull(String.format("Did not expect error from hadleUpdatedValudeData: %s", e.getMessage()), null); // Fail on error
        }
    }

    byte[] generateData(float x, float y) {
        ByteBuffer updateData = ByteBuffer.allocate(8).order(ByteOrder.LITTLE_ENDIAN);
        updateData.putFloat(x);
        updateData.putFloat(y);

        return updateData.array();
    }

    byte[] generateData(float x, float y, float z) {
        ByteBuffer updateData = ByteBuffer.allocate(12).order(ByteOrder.LITTLE_ENDIAN);
        updateData.putFloat(x);
        updateData.putFloat(y);
        updateData.putFloat(z);

        return updateData.array();
    }

    private void updateInputFormatMode(TiltSensor.TiltSensorMode mode, InputFormat.InputFormatUnit unit, int numberOfBytes) {
        byte[] data = TestStubFactory.inputFormatWriteData(1,
                tiltSensor.getConnectInfo().getConnectId(),
                tiltSensor.getConnectInfo().getType(),
                mode.getValue(),
                1,
                unit,
                true,
                numberOfBytes);

        InputFormat format = InputFormat.inputFormat(data);
        tiltSensor.handleUpdatedInputFormat(format);
    }

    @Override
    public void didUpdateDirection(TiltSensor sensor, TiltSensor.TiltSensorDirection oldDirection, TiltSensor.TiltSensorDirection newDirection) {
        directionCountDownLatch.countDown();
        notifiedDirection = newDirection;
    }

    @Override
    public void didUpdateAngle(TiltSensor sensor, TiltSensor.TiltSensorAngle oldAngle, TiltSensor.TiltSensorAngle newAngle) {
        angleCountDownLatch.countDown();
        notifiedAngle = newAngle;
    }

    @Override
    public void didUpdateCrash(TiltSensor sensor, TiltSensor.TiltSensorCrash oldCrashValue, TiltSensor.TiltSensorCrash newCrash) {
        crashCountDownLatch.countDown();
        notifiedCrash = newCrash;
    }

    @Override
    public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
        // It is easy to 'fix' the code to make this test fail when trying to 'cache' values in the TiltSensor
        // You should make sure that at the time this delegate is invoked, the value of the "direction"
        // will also return the updated value
        if (tiltSensor.getTiltSensorMode() == TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_TILT && tiltSensor.getInputFormat().getUnit() == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW) {
            assertEquals(tiltSensor.getDirection(), TiltSensor.TiltSensorDirection.fromInteger(ByteUtils.getUnsignedByte(ByteBuffer.wrap(newValue))));
        }
    }

    @Override
    public void didUpdateInputFormat(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        // Do nothing
    }
}
