package dk.lego.devicesdk.unity;

import android.content.Context;
import android.test.suitebuilder.annotation.SmallTest;

import junit.framework.TestCase;

import java.util.Arrays;
import java.util.HashMap;

import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.LegoDeviceManagerImpl;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyMapOf;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class LegoDeviceManagerWrapperTest extends TestCase {

    private LegoDeviceManagerWrapperImpl deviceManagerWrapper;
    private LegoDeviceManagerImpl deviceManagerMock;

    private LegoDevice deviceMock;
    private String deviceID;

    private LegoDevice deviceMock2;

    private WrapperSerialization serializationMock;
    private UnityInvoker unityInvokerMock;

    // --------- Setup ---------

    @Override
    public void setUp() throws Exception {
        super.setUp();

        deviceManagerMock = mock(LegoDeviceManagerImpl.class);
        serializationMock = spy(new WrapperSerialization());
        unityInvokerMock = mock(UnityInvoker.class);

        deviceManagerWrapper = new LegoDeviceManagerWrapperImpl(deviceManagerMock, null, serializationMock, unityInvokerMock);

        deviceID = "TestDevice-123";
        deviceMock = mock(LegoDevice.class);
        when(deviceMock.getDeviceId()).thenReturn(deviceID);

        deviceMock2 = mock(LegoDevice.class);
        when(deviceMock2.getDeviceId()).thenReturn("AnotherDevice-123456789");

        // we need to return some data when serializing devices since WrapperSerialization expects that internally
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeDevice(any(LegoDevice.class));
    }

    // --------- API methods ---------

    @SmallTest
    public void testStartScanning() throws Exception {
        // Setup
        Context contextMock = mock(Context.class);
        deviceManagerWrapper.setContext(contextMock);

        // Test
        deviceManagerWrapper.scan();

        // Verify
        verify(deviceManagerMock).scan(contextMock);
    }

    @SmallTest
    public void testStopScanning() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.stopScanning();

        // Verify
        verify(deviceManagerMock).stopScanning();
    }

    @SmallTest
    public void testConnectToDevice() throws Exception {
        // Setup
        Context contextMock = mock(Context.class);
        deviceManagerWrapper.setContext(contextMock);

        doReturn(Arrays.asList(deviceMock)).when(deviceManagerMock).allDevices();

        // Test
        deviceManagerWrapper.connectToDevice(deviceMock.getDeviceId());

        // Verify
        verify(deviceManagerMock).connectToDevice(contextMock, deviceMock);
    }

    @SmallTest
    public void testDisconnectDevice() throws Exception {
        // Setup
        doReturn(Arrays.asList(deviceMock)).when(deviceManagerMock).allDevices();

        // Test
        deviceManagerWrapper.disconnectDevice(deviceMock.getDeviceId());

        // Verify
        verify(deviceManagerMock).cancelDeviceConnection(deviceMock);
    }

    @SmallTest
    public void testAllDevices() throws Exception {
        // Setup
        doReturn(Arrays.asList(deviceMock, deviceMock2)).when(deviceManagerMock).allDevices();

        // Test
        String allDevices = deviceManagerWrapper.getAllDevices();

        // Verify
        assertTrue(allDevices != null && allDevices.length() > 0);
        verify(serializationMock).serializeDevice(deviceMock);
        verify(serializationMock).serializeDevice(deviceMock2);
    }

    // --------- Callbacks/events ---------

    @SmallTest
    private void verifyCallbackCall(String callbackName) {
        verify(serializationMock).serializeDevice(deviceMock);
        verify(unityInvokerMock).invokeMethod(eq(callbackName), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testOnDeviceAppeared() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDeviceAppeared(deviceMock);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDeviceDidAppear);
    }

    @SmallTest
    public void testOnDeviceDisappeared() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDeviceDisappeared(deviceMock);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDeviceDidDisappear);
    }

    @SmallTest
    public void testOnDidDisconnectFromDevice() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDidDisconnectFromDevice(deviceMock, false, null);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDidDisconnectFromDevice);
    }

    @SmallTest
    public void testOnDidFailToConnectToDevice() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDidFailToConnectToDevice(deviceMock, false, null);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDidFailToConnectToDevice);
    }

    @SmallTest
    public void testOnDidStartInterrogatingDevice() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDidStartInterrogatingDevice(deviceMock);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDidStartInterrogatingDevice);
    }

    @SmallTest
    public void testOnDidFinishInterrogatingDevice() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onDidFinishInterrogatingDevice(deviceMock);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerDidFinishInterrogatingDevice);
    }

    @SmallTest
    public void testOnWillStartConnectingToDevice() throws Exception {
        // Setup

        // Test
        deviceManagerWrapper.onWillStartConnectingToDevice(deviceMock);

        // Verify
        verifyCallbackCall(UnitySDKCallbacks.LEDeviceManagerWillStartConnectingToDevice);
    }
}
