package dk.lego.devicesdk.unity;

import android.test.suitebuilder.annotation.SmallTest;

import junit.framework.TestCase;

import java.util.HashMap;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.DeviceInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.services.LegoService;

import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyBoolean;
import static org.mockito.Matchers.anyMapOf;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class LegoDeviceWrapperTest extends TestCase {

    private LegoDeviceWrapperImpl deviceWrapper;

    private LegoDevice deviceMock;
    private String deviceID;

    private LegoService serviceMock;

    private LegoDeviceManagerWrapperImpl deviceManagerWrapperMock;
    private LegoServiceWrapperImpl serviceWrapperMock;
    private WrapperSerialization serializationMock;
    private UnityInvoker unityInvokerMock;

    // --------- Setup ---------

    @Override
    public void setUp() throws Exception {
        super.setUp();

        deviceManagerWrapperMock = mock(LegoDeviceManagerWrapperImpl.class);
        serializationMock = spy(new WrapperSerialization());
        unityInvokerMock =  mock(UnityInvoker.class);
        serviceWrapperMock = mock(LegoServiceWrapperImpl.class);

        deviceWrapper = new LegoDeviceWrapperImpl(deviceManagerWrapperMock, serviceWrapperMock, serializationMock, unityInvokerMock);

        deviceID = "TestDevice-123";
        deviceMock = mock(LegoDevice.class);
        when(deviceMock.getDeviceId()).thenReturn(deviceID);

        serviceMock = mock(LegoService.class);

        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeDevice(any(LegoDevice.class));
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeService(eq(serviceMock), anyBoolean());
    }

    // --------- API methods ---------

    @SmallTest
    public void testChangeDeviceName() throws Exception {
        // Setup
        String newName = "FooBar-1";
        when(deviceManagerWrapperMock.findDevice(deviceID)).thenReturn(deviceMock);

        // Test
        deviceWrapper.changeDeviceName(deviceMock.getDeviceId(), newName);

        // Verify
        verify(deviceMock).setName(newName);
    }

    // --------- Callbacks/events ---------

    @SmallTest
    public void testDidAddService() throws Exception {
        // Setup

        // Test
        deviceWrapper.didAddService(deviceMock, serviceMock);

        // Verify
        verify(serializationMock).serializeService(serviceMock, false);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidAddService), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidRemoveService() throws Exception {
        // Setup

        // Test
        deviceWrapper.didRemoveService(deviceMock, serviceMock);

        // Verify
        verify(serializationMock).serializeService(serviceMock, true);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidRemoveService), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidChangeButtonState() throws Exception {
        // Setup
        boolean buttonPressed = true;

        // Test
        deviceWrapper.didChangeButtonState(deviceMock, buttonPressed);

        // Verify
        verify(serializationMock).serializeDeviceButtonStateChange(deviceMock, buttonPressed);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidChangeButtonState), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidChangeName() throws Exception {
        // Setup
        String oldName = "OldName";
        String newName = "NewName";

        // Test
        deviceWrapper.didChangeNameFrom(deviceMock, oldName, newName);

        // Verify
        verify(serializationMock).serializeDeviceNameChange(deviceMock, oldName, newName);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidChangeName), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidFailToAddServiceWithError() throws Exception {
        // Setup
        LDSDKError error = LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, "some error");

        // Test
        deviceWrapper.didFailToAddServiceWithError(deviceMock, error);

        // Verify
        verify(serializationMock).serializeDeviceError(deviceMock, error);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidFailToAddServiceWithError), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidUpdateBatteryLevel() throws Exception {
        // Setup
        int batteryLevel = 72;

        // Test
        deviceWrapper.didUpdateBatteryLevel(deviceMock, batteryLevel);

        // Verify
        verify(serializationMock).serializeDeviceBatteryLevel(deviceMock, batteryLevel);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidUpdateBatteryLevel), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testDidUpdateDeviceInfo() throws Exception {
        // Setup
        DeviceInfo deviceInfo = new DeviceInfo();

        // Test
        deviceWrapper.didUpdateDeviceInfo(deviceMock, deviceInfo, null);

        // Verify
        verify(serializationMock).serializeDevice(deviceMock);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEDeviceDidUpdateDeviceInfo), anyMapOf(String.class, Object.class));
    }
}
