package dk.lego.devicesdk.unity;

import android.graphics.Color;
import android.test.suitebuilder.annotation.SmallTest;

import junit.framework.TestCase;

import org.mockito.ArgumentCaptor;

import java.util.Arrays;
import java.util.HashMap;

import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.LegoDeviceManager;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.CurrentSensor;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.MotionSensor;
import dk.lego.devicesdk.services.Motor;
import dk.lego.devicesdk.services.PiezoTonePlayer;
import dk.lego.devicesdk.services.RGBLight;
import dk.lego.devicesdk.services.TiltSensor;
import dk.lego.devicesdk.services.VoltageSensor;

import static org.mockito.Matchers.anyMapOf;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class LegoServiceWrapperTest extends TestCase {

    private LegoServiceWrapperImpl serviceWrapper;

    private LegoDevice deviceMock;
    private String deviceID;

    private LegoService serviceMock;
    private String serviceConnectID;

    private LegoDeviceManager deviceManagerMock;
    private WrapperSerialization serializationMock;
    private UnityInvoker unityInvokerMock;

    // --------- Setup ---------

    @Override
    public void setUp() throws Exception {
        super.setUp();

        deviceManagerMock = mock(LegoDeviceManager.class);
        serializationMock = spy(new WrapperSerialization());
        unityInvokerMock =  mock(UnityInvoker.class);

        serviceWrapper = new LegoServiceWrapperImpl(new LegoDeviceManagerWrapperImpl(deviceManagerMock, null, serializationMock, unityInvokerMock), serializationMock, unityInvokerMock);

        deviceID = "TestDevice-123";
        serviceConnectID = "1";
    }

    private void prepareServiceMockOfClass(Class serviceClass) {
        ConnectInfo connectInfo = new ConnectInfo(Integer.parseInt(serviceConnectID), 0, ConnectInfo.IOType.IO_TYPE_GENERIC);
        serviceMock = (LegoService)mock(serviceClass);
        when(serviceMock.getConnectInfo()).thenReturn(connectInfo);

        deviceMock = mock(LegoDevice.class);
        when(deviceMock.getDeviceId()).thenReturn(deviceID);
        when(deviceMock.getServices()).thenReturn(Arrays.asList(serviceMock));

        doReturn(Arrays.asList(deviceMock)).when(deviceManagerMock).allDevices();
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceWithData(serviceMock);
    }

    // --------- API methods ---------

    @SmallTest
    public void testUpdateServiceData() throws Exception {
        // Setup
        prepareServiceMockOfClass(LegoService.class);

        // Test
        serviceWrapper.updateServiceData(deviceID, serviceConnectID);

        // Verify
        verify(serializationMock).serializeServiceWithData(serviceMock);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEServiceUpdateServiceData), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testUpdateInputFormat() throws Exception {
        // Setup
        int mode = 2;
        int deltaInterval = 15;
        InputFormat.InputFormatUnit unit = InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW;
        boolean notificationsEnabled = true;

        prepareServiceMockOfClass(LegoService.class);

        // Test
        serviceWrapper.updateInputFormat(deviceID, serviceConnectID, Integer.toString(mode), Integer.toString(unit.getValue()), Integer.toString(deltaInterval), notificationsEnabled);

        // Verify
        ArgumentCaptor<InputFormat> argument = ArgumentCaptor.forClass(InputFormat.class);
        verify(serviceMock).updateInputFormat(argument.capture());
        assertEquals(argument.getValue().getMode(), mode);
        assertEquals(argument.getValue().getDeltaInterval(), (long)deltaInterval);
        assertEquals(argument.getValue().getUnit(), unit);
        assertEquals(argument.getValue().isNotificationsEnabled(), notificationsEnabled);
    }

    @SmallTest
    public void testMotorRun() throws Exception {
        // Setup
        Motor.MotorDirection direction = Motor.MotorDirection.MOTOR_DIRECTION_LEFT;
        int power = 77;
        prepareServiceMockOfClass(Motor.class);

        // Test
        serviceWrapper.run(deviceID, serviceConnectID, Integer.toString(direction.getValue()), Integer.toString(power));

        // Verify
        verify((Motor)serviceMock).run(direction, power);
    }

    @SmallTest
    public void testMotorBrake() throws Exception {
        // Setup
        prepareServiceMockOfClass(Motor.class);

        // Test
        serviceWrapper.brake(deviceID, serviceConnectID);

        // Verify
        verify((Motor)serviceMock).brake();
    }

    @SmallTest
    public void testMotorDrift() throws Exception {
        // Setup
        prepareServiceMockOfClass(Motor.class);

        // Test
        serviceWrapper.drift(deviceID, serviceConnectID);

        // Verify
        verify((Motor)serviceMock).drift();
    }

    @SmallTest
    public void testLightSwitchOff() throws Exception {
        // Setup
        prepareServiceMockOfClass(RGBLight.class);

        // Test
        serviceWrapper.switchOff(deviceID, serviceConnectID);

        // Verify
        verify((RGBLight)serviceMock).switchOff();
    }

    @SmallTest
    public void testLightChangeColorToDefault() throws Exception {
        // Setup
        prepareServiceMockOfClass(RGBLight.class);

        // Test
        serviceWrapper.changeColorToDefault(deviceID, serviceConnectID);

        // Verify
        verify((RGBLight)serviceMock).switchToDefaultColor();
    }

    @SmallTest
    public void testLightChangeColor() throws Exception {
        // Setup
        int color = Color.rgb(255, 120, 50);
        String colorJSONString = serializationMock.mapToJSONString(serializationMock.serializeColor(color));
        prepareServiceMockOfClass(RGBLight.class);

        // Test
        serviceWrapper.changeColor(deviceID, serviceConnectID, colorJSONString);

        // Verify
        verify((RGBLight)serviceMock).setColor(color);
    }

    @SmallTest
    public void testMotionSensorSetMode() throws Exception {
        // Setup
        MotionSensor.MotionSensorMode mode = MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT;
        prepareServiceMockOfClass(MotionSensor.class);

        // Test
        serviceWrapper.setMotionSensorMode(deviceID, serviceConnectID, Integer.toString(mode.getValue()));

        // Verify
        verify((MotionSensor)serviceMock).setMotionSensorMode(mode);
    }

    @SmallTest
    public void testTiltSensorSetMode() throws Exception {
        // Setup
        TiltSensor.TiltSensorMode mode = TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH;
        prepareServiceMockOfClass(TiltSensor.class);

        // Test
        serviceWrapper.setTiltSensorMode(deviceID, serviceConnectID, Integer.toString(mode.getValue()));

        // Verify
        verify((TiltSensor)serviceMock).setTiltSensorMode(mode);
    }

    @SmallTest
    public void testPiezoPlayerPlayFrequency() throws Exception {
        // Setup
        int frequency = 5000;
        int milliseconds = 350;
        prepareServiceMockOfClass(PiezoTonePlayer.class);

        // Test
        serviceWrapper.playFrequency(deviceID, serviceConnectID, Integer.toString(frequency), Integer.toString(milliseconds));

        // Verify
        verify((PiezoTonePlayer)serviceMock).playFrequency(frequency, milliseconds);
    }

    @SmallTest
    public void testPiezoPlayerPlayNote() throws Exception {
        // Setup
        PiezoTonePlayer.PiezoTonePlayerNote note = PiezoTonePlayer.PiezoTonePlayerNote.PIEZO_NOTE_B;
        int octave = 2;
        int milliseconds = 350;
        prepareServiceMockOfClass(PiezoTonePlayer.class);

        // Test
        serviceWrapper.playNote(deviceID, serviceConnectID, Integer.toString(note.getValue()), Integer.toString(octave), Integer.toString(milliseconds));

        // Verify
        verify((PiezoTonePlayer)serviceMock).playNote(note, octave, milliseconds);
    }

    @SmallTest
    public void testPiezoPlayerStopPlaying() throws Exception {
        // Setup
        prepareServiceMockOfClass(PiezoTonePlayer.class);

        // Test
        serviceWrapper.stopPlaying(deviceID, serviceConnectID);

        // Verify
        verify((PiezoTonePlayer)serviceMock).stopPlaying();
    }

    // --------- Callbacks/events ---------

    @SmallTest
    public void testServiceDidUpdateInputFormat() throws Exception {
        // Setup
        InputFormat oldInputFormat = InputFormat.inputFormat(Integer.parseInt(serviceConnectID), ConnectInfo.IOType.IO_TYPE_GENERIC, 3, 10, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);
        InputFormat newInputFormat = InputFormat.inputFormat(Integer.parseInt(serviceConnectID), ConnectInfo.IOType.IO_TYPE_MOTOR, 4, 30, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, false);

        prepareServiceMockOfClass(LegoService.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceInputFormatChange(serviceMock, oldInputFormat, newInputFormat);

        // Test
        serviceWrapper.didUpdateInputFormat(serviceMock, oldInputFormat, newInputFormat);

        // Verify
        verify(serializationMock).serializeServiceInputFormatChange(serviceMock, oldInputFormat, newInputFormat);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEServiceDidUpdateInputFormat), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testServiceDidUpdateValueData() throws Exception {
        // Setup
        // the exact data doesn't really matter, we're only interested in different data pointers
        byte[] oldData = {'a'};
        byte[] newData = {'b'};

        prepareServiceMockOfClass(LegoService.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceValueDataChange(serviceMock, oldData, newData);

        // Test
        serviceWrapper.didUpdateValueData(serviceMock, oldData, newData);

        // Verify
        verify(serializationMock).serializeServiceValueDataChange(serviceMock, oldData, newData);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEServiceDidUpdateValueData), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testCurrentSensorDidUpdateMilliAmps() throws Exception {
        // Setup
        float current = 3.14f;

        prepareServiceMockOfClass(CurrentSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceCurrentChange((CurrentSensor) serviceMock, current);

        // Test
        serviceWrapper.didUpdateMilliAmps((CurrentSensor) serviceMock, current);

        // Verify
        verify(serializationMock).serializeServiceCurrentChange((CurrentSensor) serviceMock, current);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LECurrentSensorDidUpdateMilliAmp), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testVoltageSensorDidUpdateMilliVolts() throws Exception {
        // Setup
        float voltage = 1234.0f;

        prepareServiceMockOfClass(VoltageSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceVoltageChange((VoltageSensor) serviceMock, voltage);

        // Test
        serviceWrapper.didUpdateMilliVolts((VoltageSensor) serviceMock, voltage);

        // Verify
        verify(serializationMock).serializeServiceVoltageChange((VoltageSensor) serviceMock, voltage);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEVoltageSensorDidUpdateMilliVolts), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testMotionSensorDidUpdateCount() throws Exception {
        // Setup
        int count = 35;

        prepareServiceMockOfClass(MotionSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceCountChange((MotionSensor) serviceMock, count);

        // Test
        serviceWrapper.didUpdateCount((MotionSensor) serviceMock, count);

        // Verify
        verify(serializationMock).serializeServiceCountChange((MotionSensor) serviceMock, count);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEMotionSensorDidUpdateCount), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testMotionSensorDidUpdateDistance() throws Exception {
        // Setup
        float oldDistance = 6;
        float newDistance = 8;

        prepareServiceMockOfClass(MotionSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceDistanceChange((MotionSensor) serviceMock, oldDistance, newDistance);

        // Test
        serviceWrapper.didUpdateDistance((MotionSensor) serviceMock, oldDistance, newDistance);

        // Verify
        verify(serializationMock).serializeServiceDistanceChange((MotionSensor) serviceMock, oldDistance, newDistance);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LEMotionSensorDidUpdateDistance), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testRGBLightDidUpdateColor() throws Exception {
        // Setup
        int oldColor = Color.rgb(255, 255, 255);
        int newColor = Color.rgb(200, 100, 50);

        prepareServiceMockOfClass(RGBLight.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceColorChange((RGBLight) serviceMock, oldColor, newColor);

        // Test
        serviceWrapper.didUpdateRGBValue((RGBLight) serviceMock, oldColor, newColor);

        // Verify
        verify(serializationMock).serializeServiceColorChange((RGBLight) serviceMock, oldColor, newColor);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LERGBLightDidUpdateColor), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testRGBLightDidUpdateColorIndex() throws Exception {
        // Setup
        int oldColorIndex = 1;
        int newColorIndex = 2;

        prepareServiceMockOfClass(RGBLight.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceColorIndexChange((RGBLight) serviceMock, oldColorIndex, newColorIndex);

        // Test
        serviceWrapper.didUpdateRGBColorIndex((RGBLight)serviceMock, oldColorIndex, newColorIndex);

        // Verify
        verify(serializationMock).serializeServiceColorIndexChange((RGBLight) serviceMock, oldColorIndex, newColorIndex);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LERGBLightDidUpdateColorIndex), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testTiltSensorDidUpdateAngle() throws Exception {
        // Setup
        TiltSensor.TiltSensorAngle oldAngle = TiltSensor.tiltSensorAngleMake(10, 10);
        TiltSensor.TiltSensorAngle newAngle = TiltSensor.tiltSensorAngleMake(20, 20);

        prepareServiceMockOfClass(TiltSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceAngleChange((TiltSensor)serviceMock, oldAngle, newAngle);

        // Test
        serviceWrapper.didUpdateAngle((TiltSensor)serviceMock, oldAngle, newAngle);

        // Verify
        verify(serializationMock).serializeServiceAngleChange((TiltSensor)serviceMock, oldAngle, newAngle);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LETiltSensorDidUpdateAngle), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testTiltSensorDidUpdateCrash() throws Exception {
        // Setup
        TiltSensor.TiltSensorCrash oldCrash = TiltSensor.tiltSensorCrashMake(1, 2, 3);
        TiltSensor.TiltSensorCrash newCrash = TiltSensor.tiltSensorCrashMake(7, 8, 9);

        prepareServiceMockOfClass(TiltSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceCrashChange((TiltSensor)serviceMock, oldCrash, newCrash);

        // Test
        serviceWrapper.didUpdateCrash((TiltSensor)serviceMock, oldCrash, newCrash);

        // Verify
        verify(serializationMock).serializeServiceCrashChange((TiltSensor)serviceMock, oldCrash, newCrash);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LETiltSensorDidUpdateCrash), anyMapOf(String.class, Object.class));
    }

    @SmallTest
    public void testTiltSensorDidUpdateDirection() throws Exception {
        // Setup
        TiltSensor.TiltSensorDirection oldDirection = TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_LEFT;
        TiltSensor.TiltSensorDirection newDirection = TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_RIGHT;

        prepareServiceMockOfClass(TiltSensor.class);
        doReturn(new HashMap<String, Object>()).when(serializationMock).serializeServiceDirectionChange((TiltSensor) serviceMock, oldDirection, newDirection);

        // Test
        serviceWrapper.didUpdateDirection((TiltSensor) serviceMock, oldDirection, newDirection);

        // Verify
        verify(serializationMock).serializeServiceDirectionChange((TiltSensor) serviceMock, oldDirection, newDirection);
        verify(unityInvokerMock).invokeMethod(eq(UnitySDKCallbacks.LETiltSensorDidUpdateDirection), anyMapOf(String.class, Object.class));
    }
}
