package dk.lego.devicesdk.unity;

import android.graphics.Color;
import android.test.suitebuilder.annotation.SmallTest;

import junit.framework.TestCase;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.device.DeviceInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.Revision;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.services.CurrentSensor;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.MotionSensor;
import dk.lego.devicesdk.services.Motor;
import dk.lego.devicesdk.services.PiezoTonePlayer;
import dk.lego.devicesdk.services.RGBLight;
import dk.lego.devicesdk.services.TiltSensor;
import dk.lego.devicesdk.services.VoltageSensor;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@SuppressWarnings("unchecked")
public class WrapperSerializationTest extends TestCase {

    private String deviceID = "DeviceID";
    private LegoDevice deviceMock;

    private String deviceID2 = "AnotherDevice-123";
    private LegoDevice deviceMock2;

    private int serviceConnectID = 2;
    private LegoService serviceMock;

    private WrapperSerialization wrapperSerialization = new WrapperSerialization();

    // --------- Setup ---------

    @Override
    public void setUp() throws Exception {
        super.setUp();

        deviceMock = prepareDeviceMock(deviceID);
        deviceMock2 = prepareDeviceMock(deviceID2);
        serviceMock = prepareServiceMock(LegoService.class);
    }

    private LegoDevice prepareDeviceMock(String deviceID) {
        DeviceInfo deviceInfo = new DeviceInfo();
        deviceInfo.setFirmwareRevisionString("1.1.1.1");
        deviceInfo.setSoftwareRevisionString("2.2.2.2");
        deviceInfo.setHardwareRevisionString("3.3.3.3");
        deviceInfo.setManufacturerName("LEGO");

        LegoDevice deviceMock = mock(LegoDevice.class);
        when(deviceMock.getDeviceInfo()).thenReturn(deviceInfo);
        when(deviceMock.getName()).thenReturn("DeviceName");
        when(deviceMock.getDeviceId()).thenReturn(deviceID);
        when(deviceMock.isButtonPressed()).thenReturn(true);
        when(deviceMock.getBatteryLevel()).thenReturn(90);
        when(deviceMock.getConnectState()).thenReturn(LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_INTERROGATING);

        return deviceMock;
    }

    private LegoService prepareServiceMock(Class serviceClass) {
        InputFormat inputFormat = InputFormat.inputFormat(1, ConnectInfo.IOType.fromInteger(1), 3, 10, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);
        InputFormat defaultInputFormat = InputFormat.inputFormat(5, ConnectInfo.IOType.fromInteger(20), 8, 30, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, false);
        ConnectInfo connectInfo = new ConnectInfo(serviceConnectID, 1, ConnectInfo.IOType.IO_TYPE_GENERIC.getValue(), Revision.createFromString("1.1.1.1"), Revision.createFromString("2.2.2.2"));

        LegoService serviceMock = (LegoService)mock(serviceClass);
        when(serviceMock.getServiceName()).thenReturn("Test service");
        when(serviceMock.getInputFormat()).thenReturn(inputFormat);
        when(serviceMock.getDefaultInputFormat()).thenReturn(defaultInputFormat);
        when(serviceMock.isInternalService()).thenReturn(true);
        when(serviceMock.getConnectInfo()).thenReturn(connectInfo);
        when(serviceMock.getDevice()).thenReturn(deviceMock);

        return serviceMock;
    }

    private void verifySerializedDevice(Map<String, Object> serializedDevice, LegoDevice mock, boolean onlyBasic) {
        assert mock == deviceMock || mock == deviceMock2;

        assertEquals(mock == deviceMock ? deviceID : deviceID2, serializedDevice.get("DeviceID"));
        if (onlyBasic) {
            return;
        }

        assertEquals("DeviceName", serializedDevice.get("DeviceName"));
        assertEquals(true, serializedDevice.get("ButtonPressed"));
        assertEquals(90, serializedDevice.get("BatteryLevel"));
        assertEquals(3, serializedDevice.get("ConnectedState"));

        Map<String, Object> serializedDeviceInfo = (Map<String, Object>)serializedDevice.get("DeviceInfo");
        assertEquals("1.1.1.1", serializedDeviceInfo.get("FirmwareRevision"));
        assertEquals("2.2.2.2", serializedDeviceInfo.get("SoftwareRevision"));
        assertEquals("3.3.3.3", serializedDeviceInfo.get("HardwareRevision"));
        assertEquals("LEGO", serializedDeviceInfo.get("ManufacturerName"));
    }

    private void verifySerializedService(Map<String, Object> serializedService, boolean onlyBasic) {
        if (onlyBasic) {
            assertEquals(deviceID, serializedService.get("DeviceID"));
            assertEquals(serviceConnectID, serializedService.get("ConnectID"));
            return;
        }

        assertEquals("Test service", serializedService.get("ServiceName"));
        assertEquals(true, serializedService.get("IsInternalService"));
        assertEquals(deviceID, serializedService.get("DeviceID"));

        Map<String, Object> connectInfo = (Map<String, Object>)serializedService.get("ConnectInfo");
        assertEquals(serviceConnectID, connectInfo.get("ConnectID"));
        assertEquals(1, connectInfo.get("HubIndex"));
        assertEquals("1.1.1.1", connectInfo.get("HardwareRevision"));
        assertEquals("2.2.2.2", connectInfo.get("SoftwareRevision"));
        assertEquals(ConnectInfo.IOType.IO_TYPE_GENERIC.getValue(), connectInfo.get("Type"));

        Map<String, Object> inputFormat = (Map<String, Object>)serializedService.get("InputFormat");
        assertEquals(0, inputFormat.get("Revision"));
        assertEquals(1, inputFormat.get("ConnectID"));
        assertEquals(1, inputFormat.get("TypeID"));
        assertEquals(3, inputFormat.get("Mode"));
        assertEquals((long)10, inputFormat.get("DeltaInterval"));
        assertEquals(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI.getValue(), inputFormat.get("Unit"));
        assertEquals(true, inputFormat.get("NotificationsEnabled"));

        Map<String, Object> defaultInputFormat = (Map<String, Object>)serializedService.get("DefaultInputFormat");
        assertEquals(0, defaultInputFormat.get("Revision"));
        assertEquals(5, defaultInputFormat.get("ConnectID"));
        assertEquals(20, defaultInputFormat.get("TypeID"));
        assertEquals(8, defaultInputFormat.get("Mode"));
        assertEquals((long)30, defaultInputFormat.get("DeltaInterval"));
        assertEquals(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW.getValue(), defaultInputFormat.get("Unit"));
        assertEquals(false, defaultInputFormat.get("NotificationsEnabled"));
    }

    // --------- API methods ---------

    @SmallTest
    public void testDevice() throws Exception {
        // Setup
        LegoDevice deviceMock = this.deviceMock;

        // Test
        Map<String, Object> serializedDevice = wrapperSerialization.serializeDevice(deviceMock);

        // Verify
        verifySerializedDevice(serializedDevice, deviceMock, false);
    }

    @SmallTest
    public void testDevices() throws Exception {
        // Setup
        LegoDevice[] deviceMocks = {this.deviceMock, this.deviceMock2};

        // Test
        List<Map<String, Object>> serializedDevices = wrapperSerialization.serializeDevices(Arrays.asList(deviceMocks));

        // Verify
        assertEquals(2, serializedDevices.size());
        verifySerializedDevice(serializedDevices.get(0), deviceMock, false);
        verifySerializedDevice(serializedDevices.get(1), deviceMock2, false);
    }

    @SmallTest
    public void testDeviceNameChange() throws Exception {
        // Setup
        LegoDevice deviceMock = this.deviceMock;
        String deviceOldName = "OldName";
        String deviceNewName = "NewName";

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeDeviceNameChange(deviceMock, deviceOldName, deviceNewName);

        // Verify
        verifySerializedDevice(serialized, deviceMock, true);
        assertEquals(deviceOldName, serialized.get("OldDeviceName"));
        assertEquals(deviceNewName, serialized.get("DeviceName"));
    }

    @SmallTest
    public void testDeviceButtonStateChange() throws Exception {
        // Setup
        LegoDevice deviceMock = this.deviceMock;
        boolean buttonState = true;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeDeviceButtonStateChange(deviceMock, buttonState);

        // Verify
        verifySerializedDevice(serialized, deviceMock, true);
        assertEquals(buttonState, serialized.get("ButtonState"));
    }

    @SmallTest
    public void testDeviceBatteryLevel() throws Exception {
        // Setup
        LegoDevice deviceMock = this.deviceMock;
        int batteryLevel = 42;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeDeviceBatteryLevel(deviceMock, batteryLevel);

        // Verify
        verifySerializedDevice(serialized, deviceMock, true);
        assertEquals(batteryLevel, serialized.get("BatteryLevel"));
    }

    @SmallTest
    public void testDeviceError() throws Exception {
        // Setup
        LegoDevice deviceMock = this.deviceMock;
        String errorDescription = "some error";
        LDSDKError error = LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, errorDescription);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeDeviceError(deviceMock, error);

        // Verify
        verifySerializedDevice(serialized, deviceMock, true);
        assertEquals(errorDescription, serialized.get("Error"));
    }

    @SmallTest
    public void testInputFormat() throws Exception {
        // Setup
        InputFormat inputFormat = InputFormat.inputFormat(1, ConnectInfo.IOType.IO_TYPE_GENERIC, 3, 10, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeInputFormat(inputFormat);

        // Verify
        assertEquals(0, serialized.get("Revision"));
        assertEquals(1, serialized.get("ConnectID"));
        assertEquals(ConnectInfo.IOType.IO_TYPE_GENERIC.getValue(), serialized.get("TypeID"));
        assertEquals(3, serialized.get("Mode"));
        assertEquals((long)10, serialized.get("DeltaInterval"));
        assertEquals(InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI.getValue(), serialized.get("Unit"));
        assertEquals(true, serialized.get("NotificationsEnabled"));
    }

    @SmallTest
    public void testServiceFull() throws Exception {
        // Setup
        LegoService serviceMock = this.serviceMock;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeService(serviceMock);

        // Verify
        verifySerializedService(serialized, false);
    }

    @SmallTest
    public void testServiceBasic() throws Exception {
        // Setup
        LegoService serviceMock = this.serviceMock;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeService(serviceMock, true);

        // Verify
        verifySerializedService(serialized, true);
    }

    @SmallTest
    public void testMotorServiceData() throws Exception {
        // Setup
        Motor motorMock = (Motor)prepareServiceMock(Motor.class);
        when(motorMock.getPower()).thenReturn(24);
        when(motorMock.getDirection()).thenReturn(Motor.MotorDirection.MOTOR_DIRECTION_LEFT);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(motorMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");
        assertEquals(24, serviceData.get("Power"));
        assertEquals(1, serviceData.get("MotorDirection"));
    }

    @SmallTest
    public void testRGBLightServiceData() throws Exception {
        // Setup
        RGBLight lightMock = (RGBLight)prepareServiceMock(RGBLight.class);
        when(lightMock.getColor()).thenReturn(Color.argb(200, 120, 130, 140));
        when(lightMock.getDefaultColor()).thenReturn(Color.argb(255, 200, 210, 220));

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(lightMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");

        Map<String, Object> color = (Map<String, Object>)serviceData.get("Color");
        assertEquals(120 / 255.0, color.get("R"));
        assertEquals(130 / 255.0, color.get("G"));
        assertEquals(140 / 255.0, color.get("B"));
        assertEquals(200 / 255.0, color.get("A"));

        Map<String, Object> defaultColor = (Map<String, Object>)serviceData.get("DefaultColor");
        assertEquals(200 / 255.0, defaultColor.get("R"));
        assertEquals(210 / 255.0, defaultColor.get("G"));
        assertEquals(220 / 255.0, defaultColor.get("B"));
        assertEquals(255 / 255.0, defaultColor.get("A"));
    }

    @SmallTest
    public void testMotionSensorServiceData() throws Exception {
        // Setup
        MotionSensor motionSensorMock = (MotionSensor)prepareServiceMock(MotionSensor.class);
        when(motionSensorMock.getCount()).thenReturn(2);
        when(motionSensorMock.getDistance()).thenReturn(3.14f);
        when(motionSensorMock.getMotionSensorMode()).thenReturn(MotionSensor.MotionSensorMode.MOTION_SENSOR_MODE_COUNT);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(motionSensorMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");
        assertEquals(2, serviceData.get("Count"));
        assertEquals(3.14f, serviceData.get("Distance"));
        assertEquals(1, serviceData.get("MotionSensorMode")); // MOTION_SENSOR_MODE_COUNT is 1
    }

    @SmallTest
    public void testTiltSensorServiceData() throws Exception {
        // Setup
        TiltSensor tiltSensorMock = (TiltSensor)prepareServiceMock(TiltSensor.class);
        when(tiltSensorMock.getDirection()).thenReturn(TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_RIGHT);
        when(tiltSensorMock.getTiltSensorMode()).thenReturn(TiltSensor.TiltSensorMode.TILT_SENSOR_MODE_CRASH);
        when(tiltSensorMock.getCrash()).thenReturn(TiltSensor.tiltSensorCrashMake(1, 2, 3));
        when(tiltSensorMock.getAngle()).thenReturn(TiltSensor.tiltSensorAngleMake(20, 30));

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(tiltSensorMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");

        assertEquals(5, serviceData.get("TiltSensorDirection")); // TILT_SENSOR_DIRECTION_RIGHT is 5
        assertEquals(2, serviceData.get("TiltSensorMode")); // TILT_SENSOR_MODE_CRASH is 2

        Map<String, Object> crash = (Map<String, Object>)serviceData.get("Crash");
        assertEquals(1, crash.get("X"));
        assertEquals(2, crash.get("Y"));
        assertEquals(3, crash.get("Z"));

        Map<String, Object> angle = (Map<String, Object>)serviceData.get("Angle");
        assertEquals(20.0f, angle.get("X"));
        assertEquals(30.0f, angle.get("Y"));
    }

    @SmallTest
    public void testVoltageSensorServiceData() throws Exception {
        // Setup
        VoltageSensor voltageSensorMock = (VoltageSensor)prepareServiceMock(VoltageSensor.class);
        when(voltageSensorMock.getValueAsMilliVolts()).thenReturn(1234.25f);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(voltageSensorMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");
        assertEquals(1234.25f, serviceData.get("MilliVolts"));
    }

    @SmallTest
    public void testCurrentSensorServiceData() throws Exception {
        // Setup
        CurrentSensor currentSensorMock = (CurrentSensor)prepareServiceMock(CurrentSensor.class);
        when(currentSensorMock.getValueAsMilliAmps()).thenReturn(1234.25f);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(currentSensorMock);

        // Verify
        verifySerializedService(serialized, false);
        Map<String, Object> serviceData = (Map<String, Object>)serialized.get("ServiceData");
        assertEquals(1234.25f, serviceData.get("MilliAmp"));
    }

    @SmallTest
    public void testPiezoTonePlayerServiceData() throws Exception {
        // Setup
        PiezoTonePlayer piezoPlayerMock = (PiezoTonePlayer)prepareServiceMock(PiezoTonePlayer.class);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(piezoPlayerMock);

        // Verify
        verifySerializedService(serialized, false);
        assertNull(serialized.get("ServiceData")); // No service data should be generated for piezo tone player, at least for now
    }

    // Custom service mock
    public class FooBarService extends LegoService {
        FooBarService() { super(null, null); }
    }

    @SmallTest
    public void testUnknownServiceData() throws Exception {
        // Setup
        FooBarService unknownServiceMock = (FooBarService)prepareServiceMock(FooBarService.class);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceWithData(unknownServiceMock);

        // Verify
        verifySerializedService(serialized, false);
        assertNull(serialized.get("ServiceData")); // No service data should be generated for unknown services
    }

    @SmallTest
    public void testServiceValueDataDidChange() throws Exception {
        // Setup
        // the exact data doesn't really matter, we're only interested in different data pointers
        byte[] oldData = {'a'};
        byte[] newData = {'b'};
        LegoService serviceMock = this.serviceMock;
        when(serviceMock.getFloatFromData(oldData)).thenReturn(1.0f);
        when(serviceMock.getFloatFromData(newData)).thenReturn(2.0f);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceValueDataChange(serviceMock, oldData, newData);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(1.0f, serialized.get("OldValue"));
        assertEquals(2.0f, serialized.get("NewValue"));
    }

    @SmallTest
    public void testServiceInputFormatDidChange() throws Exception {
        // Setup
        LegoService serviceMock = this.serviceMock;
        InputFormat oldInputFormat = InputFormat.inputFormat(1, ConnectInfo.IOType.IO_TYPE_MOTOR, 3, 10, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);
        InputFormat newInputFormat = InputFormat.inputFormat(1, ConnectInfo.IOType.IO_TYPE_MOTION_SENSOR, 4, 30, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, false);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceInputFormatChange(serviceMock, oldInputFormat, newInputFormat);

        // Verify
        verifySerializedService(serialized, true);

        Map<String, Object> serializedOldFormat = (Map<String, Object>)serialized.get("OldInputFormat");
        assertEquals(0, serializedOldFormat.get("Revision"));
        assertEquals(1, serializedOldFormat.get("ConnectID"));
        assertEquals(1, serializedOldFormat.get("TypeID")); // IO_TYPE_MOTOR is 1
        assertEquals(3, serializedOldFormat.get("Mode"));
        assertEquals((long)10, serializedOldFormat.get("DeltaInterval"));
        assertEquals(2, serializedOldFormat.get("Unit")); // INPUT_FORMAT_UNIT_SI is 2
        assertEquals(true, serializedOldFormat.get("NotificationsEnabled"));

        Map<String, Object> serializedNewFormat = (Map<String, Object>)serialized.get("NewInputFormat");
        assertEquals(0, serializedNewFormat.get("Revision"));
        assertEquals(1, serializedNewFormat.get("ConnectID"));
        assertEquals(35, serializedNewFormat.get("TypeID")); // IO_TYPE_MOTION_SENSOR is 35
        assertEquals(4, serializedNewFormat.get("Mode"));
        assertEquals((long)30, serializedNewFormat.get("DeltaInterval"));
        assertEquals(0, serializedNewFormat.get("Unit")); // INPUT_FORMAT_UNIT_RAW is 0
        assertEquals(false, serializedNewFormat.get("NotificationsEnabled"));
    }

    @SmallTest
    public void testMotionSensorDistanceDidChange() throws Exception {
        // Setup
        MotionSensor motionSensorMock = (MotionSensor)prepareServiceMock(MotionSensor.class);
        float oldDistance = 42.0f;
        float newDistance = 3.14f;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceDistanceChange(motionSensorMock, oldDistance, newDistance);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(oldDistance, serialized.get("OldDistance"));
        assertEquals(newDistance, serialized.get("NewDistance"));
    }

    @SmallTest
    public void testMotionSensorCountDidChange() throws Exception {
        // Setup
        MotionSensor motionSensorMock = (MotionSensor)prepareServiceMock(MotionSensor.class);
        int count = 4;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceCountChange(motionSensorMock, count);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(count, serialized.get("Count"));
    }

    @SmallTest
    public void testRGBLightColorDidChange() throws Exception {
        // Setup
        RGBLight lightMock = (RGBLight)prepareServiceMock(RGBLight.class);
        int oldColor = Color.argb(255, 255, 255, 255);
        int newColor = Color.argb(255, 200, 100, 150);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceColorChange(lightMock, oldColor, newColor);

        // Verify
        verifySerializedService(serialized, true);

        Map<String, Object> serializedOldColor = (Map<String, Object>)serialized.get("OldColor");
        assertEquals(Color.red(oldColor) / 255.0, serializedOldColor.get("R"));
        assertEquals(Color.green(oldColor) / 255.0, serializedOldColor.get("G"));
        assertEquals(Color.blue(oldColor) / 255.0, serializedOldColor.get("B"));
        assertEquals(Color.alpha(oldColor) / 255.0, serializedOldColor.get("A"));

        Map<String, Object> serializedNewColor = (Map<String, Object>)serialized.get("NewColor");
        assertEquals(Color.red(newColor) / 255.0, serializedNewColor.get("R"));
        assertEquals(Color.green(newColor) / 255.0, serializedNewColor.get("G"));
        assertEquals(Color.blue(newColor) / 255.0, serializedNewColor.get("B"));
        assertEquals(Color.alpha(newColor) / 255.0, serializedNewColor.get("A"));
    }

    @SmallTest
    public void testTiltSensorDirectionDidChange() throws Exception {
        // Setup
        TiltSensor tiltSensorMock = (TiltSensor)prepareServiceMock(TiltSensor.class);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceDirectionChange(tiltSensorMock, TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_LEFT, TiltSensor.TiltSensorDirection.TILT_SENSOR_DIRECTION_FORWARD);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(7, serialized.get("OldDirection")); // TILT_SENSOR_DIRECTION_LEFT is 7
        assertEquals(9, serialized.get("NewDirection")); // TILT_SENSOR_DIRECTION_FORWARD is 9
    }

    @SmallTest
    public void testTiltSensorAngleDidChange() throws Exception {
        // Setup
        TiltSensor tiltSensorMock = (TiltSensor)prepareServiceMock(TiltSensor.class);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceAngleChange(tiltSensorMock, TiltSensor.tiltSensorAngleMake(10, 20), TiltSensor.tiltSensorAngleMake(15, 15));

        // Verify
        verifySerializedService(serialized, true);

        Map<String, Object> serializedOldAngle = (Map<String, Object>)serialized.get("OldAngle");
        assertEquals(10.0f, serializedOldAngle.get("X"));
        assertEquals(20.0f, serializedOldAngle.get("Y"));

        Map<String, Object> serializedNewAngle = (Map<String, Object>)serialized.get("NewAngle");
        assertEquals(15.0f, serializedNewAngle.get("X"));
        assertEquals(15.0f, serializedNewAngle.get("Y"));
    }

    @SmallTest
    public void testTiltSensorCrashDidChange() throws Exception {
        // Setup
        TiltSensor tiltSensorMock = (TiltSensor)prepareServiceMock(TiltSensor.class);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceCrashChange(tiltSensorMock, TiltSensor.tiltSensorCrashMake(1, 2, 3), TiltSensor.tiltSensorCrashMake(3, 5, 7));

        // Verify
        verifySerializedService(serialized, true);

        Map<String, Object> serializedOldCrash = (Map<String, Object>)serialized.get("OldCrash");
        assertEquals(1, serializedOldCrash.get("X"));
        assertEquals(2, serializedOldCrash.get("Y"));
        assertEquals(3, serializedOldCrash.get("Z"));

        Map<String, Object> serializedNewCrash = (Map<String, Object>)serialized.get("NewCrash");
        assertEquals(3, serializedNewCrash.get("X"));
        assertEquals(5, serializedNewCrash.get("Y"));
        assertEquals(7, serializedNewCrash.get("Z"));
    }

    @SmallTest
    public void testVoltageSensorValueDidChange() throws Exception {
        // Setup
        VoltageSensor voltageSensorMock = (VoltageSensor)prepareServiceMock(VoltageSensor.class);
        float voltage = 2015.25f;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceVoltageChange(voltageSensorMock, voltage);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(voltage, serialized.get("MilliVolts"));
    }

    @SmallTest
    public void testCurrentSensorValueDidChange() throws Exception {
        // Setup
        CurrentSensor currentSensorMock = (CurrentSensor)prepareServiceMock(CurrentSensor.class);
        float current = 314.5f;

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeServiceCurrentChange(currentSensorMock, current);

        // Verify
        verifySerializedService(serialized, true);
        assertEquals(current, serialized.get("MilliAmp"));
    }

    @SmallTest
    public void testSerializeColor() throws Exception {
        // Setup
        int color = Color.argb(200, 50, 100, 170);

        // Test
        Map<String, Object> serialized = wrapperSerialization.serializeColor(color);

        // Verify
        assertEquals(Color.red(color) / 255.0, serialized.get("R"));
        assertEquals(Color.green(color) / 255.0, serialized.get("G"));
        assertEquals(Color.blue(color) / 255.0, serialized.get("B"));
        assertEquals(Color.alpha(color) / 255.0, serialized.get("A"));
    }

    @SmallTest
    public void testDeserializeColor() throws Exception {
        // Setup
        Map<String, Object> color = new HashMap<>();
        color.put("R", 0.5); color.put("G", 0.6); color.put("B", 0.7); color.put("A", 0.9);

        // Test
        int deserialized = wrapperSerialization.deserializeColor(color);

        // Verify
        assertEquals((int)(0.5 * 255), Color.red(deserialized));
        assertEquals((int)(0.6 * 255), Color.green(deserialized));
        assertEquals((int) (0.7 * 255), Color.blue(deserialized));
        assertEquals((int)(0.9 * 255), Color.alpha(deserialized));
    }
}
