package dk.lego.devicesdk.bluetooth;

import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;

import java.util.ArrayList;
import java.util.EnumSet;
import java.util.List;
import java.util.UUID;

public class BluetoothHelper {
    private static final String UUID_CUSTOM_BASE    = "1212-EFDE-1523-785FEABCD123";
    public static final String UUID_STANDARD_BASE  = "0000-1000-8000-00805f9b34fb";

    public static enum Permissions {
        PERMISSION_READ,                    // 1   - Characteristic read permission
        PERMISSION_READ_ENCRYPTED,          // 2   - Allow encrypted read operations
        PERMISSION_READ_ENCRYPTED_MITM,     // 4   - Allow reading with man-in-the-middle protection
        UNUSED,                             // 8   - Unused
        PERMISSION_WRITE,                   // 16  - Write permission
        PERMISSION_WRITE_ENCRYPTED,         // 32  - Allow encrypted writes
        PERMISSION_WRITE_ENCRYPTED_MITM,    // 64  - Allow encrypted writes with man-in-the-middle protection
        PERMISSION_WRITE_SIGNED,            // 128 - Allow signed write operations
        PERMISSION_WRITE_SIGNED_MITM        // 256 - Allow signed write operations with man-in-the-middle protection
    }

    public static enum Properties {
        PROPERTY_BROADCAST,                 // 1   - Characteristic is broadcastable.
        PROPERTY_READ,                      // 2   - Characteristic is readable.
        PROPERTY_WRITE_NO_RESPONSE,         // 4   - Characteristic can be written without response.
        PROPERTY_WRITE,                     // 8   - Characteristic can be written.
        PROPERTY_NOTIFY,                    // 16  - Characteristic supports notification
        PROPERTY_INDICATE,                  // 32  - Characteristic supports indication
        PROPERTY_SIGNED_WRITE,              // 64  - Characteristic supports write with signature
        PROPERTY_EXTENDED_PROPS             // 128 - Characteristic has extended properties
    }

    public static String UUIDWithPrefixCustomBase(String prefix)
    {
        String padding = addLeadingZeroes(prefix);
        return String.format("%s-%s", padding, UUID_CUSTOM_BASE);
    }

    public static String UUIDWithPrefixStandardBase(String prefix)
    {
        String padding = addLeadingZeroes(prefix);
        return String.format("%s-%s", padding, UUID_STANDARD_BASE);
    }

    public static String addLeadingZeroes(String prefix) {
        // Add leading zeroes such that prefix is 8 chars long
        String hexPrefix = "0x";
        if (prefix.startsWith(hexPrefix)) {
            prefix = prefix.substring(hexPrefix.length());
        }
        return ("00000000" + prefix).substring(prefix.length());
    }

    public static BluetoothGattCharacteristic create(UUID characteristicUUID, BluetoothGattService service)
    {
        for (BluetoothGattCharacteristic characteristic : service.getCharacteristics()) {
            if (characteristicUUID.equals(characteristic.getUuid())) {
                return characteristic;
            }
        }
        return null;
    }

    public static List<String> arrayOfStringsFromCharacteristicProperties(int properties)
    {
        List<Properties> list = new ArrayList<>();
        for (Properties value : Properties.values()) {
            if ((properties & (1 << value.ordinal())) != 0) {
                list.add(value);
            }
        }

        List<String> propStrings = new ArrayList<>();
        if (list.size() > 0) {
            EnumSet<Properties> result = EnumSet.copyOf(list);

            for (Properties property : result) {
                propStrings.add(property.name());
            }
        }

        return propStrings;
    }
}
