package dk.lego.devicesdk.bluetooth;

import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.util.SparseArray;

import java.nio.ByteBuffer;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.bluetooth.service_definitions.IOServiceDefinition;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.InputCommand;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.input_output.OutputCommand;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.utils.ByteUtils;

public class BluetoothIO extends IO {

    private final BluetoothGattService service;
    private BluetoothGattCharacteristic inputValueCharacteristic;
    private BluetoothGattCharacteristic inputFormatCharacteristic;
    private BluetoothGattCharacteristic inputCommandCharacteristic;
    private BluetoothGattCharacteristic outputCommandCharacteristic;

    IOServiceDefinition serviceDefinition;
    SparseArray<InputFormat> inputFormats = new SparseArray<>();

    public static BluetoothIO bluetoothIO(BluetoothGattService service, LegoBluetoothDevice associatedDevice) {
        assert(service != null) : "Cannot create BluetoothIO with null service";

        IOServiceDefinition serviceDefinition = IOServiceDefinition.getInstance();

        BluetoothGattCharacteristic inputValueCharacteristic = BluetoothHelper.create(serviceDefinition.getInputValue().getUuid(), service);
        BluetoothGattCharacteristic inputFormatCharacteristic = BluetoothHelper.create(serviceDefinition.getInputFormat().getUuid(), service);
        BluetoothGattCharacteristic inputCommandCharacteristic = BluetoothHelper.create(serviceDefinition.getInputCommand().getUuid(), service);
        BluetoothGattCharacteristic outputCommandCharacteristic = BluetoothHelper.create(serviceDefinition.getOutputCommand().getUuid(), service);

        if (inputValueCharacteristic == null
                || inputFormatCharacteristic == null
                || inputCommandCharacteristic == null
                || outputCommandCharacteristic == null) {
            LDSDKLogger.e("IOService missing mandatory characteristic");
            return null;
        }

        return new BluetoothIO(service,
                associatedDevice,
                serviceDefinition,
                inputValueCharacteristic,
                inputFormatCharacteristic,
                inputCommandCharacteristic,
                outputCommandCharacteristic);
    }

    private BluetoothIO(BluetoothGattService service,
                        LegoBluetoothDevice associatedDevice,
                        IOServiceDefinition definition,
                        BluetoothGattCharacteristic inputValueCharacteristic,
                        BluetoothGattCharacteristic inputFormatCharacteristic,
                        BluetoothGattCharacteristic inputCommandCharacteristic,
                        BluetoothGattCharacteristic outputCommandCharacteristic) {

        this.serviceDefinition = definition;

        this.inputValueCharacteristic = inputValueCharacteristic;
        this.inputFormatCharacteristic = inputFormatCharacteristic;
        this.inputCommandCharacteristic = inputCommandCharacteristic;
        this.outputCommandCharacteristic = outputCommandCharacteristic;

        this.service = service;
        this.associatedDevice = associatedDevice;

        this.associatedDevice.setCharacteristicNotification(inputValueCharacteristic, true);
        this.associatedDevice.setCharacteristicNotification(inputFormatCharacteristic, true);
    }

    @Override
    public void readValueForConnectId(int connectId)
    {
        writeInputCommand(InputCommand.commandReadValueForConnectId(connectId));
    }

    @Override
    public void resetStateForConnectId(int connectId)
    {
        // Byte sequence sent to sensor to reset any state (for instance, crash-count for tilt sensor)
        byte[] resetBytes = { 0x44, 0x11, (byte) 0xAA};
        writeData(resetBytes, connectId);
    }

    @Override
    public void writeInputFormat(InputFormat inputFormat, int connectId)
    {
        writeInputCommand(InputCommand.commandWriteInputFormat(inputFormat, connectId));
    }

    @Override
    public void readInputFormatForConnectId(int connectId)
    {
        writeInputCommand(InputCommand.commandReadInputFormatForConnectId(connectId));
    }

    public void writeInputCommand(InputCommand command)
    {
        LDSDKLogger.v(String.format("Writing Input Command: %s", ByteUtils.toHexString(command.getData())));
        associatedDevice.addWriteOfCharacteristicToQueue(inputCommandCharacteristic, command.getData(), BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
    }

    @Override
    public void writeMotorPower(int power, int connectId) {
        writeMotorPower(power, 0, connectId);
    }

    @Override
    public void writeMotorPower(int power, int offset, int connectId) {
        boolean isPositive = power >= 0;
        power = Math.abs(power);

        float actualPower = ((100.0f - offset) / 100.0f) * power + offset;
        int actualResultInt = Math.round(actualPower);

        if (!isPositive) {
            actualResultInt = -actualResultInt;
        }

        OutputCommand outputCommand = OutputCommand.commandWriteMotorPower(actualResultInt, connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Writing motor power command: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    @Override
    public void writePiezoToneFrequency(int frequency, int duration, int connectId) {
        OutputCommand outputCommand = OutputCommand.commandWritePiezoToneFrequency(frequency, duration, connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Writing piezo tone play command: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    @Override
    public void writePiezoToneStop(int connectId) {
        OutputCommand outputCommand = OutputCommand.commandWritePiezoToneStopForConnectId(connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Writing piezo tone stop command: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    @Override
    public void writeColor(int red, int green, int blue, int connectId) {
        OutputCommand outputCommand = OutputCommand.commandWriteRGBLight(red, green, blue, connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Writing RGB command: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    @Override
    public void writeColorIndex(int index, int connectId) {
        OutputCommand outputCommand = OutputCommand.commandWriteRGBLightIndex(index, connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Writing RGB Index command: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    @Override
    public void writeData(byte[] data, int connectId) {
        OutputCommand outputCommand = OutputCommand.commandWithDirectWriteThroughData(data, connectId);
        writeOutputCommand(outputCommand);
        LDSDKLogger.d(String.format("Direct write with data: %s", ByteUtils.toHexString(outputCommand.getData())));
    }

    public void writeOutputCommand(OutputCommand command)
    {
        associatedDevice.addWriteOfCharacteristicToQueue(outputCommandCharacteristic, command.getData(), BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE);
    }

    public void handleWriteResponseFromIOServiceWithCharacteristic(BluetoothGattCharacteristic characteristic, LDSDKError error)
    {
        if (error == null) {
            LDSDKLogger.d(String.format("Did write data %s", ByteUtils.toHexString(characteristic.getValue())));
        } else {
            LDSDKLogger.e(String.format("Failed to write data for characteristic %s", characteristic.toString()));
        }
    }

    public void handleUpdatedInputServiceCharacteristic(BluetoothGattCharacteristic characteristic) {
        if (serviceDefinition.getInputFormat().matchesCharacteristic(characteristic)) {
            this.handleUpdatedInputFormatData(characteristic.getValue());
        } else if (serviceDefinition.getInputValue().matchesCharacteristic(characteristic)) {
            this.handleUpdatedInputValueData(characteristic.getValue());
        }
    }

    void handleUpdatedInputFormatData(byte[] data) {
        InputFormat format = InputFormat.inputFormat(data);
        if (format != null) {
            LDSDKLogger.d(String.format("Did receive Input Format: %s", format.toString()));

            // If we already have input format with an earlier revision, delete all those
            // as all known formats must have the same version
            if (inputFormats.size() > 0) {
                InputFormat anyFormat = inputFormats.valueAt(0);
                if (anyFormat.getRevision() != format.getRevision()) {
                    inputFormats.clear();
                }
            }

            inputFormats.append(format.getConnectId(), format);

            callbackHelper.performDidReceiveInputFormatCallback(this, format);
        }
    }

    void handleUpdatedInputValueData(byte[] data)
    {
        ByteBuffer dataBuffer = ByteBuffer.wrap(data);

        int valueFormatRevision = ByteUtils.getUnsignedByte(dataBuffer);

        boolean hasMoreValues = true;

        SparseArray<byte[]> idToValue = new SparseArray<>();

        while (hasMoreValues) {
            int valueConnectId = ByteUtils.getUnsignedByte(dataBuffer);

            InputFormat format = inputFormats.get(valueConnectId);

            if (inputFormats.size() == 0) {
                LDSDKLogger.d("Cannot parse value - has not yet received any Input Format from device");
                readInputFormatForConnectId(valueConnectId); //HUB is sending data, so it must have an Input Format registered
                return;
            }

            if (format == null) {
                LDSDKLogger.d(String.format("No known Input Format for input with Connect ID %d", valueConnectId));
                return;
            }

            if (format.getRevision() != valueFormatRevision) {
                LDSDKLogger.d(String.format("Format revision %d in received value does not match last received Input Format revision %d", valueFormatRevision, format.getRevision()));
                readInputFormatForConnectId(valueConnectId); // HUB is sending data, so it must have an Input Format registered
                return;
            }

            byte[] value = new byte[format.getNumberOfBytes()];
            dataBuffer.get(value, 0, format.getNumberOfBytes());

            idToValue.append(valueConnectId, value);

            if (dataBuffer.position() >= dataBuffer.capacity()) {
                hasMoreValues = false;
            }
        }
        callbackHelper.performDidReceiveValueDataCallback(this, idToValue);
    }
}
