package dk.lego.devicesdk.bluetooth;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothGattService;
import android.content.Context;
import android.os.ParcelUuid;
import android.util.SparseArray;

import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.UUID;

import dk.lego.devicesdk.BuildConfig;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.android.bluetooth.le.ScanRecord;
import dk.lego.devicesdk.bluetooth.service_definitions.BatteryServiceDefinition;
import dk.lego.devicesdk.bluetooth.service_definitions.BluetoothServiceDefinition;
import dk.lego.devicesdk.bluetooth.service_definitions.DeviceInfoServiceDefinition;
import dk.lego.devicesdk.bluetooth.service_definitions.DeviceServiceDefinition;
import dk.lego.devicesdk.bluetooth.service_definitions.IOServiceDefinition;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothDeviceWrapper;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.LegoDeviceImpl;
import dk.lego.devicesdk.device.Revision;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.LegoServiceFactory;
import dk.lego.devicesdk.utils.ByteUtils;

public class LegoBluetoothDeviceImpl extends LegoDeviceImpl implements LegoBluetoothDevice {

    private static final UUID CHARACTERISTIC_UPDATE_NOTIFICATION_DESCRIPTOR_UUID = UUID.fromString("00002902-0000-1000-8000-00805f9b34fb");
    public static final int NUMBER_OF_RSSI_VALUES_TO_AVERAGE = 10;
    private static final int SIZE_OF_ATTACHED_IO_DATA = 12;

    private final AndroidBluetoothDeviceWrapper androidBluetoothDeviceWrapper;
    private ScanRecord scanRecord;
    private BluetoothGatt bluetoothGatt = null;

    //LEGO IO Service
    private BluetoothIO bluetoothIO;
    private IOServiceDefinition ioServiceDefinition;

    //LEGO Device Service
    private DeviceServiceDefinition deviceServiceDefinition;
    private BluetoothGattCharacteristic nameCharacteristic;
    private BluetoothGattCharacteristic ioCharacteristic;
    private BluetoothGattCharacteristic buttonCharacteristic;
    private BluetoothGattCharacteristic lowVoltageAlertCharacteristic;

    //BatteryService
    private BatteryServiceDefinition batteryServiceDefinition;
    private BluetoothGattCharacteristic batteryLevelCharacteristic;

    //DeviceInfoService
    private DeviceInfoServiceDefinition deviceInfoServiceDefinition;
    private BluetoothGattCharacteristic firmwareRevisionCharacteristic;
    private BluetoothGattCharacteristic hardwareRevisionCharacteristic;
    private BluetoothGattCharacteristic softwareRevisionCharacteristic;
    private BluetoothGattCharacteristic manufacturerCharacteristic;

    private boolean interrogationFinished = false;
    private boolean advertising = false;
    private Queue<Integer> rssiValues = new LinkedList<>();
    private int rssiIndex;
    private int averageRSSI;

    private Queue<SynchronousQueueElement> synchronousOperationsQueue = new LinkedList<>();

    public static LegoBluetoothDevice deviceWithWrapper(AndroidBluetoothDeviceWrapper androidBluetoothDeviceWrapper, byte[] scanRecord, int rssi) {
        return new LegoBluetoothDeviceImpl(androidBluetoothDeviceWrapper, scanRecord, rssi);
    }

    private LegoBluetoothDeviceImpl(AndroidBluetoothDeviceWrapper androidBluetoothDeviceWrapper, byte[] scanRecord, int rssi) {

        ioServiceDefinition = BluetoothServiceDefinition.ioServiceDefinition();
        deviceServiceDefinition = BluetoothServiceDefinition.deviceServiceDefinition();
        deviceInfoServiceDefinition = BluetoothServiceDefinition.deviceInfoServiceDefinition();
        batteryServiceDefinition = BluetoothServiceDefinition.batteryServiceDefinition();
        updateWithAdvertisementDataAndRSSI(scanRecord, rssi);

        this.androidBluetoothDeviceWrapper = androidBluetoothDeviceWrapper;
        this.scanRecord = ScanRecord.parseFromBytes(scanRecord);
        name = this.androidBluetoothDeviceWrapper.getName();

        resetState();
    }

    private void resetState() {
        LDSDKLogger.i("Resetting state for device: " + getName());
        interrogationFinished = false;
        rssiValues.clear();
        services.clear();
        batteryLevel = null;
        advertising = false;

        // Do not reset things set in 'advertising' as it is possible to reconnect to a device
        // without ever receiving a advertising package (e.g. on automatic reconnect)
    }

    public boolean isAdvertising() {
        return advertising;
    }

    public void setAdvertising(boolean advertising) {
        this.advertising = advertising;
        updateConnectionStateUponAdvertisingChange();
    }

    @Override
    public String getDeviceId() {
        // In theory the accessed instance variable should never be null since it's only assigned in the
        // constructor where it's guaranteed to be non-null. In practice though it sometimes seems
        // to be null, e.g. immediately after the device is disconnected.
        // Unfortunately I wasn't able to figure out why, thus this only partial fix.
        return androidBluetoothDeviceWrapper != null ? androidBluetoothDeviceWrapper.getAddress() : null;
    }

    @Override
    public void setName(String newName) {
        if (connectState != DeviceState.DEVICE_CONNECTION_STATE_INTERROGATION_FINISHED) {
            LDSDKLogger.w("Ignoring call to set device name - not connected to device");
            return;
        }

        String oldName = this.name;
        name = newName;

        byte[] dataToWrite = (name.isEmpty()) ? new byte[0] : name.getBytes(StandardCharsets.UTF_8);

        /**
         * We use an optimistic approach when setting the name - we assume that the write is successful. If
         * we get an {@link android.bluetooth.BluetoothGattCallback#onCharacteristicWrite} with an error, we will revert the value and notify through the callback listener
         */
        addWriteOfCharacteristicToQueue(nameCharacteristic, dataToWrite, BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT);

        callbackHelper.performDidChangeNameCallback(this, oldName, newName);
    }

    public void updateWithAdvertisementDataAndRSSI(byte[] scanRecord, int rssi) {
        calculateAverageRSSI(rssi);
        this.scanRecord = dk.lego.devicesdk.android.bluetooth.le.ScanRecord.parseFromBytes(scanRecord);

        String oldName = name;
        name = this.scanRecord.getDeviceName();

        if (name != null && !name.equals(oldName)) {
            callbackHelper.performDidChangeNameCallback(this, oldName, name);
        }

        SparseArray<byte[]> manufacturerData = this.scanRecord.getManufacturerSpecificData();

        if (manufacturerData == null) {
            LDSDKLogger.d("Manufacturer ScanRecord was null");
        }

        byte[] serviceDataContent = this.scanRecord.getServiceData(ParcelUuid.fromString("00001523-0000-1000-8000-00805f9b34fb"));
        if (serviceDataContent == null || serviceDataContent.length == 0) {
            LDSDKLogger.e(String.format("Did not find any ServiceData for service %s", "00001523-0000-1000-8000-00805f9b34fb"));
            return;
        }

        if (serviceDataContent.length >= 1) {
            Byte buttonState = serviceDataContent[0];
            this.updateButtonState(buttonState);
        }
        if (serviceDataContent.length >= 2) {
            category = DeviceCategory.fromInteger(serviceDataContent[1]);
        }
        if (serviceDataContent.length >= 3) {
            supportedFunctions = DeviceFunction.fromInteger(serviceDataContent[2]);
        }
        if (serviceDataContent.length >= 4) {
            lastConnectedNetworkId = serviceDataContent[3];
        }
    }

    private void updateButtonState(byte buttonState) {
        boolean oldButtonPressedState = this.buttonPressed;
        this.buttonPressed = (buttonState == 1);

        if (this.buttonPressed != oldButtonPressedState) {
            LDSDKLogger.d(String.format("Button %s", this.buttonPressed ? "pushed" : "released"));
            callbackHelper.performDidChangeButtonStateCallback(this, this.buttonPressed);
        }
    }

    public void calculateAverageRSSI(int newRSSI) {
        if (this.rssiValues.size() < NUMBER_OF_RSSI_VALUES_TO_AVERAGE) {
            // Until we have reached the first NUMBER_OF_RSSI_VALUES_TO_AVERAGE values, just add them
            this.rssiValues.add(newRSSI);
        } else {
            // After having received the 10 first values, replace existing value at rssiIndex
            this.rssiValues.add(newRSSI);
            this.rssiValues.remove();
        }

        if (this.rssiIndex <= NUMBER_OF_RSSI_VALUES_TO_AVERAGE) {
            // Sort the values in a list a pick the value in the middle of the list as the average value
            // which is a simple way to ignore outliers.
            List<Integer> sortedRSSIValues = new ArrayList<>();
            sortedRSSIValues.addAll(rssiValues);
            Collections.sort(sortedRSSIValues);
            int middleIndex = sortedRSSIValues.size() / 2; // intentionally assign to int, to round down result to nearest integer
            this.averageRSSI = sortedRSSIValues.get(middleIndex);
        }

        if (this.rssiIndex == NUMBER_OF_RSSI_VALUES_TO_AVERAGE) {
            this.rssiIndex = 0;
        }
    }

    public void updateDeviceState(DeviceState newDeviceState) {
        LDSDKLogger.i("DeviceConnectionState updated! New state: " + newDeviceState);
        connectState = newDeviceState;
    }

    @Override
    public int getRSSIValue() {
        return averageRSSI;
    }

    @Override
    public DeviceCategory getCategory() {
        return category;
    }

    @Override
    public DeviceFunction getSupportedFunctions() {
        return supportedFunctions;
    }

    @Override
    public int getLastConnectedNetworkId() {
        return lastConnectedNetworkId;
    }

    @Override
    public boolean isLowVoltage() {
        return lowVoltage;
    }

    private void addService(LegoService service) {
        services.add(service);
        Collections.sort(services, new Comparator<LegoService>() {
            @Override
            public int compare(LegoService lhs, LegoService rhs) {
                int lhsConnectID = lhs.getConnectInfo().getConnectId();
                int rhsConnectID = rhs.getConnectInfo().getConnectId();

                if (lhsConnectID == rhsConnectID)
                    return 0;
                else if (lhsConnectID > rhsConnectID)
                    return 1;
                else
                    return -1;
            }
        });
    }

    private void removeService(LegoService service) {
        services.remove(service);
    }

    protected void connectGatt(Context applicationContext, boolean autoConnect, final BluetoothGattCallback bluetoothGattCallback) {
        updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
        androidBluetoothDeviceWrapper.connectGatt(applicationContext, autoConnect, new BluetoothGattCallback() {
            @Override
            public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
                super.onConnectionStateChange(gatt, status, newState);

                bluetoothGatt = gatt;
                bluetoothGattCallback.onConnectionStateChange(gatt, status, newState);

                updateConnectionStateUponConnectionStateChange(status, newState);
            }

            @Override
            public void onServicesDiscovered(BluetoothGatt gatt, int status) {
                super.onServicesDiscovered(gatt, status);

                bluetoothGatt = gatt;
                bluetoothGattCallback.onServicesDiscovered(gatt, status);

                if (status == BluetoothGatt.GATT_SUCCESS) {
                    LDSDKLogger.d(String.format("Did discover services for device: %s", getName()));

                    if (bluetoothGatt.getServices().size() == 0) {
                        LDSDKLogger.e(String.format("Did not find any services for device %s", getName()));
                    }

                    // Loop through the newly filled services array, just in case there's more than one.
                    for (BluetoothGattService service : bluetoothGatt.getServices()) {
                        BluetoothServiceDefinition serviceDefinition = BluetoothServiceDefinition.serviceDefinitionWithUUID(service.getUuid());

                        if (serviceDefinition != null) {
                            LDSDKLogger.d(String.format("Adding service: %s", serviceDefinition.toString()));
                        } else {
                            LDSDKLogger.e(String.format("Did discover service with unknown UUID %s", service.getUuid().toString()));
                        }

                        // We start by only discovering the Bluetooth Device Info service. Once we are sure that we are talking
                        // to a Device with a supported firmware version, we discover characteristics for the remaining services
                        if (DeviceInfoServiceDefinition.getInstance().matchesService(service)) {
                            discoverCharacteristicsForService(service);
                        }
                    }
                }
            }

            @Override
            public void onCharacteristicRead(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) {
                super.onCharacteristicRead(gatt, characteristic, status);

                synchronousOperationsQueue.poll();

                bluetoothGatt = gatt;
                bluetoothGattCallback.onCharacteristicRead(gatt, characteristic, status);

                LDSDKLogger.i(String.format("Read characteristic %s with value %s", characteristic.getUuid(), Arrays.toString(characteristic.getValue())));

                handleReadOrUpdatedCharacteristic(characteristic);

                executeNextOperationQueue();
            }

            @Override
            public void onCharacteristicWrite(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic, int status) throws LDSDKError {
                super.onCharacteristicWrite(gatt, characteristic, status);

                synchronousOperationsQueue.poll();

                bluetoothGatt = gatt;
                bluetoothGattCallback.onCharacteristicWrite(gatt, characteristic, status);

                LDSDKError error = null;
                if (status != BluetoothGatt.GATT_SUCCESS) {
                    LDSDKLogger.e(String.format("Failed to write value for characteristic %s with status %d",
                            characteristic.toString(),
                            status));
                    error = LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, "Failed with status: " + status);
                }

                if (ioServiceDefinition.matchesService(characteristic.getService())) {
                    bluetoothIO.handleWriteResponseFromIOServiceWithCharacteristic(characteristic, error);
                } else if (deviceInfoServiceDefinition.matchesService(characteristic.getService())) {
                    handleWriteResponseFromDeviceServiceWithCharacteristic(characteristic, error);
                } else {
                    LDSDKLogger.w(String.format("Received a did-write characteristic value for characteristic %s, but did not find a service to deliver it to",
                            characteristic.toString()));
                }

                if (error != null) {
                    throw error;
                }

                executeNextOperationQueue();
            }

            public void handleWriteResponseFromDeviceServiceWithCharacteristic(BluetoothGattCharacteristic characteristic, LDSDKError error) {
                if (error != null && characteristic.equals(nameCharacteristic)) {
                    // If the name was not written - make sure we read and restore to the value on the device
                    addReadOfCharacteristicToQueue(nameCharacteristic);
                }
            }

            @Override
            public void onCharacteristicChanged(BluetoothGatt gatt, BluetoothGattCharacteristic characteristic) {
                super.onCharacteristicChanged(gatt, characteristic);

                bluetoothGatt = gatt;
                bluetoothGattCallback.onCharacteristicChanged(gatt, characteristic);

                handleReadOrUpdatedCharacteristic(characteristic);
            }

            @Override
            public void onDescriptorRead(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorRead(gatt, descriptor, status);

                synchronousOperationsQueue.poll();

                bluetoothGatt = gatt;
                bluetoothGattCallback.onDescriptorRead(gatt, descriptor, status);
            }

            @Override
            public void onDescriptorWrite(BluetoothGatt gatt, BluetoothGattDescriptor descriptor, int status) {
                super.onDescriptorWrite(gatt, descriptor, status);

                synchronousOperationsQueue.poll();

                bluetoothGatt = gatt;
                bluetoothGattCallback.onDescriptorWrite(gatt, descriptor, status);

                executeNextOperationQueue();
            }

            @Override
            public void onReliableWriteCompleted(BluetoothGatt gatt, int status) {
                super.onReliableWriteCompleted(gatt, status);

                bluetoothGatt = gatt;
                bluetoothGattCallback.onReliableWriteCompleted(gatt, status);
            }

            @Override
            public void onReadRemoteRssi(BluetoothGatt gatt, int rssi, int status) {
                super.onReadRemoteRssi(gatt, rssi, status);

                bluetoothGatt = gatt;
                bluetoothGattCallback.onReadRemoteRssi(gatt, rssi, status);
            }
        });
    }

    public void handleReadOrUpdatedCharacteristic(BluetoothGattCharacteristic characteristic) {
        if (ioServiceDefinition.matchesService(characteristic.getService())) {
            bluetoothIO.handleUpdatedInputServiceCharacteristic(characteristic);
        } else if (deviceServiceDefinition.matchesService(characteristic.getService())) {
            handleUpdatedDeviceServiceCharacteristic(characteristic);
        } else if (deviceInfoServiceDefinition.matchesService(characteristic.getService())) {
            handleUpdatedDeviceInfoServiceCharacteristic(characteristic);
        } else if (batteryLevelCharacteristic.equals(characteristic)) {
            handleUpdatedBatteryLevelCharacteristic(characteristic);
        } else {
            LDSDKLogger.w(String.format("Received update for unknown characteristic: %s", characteristic));
        }
    }

    @Override
    public BluetoothDevice getAndroidBluetoothDevice() {
        return androidBluetoothDeviceWrapper.getDevice();
    }

    protected void disconnectAndCloseGatt() {
        synchronousOperationsQueue.clear();

        if (bluetoothGatt != null) {
            bluetoothGatt.disconnect();
            bluetoothGatt.close();
            resetState();
        }
    }

    private void discoverCharacteristicsForService(BluetoothGattService service) {

        BluetoothServiceDefinition serviceDefinition = BluetoothServiceDefinition.serviceDefinitionWithUUID(service.getUuid());

        if (serviceDefinition != null) {
            LDSDKLogger.d(String.format("Did discover characteristics for %s", serviceDefinition.getServiceName()));
            LDSDKLogger.i(String.format("%s sucessfully added", serviceDefinition.getServiceName()));

            if (BluetoothServiceDefinition.deviceInfoServiceDefinition().matchesService(service)) {
                deviceInfoServiceDefinition = (DeviceInfoServiceDefinition) serviceDefinition;

                this.firmwareRevisionCharacteristic = BluetoothHelper.create(deviceInfoServiceDefinition.getFirmwareRevision().getUuid(), service);
                this.hardwareRevisionCharacteristic = BluetoothHelper.create(deviceInfoServiceDefinition.getHardwareRevision().getUuid(), service);
                this.softwareRevisionCharacteristic = BluetoothHelper.create(deviceInfoServiceDefinition.getSoftwareRevision().getUuid(), service);
                this.manufacturerCharacteristic = BluetoothHelper.create(deviceInfoServiceDefinition.getManufacturerName().getUuid(), service);

                addReadOfCharacteristicToQueue(this.firmwareRevisionCharacteristic);
                if (this.hardwareRevisionCharacteristic != null) {
                    addReadOfCharacteristicToQueue(this.hardwareRevisionCharacteristic);
                }
                addReadOfCharacteristicToQueue(this.softwareRevisionCharacteristic);
                addReadOfCharacteristicToQueue(this.manufacturerCharacteristic);
            } else if (BluetoothServiceDefinition.deviceServiceDefinition().matchesService(service)) {

                deviceServiceDefinition = (DeviceServiceDefinition) serviceDefinition;

                this.nameCharacteristic = BluetoothHelper.create(deviceServiceDefinition.getDeviceName().getUuid(), service);
                this.ioCharacteristic = BluetoothHelper.create(deviceServiceDefinition.getAttachedIO().getUuid(), service);
                this.buttonCharacteristic = BluetoothHelper.create(deviceServiceDefinition.getDeviceButton().getUuid(), service);
                this.lowVoltageAlertCharacteristic = BluetoothHelper.create(deviceServiceDefinition.getLowVoltageAlert().getUuid(), service);

                if (this.nameCharacteristic != null) {
                    LDSDKLogger.i("Succesfully added device 'name' characteristic");
                    addReadOfCharacteristicToQueue(this.nameCharacteristic);
                }
                if (this.ioCharacteristic != null) {
                    LDSDKLogger.i("Succesfully added device 'Attached IO' characteristic");
                    setCharacteristicNotification(ioCharacteristic, true);
                }
                if (this.buttonCharacteristic != null) {
                    LDSDKLogger.i("Succesfully added device 'Button state' characteristic");
                    setCharacteristicNotification(buttonCharacteristic, true);
                }
                if (this.lowVoltageAlertCharacteristic != null) {
                    LDSDKLogger.i("Succesfully added device 'Low Voltage Alert' characteristic");
                    setCharacteristicNotification(lowVoltageAlertCharacteristic, true);
                }
            } else if (BluetoothServiceDefinition.ioServiceDefinition().matchesService(service)) {
                bluetoothIO = BluetoothIO.bluetoothIO(service, this);
                if (!interrogationFinished) {
                    interrogationFinished = true;
                    updateConnectionStateUponServicesDiscovered();
                }
            } else if (BluetoothServiceDefinition.batteryServiceDefinition().matchesService(service)) {
                batteryServiceDefinition = (BatteryServiceDefinition) serviceDefinition;
                batteryLevelCharacteristic = BluetoothHelper.create(batteryServiceDefinition.getBatteryLevel().getUuid(), service);
                addReadOfCharacteristicToQueue(batteryLevelCharacteristic);
                bluetoothGatt.setCharacteristicNotification(batteryLevelCharacteristic, true);
            } else {
                LDSDKLogger.w(String.format("Discovered characteristics for unknown service with UUID %s", service.getUuid().toString()));
            }
        }
    }

    private void addReadOfCharacteristicToQueue(BluetoothGattCharacteristic characteristic) {
        addOperationToSynchronousQueue(characteristic, null);
    }

    public void addWriteOfCharacteristicToQueue(BluetoothGattCharacteristic characteristic, byte[] data, int writeType) {
        characteristic.setWriteType(writeType);
        characteristic.setValue(data);

        addOperationToSynchronousQueue(characteristic, null);
    }

    private synchronized void addOperationToSynchronousQueue(BluetoothGattCharacteristic characteristic, BluetoothGattDescriptor descriptor) {
        if (characteristic != null) {
            synchronousOperationsQueue.add(new SynchronousQueueElement(characteristic, characteristic.getValue()));
        } else {
            synchronousOperationsQueue.add(new SynchronousQueueElement(descriptor, descriptor.getValue()));
        }

        if (synchronousOperationsQueue.size() == 1) {
            executeNextOperationQueue();
        }
    }

    private synchronized void executeNextOperationQueue() {
        if (synchronousOperationsQueue.size() > 0) {
            SynchronousQueueElement element = synchronousOperationsQueue.element();

            if (element.characteristic != null) {
                if (element.value != null) {
                    element.characteristic.setValue(element.value);
                    bluetoothGatt.writeCharacteristic(element.characteristic);
                } else {
                    bluetoothGatt.readCharacteristic(element.characteristic);
                }
            } else {
                if (element.value != null) {
                    element.descriptor.setValue(element.value);
                    bluetoothGatt.writeDescriptor(element.descriptor);
                } else {
                    bluetoothGatt.readDescriptor(element.descriptor);
                }
            }
        }
    }

    public void setCharacteristicNotification(BluetoothGattCharacteristic characteristic, boolean enable) {
        LDSDKLogger.d(String.format("Notifications for characteristic %s set to %b", characteristic.getUuid(), enable));
        bluetoothGatt.setCharacteristicNotification(characteristic, enable);
        BluetoothGattDescriptor descriptor = characteristic.getDescriptor(CHARACTERISTIC_UPDATE_NOTIFICATION_DESCRIPTOR_UUID);
        descriptor.setValue(enable ? BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE : new byte[]{0x00, 0x00});
        addOperationToSynchronousQueue(null, descriptor);
    }

    @Override
    public dk.lego.devicesdk.android.bluetooth.le.ScanRecord getScanRecord() {
        return scanRecord;
    }

    public void updateConnectionStateUponConnectionStateChange(int status, int newState) {
        if (status != BluetoothGatt.GATT_SUCCESS) {
            updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
            return;
        }

        switch (newState) {
            case BluetoothGatt.STATE_CONNECTED:
                updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_INTERROGATING);
                break;
            case BluetoothGatt.STATE_CONNECTING:
                updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_CONNECTING);
                break;
            case BluetoothGatt.STATE_DISCONNECTING:
                if (isAdvertising()) {
                    updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING);
                } else {
                    updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
                }
                break;
            case BluetoothGatt.STATE_DISCONNECTED:
                updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
                break;
            default:
                LDSDKLogger.e("Reached default case when trying to determine new LegeBluetoothDevice state. Unknown state was: " + newState);
        }
    }

    private void updateConnectionStateUponServicesDiscovered() {
        if (connectState != DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING) {
            updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_INTERROGATION_FINISHED);
        }
    }

    private void updateConnectionStateUponAdvertisingChange() {
        if (connectState == DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING && advertising) {
            updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING);
        }

        if (connectState == DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING && !advertising) {
            updateDeviceState(DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING);
        }
    }

    private void handleUpdatedDeviceServiceCharacteristic(BluetoothGattCharacteristic characteristic) {
        if (deviceServiceDefinition.getDeviceName().matchesCharacteristic(characteristic)) {
            String oldName = getName();
            String newName = new String(characteristic.getValue(), StandardCharsets.UTF_8);

            LDSDKLogger.d(String.format("Received new device name: %s", newName));
            this.name = newName;

            if (!newName.equals(oldName)) {
                callbackHelper.performDidChangeNameCallback(this, oldName, newName);
            }
        } else if (deviceServiceDefinition.getAttachedIO().matchesCharacteristic(characteristic)) {
            LDSDKLogger.d(String.format("Received characteristics for attacthed IOs: %s", ByteUtils.toHexString(characteristic.getValue())));
            handleAttachedIOData(characteristic.getValue());
        } else if (deviceServiceDefinition.getDeviceButton().matchesCharacteristic(characteristic)) {
            LDSDKLogger.d(String.format("Received update for button state characteristic: %s", ByteUtils.toHexString(characteristic.getValue())));
            if (characteristic.getValue().length != 1) {
                LDSDKLogger.e(String.format("Unexpected lenght of button state characteristic, expected 1 but was %d", characteristic.getValue().length));
                return;
            }

            int newStateInt = characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_UINT8, 0);

            boolean oldState = buttonPressed;
            buttonPressed = (newStateInt == 1);

            if (oldState != buttonPressed) {
                callbackHelper.performDidChangeButtonStateCallback(this, buttonPressed);
            }
        } else if (deviceServiceDefinition.getLowVoltageAlert().matchesCharacteristic(characteristic)) {
            LDSDKLogger.d(String.format("Received updated value for lowVoltageAlert: %s", ByteUtils.toHexString(characteristic.getValue())));
            handleUpdatedLowVoltageAlertCharacteristic(characteristic);
        } else {
            LDSDKLogger.w(String.format("Received updated value for unknown characteristic %s", characteristic));
        }
    }

    private void handleAttachedIOData(byte[] data) {
        //Format for attach: {0: deviceId, 1: attached|detached, 2: hubIndex, 3: ioType, 4-8: hwVersion, 8-12: fwVersion}
        //Format for detach: {0: deviceId, 1: attached|detached }
        if (data.length < 2) {
            String errorMessage = String.format("Did receive IO attached/detached notification from Device with size %d, must be size %d for attached and size 2 for detached",
                    data.length, SIZE_OF_ATTACHED_IO_DATA);
            LDSDKLogger.e(String.format("%s", errorMessage));

            LDSDKError error = LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, errorMessage);
            callbackHelper.performDidFailToAddServiceWithErrorCallback(this, error);
            return;
        }

        int connectID = Arrays.copyOfRange(data, 0, 1)[0];
        int attached = Arrays.copyOfRange(data, 1, 2)[0];

        if (attached == 0) {
            LegoService service = serviceWithConnectID(connectID);
            if (service != null) {
                removeService(service);
                callbackHelper.performDidRemoveServiceCallback(this, service);
                LDSDKLogger.i(String.format("Removed input/output: %s", service.getConnectInfo()));
            } else {
                LDSDKLogger.w(String.format("Recieved notification for detached IO with unknown connect id %d", connectID));
            }
        } else {
            if (data.length != SIZE_OF_ATTACHED_IO_DATA) {
                String errorMessage = String.format("Did receive IO attached notification from Device with size %d, must be size %d",
                        data.length,
                        SIZE_OF_ATTACHED_IO_DATA);
                LDSDKLogger.e(String.format("%s", errorMessage));

                LDSDKError error = LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, errorMessage);
                callbackHelper.performDidFailToAddServiceWithErrorCallback(this, error);
                return;
            }

            int hubIndex = Arrays.copyOfRange(data, 2, 3)[0];
            int ioType = Arrays.copyOfRange(data, 3, 4)[0];

            byte[] hwRevision = Arrays.copyOfRange(data, 4, 8);
            byte[] fwRevision = Arrays.copyOfRange(data, 8, 12);

            ConnectInfo connectInfo = new ConnectInfo(connectID, hubIndex, ioType, Revision.createFromDataData(hwRevision), Revision.createFromDataData(fwRevision));

            LegoService service = LegoServiceFactory.create(connectInfo, bluetoothIO, this);
            if (services.contains(service)) {
                LDSDKLogger.w(String.format("Will ignore duplicate notification for attached IO: %s", connectInfo));
                return;
            }

            if (service != null) {
                addService(service);
                LDSDKLogger.i(String.format("Added input/output %s", service.getConnectInfo().toString()));
                callbackHelper.performDidAddServiceCallback(this, service);

                if (service.getDefaultInputFormat() != null) {
                    service.updateInputFormat(service.getDefaultInputFormat());
                }
            }
        }
    }

    private void handleUpdatedDeviceInfoServiceCharacteristic(BluetoothGattCharacteristic characteristic) {
        String valueString = new String(characteristic.getValue(), StandardCharsets.UTF_8);

        if (deviceInfoServiceDefinition.getFirmwareRevision().matchesCharacteristic(characteristic)) {
            deviceInfo.setFirmwareRevisionString(valueString);
        } else if (deviceInfoServiceDefinition.getHardwareRevision().matchesCharacteristic(characteristic)) {
            deviceInfo.setHardwareRevisionString(valueString);
        } else if (deviceInfoServiceDefinition.getSoftwareRevision().matchesCharacteristic(characteristic)) {
            deviceInfo.setSoftwareRevisionString(valueString);
        } else if (deviceInfoServiceDefinition.getManufacturerName().matchesCharacteristic(characteristic)) {
            deviceInfo.setManufacturerName(valueString);
        }

        // We have received values for all required device info characteristics
        LDSDKError error = null;
        if (deviceInfo.isComplete()) {
            Revision firmwareRevisionTestedWithSDK = Revision.createFromString(BuildConfig.BLUETOOTH_DEVICE_FIRMWARE_VERSION);
            Revision deviceFirmwareRevision = deviceInfo.getFirmwareRevision();

            if (deviceFirmwareRevision.getMajorVersion() <= firmwareRevisionTestedWithSDK.getMajorVersion()) {
                LDSDKLogger.i("Did find Device with firmware supported by SDK - querying characteristics for remaining services");

                for (BluetoothGattService service : bluetoothGatt.getServices()) {
                    //The Device Firmware is supported - query for characteristics for all remaining discovered services
                    if (!(DeviceInfoServiceDefinition.getInstance().matchesService(service))) {
                        discoverCharacteristicsForService(service);
                    }
                }

                if (deviceFirmwareRevision.getMajorVersion() == firmwareRevisionTestedWithSDK.getMajorVersion() &&
                    deviceFirmwareRevision.getMinorVersion() > firmwareRevisionTestedWithSDK.getMinorVersion()) {
                    LDSDKLogger.w(String.format("SDK is tested to work with Device Firmware version %s but is connected to a Device with Firmware version %s.%n "
                            + "All Device features may not be supported by SDK.", firmwareRevisionTestedWithSDK, deviceFirmwareRevision));
                }
            } else {
                LDSDKLogger.e(String.format("Did connect to Device with firmware version %s where minimum required major version is %d",
                        deviceFirmwareRevision,
                        firmwareRevisionTestedWithSDK.getMajorVersion()));

                error = LDSDKError.deviceError(
                        LDSDKError.ErrorCode.BLUETOOTH_UNSUPPORTED_FIRMWARE_VERSION,
                        String.format("Device firmware version %s not supported by SDK, SDK requires firmware major version %d", deviceFirmwareRevision, firmwareRevisionTestedWithSDK.getMajorVersion()));
            }
            callbackHelper.performDidUpdateDeviceInfoCallack(this, deviceInfo, error);
        }
    }

    private void handleUpdatedBatteryLevelCharacteristic(BluetoothGattCharacteristic characteristic) {
        LDSDKLogger.d(String.format("Received new batterylevel %s", Arrays.asList(characteristic.getValue())));
        if (characteristic.getValue().length == 1) {
            batteryLevel = characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_UINT8, 0);
            callbackHelper.performDidUpdateBatteryLevelCallback(this, batteryLevel);
        }
    }

    private void handleUpdatedLowVoltageAlertCharacteristic(BluetoothGattCharacteristic characteristic) {
        LDSDKLogger.d(String.format("Received new low voltage alert %s", ByteUtils.toHexString(characteristic.getValue())));
        if (characteristic.getValue().length == 1) {
            lowVoltage = characteristic.getIntValue(BluetoothGattCharacteristic.FORMAT_UINT8, 0) == 1;
            callbackHelper.performDidUpdateLowVoltageStateCallback(this, lowVoltage);
        }
    }

    private LegoService serviceWithConnectID(int connectID) {

        for (LegoService service : services) {
            if (service.getConnectInfo().getConnectId() == connectID) {
                return service;
            }
        }
        return null;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        LegoBluetoothDeviceImpl other = (LegoBluetoothDeviceImpl) o;

        String otherAddress = other.androidBluetoothDeviceWrapper.getAddress();
        if (androidBluetoothDeviceWrapper.getAddress() != null ? !androidBluetoothDeviceWrapper.getAddress().equals(otherAddress) : otherAddress != null)
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        return androidBluetoothDeviceWrapper.getAddress() != null ? androidBluetoothDeviceWrapper.getAddress().hashCode() : 0;
    }

    public void cleanUp()
    {
        LDSDKLogger.d(String.format("Cleaning up device %s", name));
        resetState();
    }

    private static class SynchronousQueueElement {
        private BluetoothGattCharacteristic characteristic = null;
        private BluetoothGattDescriptor descriptor = null;
        private byte[] value = null;

        public SynchronousQueueElement(BluetoothGattCharacteristic characteristic, byte[] value) {
            this.characteristic = characteristic;
            this.value = value;
        }

        public SynchronousQueueElement(BluetoothGattDescriptor descriptor, byte[] value) {
            this.descriptor = descriptor;
            this.value = value;
        }
    }
}
