package dk.lego.devicesdk.bluetooth;

import android.content.Context;

import java.util.List;

import dk.lego.devicesdk.device.LegoDevice;

/**
 * This class handles the responsibility of scanning for LEGO specific Bluetooth devices
 */
public interface LegoBluetoothDeviceManager {
    /**
     * Start scanning for LEGO BLE devices
     *
     * @param context the Android context
     * @throws IllegalStateException when Bluetooth is not available or enabled on device
     */
    void scan(Context context) throws IllegalStateException;

    /**
     * Stop scanning for LEGO BLE devices
     */
    void stopScanning();

    /**
     * Connect to a @class LegoBluetoothDevice. If a connection is not established within the connectRequestTimeoutInterval the
     * connection attempt is cancelled and the @see DeviceManagerCallbackListener#onDidFailToConnectToDevice is invoked.
     * @param context The Android context
     * @param device The device to establish a connection to
     */
    public void connectToDevice(Context context, final LegoBluetoothDevice device);

    /**
     * @return true if automatic reconnect is enabled, otherwise false
     */
    public boolean isAutomaticReconnectOnConnectionLostEnabled();

    /**
     * @param enabled whether automatic reconnect attempts should be performed or not
     */
    public void setAutomaticReconnectOnConnectionLostEnabled(boolean enabled);

    /**
     * @return the connect timeout interval used when connecting to devices
     */
    public long getConnectRequestTimeoutInterval();

    /**
     * @param interval the timeout interval used for device connect attempts
     */
    public void setConnectRequestTimeoutInterval(long interval);

    /**
     * Disconnect from a LEGO Bluetooth LE Device.
     *
     * @param device The device to disconnect from
     */
    public void cancelDeviceConnection(LegoBluetoothDevice device);

    /**
     * Add a listener to receive device discovery and connection events.
     * Note, that callbacks are not guaranteed to be delivered in the same order as the listeners are added.
     *
     * @param listener The listener to add
     */
    public void registerCallbackListener(LegoBluetoothDeviceManagerCallbackListener listener);

    /**
     * Remove a listener
     *
     * @param listener The listener to remove
     */
    public void unregisterCallbackListener(LegoBluetoothDeviceManagerCallbackListener listener);

    /**
     * @return The interval in milliseconds within which devices should advertise before being ruled out
     */
    public long getAdvertisingDevicesInterval();

    /**
     * Update the Advertising interval - determines the interval in milliseconds within which devices should advertise before being ruled out
     * @param interval Interval to set
     */
    public void updateAdvertisingDevicesInterval(long interval);

    /**
     * Returns a list of devices in the specified device state. Use this, for example, to retrieve a list
     * of all advertising but non-connected devices, i.e. DEVICE_CONNECTION_STATE_DISCONNECTED.
     *
     * @param connectionState The state of the devices.
     */
    public List<LegoBluetoothDevice> devicesInState(LegoDevice.DeviceState connectionState);

    /**
     Returns a list with all known devices regardless of their current connect state.

     For BLE devices, the list is ordered according to relative distance, based of the RSSI values received for each Device.
     The Device closest will be first in the list. Note, the RSSI value is only updated while in advertising mode.

     */
    public List<LegoBluetoothDevice> allDevices();
}
