package dk.lego.devicesdk.bluetooth;

import dk.lego.devicesdk.LDSDKError;

/**
 * Callback used to signal whenever a LEGO Bluetooth device has appeared or disappeared.
 * Note, that callbacks are not guaranteed to be delivered in the same order as the listeners are added.
 */
public interface LegoBluetoothDeviceManagerCallbackListener {

    /**
     * Invoked when a Bluetooth device advertising a LEGO Device service UUID is discovered.
     *
     * @param bluetoothDevice The discovered device
     */
    public void onDeviceDidAppear(LegoBluetoothDevice bluetoothDevice);

    /**
     * Invoked when a Bluetooth device stops advertising a LEGO Device service.
     * The DeviceManager will check at small refresh-intervals if an advertising
     * packet was received during the refresh-interval. If not, this method is invoked.
     *
     * @param bluetoothDevice The device that stopped advertising
     */
    public void onDeviceDidDisappear(LegoBluetoothDevice bluetoothDevice);

    /**
     * Invoked when starting a device connect attempt. Normally, this will happen right after calling see {@link dk.lego.devicesdk.device.LegoDeviceManager#connectToDevice(android.content.Context, dk.lego.devicesdk.device.LegoDevice)}.
     * However, it may also happen in relation to an automatic reconnect attempt.
     * @param device The device to establish a connection to
     */
    void onWillStartConnectingToDevice(LegoBluetoothDevice device);

    /**
     Invoked when a device fails to connect, of if a connection request times out.

     @param device         The device that failed to connect
     @param autoReconnect  true if an automatic reconnect will be attempted, see {@link dk.lego.devicesdk.device.LegoDeviceManager#isAutomaticReconnectOnConnectionLostEnabled()}.
     @param error          The cause of the failure.
     */
    void onDidFailToConnectToDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error);

    /**
     Invoked when a device is disconnected.

     @param device         The disconnected device
     @param autoReconnect  true if an automatic reconnect will be attempted, see {@link dk.lego.devicesdk.device.LegoDeviceManager#isAutomaticReconnectOnConnectionLostEnabled()}.
     @param error          If an error occurred, the cause of the failure.
     */
    void onDidDisconnectFromDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error);

    /**
     * Invoked when a connection to a device is established, and the interrogation of the device for required services begins.
     * A connection is established at this point but the device is not yet ready to be used.
     * @param device The connected device
     */
    void onDidStartInterrogatingDevice(LegoBluetoothDevice device);

    /**
     * Invoked when a connection to a device is established and all required services has been discovered. At this point the device is ready for use.
     * @param device The connected device
     */
    void onDidFinishInterrogatingDevice(LegoBluetoothDevice device);
}
