package dk.lego.devicesdk.bluetooth;

import android.bluetooth.BluetoothGatt;
import android.content.Context;
import android.os.CountDownTimer;
import android.os.Handler;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.bluetooth.service_definitions.BluetoothServiceDefinition;
import dk.lego.devicesdk.bluetooth.service_definitions.DeviceServiceDefinition;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothAdapterWrapper;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothDeviceWrapper;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothScanCallbackListener;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.LegoDeviceManager;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.utils.HandlerHelper;

public class LegoBluetoothDeviceManagerImpl implements LegoBluetoothDeviceManager {

    private static final long REFRESH_DEVICE_LIST_DEFAULT_INTERVAL = 2000;  // [ms]
    private static final long REFRESH_DEVICE_LIST_MINIMUM_INTERVAL = 1000;   // [ms]
    private static final long START_STOP_SCANNING_INTERVAL = 600;   // [ms]

    private Context applicationContext = null;

    private final AndroidBluetoothAdapterWrapper bluetoothAdapter;

    private final Set<LegoBluetoothDevice> devices = new HashSet<>();
    private final Map<String, Long> deviceIdentifierToLastSeenTimeStamp = new HashMap<>();

    private final Map<String, DeviceConnectTimeoutCountDownTimer> deviceToConnectTimerDic = new HashMap<>();
    private final Map<String, Long> deviceToReconnectCount = new HashMap<>();

    private long updateAdvertisingDevicesInterval = REFRESH_DEVICE_LIST_DEFAULT_INTERVAL;

    private List<UUID> primaryServiceUUIds = new ArrayList<>();

    private long connectRequestTimeoutInterval = LegoDeviceManager.DEFAULT_CONNECT_REQUEST_TIMEOUT;
    private boolean automaticReconnectOnConnectionLostEnabled = false;
    private boolean userInitiatedDisconnect = false;

    private BluetoothDeviceManagerCallbackHelper callbackHelper = new BluetoothDeviceManagerCallbackHelper();

    private static final Comparator<LegoBluetoothDevice> rssiComparatorDescending = new Comparator<LegoBluetoothDevice>() {
        @Override
        public int compare(LegoBluetoothDevice lhs, LegoBluetoothDevice rhs) {
            int lhsRSSI = lhs.getRSSIValue();
            int rhsRSSI = rhs.getRSSIValue();

            if (lhsRSSI == rhsRSSI)
                return 0;
            else if (lhsRSSI < rhsRSSI)
                return 1;
            else
                return -1;
        }
    };

    /**
     * Two Runnables are created to start and stop scanning of BLE devices.
     * This solution is made on the basis of the following thread:
     * http://stackoverflow.com/questions/20058450/how-to-continuously-get-rssi-without-connecting-to-the-ble-device
     *
     * The problem is, that the frequency of scanning results is device dependent.
     * Thus, we cannot rely on a single scan to provide enough "pings" and RSSI measurements,
     * in order for the device filtering to function properly.
     */
    private Runnable startScan = new Runnable() {
        @Override
        public void run() {
            //bluetoothAdapter.startLeScan(UUIDS, mLeScanCallback); // Not working - Android Issue: https://code.google.com/p/android/issues/detail?id=59490
            bluetoothAdapter.startLeScan();
            HandlerHelper.getInstance().postDelayed(stopScan, START_STOP_SCANNING_INTERVAL);
        }
    };

    private Runnable stopScan = new Runnable() {
        @Override
        public void run() {
            bluetoothAdapter.stopLeScan();
            HandlerHelper.getInstance().postDelayed(startScan, START_STOP_SCANNING_INTERVAL);
        }
    };

    private final Runnable cleanUpListOfDiscoveredDevices = new Runnable() {
        @Override
        public void run() {
            Iterator<LegoBluetoothDevice> i = devices.iterator();
            while (i.hasNext()) {
                LegoBluetoothDeviceImpl device = (LegoBluetoothDeviceImpl) i.next();

                if (device != null) {
                    if (deviceIdentifierToLastSeenTimeStamp.containsKey(device.getDeviceId())) {
                        long lastSeenDateInMillis = deviceIdentifierToLastSeenTimeStamp.get(device.getDeviceId());
                        boolean lastAdvertisingPackageTooOld = Calendar.getInstance().getTimeInMillis() - lastSeenDateInMillis > updateAdvertisingDevicesInterval;
                        if (lastAdvertisingPackageTooOld) {
                            device.setAdvertising(false);
                        }
                        if ((device.getConnectState() == LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING
                                || device.getConnectState() == LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING)
                             && lastAdvertisingPackageTooOld) {
                            deviceIdentifierToLastSeenTimeStamp.remove(device.getDeviceId());
                            LDSDKLogger.d("cleanUpListOfDiscoveredDevices Filtering out " + device.getName());
                            i.remove();
                            callbackHelper.performDeviceDidDisappearCallback(device);
                        }
                    }
                }
            }
            HandlerHelper.getInstance().postDelayed(cleanUpListOfDiscoveredDevices,updateAdvertisingDevicesInterval);
        }
    };

    public LegoBluetoothDeviceManagerImpl(AndroidBluetoothAdapterWrapper adapter) {

        // Add primary service UUIDs
        primaryServiceUUIds.add(BluetoothServiceDefinition.ioServiceDefinition().getServiceUUID());
        primaryServiceUUIds.add(BluetoothServiceDefinition.deviceServiceDefinition().getServiceUUID());
        primaryServiceUUIds.add(BluetoothServiceDefinition.batteryServiceDefinition().getServiceUUID());
        primaryServiceUUIds.add(BluetoothServiceDefinition.deviceInfoServiceDefinition().getServiceUUID());

        bluetoothAdapter = adapter;
        AndroidBluetoothScanCallbackListener androidBluetoothScanCallbackListener = new AndroidBluetoothScanCallbackListener() {
            @Override
            public void onLeScan(AndroidBluetoothDeviceWrapper device, int rssi, byte[] scanRecord) {
                byte[] scanRecordCopy = scanRecord.clone();

                List<UUID> advertisedServiceUUIDS = parseUuids(scanRecordCopy);
                if (advertisedServiceUUIDS.contains(DeviceServiceDefinition.getInstance().getServiceUUID())) {

                    LegoBluetoothDeviceImpl legoBluetoothDevice = (LegoBluetoothDeviceImpl) getLegoBluetoothDeviceWithAddress(device.getAddress());

                    if (legoBluetoothDevice == null) {
                        LDSDKLogger.d("Adding new peripheral " + device.getName());
                        legoBluetoothDevice = (LegoBluetoothDeviceImpl) LegoBluetoothDeviceImpl.deviceWithWrapper(device, scanRecordCopy, rssi);
                        devices.add(legoBluetoothDevice);
                        callbackHelper.performDeviceDidAppearCallback(legoBluetoothDevice);
                    } else {
                        legoBluetoothDevice.updateWithAdvertisementDataAndRSSI(scanRecordCopy, rssi);
                    }

                    if (!device.getAddress().isEmpty()) {
                        deviceIdentifierToLastSeenTimeStamp.put(device.getAddress(), Calendar.getInstance().getTimeInMillis());
                    }

                    legoBluetoothDevice.setAdvertising(true);
                }
            }

            LegoBluetoothDevice getLegoBluetoothDeviceWithAddress(String newDeviceAddress) {
                for (LegoBluetoothDevice device : devices) {
                    if (device.getDeviceId().equals(newDeviceAddress)) {
                        return device;
                    }
                }
                return null;
            }
        };
        bluetoothAdapter.connectCallbackListener(androidBluetoothScanCallbackListener);
    }

    public void scan(Context context) throws IllegalStateException {
        // Create a handler that can be used to post to the main thread
        HandlerHelper.getInstance().setMainHandler(new Handler(context.getApplicationContext().getMainLooper()));

        // Ensure that Bluetooth is available on the device and it is enabled
        if (!isBluetoothAvailable()) {
            throw new IllegalStateException("Bluetooth is not available on device");
        }

        if (!isBluetoothEnabled()) {
            throw new IllegalStateException("Bluetooth is not enabled");
        }

        startRepeatingScanning();
        startRepeatingCleanUpListOfDiscoveredDevices();
    }

    public void stopScanning() {
        LDSDKLogger.i("Stop scanning...");
        stopRepeatingDeviceDisappearanceDiscovery();
        stopRepeatingScanning();
    }

    @Override
    public void connectToDevice(Context context, final LegoBluetoothDevice device) {
        callbackHelper.performWillStartConnectingToDeviceCallback(device);

        addConnectTimeoutTimerForDevice(device, connectRequestTimeoutInterval);

        LDSDKLogger.d("Connecting to device: " + device.getAndroidBluetoothDevice().getAddress());

        if (context != null && applicationContext != null) {
            applicationContext = context.getApplicationContext();
        }
        ((LegoBluetoothDeviceImpl) device).connectGatt(applicationContext, false, new BaseBluetoothGattCallback() {
            @Override
            public void onConnectionStateChange(BluetoothGatt gatt, int status, int newState) {
                super.onConnectionStateChange(gatt, status, newState);

                if (status != BluetoothGatt.GATT_SUCCESS) {
                    LDSDKLogger.i("onConnectionStateChange, not success");
                    bailOut("Unable to connect. Failed with status: " + status, status);
                    return;
                }

                if (newState != BluetoothGatt.STATE_CONNECTED) {
                    LDSDKLogger.i("onConnectionStateChange, not connected, newState is " + newState);
                    bailOut("Disconnected from device", status);
                    return;
                }

                deviceDidConnect(device);

                if (!gatt.discoverServices()) {
                    bailOut("Discovering services failed with status: " + status, status);
                    //noinspection UnnecessaryReturnStatement
                    return;
                }
            }

            @Override
            public void onServicesDiscovered(final BluetoothGatt gatt, int status) {
                super.onServicesDiscovered(gatt, status);

                if (status != BluetoothGatt.GATT_SUCCESS) {
                    bailOut("Fetching services failed with status: " + status, status);
                    //noinspection UnnecessaryReturnStatement
                    return;
                } else {
                    deviceInterrogationFinished(device);

                }
            }

            private void bailOut(String message, int status) {
                LDSDKLogger.d(message);
                if (!userInitiatedDisconnect) {
                    if (device.getConnectState() == LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_INTERROGATION_FINISHED) {
                        deviceDidDisconnect(device, status);
                    } else {
                        deviceDidFailToConnect(device, status);
                    }
                }

                // Reset flag
                userInitiatedDisconnect = false;
            }
        });
    }

    private void deviceDidDisconnect(LegoBluetoothDevice device, int status) {
        ((LegoBluetoothDeviceImpl) device).cleanUp();

        String errorMessage;
        if (status == BluetoothGatt.GATT_SUCCESS) {
            errorMessage = "";
        } else if (status == 8) {
            // Connection lost e.g. due to out of range or Hub lid removed
            errorMessage = String.format("Connection lost to device: %s", device.getName());
            LDSDKLogger.d(errorMessage);
        } else if (status == 19) {
            // We do not want to treat this as an error
            errorMessage = "";
            //Happens e.g. when the user holds down the button on the device for a few seconds
            LDSDKLogger.d(String.format("Connection closed from device: %s", device.getName()));
        } else {
            errorMessage = String.format("Connection closed unexpectedly to device: %s", device.getName());
            LDSDKLogger.w(errorMessage);
        }

        LDSDKError error;
        if (!errorMessage.isEmpty()) {
            error = LDSDKError.deviceError(LDSDKError.ErrorCode.BLUETOOTH_CONNECTION_ERROR, errorMessage);
        } else {
            error = null;
        }

        LDSDKLogger.d(String.format("Device did disconnect with status %d", status));

        boolean shouldAttemptReconnect = (shouldAllowReconnectToDevice(device));

        callbackHelper.performDidDisconnectFromDeviceCallback(device, shouldAttemptReconnect, error);

        performReconnect(device, shouldAttemptReconnect);
    }

    private void deviceDidFailToConnect(LegoBluetoothDevice device, int status) {

        LDSDKLogger.d(String.format("Device did fail to connect with status %d", status));

        ((LegoBluetoothDeviceImpl) device).cleanUp();

        removeConnectTimeoutTimerForDevice(device);

        boolean shouldAttemptReconnect = shouldAllowReconnectToDevice(device);
        callbackHelper.performDidFailToConnectToDeviceCallback(device, shouldAttemptReconnect, null);

        performReconnect(device, shouldAttemptReconnect);
    }

    private void performReconnect(LegoBluetoothDevice device, boolean shouldAttemptReconnect) {
        if (shouldAttemptReconnect) {
            reconnectToDevice(device);
        } else {
            resetReconnectCountForDevice(device);
            ((LegoBluetoothDeviceImpl) device).disconnectAndCloseGatt();
            devices.remove(device);
        }
    }

    private void deviceDidConnect(LegoBluetoothDevice device) {
        removeConnectTimeoutTimerForDevice(device);
        callbackHelper.performDidStartInterrogatingDeviceCallback(device);
    }

    private void deviceInterrogationFinished(LegoBluetoothDevice device) {
        resetReconnectCountForDevice(device);
        callbackHelper.performDidFinishInterrogatingDeviceCallback(device);
    }

    @Override
    public boolean isAutomaticReconnectOnConnectionLostEnabled() {
        return automaticReconnectOnConnectionLostEnabled;
    }

    @Override
    public void setAutomaticReconnectOnConnectionLostEnabled(boolean enabled) {
        automaticReconnectOnConnectionLostEnabled = enabled;
    }

    @Override
    public long getConnectRequestTimeoutInterval() {
        return connectRequestTimeoutInterval;
    }

    @Override
    public void setConnectRequestTimeoutInterval(long interval) {
        connectRequestTimeoutInterval = interval;
    }

    private void addConnectTimeoutTimerForDevice(LegoBluetoothDevice device, long timeout)
    {
        DeviceConnectTimeoutCountDownTimer timer = new DeviceConnectTimeoutCountDownTimer(device, timeout);
        timer.start();

        deviceToConnectTimerDic.put(device.getAndroidBluetoothDevice().getAddress(), timer);
    }

    public void removeConnectTimeoutTimerForDevice(LegoBluetoothDevice device)
    {
        DeviceConnectTimeoutCountDownTimer timer = deviceToConnectTimerDic.get(device.getAndroidBluetoothDevice().getAddress());
        if (timer != null) {
            timer.cancel();
        }
        deviceToConnectTimerDic.remove(device.getAndroidBluetoothDevice().getAddress());
    }

    @Override
    public void registerCallbackListener(LegoBluetoothDeviceManagerCallbackListener listener) {
        callbackHelper.registerCallbackListener(listener);
    }

    @Override
    public void unregisterCallbackListener(LegoBluetoothDeviceManagerCallbackListener listener) {
        callbackHelper.unregisterCallbackListener(listener);
    }

    @Override
    public long getAdvertisingDevicesInterval() {
        return updateAdvertisingDevicesInterval;
    }

    @Override
    public void updateAdvertisingDevicesInterval(long interval) {
        if (interval > REFRESH_DEVICE_LIST_MINIMUM_INTERVAL) {
            updateAdvertisingDevicesInterval = interval;
        } else {
            updateAdvertisingDevicesInterval = REFRESH_DEVICE_LIST_MINIMUM_INTERVAL;
        }
    }

    @Override
    public void cancelDeviceConnection(LegoBluetoothDevice device) {
        userInitiatedDisconnect = true;
        ((LegoBluetoothDeviceImpl) device).disconnectAndCloseGatt();
        devices.remove(device);
    }

    @Override
    public List<LegoBluetoothDevice> devicesInState(LegoDevice.DeviceState connectionState) {
        List<LegoBluetoothDevice> devicesInState = new ArrayList<>();

        for (LegoBluetoothDevice device : devices) {
            if (device.getConnectState() == connectionState) {
                devicesInState.add(device);
            }
        }

        if (devicesInState.size() > 0) {
            if (connectionState == LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING || connectionState == LegoDevice.DeviceState.DEVICE_CONNECTION_STATE_CONNECTING) {
                Collections.sort(devicesInState, rssiComparatorDescending);
            }
        }

        return devicesInState;
    }

    @Override
    public List<LegoBluetoothDevice> allDevices() {
        List<LegoBluetoothDevice> devices = new ArrayList<>(this.devices);
        Collections.sort(devices, rssiComparatorDescending);

        return devices;
    }

    public boolean isBluetoothEnabled() {
        return bluetoothAdapter.isBluetoothAvailable() && bluetoothAdapter.isBluetoothEnabledOnDevice();
    }

    public boolean isBluetoothAvailable() {
        return bluetoothAdapter.isBluetoothAvailable();
    }

    private void resetReconnectCountForDevice(LegoBluetoothDevice device)
    {
        deviceToReconnectCount.remove(device.getDeviceId());
    }

    private void reconnectToDevice(final LegoBluetoothDevice device)
    {
        long reconnectCount = 0;
        if (deviceToReconnectCount.containsKey(device.getAndroidBluetoothDevice().getAddress())) {
            reconnectCount = deviceToReconnectCount.get(device.getAndroidBluetoothDevice().getAddress());
        }
        deviceToReconnectCount.put(device.getAndroidBluetoothDevice().getAddress(), reconnectCount + 1);

        HandlerHelper.getInstance().post(new Runnable() {
            @Override
            public void run() {
                ((LegoBluetoothDeviceImpl) device).disconnectAndCloseGatt();
                connectToDevice(null, device);
            }
        });
    }

    private boolean shouldAllowReconnectToDevice(LegoBluetoothDevice device)
    {
        if (!automaticReconnectOnConnectionLostEnabled) {
            return false;
        }

        if (deviceToReconnectCount.size() > 0) {
            long reconnectCount = deviceToReconnectCount.get(device.getAndroidBluetoothDevice().getAddress());
            return (reconnectCount < 1);
        } else {
            return true;
        }
    }

    private List<UUID> parseUuids(final byte[] advertisedData) {
        List<UUID> uuids = new ArrayList<>();

        ByteBuffer buffer = ByteBuffer.wrap(advertisedData).order(ByteOrder.LITTLE_ENDIAN);
        while (buffer.remaining() > 2) {
            byte length = buffer.get();
            if (length == 0) break;

            byte type = buffer.get();
            switch (type) {
                case 0x02: // Partial list of 16-bit UUIDs
                case 0x03: // Complete list of 16-bit UUIDs
                    while (length >= 2) {
                        uuids.add(UUID.fromString(String.format("%08x-%s", buffer.getShort(), BluetoothHelper.UUID_STANDARD_BASE)));
                        length -= 2;
                    }
                    break;

                case 0x06: // Partial list of 128-bit UUIDs
                case 0x07: // Complete list of 128-bit UUIDs
                    while (length >= 16) {
                        long lsb = buffer.getLong();
                        long msb = buffer.getLong();
                        uuids.add(new UUID(msb, lsb));
                        length -= 16;
                    }
                    break;

                default:
                    buffer.position(buffer.position() + length - 1);
                    break;
            }
        }

        return uuids;
    }

    private void startRepeatingScanning() {
        HandlerHelper.getInstance().post(startScan);
    }

    private void stopRepeatingScanning() {
        HandlerHelper.getInstance().removeCallbacks(startScan);
        HandlerHelper.getInstance().removeCallbacks(stopScan);

        bluetoothAdapter.stopLeScan();
    }

    private void startRepeatingCleanUpListOfDiscoveredDevices() {
        HandlerHelper.getInstance().post(cleanUpListOfDiscoveredDevices);
    }

    private void stopRepeatingDeviceDisappearanceDiscovery() {
        HandlerHelper.getInstance().removeCallbacks(cleanUpListOfDiscoveredDevices);
    }

    private class DeviceConnectTimeoutCountDownTimer extends CountDownTimer {

        private LegoBluetoothDevice device = null;

        public DeviceConnectTimeoutCountDownTimer(LegoBluetoothDevice device, long millisInFuture) {
            this(millisInFuture, 1);
            this.device = device;
        }

        /**
         * @param millisInFuture    The number of millis in the future from the call
         *                          to {@link #start()} until the countdown is done and {@link #onFinish()}
         *                          is called.
         * @param countDownInterval The interval along the way to receive
         *                          {@link #onTick(long)} callbacks.
         */
        private DeviceConnectTimeoutCountDownTimer(long millisInFuture, long countDownInterval) {
            super(millisInFuture, countDownInterval);
        }

        @Override
        public void onTick(long millisUntilFinished) {
            // Do nothing
        }

        @Override
        public void onFinish() {
            if (device == null) {
                LDSDKLogger.e("Received a connection time out but could not find device to cancel connect request for");
                return;
            }
            LDSDKLogger.w(String.format("Connect request timed out for device: %s", device.getName()));
            removeConnectTimeoutTimerForDevice(device);
            cancelDeviceConnection(device);

            LDSDKError error = LDSDKError.deviceError(LDSDKError.ErrorCode.BLUETOOTH_CONNECTION_TIMEOUT, "Connect request timed out");

            callbackHelper.performDidFailToConnectToDeviceCallback(device, false, error);
        }
    }
}
