package dk.lego.devicesdk.bluetooth.service_definitions;

import android.bluetooth.BluetoothGattCharacteristic;

import java.util.List;
import java.util.UUID;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.bluetooth.BluetoothHelper;
import dk.lego.devicesdk.logging.LDSDKLogger;

public class CharacteristicDefinition {

    private BluetoothServiceDefinition serviceDefinition;
    private String name;
    private UUID uuid;
    private boolean mandatory;
    private int mandatoryProperties;
    private int recommendedProperties;
    private int permissions;

    public BluetoothServiceDefinition getServiceDefinition() {
        return serviceDefinition;
    }

    public String getName() {
        return name;
    }

    public UUID getUuid() {
        return uuid;
    }

    public boolean isMandatory() {
        return mandatory;
    }

    public int getMandatoryProperties() {
        return mandatoryProperties;
    }

    public int getRecommendedProperties() {
        return recommendedProperties;
    }

    public int getPermissions() {
        return permissions;
    }

    public CharacteristicDefinition(String name,
                                    BluetoothServiceDefinition serviceDefinition,
                                    UUID uuid,
                                    boolean mandatory,
                                    int mandatoryProperties,
                                    int recommendedProperties,
                                    int permissions) {
        this.name = name;
        this.serviceDefinition = serviceDefinition;
        this.uuid = uuid;
        this.mandatory = mandatory;
        this.mandatoryProperties = mandatoryProperties;
        this.recommendedProperties = recommendedProperties;
        this.permissions = permissions;
    }

    public void validateDefinitionIsSatisfiedByCharacteristic(BluetoothGattCharacteristic characteristic) throws LDSDKError {

        if ((mandatoryProperties != 0) && ((characteristic.getProperties() & mandatoryProperties) != 0)) {
            String errorMessage = String.format("Characteristic %s with properties %s does not include mandatory properties %s",
                    toStringShort(), stringsFromProperties(characteristic.getProperties()), stringsFromProperties(mandatoryProperties));
            throw LDSDKError.deviceError(LDSDKError.ErrorCode.BLUETOOTH_INVALID_CHARACTERISTIC_PROPERTIES, errorMessage);
        }

        if ((recommendedProperties != 0) && ((characteristic.getProperties() & recommendedProperties) != 0)) {
            LDSDKLogger.w(String.format("Characteristic %s with properties %s does not include recommended properties %s",
                    toStringShort(),
                    stringsFromProperties(characteristic.getProperties()),
                    stringsFromProperties(recommendedProperties)));
        }
    }

    public boolean matchesCharacteristic(BluetoothGattCharacteristic characteristic) {
        return (uuid.equals(characteristic.getUuid()));
    }

    @Override
    public String toString() {

        return String.format("<%s: ",
                this.getClass().getSimpleName())
                + String.format(", serviceName=%s", this.serviceDefinition.serviceName)
                + String.format(", UUID=%s", this.uuid.toString())
                + String.format(", isMandatory=%b", this.mandatory)
                + String.format(", mandatoryProperties=%s", BluetoothHelper.arrayOfStringsFromCharacteristicProperties(this.mandatoryProperties))
                + String.format(", recommendedProperties=%s", BluetoothHelper.arrayOfStringsFromCharacteristicProperties(this.recommendedProperties))
                + String.format(">");
    }

    public String toStringShort() {
        return String.format("%s.%s(%s)", serviceDefinition.serviceName, name, uuid.toString());
    }

    private List stringsFromProperties(int properties) {
        return BluetoothHelper.arrayOfStringsFromCharacteristicProperties(properties);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        CharacteristicDefinition that = (CharacteristicDefinition) o;

        if (mandatory != that.mandatory) return false;
        if (mandatoryProperties != that.mandatoryProperties) return false;
        if (recommendedProperties != that.recommendedProperties) return false;
        if (name != null ? !name.equals(that.name) : that.name != null) return false;
        if (uuid != null ? !uuid.equals(that.uuid) : that.uuid != null) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = name != null ? name.hashCode() : 0;
        result = 31 * result + (uuid != null ? uuid.hashCode() : 0);
        result = 31 * result + (mandatory ? 1 : 0);
        result = 31 * result + mandatoryProperties;
        result = 31 * result + recommendedProperties;
        return result;
    }
}
