package dk.lego.devicesdk.device;

import java.util.ArrayList;
import java.util.List;

/**
 * The ConnectInfo class represents generic info about an IO (service) attached to a device.
 */
public class ConnectInfo {
    private List<Integer> validTypes;

    /**
     * Represent the type of an attached IO (motor, sensor, etc).
     */
    public enum IOType {
        /**
         * A Motor - use the Motor to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.Motor}
         */
        IO_TYPE_MOTOR(1),  //0x01
        /**
         * A Voltage Sensor - use the VoltageSensor to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.VoltageSensor}
         */
        IO_TYPE_VOLTAGE(20), //0x14
        /**
         * A Current Sensor - use the CurrentSensor to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.CurrentSensor}
         */
        IO_TYPE_CURRENT(21), //0x15
        /**
         * A Piezo Tone player - use the PiezoTonePlayer to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.PiezoTonePlayer}
         */
        IO_TYPE_PIEZO_TONE_PLAYER(22), //0x16
        /**
         * An RGB light - use the RGBLight to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.RGBLight}
         */
        IO_TYPE_RGB_LIGHT(23), //0x17
        /**
         * A Tilt Sensor - use the TiltSensor to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.TiltSensor}
         */
        IO_TYPE_TILT_SENSOR(34), //0x22
        /**
         * A Motion Sensor (aka. Detect Sensor) - use the MotionSensor to communicate with this type of IO.
         * See {@link dk.lego.devicesdk.services.MotionSensor}
         */
        IO_TYPE_MOTION_SENSOR(35), //0x23
        /**
         * A type unknown to the SDK - use the {@link dk.lego.devicesdk.services.GenericService} to communicate with this type of IO.
         */
        IO_TYPE_GENERIC(0);

        private int value;

        IOType(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static IOType fromInteger(int value) {
            switch (value) {
                case 1:
                    return IO_TYPE_MOTOR;
                case 20:
                    return IO_TYPE_VOLTAGE;
                case 21:
                    return IO_TYPE_CURRENT;
                case 22:
                    return IO_TYPE_PIEZO_TONE_PLAYER;
                case 23:
                    return IO_TYPE_RGB_LIGHT;
                case 34:
                    return IO_TYPE_TILT_SENSOR;
                case 35:
                    return IO_TYPE_MOTION_SENSOR;
                default:
                    return IO_TYPE_GENERIC;
            }
        }
    }

    /**
     * An identifier used to uniquely identify and address the service. The device is guaranteed not to have two services with the same connectId at the same time
     */
    private int connectId;

    /**
     * The index of the port on the Hub the IO is attached to. If the index is higher than or equal to 50 the service is an internal service
     */
    private int hubIndex;

    /**
     * The raw type number of the IO as received from the Device. Use the typeEnum property to get the type as an {@link dk.lego.devicesdk.device.ConnectInfo.IOType}
     */
    private int type;

    /**
     * The hardware revision of the attached IO as received from the device
     */
    private Revision hardwareVersion;

    /**
     * The firmware revision of the attached IO as received from the device
     */
    private Revision firmwareVersion;

    public ConnectInfo(int connectId, int hubIndex, int type, Revision hardwareVersion, Revision firmwareVersion) {
        this.connectId = connectId;
        this.hubIndex = hubIndex;
        this.type = type;
        this.hardwareVersion = hardwareVersion;
        this.firmwareVersion = firmwareVersion;

        populateValidTypes();
    }

    public ConnectInfo(int connectId, int hubIndex, IOType type) {
        this(connectId, hubIndex, type.getValue(), null, null);
    }

    private void populateValidTypes() {
        validTypes = new ArrayList<>();
        IOType ioType = IOType.IO_TYPE_MOTOR;

        //We use a enum with fall-through (no break statements) to make sure we get all types
        //this will also give a compile warning if a new type is added that is not in the cases below
        switch (ioType) {
            case IO_TYPE_MOTOR:
                validTypes.add(IOType.IO_TYPE_MOTOR.getValue());
            case IO_TYPE_VOLTAGE:
                validTypes.add(IOType.IO_TYPE_VOLTAGE.getValue());
            case IO_TYPE_CURRENT:
                validTypes.add(IOType.IO_TYPE_CURRENT.getValue());
            case IO_TYPE_PIEZO_TONE_PLAYER:
                validTypes.add(IOType.IO_TYPE_PIEZO_TONE_PLAYER.getValue());
            case IO_TYPE_RGB_LIGHT:
                validTypes.add(IOType.IO_TYPE_RGB_LIGHT.getValue());
            case IO_TYPE_TILT_SENSOR:
                validTypes.add(IOType.IO_TYPE_TILT_SENSOR.getValue());
            case IO_TYPE_MOTION_SENSOR:
                validTypes.add(IOType.IO_TYPE_MOTION_SENSOR.getValue());
            case IO_TYPE_GENERIC:
                validTypes.add(IOType.IO_TYPE_GENERIC.getValue());
        }
    }

    public int getConnectId() {
        return connectId;
    }

    public int getHubIndex() {
        return hubIndex;
    }

    public IOType getType() {
        return IOType.fromInteger(type);
    }

    public Revision getHardwareVersion() {
        return hardwareVersion;
    }

    public Revision getFirmwareVersion() {
        return firmwareVersion;
    }

    public String getTypeString() {
        return getTypeEnum().name();
    }

    public IOType getTypeEnum() {
        if (validTypes.contains(type)) {
            return IOType.fromInteger(type);
        } else {
            return IOType.IO_TYPE_GENERIC;
        }
    }

    @Override
    public String toString() {
        return "ConnectInfo{" +
                ", connectId=" + connectId +
                ", hubIndex=" + hubIndex +
                ", type=" + type +
                '}';
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        ConnectInfo that = (ConnectInfo) o;

        if (connectId != that.connectId) return false;
        if (hubIndex != that.hubIndex) return false;
        if (type != that.type) return false;
        if (firmwareVersion != null ? !firmwareVersion.equals(that.firmwareVersion) : that.firmwareVersion != null)
            return false;
        if (hardwareVersion != null ? !hardwareVersion.equals(that.hardwareVersion) : that.hardwareVersion != null)
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = connectId;
        result = 31 * result + hubIndex;
        result = 31 * result + type;
        result = 31 * result + (hardwareVersion != null ? hardwareVersion.hashCode() : 0);
        result = 31 * result + (firmwareVersion != null ? firmwareVersion.hashCode() : 0);
        return result;
    }
}
