package dk.lego.devicesdk.device;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.services.LegoService;

/**
 * Implement this listener to be notified about changes to the attributes of the device.
 * As a listener you will be notified when an IO such as a motor or sensor is attached or detached from the Device.
 * Furthermore, you will be notified about changes to the device name, battery level and the connect-button state (pressed/released).
 */
public interface DeviceCallbackListener {

    /**
     * Invoked when a new DeviceInfo with info about the software and firmware revision is received from the device.
     * @param device     The device
     * @param deviceInfo Info about software and firmware revision of the device
     * @param error      If an error occurred, the cause of the failure.
     */
    void didUpdateDeviceInfo(LegoDevice device, DeviceInfo deviceInfo, LDSDKError error);

    /**
     * Invoked if the device sends an updated device name.
     *
     * @param device  The device
     * @param oldName The previous name of the device
     * @param newName The new name of the device
     */
    void didChangeNameFrom(LegoDevice device, String oldName, String newName);

    /**
     * Invoked when the user press or release the connect-button on the device
     *
     * @param device  The device
     * @param pressed true if the button is pressed, false otherwise.
     */
    void didChangeButtonState(LegoDevice device, boolean pressed);

    /**
     * Invoked when a device sends an updated battery level.
     *
     * @param device   The device
     * @param newLevel The new battery level as a number between 0 and 100.
     */
    void didUpdateBatteryLevel(LegoDevice device, int newLevel);

    /**
     * Invoked when a device sends a low voltage notification.
     *
     * @param device   The device
     * @param lowVoltage true if the battery has 'low voltage', false otherwise.
     */
    void didUpdateLowVoltageState(LegoDevice device, boolean lowVoltage);

    /**
     * Invoked when the a new motor, sensor or other service is attached to the device (Hub).
     * <p/>
     * Before using the LegoService you may check whether the LegoService is an instance of
     * one of the known service types, such as Moto, TiltSensor or MotionSensor just to mention a few.
     * This will allow you to use a 'safer' and more convenient interface to address the underlying IO.
     *
     * @param device  The device
     * @param service The attached service
     */
    void didAddService(LegoDevice device, LegoService service);

    /**
     * Invoked when a new motor, sensor or other service is detached from the device (Hub).
     *
     * @param device  The device
     * @param service The detached service
     */
    void didRemoveService(LegoDevice device, LegoService service);

    /**
     * Invoked when an update from the device about attached services (sensor, motors, etc) could not
     * be understood.
     *
     * @param device The device
     * @param error  The cause of the failure.
     */
    void didFailToAddServiceWithError(LegoDevice device, LDSDKError error);
}