package dk.lego.devicesdk.device;

import dk.lego.devicesdk.LDSDKError;

/**
 * Callback used to signal whenever a LEGO device has appeared or disappeared.
 * Note, that callbacks are not guaranteed to be delivered in the same order as the listeners are added.
 */
public interface DeviceManagerCallbackListener {

    /**
     * Invoked when a device advertising a LEGO Device service UUID is discovered.
     *
     * @param device The discovered device
     */
    public void onDeviceAppeared(LegoDevice device);

    /**
     * Invoked when a device stops advertising a LEGO Device service.
     * The DeviceManager will check at small refresh-intervals if an advertising
     * packet was received during the refresh-interval. If not, this method is invoked.
     *
     * @param device The device that stopped advertising
     */
    public void onDeviceDisappeared(LegoDevice device);

    /**
     * Invoked when starting a device connect attempt. Normally, this will happen right after calling see {@link dk.lego.devicesdk.device.LegoDeviceManager#connectToDevice(android.content.Context, LegoDevice)}.
     * However, it may also happen in relation to an automatic reconnect attempt.
     *
     * @param device The device to establish a connection to
     */
    public void onWillStartConnectingToDevice(LegoDevice device);

    /**
     Invoked when a device fails to connect, of if a connection request times out.

     @param device         The device that failed to connect
     @param autoReconnect  true if an automatic reconnect will be attempted, see {@link LegoDeviceManager#isAutomaticReconnectOnConnectionLostEnabled()}.
     @param error          The cause of the failure.
     */
    public void onDidFailToConnectToDevice(LegoDevice device, boolean autoReconnect, LDSDKError error);

    /**
     Invoked when a device is disconnected.

     @param device         The disconnected device
     @param autoReconnect  true if an automatic reconnect will be attempted, see {@link LegoDeviceManager#isAutomaticReconnectOnConnectionLostEnabled()}.
     @param error          If an error occurred, the cause of the failure.
     */
    public void onDidDisconnectFromDevice(LegoDevice device, boolean autoReconnect, LDSDKError error);

    /**
     * Invoked when a connection to a device is established, and the interrogation of the device for required services begins.
     * A connection is established at this point but the device is not yet ready to be used.
     *
     * @param device The connected device
     */
    public void onDidStartInterrogatingDevice(LegoDevice device);

    /**
     * Invoked when a connection to a device is established and all required services has been discovered. At this point the device is ready for use.
     *
     * @param device The connected device
     */
    public void onDidFinishInterrogatingDevice(LegoDevice device);
}
