package dk.lego.devicesdk.device;

import java.util.List;

import dk.lego.devicesdk.services.LegoService;

/**
 * A device represents the physical device / Hub.
 * The device may have a number of services (inputs, motors, etc).
 * <p/>
 * The {@link dk.lego.devicesdk.device.LegoDeviceManager}  can be used to scan for and connect to a {@link dk.lego.devicesdk.device.LegoDevice}.
 * <p/>
 * Implement the {@link dk.lego.devicesdk.device.DeviceCallbackListener}  to be notified about changes to the device.
 */
public interface LegoDevice {
    /**
     * The connect state of the Device
     */
    public enum DeviceState {
        /**
         * The Device is disconnected and no connection attempt is in progress
         */
        DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING,
        /**
         * The Device is disconnected and no longer advertising - the device also returns to this state after a successful connection has been closed
         */
        DEVICE_CONNECTION_STATE_DISCONNECTED_NOT_ADVERTISING,
        /**
         * A connection attempt is in progress
         */
        DEVICE_CONNECTION_STATE_CONNECTING,
        /**
         * Connected and interrogating Device for required services
         */
        DEVICE_CONNECTION_STATE_INTERROGATING,
        /**
         * Connected and interrogation complete - device is ready for use
         */
        DEVICE_CONNECTION_STATE_INTERROGATION_FINISHED
    }

    /**
     * The Category of the Device
     */
    public enum DeviceCategory {
        /**
         * WeDo
         */
        DEVICE_CATEGORY_WEDO(0),
        /**
         * Duplo
         */
        DEVICE_CATEGORY_DUPLO(1),
        /**
         * System
         */
        DEVICE_CATEGORY_SYSTEM(2),
        /**
         * Technic
         */
        DEVICE_CATEGORY_TECHNIC(3),
        /**
         * MindStorms
         */
        DEVICE_CATEGORY_MINDSTORMS(4),
        /**
         * Unknown
         */
        DEVICE_CATEGORY_UNKNOWN(5);

        private int value;

        private DeviceCategory(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static DeviceCategory fromInteger(int value) {
            switch (value) {
                case 0:
                    return DEVICE_CATEGORY_WEDO;
                case 1:
                    return DEVICE_CATEGORY_DUPLO;
                case 2:
                    return DEVICE_CATEGORY_SYSTEM;
                case 3:
                    return DEVICE_CATEGORY_TECHNIC;
                case 4:
                    return DEVICE_CATEGORY_MINDSTORMS;
                case 5:
                default:
                    return DEVICE_CATEGORY_UNKNOWN;
            }
        }
    }

    /**
     * The Function a Device will support
     */
    public enum DeviceFunction {
        /**
         * Central Mode
         */
        DEVICE_FUNCTION_CENTRAL_MODE(0),
        /**
         * Peripheral Mode
         */
        DEVICE_FUNCTION_PERIPHERAL_MODE(1),
        /**
         * Device IO Ports mode
         */
        DEVICE_FUNCTION_IO_PORTS(2),
        /**
         * Device acts as a remote controller
         */
        DEVICE_FUNCTION_ACTS_AS_REMOTE_CONTROLLER(3),
        /**
         * Unknown
         */
        DEVICE_FUNCTION_UNKNOWN(4);

        private int value;

        private DeviceFunction(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static DeviceFunction fromInteger(int value) {
            switch (value) {
                case 0:
                    return DEVICE_FUNCTION_CENTRAL_MODE;
                case 1:
                    return DEVICE_FUNCTION_PERIPHERAL_MODE;
                case 2:
                    return DEVICE_FUNCTION_IO_PORTS;
                case 3:
                    return DEVICE_FUNCTION_ACTS_AS_REMOTE_CONTROLLER;
                case 4:
                default:
                    return DEVICE_FUNCTION_UNKNOWN;
            }
        }
    }

    /**
     * The currently available inputs and outputs
     */
    public List<LegoService> getServices();

    /**
     * An internal service is a service that is inherent to the Device - something that can never be 'detached'.
     * Examples include the VoltageSensor and CurrentSensor that
     */
    public List<LegoService> getInternalServices();

    /**
     * An external service is a service that represent and IO that can be attached to the device Hub.
     * Examples include the Motor, TiltSensor and MotionSensor
     */
    public List<LegoService> getExternalServices();

    /**
     * Returns the current state of the connection.
     */
    public DeviceState getConnectState();

    /**
     * A unique identifier for the device
     */
    public String getDeviceId();

    /**
     * The most recent value of the name property read from the Hub.
     */
    public String getName();

    /**
     * Writing a new name will immediately update the property value, even though the actual write to the
     * hardware is asynchronously, and may potentially fail. When the write completes the {@link dk.lego.devicesdk.device.DeviceCallbackListener#didChangeNameFrom}
     * is invoked.
     */
     public void setName(String name);

    /**
     * The most recent button pressed state read from the Device.
     */
    public boolean isButtonPressed();

    /**
     * The System Category of the connected Device
     */
    public DeviceCategory getCategory();

    /**
     * The Fuction(s) supported by the connected Device
     */
    public DeviceFunction getSupportedFunctions();

    /**
     * The ID of the network this Device was connected to last
     */
    public int getLastConnectedNetworkId();

    /**
     * The battery level of the device in percentage
     * If no battery level has been received from the Device, the value is null
     */
    public Integer getBatteryLevel();

    /**
     * True if the a low voltage alert has been received from the Device, indicating that batteries should be changed/charged
     */
    public boolean isLowVoltage();

    /**
     * Info about the device hardware, firmware, and software revision
     */
    public DeviceInfo getDeviceInfo();

    /**
     * If a callback listener is registered it receives callbacks on changes to offered services,
     * as well as properties of the device like name and color.
     *
     * @param listener The delegate to add
     */
    public void registerCallbackListener(DeviceCallbackListener listener);

    /**
     * Remove callback listener from this device
     *
     * @param listener The delegate to remove
     */
    public void unregisterCallbackListener(DeviceCallbackListener listener);
}
