package dk.lego.devicesdk.device;

import android.content.Context;

import java.util.List;

/**
 * This class is the main entry point for connecting and communicating with a LEGO Device.
 * You must implement the {@link dk.lego.devicesdk.device.DeviceManagerCallbackListener} and set the listener property
 * before scanning for and connecting to devices.
 * <p/>
 * <b>Note:</b> Remember to remove attached listener(s) again to avoid memory leaks
 */
public interface LegoDeviceManager {
    /**
     * The default connect request timeout in milliseconds
     */
    static final long DEFAULT_CONNECT_REQUEST_TIMEOUT = 10000;

    /**
     * Start scanning for LEGO devices
     *
     * @param context The Android context
     */
    public void scan(Context context);

    /**
     * Stop scanning for LEGO BLE devices
     */
    public void stopScanning();

    /**
     * Connect to a LEGO Device.
     * If a connection is not established within the connectRequestTimeoutInterval the connection attempt is cancelled
     * and the {@link DeviceManagerCallbackListener#onDidFailToConnectToDevice} is invoked.
     *
     * @param context The Android context
     * @param device  The device to establish a connection to
     */
    public void connectToDevice(Context context, LegoDevice device);

    /**
     * If a connect request is not successful within this time interval the connection attempt is cancelled
     * and the {@link dk.lego.devicesdk.device.DeviceManagerCallbackListener#onDidFailToConnectToDevice(LegoDevice, boolean, dk.lego.devicesdk.LDSDKError)} is invoked.
     * The default value is 10 seconds.
     *
     * @param timeoutInterval The timeout interval in milliseconds
     */
    public void setConnectRequestTimeoutInterval(long timeoutInterval);

    /**
     * @return the connection attempt timeout value interval
     */
    public long getConnectRequestTimeoutInterval();

    /**
     * If enabled, the LegoDeviceManager will attempt to reconnect in case of a connection loss, but only if the connection was not closed by the user,
     * the default value is false.
     * <p/>
     * If this property is set to true an automatic attempt to connect is made in the following cases:
     * <p/>
     * If a connection requests fails with an 'unexpected error' (e.g. not due to a time-out). If the second connection attempt also fails,
     * no further attempts are made to connect automatically.
     * If a connection in lost in state {@link dk.lego.devicesdk.device.LegoDevice.DeviceState#DEVICE_CONNECTION_STATE_INTERROGATING} with an 'unexpected error' (e.g. not closed by user).
     * If the second connection attempt also fails, no further attempts are made to connect automatically.
     * If a connection is lost after successful connection (i.e. in state {@link dk.lego.devicesdk.device.LegoDevice.DeviceState#DEVICE_CONNECTION_STATE_INTERROGATION_FINISHED})
     * <p/>
     * If the connect-request does not succeed within the connectRequestTimeout no attempt is made to automatically reconnect.
     *
     * @param enabled Whether automatic reconnect should be enabled or not
     */
    public void setAutomaticReconnectOnConnectionLostEnabled(boolean enabled);

    /**
     * @return true if automatic reconnect on connection lost is enabled, otherwise false
     */
    public boolean isAutomaticReconnectOnConnectionLostEnabled();

    /**
     * Disconnect from a LEGO Device.
     *
     * @param device The device to disconnect from
     */
    public void cancelDeviceConnection(LegoDevice device);


    /**
     * Returns a list of devices in the specified device state. Use this, for example, to retrieve a list
     * of all advertising but non-connected devices, i.e. {@link dk.lego.devicesdk.device.LegoDevice.DeviceState#DEVICE_CONNECTION_STATE_DISCONNECTED_ADVERTISING}.
     *
     * @param connectionState The state of the devices.
     */
    public List<? extends LegoDevice> devicesInState(LegoDevice.DeviceState connectionState);

    /**
     * Returns a list with all known devices regardless of their current connect state.
     * <p/>
     * For BLE devices, the list is ordered according to relative distance, based of the RSSI values received for each Device.
     * The Device closest will be first in the list. Note, the RSSI value is only updated while in advertising mode.
     */
    public List<? extends LegoDevice> allDevices();

    /**
     * Add a listener to receive device discovery and connection events
     * Note, that callbacks are not guaranteed to be delivered in the same order as the listeners are added.
     *
     * @param listener The listener to add
     */
    public void registerCallbackListener(DeviceManagerCallbackListener listener);

    /**
     * Remove a listener
     *
     * @param listener The listener to remove
     */
    public void unregisterCallbackListener(DeviceManagerCallbackListener listener);

    /**
     * @return the current version of the LegoDeviceSDK
     */
    public String getSDKVersion();
}
