package dk.lego.devicesdk.device;

import android.content.Context;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import dk.lego.devicesdk.BuildConfig;
import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDevice;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDeviceManagerCallbackListener;
import dk.lego.devicesdk.bluetooth.LegoBluetoothDeviceManagerImpl;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothAdapterWrapper;
import dk.lego.devicesdk.bluetooth.wrappers.AndroidBluetoothAdapterWrapperImpl;
import dk.lego.devicesdk.logging.LDSDKLogger;

public class LegoDeviceManagerImpl implements LegoDeviceManager, LegoBluetoothDeviceManagerCallbackListener {

    private LegoBluetoothDeviceManagerImpl bluetoothDeviceManager;
    private final Set<LegoBluetoothDevice> devices = new HashSet<>();

    private static LegoDeviceManager instance = null;
    private long connectRequestTimeoutInterval = LegoDeviceManager.DEFAULT_CONNECT_REQUEST_TIMEOUT;
    private boolean automaticReconnectOnConnectionLostEnabled;

    private DeviceManagerCallbackHelper callbackHelper = new DeviceManagerCallbackHelper();

    public static LegoDeviceManager getInstance() {
        if (instance == null) {
            instance = new LegoDeviceManagerImpl();
        }
        return instance;
    }

    private LegoDeviceManagerImpl() {
    }

    @Override
    public void scan(Context context) throws IllegalStateException {
        if (bluetoothDeviceManager == null) {
            bluetoothDeviceManager = new LegoBluetoothDeviceManagerImpl(getAndroidBluetoothAdapter(context.getApplicationContext()));
        }

        bluetoothDeviceManager.setAutomaticReconnectOnConnectionLostEnabled(automaticReconnectOnConnectionLostEnabled);
        bluetoothDeviceManager.setConnectRequestTimeoutInterval(connectRequestTimeoutInterval);

        bluetoothDeviceManager.registerCallbackListener(this);
        bluetoothDeviceManager.scan(context);
    }

    @Override
    public void stopScanning() {
        if (bluetoothDeviceManager != null) {
            bluetoothDeviceManager.stopScanning();
        }
    }

    @Override
    public void connectToDevice(Context context, LegoDevice device) {
        if (device instanceof LegoBluetoothDevice) {
            bluetoothDeviceManager.connectToDevice(context.getApplicationContext(), (LegoBluetoothDevice) device);
        } else {
            LDSDKLogger.e("Can only connect to devices of type LegoBluetoothDevice");
        }
    }

    @Override
    public void setConnectRequestTimeoutInterval(long timeoutInterval) {
        connectRequestTimeoutInterval = timeoutInterval;
        if (bluetoothDeviceManager != null) {
            bluetoothDeviceManager.setConnectRequestTimeoutInterval(connectRequestTimeoutInterval);
        }
    }

    @Override
    public long getConnectRequestTimeoutInterval() {
        return connectRequestTimeoutInterval;
    }

    @Override
    public void setAutomaticReconnectOnConnectionLostEnabled(boolean enabled) {
        automaticReconnectOnConnectionLostEnabled = enabled;
        if (bluetoothDeviceManager != null) {
            bluetoothDeviceManager.setAutomaticReconnectOnConnectionLostEnabled(automaticReconnectOnConnectionLostEnabled);
        }
    }

    @Override
    public boolean isAutomaticReconnectOnConnectionLostEnabled() {
        return automaticReconnectOnConnectionLostEnabled;
    }

    @Override
    public void cancelDeviceConnection(LegoDevice device) {
        if (device instanceof LegoBluetoothDevice) {
            bluetoothDeviceManager.cancelDeviceConnection((LegoBluetoothDevice) device);
        } else {
            LDSDKLogger.e("Can only connect to devices of type LegoBluetoothDevice");
        }
    }

    @Override
    public List<? extends LegoDevice> devicesInState(LegoDevice.DeviceState connectionState) {
        return bluetoothDeviceManager.devicesInState(connectionState);
    }

    @Override
    public List<? extends LegoDevice> allDevices() {
        if (bluetoothDeviceManager != null) {
            return bluetoothDeviceManager.allDevices();
        } else {
            return Collections.emptyList();
        }
    }

    @Override
    public void registerCallbackListener(DeviceManagerCallbackListener callbackListener) {
        callbackHelper.registerCallbackListener(callbackListener);
    }

    @Override
    public void unregisterCallbackListener(DeviceManagerCallbackListener callbackListener) {
        callbackHelper.unregisterCallbackListener(callbackListener);
    }

    @Override
    public String getSDKVersion() {
        return BuildConfig.VERSION_NAME;
    }

    @Override
    public void onDeviceDidAppear(LegoBluetoothDevice device) {
        devices.add(device);
        callbackHelper.performDeviceAppearedCallback(device);
    }

    @Override
    public void onDeviceDidDisappear(LegoBluetoothDevice device) {
        devices.remove(device);
        callbackHelper.performDeviceDisappearedCallback(device);
    }

    @Override
    public void onWillStartConnectingToDevice(LegoBluetoothDevice device) {
        callbackHelper.performWillStartConnectingToDeviceCallback(device);
    }

    @Override
    public void onDidFailToConnectToDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error) {
        callbackHelper.performDidFailToConnectToDevice(device, autoReconnect, error);
    }

    @Override
    public void onDidDisconnectFromDevice(LegoBluetoothDevice device, boolean autoReconnect, LDSDKError error) {
        callbackHelper.performDidDisconnectFromDevice(device, autoReconnect, error);
    }

    @Override
    public void onDidStartInterrogatingDevice(LegoBluetoothDevice device) {
        callbackHelper.performDidStartInterrogatingDevice(device);
    }

    @Override
    public void onDidFinishInterrogatingDevice(LegoBluetoothDevice device) {
        callbackHelper.performDidFinishInterrogratingDevice(device);
    }

    private AndroidBluetoothAdapterWrapper getAndroidBluetoothAdapter(Context applicationContext) throws IllegalStateException {
        return new AndroidBluetoothAdapterWrapperImpl(applicationContext);
    }
}
