package dk.lego.devicesdk.device;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Holds revision info to represent for example the hardware and firmware revisions
 * of a device and attached IOs (services).
 */
public class Revision {
    /**
     * The major version number
     */
    private int majorVersion;
    /**
     * The minor version number
     */
    private int minorVersion;
    /**
     * The bug fix version number
     */
    private int bugFixVersion;
    /**
     * The build number
     */
    private int buildNumber;

    public static Revision createFromString(String revisionString) {
        return new Revision(revisionString);
    }

    public static Revision createFromDataData(byte[] revisionData) {
        return new Revision(revisionData);
    }

    private Revision(String revisionString) {
        parseRevision(revisionString);
    }

    private Revision(byte[] revisionData) {
        if (revisionData.length >= 1) {
            majorVersion = revisionData[0];
        }
        if (revisionData.length >= 2) {
            minorVersion = revisionData[1];
        }
        if (revisionData.length >= 3) {
            bugFixVersion = revisionData[2];
        }
        if (revisionData.length >= 4) {
            buildNumber = revisionData[3];
        }
    }

    private void parseRevision(String revision) {
        Pattern pattern = Pattern.compile("(^\\d+(\\.\\d+)*$)");
        Matcher matcher = pattern.matcher(revision);
        boolean matchFound = matcher.find();

        if (matchFound) {
            String[] versionComponents = matcher.group(1).split("\\.");
            if (versionComponents.length >= 1) {
                majorVersion = Integer.parseInt(versionComponents[0]);
            }
            if (versionComponents.length >= 2) {
                minorVersion = Integer.parseInt(versionComponents[1]);
            }
            if (versionComponents.length >= 3) {
                bugFixVersion = Integer.parseInt(versionComponents[2]);
            }
            if (versionComponents.length >= 4) {
                buildNumber = Integer.parseInt(versionComponents[3]);
            }
        } else {
            // If string contains anything but numbers and '.' then set all properties to zero
            majorVersion = 0;
            minorVersion = 0;
            bugFixVersion = 0;
            buildNumber = 0;
        }
    }

    public int getMajorVersion() {
        return majorVersion;
    }

    public int getMinorVersion() {
        return minorVersion;
    }

    public int getBugFixVersion() {
        return bugFixVersion;
    }

    public int getBuildNumber() {
        return buildNumber;
    }

    @Override
    public String toString() {
        return String.format("%d.%d.%d.%d", majorVersion, minorVersion, bugFixVersion, buildNumber);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Revision revision = (Revision) o;

        if (bugFixVersion != revision.bugFixVersion) return false;
        if (buildNumber != revision.buildNumber) return false;
        if (majorVersion != revision.majorVersion) return false;
        if (minorVersion != revision.minorVersion) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = majorVersion;
        result = 31 * result + minorVersion;
        result = 31 * result + bugFixVersion;
        result = 31 * result + buildNumber;
        return result;
    }
}
