package dk.lego.devicesdk.input_output;

/**
 * This class contains info detailing how the data received for a given service (typically a sensor of some kind) should be interpreted.
 * <p/>
 * For senors types recognized by the SDK (like the tilt and motion sensor) you will not need to know about
 * the details of this class as the implementation of these services in the SDK handles this for you.
 * <p/>
 * If you need to access an Input that is not recognized by the SDK you should create and set one
 * or more valid input formats for a {@link dk.lego.devicesdk.services.LegoService}. See the {@link dk.lego.devicesdk.services.GenericService} documentation for an example.
 */
public class DataFormat {
    /**
     * The name sensor mode (fx. crash, tilt or angle for a tilt sensor)
     */
    String modeName;

    /**
     * The sensor mode
     */
    int mode;

    /**
     * The sensor unit
     */
    InputFormat.InputFormatUnit unit;

    /**
     * The data set size - fx. 4 if the data is a four byte float
     */
    int dataSetSize;

    /**
     * The data set count (fx. size 3 for a sensor that produce a value in x, y and z-direction)
     */
    int dataSetCount;

    /**
     * Create and initialize a new instance of an DataFormat.
     * <p/>
     * Example: When a tilt sensor is in mode 'angle' it will create readings of in the x, y and z-dimension. If the mode is set to SI each
     * angle will be a float representing with a value between 0 and 90 degrees. To create a data set that tells the SDK how to interpret values
     * for til tilt sensor in this mode you would write.
     * <p/>
     * <code>DataFormat tiltSensorFormat = DataFormat.create("Angle", 0, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 3, 4);</code>
     *
     * @param modeName         The name of the mode
     * @param mode             The sensor mode
     * @param unit             The sensor unit
     * @param numberOfBytes    The number of bytes in a data set
     * @param numberOfDataSets The number of data sets
     */
    public static DataFormat create(String modeName, int mode, InputFormat.InputFormatUnit unit, int numberOfBytes, int numberOfDataSets) {
        return new DataFormat(modeName, mode, unit, numberOfBytes, numberOfDataSets);
    }

    private DataFormat(String modeName, int mode, InputFormat.InputFormatUnit unit, int numberOfBytes, int numberOfDataSets) {
        this.modeName = modeName;
        this.mode = mode;
        this.unit = unit;
        this.dataSetSize = numberOfBytes;
        this.dataSetCount = numberOfDataSets;
    }

    public String getModeName() {
        return modeName;
    }

    public int getMode() {
        return mode;
    }

    public InputFormat.InputFormatUnit getUnit() {
        return unit;
    }

    public int getDataSetSize() {
        return dataSetSize;
    }

    public int getDataSetCount() {
        return dataSetCount;
    }

    @Override
    public String toString() {
        return "DataFormat{" +
                "modeName='" + modeName + '\'' +
                ", unit='" + unit.name() + '\'' +
                ", dataSetSize=" + dataSetSize +
                ", dataSetCount=" + dataSetCount +
                '}';
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        DataFormat that = (DataFormat) o;

        if (mode != that.mode) return false;
        if (unit != that.unit) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = mode;
        result = 31 * result + (unit != null ? unit.hashCode() : 0);
        return result;
    }
}