package dk.lego.devicesdk.input_output;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.utils.ByteUtils;

/**
 * This class represent a configuration of a Input (sensor). At any time a sensor can be in just one mode,
 * and the details of this mode is captured by this class.
 * <p/>
 * For senors types recognized by the SDK (like the tilt and motion sensor) you will not need to know about
 * the details of this class as the implementation of these services in the SDK handles this for you.
 * <p/>
 * If you need to access an Input that is not recognized by the SDK you will need to create and send
 * an input format for the corresponding service. See the {@link dk.lego.devicesdk.services.GenericService} documentation for an example.
 */
public class InputFormat {

    public static final int INPUT_FORMAT_PACKAGE_SIZE = 11;
    private static final int FORMAT_DATA_SIZE = 8;

    /**
     * The input format unit
     */
    public enum InputFormatUnit {
        /**
         * Raw
         */
        INPUT_FORMAT_UNIT_RAW(0),
        /**
         * Percentage
         */
        INPUT_FORMAT_UNIT_PERCENTAGE(1),
        /**
         * SI
         */
        INPUT_FORMAT_UNIT_SI(2),
        /**
         * Unknown
         */
        INPUT_FORMAT_UNIT_UNKNOWN(3);

        private int value;

        private InputFormatUnit(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static InputFormatUnit fromInteger(int value) {
            switch (value) {
                case 0:
                    return INPUT_FORMAT_UNIT_RAW;
                case 1:
                    return INPUT_FORMAT_UNIT_PERCENTAGE;
                case 2:
                    return INPUT_FORMAT_UNIT_SI;
                case 3:
                default:
                    return INPUT_FORMAT_UNIT_UNKNOWN;
            }
        }

        @Override
        public String toString() {
            switch (value) {
                case 0:
                    return "RAW";
                case 1:
                    return "%";
                case 2:
                    return "SI";
                case 3:
                default:
                    return "?";
            }
        }
    }

    /**
     * The revision of the Input Format (set by the Device).
     * When a new Input Format is set for a service the Device will send the updated Input Format through the {@link dk.lego.devicesdk.services.ServiceCallbackListener#didUpdateInputFormat(dk.lego.devicesdk.services.LegoService, InputFormat, InputFormat)}.
     * The Device will assign a revision number to the new Input Format. The revision number is matched against the revision format when receiving values for the
     * corresponding service.
     */
    int revision;

    /**
     * The connectId of the corresponding service, see {@link dk.lego.devicesdk.services.LegoService#connectInfo}
     */
    int connectId;

    /**
     * The typeId of the corresponding service, see {@link dk.lego.devicesdk.services.LegoService#connectInfo}
     */
    int typeId;

    /**
     * The mode of the Input
     */
    int mode;

    /**
     * The delta interval. When notifications are enabled the service will only receive updates if the value has change with 'delta interval' or more since last reading
     */
    long deltaInterval;

    /**
     * The unit the values are delivered in (as raw values, or as SI values)
     */
    int unit;

    /**
     * true if new values are send whenever the value of the Input changes beyond delta interval
     */
    boolean notificationsEnabled;

    /**
     * The number of bytes to be expected in the Input data payload (set by the Device)
     */
    int numberOfBytes;

    private InputFormat(int connectId, int typeId, int mode, long deltaInterval, int unit, boolean notificationsEnabled) {
        this.connectId = connectId;
        this.typeId = typeId;
        this.mode = mode;
        this.deltaInterval = deltaInterval;
        this.unit = unit;
        this.notificationsEnabled = notificationsEnabled;
    }

    private InputFormat(byte[] data) {
        ByteBuffer dataBuffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);

        revision = dataBuffer.get();
        connectId = dataBuffer.get();
        typeId = dataBuffer.get();
        mode = dataBuffer.get();
        deltaInterval = ByteUtils.getUnsignedInt(dataBuffer);
        unit = dataBuffer.get();
        notificationsEnabled = dataBuffer.get() == 1;
        numberOfBytes = dataBuffer.get();
    }

    public static InputFormat inputFormat(byte[] data) {
        if (data == null) {
            LDSDKLogger.e("Cannot instantiate InputFormat with null data");
            return null;
        }

        // First byte is for the revision number, the rest of the bytes should be one or more input format sub-packages.
        if (data.length != INPUT_FORMAT_PACKAGE_SIZE) {
            LDSDKLogger.e(String.format("Cannot create InputFormat from package with size %d, expected size to be %d",
                    data.length, INPUT_FORMAT_PACKAGE_SIZE));
            return null;
        }

        return new InputFormat(data);
    }

    /**
     * Create a new instance of an InputFormat.
     *
     * @param connectId            The connectId of the service, see {@link dk.lego.devicesdk.services.LegoService#connectInfo}
     * @param type                 The type of the IO, see {@link dk.lego.devicesdk.services.LegoService#connectInfo}
     * @param mode                 The mode of the IO (Inputs/Senors may support a number of different modes)
     * @param deltaInterval        The delta interval
     * @param unit                 The unit the sensor should return values in
     * @param notificationsEnabled true if the device should send updates when the value changes, otherwise false
     */
    public static InputFormat inputFormat(int connectId,
                                          ConnectInfo.IOType type,
                                          int mode,
                                          int deltaInterval,
                                          InputFormatUnit unit,
                                          boolean notificationsEnabled) {
        return new InputFormat(connectId, type.getValue(), mode, deltaInterval, unit.getValue(), notificationsEnabled);
    }


    /**
     * Creates a copy of this input format with a new mode
     *
     * @param mode The new mode
     */
    public InputFormat inputFormatBySettingMode(int mode) {
        return new InputFormat(this.connectId, this.typeId, mode, this.deltaInterval, this.unit, this.notificationsEnabled);
    }

    /**
     * Creates a copy of this input format with a new mode and unit
     *
     * @param mode The new mode
     * @param unit The new unit
     */
    public InputFormat inputFormatBySettingModeAndUnit(int mode, InputFormatUnit unit) {
        return new InputFormat(this.connectId, this.typeId, mode, this.deltaInterval, unit.getValue(), this.notificationsEnabled);
    }

    /**
     * Creates a copy of this input format with a new delta interval
     *
     * @param deltaInterval The new delta interval
     */
    public InputFormat inputFormatBySettingDeltaInterval(byte deltaInterval) {
        return new InputFormat(this.connectId, this.typeId, this.mode, deltaInterval, this.unit, this.notificationsEnabled);
    }

    /**
     * Creates a copy of this input format with a new value for notifications enabled
     *
     * @param notificationsEnabled true if the sensor should send updates when the value changes, otherwise false
     */
    public InputFormat inputFormatBySettingNotificationsEnabled(boolean notificationsEnabled) {
        return new InputFormat(this.connectId, this.typeId, mode, this.deltaInterval, this.unit, notificationsEnabled);
    }

    public int getRevision() {
        return revision;
    }

    public int getConnectId() {
        return connectId;
    }

    public ConnectInfo.IOType getType() {
        return ConnectInfo.IOType.fromInteger(typeId);
    }

    public int getMode() {
        return mode;
    }

    public long getDeltaInterval() {
        return deltaInterval;
    }

    public InputFormatUnit getUnit() {
        return InputFormatUnit.fromInteger(unit);
    }

    public boolean isNotificationsEnabled() {
        return notificationsEnabled;
    }

    public int getNumberOfBytes() {
        return numberOfBytes;
    }

    public byte[] writeFormatData() {
        ByteBuffer writeData = ByteBuffer.allocate(FORMAT_DATA_SIZE).order(ByteOrder.LITTLE_ENDIAN);
        ByteUtils.putUnsignedByte(writeData, typeId);
        ByteUtils.putUnsignedByte(writeData, mode);
        ByteUtils.putUnsignedInt(writeData, deltaInterval);
        ByteUtils.putUnsignedByte(writeData, unit);
        ByteUtils.putUnsignedByte(writeData, (notificationsEnabled ? 1 : 0));

        return writeData.array();
    }

    @Override
    public String toString() {
        return "InputFormat{" +
                "revision=" + revision +
                ", connectId=" + connectId +
                ", typeId=" + typeId +
                ", mode=" + mode +
                ", deltaInterval=" + deltaInterval +
                ", unit=" + unit +
                ", notificationsEnabled=" + notificationsEnabled +
                ", numberOfBytes=" + numberOfBytes +
                '}';
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        InputFormat that = (InputFormat) o;

        if (connectId != that.connectId) return false;
        if (deltaInterval != that.deltaInterval) return false;
        if (mode != that.mode) return false;
        if (notificationsEnabled != that.notificationsEnabled) return false;
        if (numberOfBytes != that.numberOfBytes) return false;
        if (revision != that.revision) return false;
        if (typeId != that.typeId) return false;
        if (unit != that.unit) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = revision;
        result = 31 * result + connectId;
        result = 31 * result + typeId;
        result = 31 * result + mode;
        result = 31 * result + (int) (deltaInterval ^ (deltaInterval >>> 32));
        result = 31 * result + unit;
        result = 31 * result + (notificationsEnabled ? 1 : 0);
        result = 31 * result + numberOfBytes;
        return result;
    }
}
