package dk.lego.devicesdk.logging;

/**
 * Use this class to configure the log level that you want the LEGO Device SDK to use.
 * <p/>
 * You may provide your own {@link dk.lego.devicesdk.logging.CustomLogger} that all log from the LEGO Device SDK will be written to,
 * by calling the {@link dk.lego.devicesdk.logging.LDSDKLogger#setCustomLogger}
 */
public final class LDSDKLogger {
    /**
     * Log level
     */
    public enum LoggerLevel {
        /**
         * Verbose Level
         */
        LOGGER_LEVEL_VERBOSE(0),
        /**
         * Debug Level
         */
        LOGGER_LEVEL_DEBUG(1),
        /**
         * Info Level
         */
        LOGGER_LEVEL_INFO(2),
        /**
         * Warn Level
         */
        LOGGER_LEVEL_WARN(3),
        /**
         * Error Level
         */
        LOGGER_LEVEL_ERROR(4);

        private int value;

        LoggerLevel(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static LoggerLevel fromInteger(int value) {
            switch (value) {
                case 0:
                    return LOGGER_LEVEL_VERBOSE;
                case 1:
                    return LOGGER_LEVEL_DEBUG;
                case 2:
                    return LOGGER_LEVEL_INFO;
                case 3:
                    return LOGGER_LEVEL_WARN;
                case 4:
                    return LOGGER_LEVEL_ERROR;
                default:
                    LDSDKLogger.w(String.format("Reached default case when trying to determine LoggerLevel for value: %s. Level set to %s", value, LOGGER_LEVEL_VERBOSE.name()));
                    return LOGGER_LEVEL_VERBOSE;
            }
        }
    }

    private static LDSDKLogger instance;
    private static CustomLogger customLogger = null;

    /**
     * The log level. Default is {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_ERROR}.
     */
    private static LoggerLevel loggerLevel = LoggerLevel.LOGGER_LEVEL_ERROR;

    public static synchronized LDSDKLogger getInstance() {
        if (instance == null) {
            instance = new LDSDKLogger();
        }

        return instance;
    }

    /**
     * You may set a custom log writer that all log from the library will be written to.
     * If set to null (the default) {@link dk.lego.devicesdk.logging.DefaultAndroidCustomLogger}
     * using the standard Android {@link android.util.Log} will be used
     */
    public void setCustomLogger(CustomLogger customLogger) {
        LDSDKLogger.customLogger = customLogger;
    }

    /**
     * @param level The new log level
     */
    public void setLogLevel(LoggerLevel level) {
        loggerLevel = level;
    }

    /**
     * @return The current log level
     */
    public LoggerLevel getLoggerLevel() {
        return loggerLevel;
    }

    /**
     * @param msg Verbose log message (requires minimum log level of {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_VERBOSE})
     */
    public static void v(String msg) {
        log(LoggerLevel.LOGGER_LEVEL_VERBOSE, msg);
    }

    /**
     * @param msg Debug log message (requires log level of at least {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_DEBUG})
     */
    public static void d(String msg) {
        log(LoggerLevel.LOGGER_LEVEL_DEBUG, msg);
    }

    /**
     * @param msg Info log message (requires log level of at least {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_INFO})
     */
    public static void i(String msg) {
        log(LoggerLevel.LOGGER_LEVEL_INFO, msg);
    }

    /**
     * @param msg Warn log message (requires log level of at least {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_WARN})
     */
    public static void w(String msg) {
        log(LoggerLevel.LOGGER_LEVEL_WARN, msg);
    }

    /**
     * @param msg Error log message (requires log level of at least {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_ERROR})
     */
    public static void e(String msg) {
        log(LoggerLevel.LOGGER_LEVEL_ERROR, msg);
    }

    /**
     * @param e Exception log message (requires log level of at least {@link dk.lego.devicesdk.logging.LDSDKLogger.LoggerLevel#LOGGER_LEVEL_ERROR})
     */
    public static void e(Exception e) {
        if (loggerLevel.ordinal() <= LoggerLevel.LOGGER_LEVEL_ERROR.ordinal()) {
            customLogger.error(e);
        }
    }

    private static void log(LoggerLevel level, String msg) {

        if (customLogger == null) {
            customLogger = new DefaultAndroidCustomLogger();
        }

        if (loggerLevel.ordinal() <= level.ordinal()) {
            switch (level) {
                case LOGGER_LEVEL_VERBOSE:
                    customLogger.verbose(msg);
                    break;
                case LOGGER_LEVEL_DEBUG:
                    customLogger.debug(msg);
                    break;
                case LOGGER_LEVEL_INFO:
                    customLogger.info(msg);
                    break;
                case LOGGER_LEVEL_WARN:
                    customLogger.warn(msg);
                    break;
                case LOGGER_LEVEL_ERROR:
                    customLogger.error(msg);
                    break;
            }
        }
    }
}
