package dk.lego.devicesdk.services;

import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.InputFormat;

/**
 * The SDK will create instances of this class for IOs with an unknown IOType.
 * <p/>
 * The {@link GenericService} is a 'tagging' interface and does not offer any extra methods or properties other than those available
 * from its parent class {@link dk.lego.devicesdk.services.LegoService}.
 * <p/>
 * As opposed to the known service types (tilt, motion, etc) a generic service does not have a predefined default input format (@see LegoService.getDefaultInputFormat()).
 * Therefore, when the SDK discovers a service with a type it does not recognize it does not automatically send an {@link dk.lego.devicesdk.input_output.InputFormat}
 * to the {@link dk.lego.devicesdk.device.LegoDevice}. Without a configured input format the Device will not send any updates for the SDK when the value readings of the sensor
 * value changes for that service.
 * <p/>
 * To receive IO value reading updates you must create and send an InputFormat to the device to be used for the service. You may look for
 * inspiration on how to do this in one of the concrete service classes, like the {@link dk.lego.devicesdk.services.MotionSensor} or {@link dk.lego.devicesdk.services.VoltageSensor}. Below an example
 * is given on how you could configure and use a new Temperature Sensor yet unknown to the SDK.
 * <p/>
 * <pre>
 * {@code
 *   if (service instanceof GenericService && service.getConnectInfo().getType() == TEMPERATURE_SENSOR_TYPE_NUMBER) {
 *     GenericService temperatureSensor = (GenericService) service;
 *
 *     // The temperature sensor is yet unknown to the SDK (there is no TemperatureSensor class) so we need to configure
 *     // the service ourselves.
 *
 *     // As a generic sensor does not how a defaultInputFormat is defined
 *     // we must create one and send it to the device.
 *     // Look in the documentation for the temperature sensor to see which modes it supports.
 *
 *     InputFormat inputFormat = InputFormat.inputFormat(
 *       getConnectInfo().getConnectId(),
 *       getConnectInfo().getType(),
 *       0, // See the documentation for the sensor for supported modes
 *       1, // Receive updates when the value changes with delta 1
 *       InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI,
 *       true); // Notifications are enabled
 *
 *     // Tell the device to configure the sensor with the input format
 *     temperatureSensor.updateInputFormat(inputFormat);
 *
 *     // We know from the documentation that the temperature sensor produces readings in
 *     // Kelvin as 4 byte floats when in mode 0 with unit set to SI.
 *     DataFormat dataFormat = DataFormat.create(
 *       "Kelvin",
 *       0, // must match the mode for the inputFormat
 *       InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, // must match the unit for the inputFormat
 *       4, // a 4 byte float
 *       1); //only one value in the data set (the temperature)
 *
 *     temperatureSensor.addValidDataFormat(dataFormat);
 *
 *     // Now, add a callback listener to receive notifications when the service has a new temperature reading
 *     temperatureSensor.registerCallbackListener(this);
 *   }
 * }
 * </pre>
 *
 * Now, when the service receives an updated value from the temperature sensor you will receive
 * a notification through the listener.
 *
 * <pre>
 * {@code
 *   public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
 *     //As we have defined a valid data format stating that the received value can be parsed as a 4 byte
 *     //float we can use the convenience method to retrieve the value as a float.
 *     Float32 temperatureReading = service.valueAsFloat;
 *   }
 * }
 * </pre>
 *
 * It is not required to add a valid data format to the {@link dk.lego.devicesdk.services.GenericService}, but it is recommended to
 * do so as this will also help the SDK validate all received data according to the defined valid data formats
 * and write any inconsistencies to the {@link dk.lego.devicesdk.logging.LDSDKLogger}.
 */
public class GenericService extends LegoService {

    private static final String SERVICE_GENERIC_NAME = "Generic IO";

    private GenericService(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
    }

    @Override
    public String getServiceName() {
        return SERVICE_GENERIC_NAME;
    }

    @Override
    public InputFormat getDefaultInputFormat() {
        return null;
    }

    public static GenericService createService(ConnectInfo connectInfo, IO io) {
        return new GenericService(connectInfo, io);
    }
}
