package dk.lego.devicesdk.services;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.input_output.DataFormat;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.IOCallbackListener;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.logging.LDSDKLogger;

/**
 * A LegoService represent an IO of some kind, for example a motor or sensor. It could also be an internal IO, such as Voltage sensor build into the device.
 * <p/>
 * The LegoService has a number of sub-classes for known IO types. This includes Motor, TiltSensor and MotionSensor just to mention a few.
 * <p/>
 * Add an instance of a {@link dk.lego.devicesdk.services.ServiceCallbackListener} using {@link dk.lego.devicesdk.services.LegoService#registerCallbackListener} to be notified when a service receives an updated value.
 */
public class LegoService implements IOCallbackListener {

    private static final int FIRST_INTERNAL_HUB_INDEX = 50;

    /**
     * General info about the connected service
     */
    protected ConnectInfo connectInfo;

    /**
     * The Device this service is related to
     */
    private LegoDevice device;

    protected IO io;
    private byte[] valueData;
    private HashSet<DataFormat> validDataFormats = new HashSet<>();
    protected InputFormat inputFormat;

    protected LegoServiceCallbackHelper callbackHelper = new LegoServiceCallbackHelper();

    protected LegoService(ConnectInfo connectInfo, IO io) {
        assert connectInfo != null : "Cannot instantiate service with null ConnectInfo";
        assert io != null : "Cannot instantiate service with null IO";

        this.io = io;
        this.io.registerCallbackListener(this);
        this.connectInfo = connectInfo;
    }

    public static LegoService createService(ConnectInfo connectInfo, IO io) {
        return new LegoService(connectInfo, io);
    }

    public void setDevice(LegoDevice device) {
        this.device = device;
    }

    public LegoDevice getDevice() {
        return device;
    }

    public IO getIo() {
        return io;
    }

    public boolean isInternalService() {
        return this.connectInfo.getHubIndex() >= FIRST_INTERNAL_HUB_INDEX;
    }

    @Override
    public ConnectInfo didRequestConnectInfo(IO io) {
        return this.connectInfo;
    }

    /**
     * Verify received data
     *
     * @param data received data
     * @throws LDSDKError if data could not bve verified
     */
    private void verifyData(byte[] data) throws LDSDKError {
        /**
         *  No data is always ok
         *  No InputDataFormats defined - means that all received data is accepted
         */
        if (data != null && this.validDataFormats.size() != 0) {

            /**
             * If one or more InputDataFormats are defined, we look at the latest received InputFormat from the device
             * For a received value to be accepted, there:
             * 1. Must exists a DataFormat that matches the latest received InputFormat from device
             * 2. The received valueData length must match this DataFormat */
            DataFormat dataFormat = dataFormatForInputFormat(this.inputFormat);
            if (dataFormat == null) {
                String msg = String.format("Did not find a valid input data format.%nThe input format received from device is: %s.%nSupported formats: %s", this.inputFormat.toString(), this.validDataFormats.toString());
                throw LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, msg);
            }

            this.verifyData(data, dataFormat);
        }
    }

    public ConnectInfo getConnectInfo() {
        return connectInfo;
    }

    private void verifyData(byte[] data, DataFormat format) throws LDSDKError {
        if (!(data.length == (format.getDataSetSize() * format.getDataSetCount()))) {
            String msg = String.format("Value for service %s in mode %s unit %s is expected to have %d data sets each with size %d bytes, but did receive package with length %d",
                    this.getServiceName(), format.getModeName(), format.getUnit().name(), format.getDataSetCount(), format.getDataSetSize(), data.length);
            throw LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, msg);
        }
    }

    private DataFormat dataFormatForInputFormat(InputFormat inputFormat) {
        for (DataFormat dataFormat : validDataFormats) {
            if (dataFormat.getMode() == inputFormat.getMode() && dataFormat.getUnit() == inputFormat.getUnit()) {
                if ((dataFormat.getDataSetSize() * dataFormat.getDataSetCount()) != inputFormat.getNumberOfBytes()) {
                    LDSDKLogger.e(String.format("%s in mode %s (%s): expected data length is %d data sets of %d bytes input format received from device says %d number of bytes",
                            this.getServiceName(), dataFormat.getModeName(), dataFormat.getUnit().name(), dataFormat.getDataSetCount(), dataFormat.getDataSetSize(), inputFormat.getNumberOfBytes()));
                    return null;
                }
                return dataFormat;
            }
        }
        return null;
    }

    public String getServiceName() {
        return "Undefined";
    }

    /**
     * The default input format that will be uploaded to the device for this service upon discovery of the service.
     * Only the known service types (Motor, TiltSensor, etc.) has a default input format.
     * Should be overwritten in sub-classes for known (non generic) IO types!
     *
     * @return the default {@link dk.lego.devicesdk.input_output.InputFormat} of the service
     */
    public InputFormat getDefaultInputFormat() {
        return null;
    }

    /**
     * The current input format for this service.
     */
    public InputFormat getInputFormat() {
        return inputFormat;
    }

    /**
     * Send an updated input format for this service to the device.
     * If successful this will trigger an invocation of the delegate callback method {@link dk.lego.devicesdk.services.ServiceCallbackListener#didUpdateInputFormat}.
     *
     * @param newFormat The input format to send to the device.
     */
    public void updateInputFormat(InputFormat newFormat) {
        io.writeInputFormat(newFormat, connectInfo.getConnectId());
    }

    /**
     * Convenience method that will return the mode of the current inputFormat.
     * If no inputFormat is set, it will return the mode of the defaultInputFormat.
     * If no inputFormat or defaultInputFormat is set, it will return mode 0.
     */
    public int getInputFormatMode() {
        if (getInputFormat() != null) {
            return getInputFormat().getMode();
        } else if (getDefaultInputFormat() != null) {
            return getDefaultInputFormat().getMode();
        }
        LDSDKLogger.d("No input format set, returning mode 0");
        return 0;
    }

    /**
     * Sends and updates input format with newMode for this service to the device.
     * If no current inputFormat is set, the message will be based on the defaultInputFormat of this service.
     * If no defaultInputFormat is set, the call to this method is ignored.
     * @param newMode The new service mode
     */
    public void updateCurrentInputFormatWithNewMode(int newMode) {
        if (getInputFormat() != null) {
            updateInputFormat(getInputFormat().inputFormatBySettingMode(newMode));
        } else if (getDefaultInputFormat() != null) {
            updateInputFormat(getDefaultInputFormat().inputFormatBySettingMode(newMode));
        } else {
            LDSDKLogger.e("Tried to update input format with new mode, but no current inputFormat or defaultInputFormat is set");
        }
    }

    /** @name Configure and read valid data formats */
    /**
     * The data formats that this service may use to parse received data.
     * <p/>
     * When a new value for a service is received from the device, the SDK
     * will look for a DataFormat among the {@link LegoService#validDataFormats} that matches the
     * {@link dk.lego.devicesdk.input_output.InputFormat#unit} and {@link dk.lego.devicesdk.input_output.InputFormat#unit}.
     * If a match is found the DataFormat is used to parse the received data correctly (as a float, integer).
     * <p/>
     * The known service types such as {@link dk.lego.devicesdk.services.TiltSensor} and {@link dk.lego.devicesdk.services.MotionSensor} etc. comes with a set
     * of predefined {@link dk.lego.devicesdk.services.LegoService#validDataFormats} and you do not need to add valid data formats yourself.
     * However, if you wish to use the LegoService with a service type unknown to the SDK
     * you must add the valid data formats using the {@link dk.lego.devicesdk.services.LegoService#addValidDataFormat} method.
     */
    public HashSet<DataFormat> getValidDataFormats() {
        return validDataFormats;
    }

    /**
     * Add a new valid data format
     *
     * @param dataFormat The data format to add (see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats})
     */
    public void addValidDataFormat(DataFormat dataFormat) {
        assert dataFormat != null : "DateFormat cannot be null";

        if (this.validDataFormats == null) {
            validDataFormats = new HashSet<>();
            validDataFormats.add(dataFormat);
        } else {
            validDataFormats.add(dataFormat);
        }
    }

    /**
     * Remove a valid data format
     *
     * @param dataFormat The data format to remove (see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats})
     */
    public void removeValidDataFormat(DataFormat dataFormat) {
        assert dataFormat != null : "DateFormat cannot be null";

        if (this.validDataFormats == null) {
            return;
        }
        validDataFormats.remove(dataFormat);
    }

    /**
     * The latest received value from the service as raw data
     */
    public byte[] getValueData() {
        return (valueData == null) ? null : valueData.clone();
    }

    /**
     * The latest received value from the service as a {@link java.lang.Number}.
     * If the <code>inputFormat</code> specifies a data set with more than one element, the first element is returned.
     * If no valid data format is found to parse the data nil is returned, see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats}.
     */
    public Number getNumberFromValueData() {
        return getNumberFromValueData(valueData);
    }

    /**
     * The value representation of data from the service as an {@link java.lang.Number}.
     * If the <code>inputFormat</code> specifies a data set with more than one element, the first element is returned.
     * If no valid data format is found to parse the data nil is returned, see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats}.
     * @param data The data to parse.
     */
    public Number getNumberFromValueData(byte[] data) {
        List<? extends Number> valuesAsNumbers = getNumbersFromValueDataSet(data);
        if (valuesAsNumbers == null) {
            return null;
        }

        if (valuesAsNumbers.size() != 1) {
            LDSDKLogger.w(String.format("Cannot get value from service %s as as a number, the active data format is %s",
                    getServiceName(),
                    dataFormatForInputFormat(inputFormat)));
            return null;
        }

        return valuesAsNumbers.get(0);
    }

    /**
     * The latest received value from the service as an array of {@link java.lang.Number}.
     * This makes sense if the data set count of the <code>inputFormat</code> is higher than 1. For example, if the data set
     * count is two the <code>valueData</code> consists of two numbers.
     * If no valid data format is found to parse the data nil is returned, see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats}.
     */
    public List<Number> getNumbersFromValueDataSet() {
        return getNumbersFromValueDataSet(valueData);
    }

    /**
     * The value data representation from the service as an array of {@link java.lang.Number}.
     * This makes sense if the data set count of the <code>inputFormat</code> is higher than 1. For example, if the data set
     * count is two the <code>valueData</code> consists of two numbers.
     * If no valid data format is found to parse the data nil is returned, see {@link dk.lego.devicesdk.services.LegoService#getValidDataFormats}.
     * @param dataSet The data to parse.
     */
    public List<Number> getNumbersFromValueDataSet(byte[] dataSet) {
        if (dataSet == null) {
            return null;
        }

        DataFormat dataFormat = dataFormatForInputFormat(inputFormat);
        if (dataFormat == null) {
            return null;
        }

        try {
            verifyData(dataSet, dataFormat);
            List<Number> results = new ArrayList<>(dataFormat.getDataSetCount());

            int currentIndex;
            for (int i = 0; i < dataFormat.getDataSetCount(); ++i) {
                currentIndex = i * dataFormat.getDataSetSize();
                byte[] dataSetBytes = Arrays.copyOfRange(dataSet, currentIndex, currentIndex + dataFormat.getDataSetSize());

                if (dataFormat.getUnit() == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW || dataFormat.getUnit() == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE) {
                    results.add(getIntegerFromData(dataSetBytes));
                } else {
                    results.add(getFloatFromData(dataSetBytes));
                }
            }

            return results;
        } catch (LDSDKError ldsdkError) {
            return null;
        }
    }

    /**
     * The value data representation from the service as a floating-point number.
     * If no valid data format is found to parse the data zero is returned, see validDataFormats.
     * @param data The data to parse.
     */
    public float getFloatFromData(byte[] data) {
        if (data.length > 4) {
            return 0f;
        }
        return ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN).getFloat();
    }

    /**
     * The value data representation from the service as an integer.
     * If no valid data format is found to parse the data zero is returned, see validDataFormats.
     * @param data The data to parse.
     */
    public Integer getIntegerFromData(byte[] data) {
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);

        if (data.length == 1) {
            return (int) buffer.get();
        } else if (data.length == 2) {
            return (int) buffer.getShort();
        } else if (data.length == 4) {
            return buffer.getInt();
        } else {
            LDSDKLogger.w(String.format("Cannot parse service value as singned int, data with size %d too big", data.length));
            return 0;
        }
    }

    /**
     * The latest received value from the service as an integer.
     * If no valid data format is found to parse the data zero is returned, see validDataFormats.
     */
    public long getValueAsInteger() {
        return getIntegerFromData(valueData);
    }

    /**
     * The latest received value from the service as an integer.
     * If no valid data format is found to parse the data zero is returned, see validDataFormats.
     */
    public float getValueAsFloat() {
        return getFloatFromData(valueData);
    }

    /**
     * If the notifications is disabled for the service in the inputFormat through {@link dk.lego.devicesdk.input_output.InputFormat#notificationsEnabled}
     * you will have to use this method to request an updated value for the service.
     * The updated value will be delivered through the delegate method {@link dk.lego.devicesdk.services.ServiceCallbackListener#didUpdateValueData}.
     */
    public void sendReadValueRequest() {
        io.readValueForConnectId((byte) connectInfo.getConnectId());
    }

    /**
     * This will send a reset command to the Device for this service. Some services has state,
     * such as a bump-count for a tilt sensor that you may wish to reset.
     */
    public void sendResetStateRequest() {
        io.resetStateForConnectId((byte) connectInfo.getConnectId());
    }

    /**
     * Will send data to the IO backed by this service. Useful to write data to an output unknown to the SDK.
     *
     * @param data The data to write.
     */
    public void writeData(byte[] data) {
        io.writeData(data, (byte) this.inputFormat.getConnectId());
    }

    @Override
    public void didReceiveInputFormat(IO io, InputFormat inputFormat) {
        if (this.inputFormat == null || !this.inputFormat.equals(inputFormat)) {
            handleUpdatedInputFormat(inputFormat);
        }
    }

    //May be overwritten in sub-classes
    public void handleUpdatedInputFormat(InputFormat inputFormat) {
        InputFormat oldFormat = this.inputFormat;
        this.inputFormat = inputFormat;

        callbackHelper.performDidUpdateInputFormatCallback(this, oldFormat, this.inputFormat);

        //After having received a new input format, we want the newest update value according to that format
        sendReadValueRequest();
    }

    @Override
    public void didReceiveValueData(IO io, byte[] valueData) {
        if (inputFormat != null && valueData != null && !Arrays.equals(this.valueData, valueData)) {

            try {
                handleUpdatedValueData(valueData);
            } catch (LDSDKError e) {
                LDSDKLogger.e(String.format("%s", e.getMessage()));
            }
        }
    }

    /**
     * Handle new value data from device (may be overwritten in sub-classes)
     *
     * @param valueData received value data
     * @throws LDSDKError if received data could not be verified
     */
    public void handleUpdatedValueData(byte[] valueData) throws LDSDKError {
        this.verifyData(valueData);
        byte[] oldData = this.valueData;
        this.valueData = Arrays.copyOf(valueData, valueData.length);

        callbackHelper.performDidUpdateValueDataCallback(this, oldData, this.valueData);
    }

    /**
     * Add a callback listener to receive service updates.
     *
     * @param listener The delegate to add
     */
    public void registerCallbackListener(ServiceCallbackListener listener) {
        callbackHelper.registerCallbackListener(listener);
    }

    /**
     * Remove a callback listener
     *
     * @param listener The delegate to remove
     */
    public void unregisterCallbackListener(ServiceCallbackListener listener) {
        callbackHelper.unregisterCallbackListener(listener);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        LegoService that = (LegoService) o;

        if (connectInfo != null ? !connectInfo.equals(that.connectInfo) : that.connectInfo != null)
            return false;

        return true;
    }

    @Override
    public int hashCode() {
        return connectInfo != null ? connectInfo.hashCode() : 0;
    }
}
