package dk.lego.devicesdk.services;

import java.util.Collections;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.utils.HandlerHelper;

class LegoServiceCallbackHelper {
    private final Set<ServiceCallbackListener> callbackListeners = Collections.newSetFromMap(new ConcurrentHashMap<ServiceCallbackListener, Boolean>());

    protected void registerCallbackListener(ServiceCallbackListener callback) {
        callbackListeners.add(callback);
    }

    protected void unregisterCallbackListener(ServiceCallbackListener callback) {
        callbackListeners.remove(callback);
    }

    public Set<ServiceCallbackListener> getCallbackListeners() {
        return callbackListeners;
    }

    protected void performDidUpdateInputFormatCallback(final LegoService service, final InputFormat oldFormat, final InputFormat inputFormat) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    callbackListener.didUpdateInputFormat(service, oldFormat, inputFormat);
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    protected void performDidUpdateValueDataCallback(final LegoService service, final byte[] oldData, final byte[] valueData) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    callbackListener.didUpdateValueData(service, oldData, valueData);
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    private void postRunnableFromSuitedHandler(Runnable runnable) {
        HandlerHelper.getInstance().post(runnable);
    }

    public void performDidUpdateMotionSensorDistanceCallback(final MotionSensor motionSensor, final float oldDistance, final float newDistance) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof MotionSensorCallbackListener) {
                        ((MotionSensorCallbackListener) callbackListener).didUpdateDistance(motionSensor, oldDistance, newDistance);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateMotionSensorCountCallback(final MotionSensor motionSensor, final int count) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof MotionSensorCallbackListener) {
                        ((MotionSensorCallbackListener) callbackListener).didUpdateCount(motionSensor, count);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performRGBLightDefaultColorInitializedCallback(final RGBLight rgbLight) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof RGBLightCallbackListener) {
                        ((RGBLightCallbackListener) callbackListener).didReadInitialColor(rgbLight);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performRGBLightDidUpdateValueCallback(final RGBLight rgbLight, final Integer oldColor, final Integer color) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof RGBLightCallbackListener) {
                        ((RGBLightCallbackListener) callbackListener).didUpdateRGBValue(rgbLight, oldColor, color);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performRGBLightDidUpdateColorIndexCallback(final RGBLight rgbLight, final int oldColorIndex, final int newColorIndex) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof RGBLightCallbackListener) {
                        ((RGBLightCallbackListener) callbackListener).didUpdateRGBColorIndex(rgbLight, oldColorIndex, newColorIndex);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateTiltDirection(final TiltSensor tiltSensor, final TiltSensor.TiltSensorDirection oldDirection, final TiltSensor.TiltSensorDirection direction) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof TiltSensorCallbackListener) {
                        ((TiltSensorCallbackListener) callbackListener).didUpdateDirection(tiltSensor, oldDirection, direction);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateTiltAngle(final TiltSensor tiltSensor, final TiltSensor.TiltSensorAngle oldAngle, final TiltSensor.TiltSensorAngle angle) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof TiltSensorCallbackListener) {
                        ((TiltSensorCallbackListener) callbackListener).didUpdateAngle(tiltSensor, oldAngle, angle);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateTiltCrash(final TiltSensor tiltSensor, final TiltSensor.TiltSensorCrash oldCrash, final TiltSensor.TiltSensorCrash crash) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof TiltSensorCallbackListener) {
                        ((TiltSensorCallbackListener) callbackListener).didUpdateCrash(tiltSensor, oldCrash, crash);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateMilliVolts(final VoltageSensor voltageSensor, final float milliVolts) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof VoltageSensorCallbackListener) {
                        ((VoltageSensorCallbackListener) callbackListener).didUpdateMilliVolts(voltageSensor, milliVolts);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }

    public void performDidUpdateMilliAmps(final CurrentSensor currentSensor, final float milliAmps) {
        Runnable runnable = new Runnable() {
            @Override
            public void run() {
                for (ServiceCallbackListener callbackListener : callbackListeners) {
                    if (callbackListener instanceof CurrentSensorCallbackListener) {
                        ((CurrentSensorCallbackListener) callbackListener).didUpdateMilliAmps(currentSensor, milliAmps);
                    }
                }
            }
        };

        postRunnableFromSuitedHandler(runnable);
    }
}
