package dk.lego.devicesdk.services;


import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.DataFormat;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.InputFormat;

/**
 * This service provides readings from an motion sensor (aka. detect sensor).
 * <p/>
 * Add a instance of a {@link dk.lego.devicesdk.services.MotionSensorCallbackListener} using {@link dk.lego.devicesdk.services.LegoService#registerCallbackListener} to be notified when a service receives an updated value.
 */
public class MotionSensor extends LegoService {
    private static final String SERVICE_MOTION_SENSOR_NAME = "Motion Sensor";

    public static final int MAX_DISTANCE = 10;
    public static final int MIN_DISTANCE = 0;


    /**
     * Supported modes for the motion sensor
     */
    public enum MotionSensorMode {
        /**
         * Detect mode - produces value that reflect the relative distance from the sensor to objects in front of it
         */
        MOTION_SENSOR_MODE_DETECT(0),
        /**
         * Count mode - produces values that reflect how many times the sensor has been activated
         */
        MOTION_SENSOR_MODE_COUNT(1),
        /**
         * Unknown (unsupported) mode
         */
        MOTION_SENSOR_MODE_UNKNOWN(2);

        private int value;

        MotionSensorMode(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static MotionSensorMode fromInteger(int mode) {
            switch (mode) {
                case 0:
                    return MOTION_SENSOR_MODE_DETECT;
                case 1:
                    return MOTION_SENSOR_MODE_COUNT;
                case 2:
                default:
                    return MOTION_SENSOR_MODE_UNKNOWN;
            }
        }
    }

    private MotionSensor(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
        addValidDataFormats();
    }

    public static MotionSensor createService(ConnectInfo connectInfo, IO io) {
        return new MotionSensor(connectInfo, io);
    }

    @Override
    public String getServiceName() {
        return SERVICE_MOTION_SENSOR_NAME;
    }

    @Override
    public InputFormat getDefaultInputFormat() {
        return InputFormat.inputFormat(getConnectInfo().getConnectId(), getConnectInfo().getType(), 0, 1, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);
    }

    private void addValidDataFormats() {
        addValidDataFormat(DataFormat.create("Detect", MotionSensorMode.MOTION_SENSOR_MODE_DETECT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 1));
        addValidDataFormat(DataFormat.create("Detect", MotionSensorMode.MOTION_SENSOR_MODE_DETECT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 1));
        addValidDataFormat(DataFormat.create("Detect", MotionSensorMode.MOTION_SENSOR_MODE_DETECT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 1));
        addValidDataFormat(DataFormat.create("Count", MotionSensorMode.MOTION_SENSOR_MODE_COUNT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 4, 1));
        addValidDataFormat(DataFormat.create("Count", MotionSensorMode.MOTION_SENSOR_MODE_COUNT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 1));
        addValidDataFormat(DataFormat.create("Count", MotionSensorMode.MOTION_SENSOR_MODE_COUNT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 1));
    }

    public float getDistance() {
        if (getMotionSensorMode() != MotionSensorMode.MOTION_SENSOR_MODE_DETECT) {
            return 0;
        }

        Number number = getNumberFromValueData();
        if (number != null) {
            return number.floatValue();
        } else {
            return 0;
        }
    }


    public int getCount() {
        if (getMotionSensorMode() != MotionSensorMode.MOTION_SENSOR_MODE_COUNT) {
            return 0;
        }

        Number number = getNumberFromValueData();
        if (number != null) {
            return number.intValue();
        } else {
            return 0;
        }
    }

    /**
     * @return The current mode of the motion sensor
     */
    public MotionSensorMode getMotionSensorMode() {
        return MotionSensorMode.fromInteger(getInputFormatMode());
    }

    public void setMotionSensorMode(MotionSensorMode motionSensorMode) {
        updateCurrentInputFormatWithNewMode(motionSensorMode.getValue());
    }

    @Override
    public void handleUpdatedValueData(byte[] valueData) throws LDSDKError {
        float oldDistance = getDistance();
        super.handleUpdatedValueData(valueData);

        if (getMotionSensorMode() == MotionSensorMode.MOTION_SENSOR_MODE_DETECT) {
            callbackHelper.performDidUpdateMotionSensorDistanceCallback(this, oldDistance, getDistance());
        } else {
            callbackHelper.performDidUpdateMotionSensorCountCallback(this, getCount());
        }
    }

    @Override
    public boolean equals(Object o) {
        return super.equals(o);
    }

    @Override
    public int hashCode() {
        return super.hashCode();
    }
}
