package dk.lego.devicesdk.services;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.logging.LDSDKLogger;

public class Motor extends LegoService {

    public static final int MOTOR_MIN_SPEED = 1;
    public static final int MOTOR_MAX_SPEED = 100;
    public static final int MOTOR_POWER_DRIFT = 0;
    public static final int MOTOR_POWER_BRAKE = 127;
    // Only send values in the range 35-100.
    // An offset is needed as values below 35 is not enough power to actually make the motor turn.
    public static final int MOTOR_POWER_OFFSET = 35;

    private static final String SERVICE_MOTOR_NAME = "Motor";

    private MotorDirection direction;
    private int mostRecentSendPower;

    protected Motor(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
    }

    public static Motor createService(ConnectInfo connectInfo, IO io) {
        return new Motor(connectInfo, io);
    }

    /**
     * The direction of a motor
     */
    public enum MotorDirection {
        /**
         * Drifting (Floating)
         */
        MOTOR_DIRECTION_DRIFTING(0),
        /**
         * Running left
         */
        MOTOR_DIRECTION_LEFT(1),
        /**
         * Running right
         */
        MOTOR_DIRECTION_RIGHT(2),
        /**
         * Brake
         */
        MOTOR_DIRECTION_BRAKING(3);

        private int value;

        MotorDirection(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static MotorDirection fromInteger(int value) {
            switch (value) {
                case 0:
                    return MOTOR_DIRECTION_DRIFTING;
                case 1:
                    return MOTOR_DIRECTION_LEFT;
                case 2:
                    return MOTOR_DIRECTION_RIGHT;
                case 3:
                    return MOTOR_DIRECTION_BRAKING;
                default:
                    LDSDKLogger.w(String.format("Reached default case when trying to determine MotorDirection for value: %s direction set to %s", value, MOTOR_DIRECTION_DRIFTING.name()));
                    return MOTOR_DIRECTION_DRIFTING;
            }
        }
    }

    @Override
    public String getServiceName() {
        return SERVICE_MOTOR_NAME;
    }

    /**
     * The power the motor is currently running with (0 if braking or drifting).
     */
    public int getPower() {
        if (mostRecentSendPower == MOTOR_POWER_BRAKE || mostRecentSendPower == MOTOR_POWER_DRIFT) {
            return 0;
        }
        return Math.abs(mostRecentSendPower);
    }

    /**
     * The current running direction of the motor
     */
    public MotorDirection getDirection() {
        return direction;
    }

    /**
     * true if the motor is currently braking (not running)
     */
    public boolean isBraking() {
        return mostRecentSendPower == MOTOR_POWER_BRAKE;
    }

    /**
     * true if the motor is currently drifting / floating.
     * When floating the motor axis can be turned without resistance.
     */
    public boolean isDrifting() {
        return mostRecentSendPower == MOTOR_POWER_DRIFT;
    }

    /**
     * Send a command to run the motor at a given power in a given direction.
     * The minimum speed is 0 and the maximum speed is 100.
     *
     * @param direction The direction to run the motor
     * @param power     The power to run the motor with.
     */
    public void run(MotorDirection direction, int power) {
        if (power == MOTOR_POWER_DRIFT) {
            drift();
        } else {
            sendPower(convertUnsignedMotorPowerToSigned(power, direction));
            this.direction = direction;
        }
    }

    /**
     * Send a command to stop (brake) the motor
     */
    public void brake() {
        sendPower(MOTOR_POWER_BRAKE);
        direction = MotorDirection.MOTOR_DIRECTION_BRAKING;
    }

    /**
     * Send a command to stop (drift/float) the motor
     */
    public void drift() {
        sendPower(MOTOR_POWER_DRIFT);
        direction = MotorDirection.MOTOR_DIRECTION_DRIFTING;
    }

    private void sendPower(int power) {
        LDSDKLogger.d(String.format("Setting motor power %d for connectID %d", power, connectInfo.getConnectId()));

        if (power == MOTOR_POWER_BRAKE || power == MOTOR_POWER_DRIFT) {
            // Brake and Float should not be affected by the offset
            io.writeMotorPower(power, connectInfo.getConnectId());
        } else {
            int offset = 0;
            try {
                if (getDevice().getDeviceInfo().getFirmwareRevision().getMajorVersion() >= 1) {
                    // On version 0.x of the firmware, PVM offset is handled in the firmware
                    offset = MOTOR_POWER_OFFSET;
                }
            } catch (NullPointerException e) {}
            io.writeMotorPower(power, offset, connectInfo.getConnectId());
        }

        mostRecentSendPower  = power;

        try {
            handleUpdatedValueData(ByteBuffer.allocate(4).order(ByteOrder.LITTLE_ENDIAN).putInt(mostRecentSendPower).array());
        } catch (LDSDKError e) {
            LDSDKLogger.e(String.format("%s", e.getMessage()));
        }
    }


    private int convertUnsignedMotorPowerToSigned(int power, MotorDirection direction)
    {
        int resultPower = ((power < MOTOR_MAX_SPEED) ? power : MOTOR_MAX_SPEED);

        if (direction == MotorDirection.MOTOR_DIRECTION_LEFT) {
            resultPower = -resultPower;
        }

        return resultPower;
    }

    @Override
    public boolean equals(Object o) {
        return super.equals(o);
    }

    @Override
    public int hashCode() {
        return super.hashCode();
    }
}
