package dk.lego.devicesdk.services;

import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.logging.LDSDKLogger;

public class PiezoTonePlayer extends LegoService {

    public static final int PIEZO_TONE_MAX_FREQUENCY = 1500;
    public static final int PIEZO_TONE_MAX_DURATION = 65536;

    private static final String SERVICE_PIEZO_TONE_PLAYER_NAME = "Piezo";

    /**
     * Tones that can be played using the PiezoTonePlayer
     */
    public enum PiezoTonePlayerNote {
        /**
         * C
         */
        PIEZO_NOTE_C(1),
        /**
         * C#
         */
        PIEZO_NOTE_CIS(2),
        /**
         * D
         */
        PIEZO_NOTE_D(3),
        /**
         * D#
         */
        PIEZO_NOTE_DIS(4),
        /**
         * E
         */
        PIEZO_NOTE_E(5),
        /**
         * F
         */
        PIEZO_NOTE_F(6),
        /**
         * F#
         */
        PIEZO_NOTE_FIS(7),
        /**
         * G
         */
        PIEZO_NOTE_G(8),
        /**
         * G#
         */
        PIEZO_NOTE_GIS(9),
        /**
         * A
         */
        PIEZO_NOTE_A(10),
        /**
         * A#
         */
        PIEZO_NOTE_AIS(11),
        /**
         * B
         */
        PIEZO_NOTE_B(12);

        private int value;

        PiezoTonePlayerNote(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static PiezoTonePlayerNote fromInteger(int note) {
            switch (note) {
                case 0:
                    return PIEZO_NOTE_C;
                case 1:
                    return PIEZO_NOTE_CIS;
                case 2:
                    return PIEZO_NOTE_D;
                case 3:
                    return PIEZO_NOTE_DIS;
                case 4:
                    return PIEZO_NOTE_E;
                case 5:
                    return PIEZO_NOTE_F;
                case 6:
                    return PIEZO_NOTE_F;
                case 7:
                    return PIEZO_NOTE_FIS;
                case 8:
                    return PIEZO_NOTE_G;
                case 9:
                    return PIEZO_NOTE_GIS;
                case 10:
                    return PIEZO_NOTE_A;
                case 11:
                    return PIEZO_NOTE_AIS;
                case 12:
                    return PIEZO_NOTE_B;
                default:
                    LDSDKLogger.w(String.format("Reached default case when trying to determine Piezo Node for value: %s note was set to %s", note, PIEZO_NOTE_C.name()));
                    return PIEZO_NOTE_C;
            }
        }
    }

    protected PiezoTonePlayer(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
    }

    public static PiezoTonePlayer createService(ConnectInfo connectInfo, IO io) {
        return new PiezoTonePlayer(connectInfo, io);
    }

    @Override
    public String getServiceName() {
        return SERVICE_PIEZO_TONE_PLAYER_NAME;
    }

    /**
     * Play a frequency for the given duration in ms
     *
     * @param frequency The frequency to play (max allowed frequency is 1500)
     * @param duration  The duration to play (max supported is 65536 milliseconds).
     */
    public void playFrequency(int frequency, int duration) {
        if (frequency > PIEZO_TONE_MAX_FREQUENCY) {
            LDSDKLogger.w(String.format("Cannot play frequenzy %d, max supported frequency is %d", frequency, PIEZO_TONE_MAX_FREQUENCY));
            frequency = PIEZO_TONE_MAX_FREQUENCY;
        }
        if (duration > PIEZO_TONE_MAX_DURATION) {
            LDSDKLogger.w(String.format("Cannot play piezo tone with duration %d ms, max supported frequency is %d ms", duration, PIEZO_TONE_MAX_DURATION));
            duration = PIEZO_TONE_MAX_DURATION;
        }

        getIo().writePiezoToneFrequency(frequency, duration, getConnectInfo().getConnectId());
    }

    /**
     * Play a note.
     * The highest supported note is F# in 6th octave
     *
     * @param note     The note to play
     * @param octave   The octave in which to play the note
     * @param duration The duration to play (max supported is 65536 milli seconds).
     */
    public void playNote(PiezoTonePlayerNote note, int octave, int duration) {
        if (octave > 6) {
            LDSDKLogger.w(String.format("Highest supported note is F# in 6th octave - invalid octave: %d", octave));
        }
        if (octave == 6 && note.getValue() > PiezoTonePlayerNote.PIEZO_NOTE_FIS.getValue()) {
            LDSDKLogger.w("Cannot play note. Highest supported note is F# in 6th octave");
        }

        /**
         * The basic formula for the frequencies of the notes of the equal tempered scale is given by
         * fn = f0 * (a)n
         * where
         * f0 = the frequency of one fixed note which must be defined. A common choice is setting the A above middle C (A4) at f0 = 440 Hz.
         * n = the number of half steps away from the fixed note you are. If you are at a higher note, n is positive. If you are on a lower note, n is negative.
         * fn = the frequency of the note n half steps away.
         * a = (2)1/12 = the twelfth root of 2 = the number which when multiplied by itself 12 times equals 2 = 1.059463094359...
         */

        double base = 440.0;
        int octavesAboveMiddle = octave - 4;
        float halfStepsAwayFromBase = (float) note.getValue() - (float) PiezoTonePlayerNote.PIEZO_NOTE_A.getValue() + (octavesAboveMiddle * 12);
        double frequency = base * Math.pow(Math.pow(2.0, 1.0 / 12), halfStepsAwayFromBase);

        playFrequency((int) Math.round(frequency), duration);
    }

    /**
     * Stop playing any currently playing tone.
     */
    public void stopPlaying() {
        getIo().writePiezoToneStop(getConnectInfo().getConnectId());
    }

}
