package dk.lego.devicesdk.services;

import android.graphics.Color;

import java.nio.ByteBuffer;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.DataFormat;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.utils.ByteUtils;

/**
 * This service allows for setting the colour of the RGB light on the device
 */
public class RGBLight extends LegoService {

    /*
     * Mode of the RGB light
     */
    public enum RGBLightMode {
        /*
         * Discrete mode allows selecting a color index from a set of predefined colors
         */
        RGB_LIGHT_MODE_DISCRETE(0),
        /*
         * Absolute mode allows selecting any color by specifying its RGB component values
         */
        RGB_LIGHT_MODE_ABSOLUTE(1),
        /*
         * Unknown
         */
        RGB_LIGHT_MODE_UNKNOWN(2);

        private int value;

        RGBLightMode(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static RGBLightMode fromInteger(int value) {
            switch (value) {
                case 0:
                    return RGB_LIGHT_MODE_DISCRETE;
                case 1:
                    return RGB_LIGHT_MODE_ABSOLUTE;
                case 2:
                default:
                    return RGB_LIGHT_MODE_UNKNOWN;
            }
        }
    }

    public static final String SERVICE_RGB_LIGHT_NAME = "RGB Light";
    /**
     * The color of the RGB light on the device.
     */
    private Integer color = null;

    /**
     * The index of the currently set color  when in the discrete mode
     */
    private int colorIndex = 0;

    protected RGBLight(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
        addValidDataFormats();
    }

    public static RGBLight createService(ConnectInfo connectInfo, IO io) {
        return new RGBLight(connectInfo, io);
    }

    @Override
    public String getServiceName() {
        return SERVICE_RGB_LIGHT_NAME;
    }

    @Override
    public InputFormat getDefaultInputFormat() {
        return InputFormat.inputFormat(getConnectInfo().getConnectId(), getConnectInfo().getType(), RGBLightMode.RGB_LIGHT_MODE_DISCRETE.getValue(), 1, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, true);
    }

    /*
     * Sets the color of the RGB light on the device (absolute mode)
     * @param rgbColor The new color of the RGB light
     */
    public void setColor(int rgbColor) {
        if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE) {
            this.color = rgbColor;

            int red = Color.red(rgbColor);
            int green = Color.green(rgbColor);
            int blue = Color.blue(rgbColor);

            getIo().writeColor(red, green, blue, getConnectInfo().getConnectId());
        } else {
            LDSDKLogger.w("Ignoring attempt to set RGB color. It is only supported when RGB is in mode Absolute.");
        }
    }

    /*
     * The current color of the RGB light on the device
     */
    public Integer getColor() {
        return color;
    }

    /*
     * The default color of the RGB light (absolute mode)
     */
    public Integer getDefaultColor() {
        // We have no reliable way of reading the default color of the Hub, so it is hardcoded here
        return Color.argb(0xFF, 0x00, 0x00, 0xFF);
    }

    /*
     * The index of the currently selected color (discrete mode)
     */
    public int getColorIndex() {
        return colorIndex;
    }

    /*
     * Sets the index of the currently selected color (discrete mode)
     */
    public void setColorIndex(int index) {
        if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_DISCRETE) {
            this.colorIndex = index;
            getIo().writeColorIndex(index, getConnectInfo().getConnectId());
        } else {
            LDSDKLogger.w("Ignoring attempt to set RGB color index. It is only supported when RGB is in mode Discrete.");
        }
    }

    /*
     * The default color index of the RGB, when in the discrete mode
     */
    public int getDefaultColorIndex() {
        // We have no reliable way of reading the default color of the Hub, so it is hardcoded here
        return 3;
    }

    /**
     *  The current mode of the RGB
     */
    public RGBLightMode getRGBMode() {
        return RGBLightMode.fromInteger(getInputFormatMode());
    }
    /**
     *  Sets the current mode of the RGB
     *  @param rgbMode The new RGB mode
     */
    public void setRGBMode(RGBLightMode rgbMode) {
        updateCurrentInputFormatWithNewMode(rgbMode.getValue());
    }

    /**
     * Switch off the RGB light on the device
     */
    public void switchOff() {
        if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE) {
            setColor(Color.rgb(0, 0, 0));
        } else if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_DISCRETE) {
            setColorIndex(0);
        } else {
            LDSDKLogger.w(String.format("Cannot switch off RGB - unknown mode: %i", getInputFormatMode()));
        }
    }

    /**
     * Switch to the default Color (i.e. the same color as the device has right after a successful connection has been established)
     */
    public void switchToDefaultColor() {
        if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE) {
            setColor(getDefaultColor());
        } else if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_DISCRETE) {
            setColorIndex(getDefaultColorIndex());
        } else {
            LDSDKLogger.w(String.format("Cannot switch to default color - unknown mode: %i", getInputFormatMode()));
        }
    }

    @Override
    public void handleUpdatedValueData(byte[] valueData) throws LDSDKError
    {
        if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE) {
            Integer oldColor = color;
            color = colorFromData(valueData);

            super.handleUpdatedValueData(valueData);
            callbackHelper.performRGBLightDidUpdateValueCallback(this, oldColor, this.color);
        } else if (getRGBMode() == RGBLightMode.RGB_LIGHT_MODE_DISCRETE) {
            int oldColorIndex = colorIndex;
            colorIndex = getIntegerFromData(valueData);

            super.handleUpdatedValueData(valueData);
            callbackHelper.performRGBLightDidUpdateColorIndexCallback(this, oldColorIndex, this.colorIndex);
        } else {
            LDSDKLogger.e(String.format("Cannot handle response for RGB in unknown mode %i", getInputFormatMode()));
        }
    }

    private Integer colorFromData(byte[] data) {
        if (data.length == 3) {
            ByteBuffer dataBuffer = ByteBuffer.wrap(data);
            return Color.rgb(ByteUtils.getUnsignedByte(dataBuffer), ByteUtils.getUnsignedByte(dataBuffer), ByteUtils.getUnsignedByte(dataBuffer));
        }

        LDSDKLogger.e(String.format("Cannot create color from data %s", ByteUtils.toHexString(data)));
        return null;
    }

    private void addValidDataFormats() {
        addValidDataFormat(DataFormat.create("Discrete", RGBLightMode.RGB_LIGHT_MODE_DISCRETE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 1));
        addValidDataFormat(DataFormat.create("Discrete", RGBLightMode.RGB_LIGHT_MODE_DISCRETE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 1));
        addValidDataFormat(DataFormat.create("Discrete", RGBLightMode.RGB_LIGHT_MODE_DISCRETE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 1));

        addValidDataFormat(DataFormat.create("Absolute", RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 3));
        addValidDataFormat(DataFormat.create("Absolute", RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 3));
        addValidDataFormat(DataFormat.create("Absolute", RGBLightMode.RGB_LIGHT_MODE_ABSOLUTE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 3));
    }

    @Override
    public boolean equals(Object o) {
        return super.equals(o);
    }

    @Override
    public int hashCode() {
        return super.hashCode();
    }
}
