package dk.lego.devicesdk.services;

import java.util.List;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.ConnectInfo;
import dk.lego.devicesdk.input_output.DataFormat;
import dk.lego.devicesdk.input_output.IO;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.logging.LDSDKLogger;

/**
 * This service provides readings from a tilt sensor.
 * <p/>
 * Add an instance of the {@link dk.lego.devicesdk.services.TiltSensorCallbackListener} using registerCallbackListener in order to be notified when the sensor receives an updated value.
 */
public class TiltSensor extends LegoService {

    private static final String SERVICE_TILT_SENSOR_NAME = "Tilt Sensor";

    /**
     * Direction of tilt sensor
     */
    public enum TiltSensorDirection {
        /**
         * Neutral
         */
        TILT_SENSOR_DIRECTION_NEUTRAL(0),
        /**
         * Backward
         */
        TILT_SENSOR_DIRECTION_BACKWARD(3),
        /**
         * Right
         */
        TILT_SENSOR_DIRECTION_RIGHT(5),
        /**
         * Left
         */
        TILT_SENSOR_DIRECTION_LEFT(7),
        /**
         * Forward
         */
        TILT_SENSOR_DIRECTION_FORWARD(9),
        /**
         * Unknown
         */
        TILT_SENSOR_DIRECTION_UNKNOWN(10);

        private int value;

        TiltSensorDirection(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static TiltSensorDirection fromInteger(int value) {
            switch (value) {
                case 0:
                    return TILT_SENSOR_DIRECTION_NEUTRAL;
                case 3:
                    return TILT_SENSOR_DIRECTION_BACKWARD;
                case 5:
                    return TILT_SENSOR_DIRECTION_RIGHT;
                case 7:
                    return TILT_SENSOR_DIRECTION_LEFT;
                case 9:
                    return TILT_SENSOR_DIRECTION_FORWARD;
                case 10:
                default:
                    LDSDKLogger.w(String.format("Reached default case when trying to determine TitlSensorDirection for value: %s direction set to %s", value, TILT_SENSOR_DIRECTION_UNKNOWN.name()));
                    return TILT_SENSOR_DIRECTION_UNKNOWN;
            }
        }


        @Override
        public String toString() {
            switch (TiltSensorDirection.fromInteger(value)) {
                case TILT_SENSOR_DIRECTION_NEUTRAL:
                    return "Neutral";
                case TILT_SENSOR_DIRECTION_BACKWARD:
                    return "Backward";
                case TILT_SENSOR_DIRECTION_FORWARD:
                    return "Forward";
                case TILT_SENSOR_DIRECTION_LEFT:
                    return "Left";
                case TILT_SENSOR_DIRECTION_RIGHT:
                    return "Right";
                case TILT_SENSOR_DIRECTION_UNKNOWN:
                default:
                    return "Unknown";
            }
        }
    }

    /**
     * Tilt sensor mode
     */
    public enum TiltSensorMode {
        /**
         * Angle
         */
        TILT_SENSOR_MODE_ANGLE(0),
        /**
         * Tilt
         */
        TILT_SENSOR_MODE_TILT(1),
        /**
         * Crash
         */
        TILT_SENSOR_MODE_CRASH(2),
        /**
         * Unknown
         */
        TILT_SENSOR_MODE_UNKNOWN(4);

        private int value;

        TiltSensorMode(int value) {
            this.value = value;
        }

        public int getValue() {
            return value;
        }

        public static TiltSensorMode fromInteger(int value) {
            switch (value) {
                case 0:
                    return TILT_SENSOR_MODE_ANGLE;
                case 1:
                    return TILT_SENSOR_MODE_TILT;
                case 2:
                    return TILT_SENSOR_MODE_CRASH;
                case 3:
                default:
                    LDSDKLogger.w(String.format("Reached default case when trying to determine TiltSensorMode for value: %s mode set to %s", value, TILT_SENSOR_MODE_UNKNOWN.name()));
                    return TILT_SENSOR_MODE_UNKNOWN;
            }
        }
    }

    public static final class TiltSensorAngle {
        public float x = 0.0f;
        public float y = 0.0f;

        @Override
        public String toString() {
            return String.format("[%s, %s]", x, y);
        }
    }

    public static final class TiltSensorCrash {
        public int x = 0;
        public int y = 0;
        public int z = 0;

        @Override
        public String toString() {
            return String.format("[%s, %s, %s]", x, y, z);
        }
    }

    public static TiltSensorAngle tiltSensorAngleMake(float x, float y) {
        TiltSensorAngle angle = new TiltSensorAngle();
        angle.x = x;
        angle.y = y;
        return angle;
    }

    public static boolean tiltSensorAngleEqualToAngle(TiltSensorAngle angle1, TiltSensorAngle angle2) {
        float acceptableDiff = 0.01f;
        return (angle1.x - angle2.x < acceptableDiff && angle1.y - angle2.y < acceptableDiff);

    }

    public static TiltSensorCrash tiltSensorCrashMake(int x, int y, int z) {
        TiltSensorCrash crash = new TiltSensorCrash();
        crash.x = x;
        crash.y = y;
        crash.z = z;
        return crash;
    }

    public static  boolean tiltSensorCrashEqualToCrash(TiltSensorCrash crash1, TiltSensorCrash crash2) {
        return (crash1.x == crash2.x && crash1.y == crash2.y && crash1.z == crash2.z);
    }

    TiltSensorAngle TiltSensorAngleZero = new TiltSensorAngle();
    TiltSensorCrash TiltSensorCrashZero = new TiltSensorCrash();

    /**
     * The most recent direction reading from the sensor.
     * <p/>
     * If no direction reading has been received, of if the sensor is not in mode {@link dk.lego.devicesdk.services.TiltSensor.TiltSensorMode#TILT_SENSOR_MODE_TILT}
     * the value of this property will be {@link dk.lego.devicesdk.services.TiltSensor.TiltSensorDirection#TILT_SENSOR_DIRECTION_UNKNOWN}.
     */
    TiltSensorDirection direction;

    /**
     * The most recent angle reading from the sensor. The angle represents the angle the sensor is tilted in the x, y and z-direction.
     * <p/>
     * If no angle reading has been received, of if the sensor is not in mode {@link dk.lego.devicesdk.services.TiltSensor.TiltSensorMode#TILT_SENSOR_MODE_ANGLE}
     * the value of this property will be {@link #TiltSensorAngleZero}
     */
    TiltSensorAngle angle;

    /**
     * The most recent crash reading from the sensor. The value represents the number of times the sensor has been 'bumped' in
     * the x, y, and z-direction. The value can be reset by sending the {@link dk.lego.devicesdk.services.LegoService#sendResetStateRequest}.
     * <p/>
     * If no angle reading has been received, of if the sensor is not in mode {@link dk.lego.devicesdk.services.TiltSensor.TiltSensorMode#TILT_SENSOR_MODE_CRASH}
     * the value of this property will be {@link #TiltSensorCrashZero}
     */
    TiltSensorCrash crash;

    /**
     * The current mode of the tilt sensor
     */
    TiltSensorMode tiltSensorMode;

    protected TiltSensor(ConnectInfo connectInfo, IO io) {
        super(connectInfo, io);
        tiltSensorMode = TiltSensorMode.fromInteger(getDefaultInputFormat().getMode());
        addValidDataFormats();
    }

    public static TiltSensor createService(ConnectInfo connectInfo, IO io) {
        return new TiltSensor(connectInfo, io);
    }

    public TiltSensorDirection getDirection() {
        if (getInputFormat().getMode() != TiltSensorMode.TILT_SENSOR_MODE_TILT.getValue()) {
            return TiltSensorDirection.TILT_SENSOR_DIRECTION_UNKNOWN;
        }

        int directionInt = getNumberFromValueData().intValue();
        if (directionInt < 10) {
            return TiltSensorDirection.fromInteger(directionInt);
        } else {
            return TiltSensorDirection.TILT_SENSOR_DIRECTION_UNKNOWN;
        }
    }

    public TiltSensorAngle getAngle() {
        if (getInputFormat().getMode() != TiltSensorMode.TILT_SENSOR_MODE_ANGLE.getValue()) {
            return TiltSensorAngleZero;
        }

        List<Number> dataSetNumbers = getNumbersFromValueDataSet();
        if (dataSetNumbers.size() == 2) {
            return tiltSensorAngleMake(dataSetNumbers.get(0).floatValue(), dataSetNumbers.get(1).floatValue());
        }
        return TiltSensorAngleZero;
    }

    public TiltSensorCrash getCrash() {
        if (getInputFormat().getMode() != TiltSensorMode.TILT_SENSOR_MODE_CRASH.getValue()) {
            return TiltSensorCrashZero;
        }

        List<Number> dataSetNumbers = getNumbersFromValueDataSet();
        if (dataSetNumbers.size() == 3) {
            return tiltSensorCrashMake(dataSetNumbers.get(0).byteValue() & 0xff, dataSetNumbers.get(1).byteValue() & 0xff, dataSetNumbers.get(2).byteValue() & 0xff);
        }
        return TiltSensorCrashZero;
    }


    public TiltSensorMode getTiltSensorMode() {
        return TiltSensorMode.fromInteger(getInputFormatMode());
    }

    public void setTiltSensorMode(TiltSensorMode tiltSensorMode) {
        updateCurrentInputFormatWithNewMode(tiltSensorMode.getValue());
    }

    private void addValidDataFormats() {
        addValidDataFormat(DataFormat.create("Angle", TiltSensorMode.TILT_SENSOR_MODE_ANGLE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 2));
        addValidDataFormat(DataFormat.create("Angle", TiltSensorMode.TILT_SENSOR_MODE_ANGLE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 2));
        addValidDataFormat(DataFormat.create("Angle", TiltSensorMode.TILT_SENSOR_MODE_ANGLE.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 2));

        addValidDataFormat(DataFormat.create("Tilt", TiltSensorMode.TILT_SENSOR_MODE_TILT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 1));
        addValidDataFormat(DataFormat.create("Tilt", TiltSensorMode.TILT_SENSOR_MODE_TILT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 1));
        addValidDataFormat(DataFormat.create("Tilt", TiltSensorMode.TILT_SENSOR_MODE_TILT.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 1));

        addValidDataFormat(DataFormat.create("Crash", TiltSensorMode.TILT_SENSOR_MODE_CRASH.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW, 1, 3));
        addValidDataFormat(DataFormat.create("Crash", TiltSensorMode.TILT_SENSOR_MODE_CRASH.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE, 1, 3));
        addValidDataFormat(DataFormat.create("Crash", TiltSensorMode.TILT_SENSOR_MODE_CRASH.getValue(), InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, 4, 3));
    }

    @Override
    public String getServiceName() {
        return SERVICE_TILT_SENSOR_NAME;
    }

    @Override
    public InputFormat getDefaultInputFormat() {
        return InputFormat.inputFormat(getConnectInfo().getConnectId(), getConnectInfo().getType(), TiltSensorMode.TILT_SENSOR_MODE_TILT.getValue(), 1, InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI, true);
    }

    @Override
    public void addValidDataFormat(DataFormat dataFormat) {
        super.addValidDataFormat(dataFormat);
    }

    @Override
    public void handleUpdatedValueData(byte[] valueData) throws LDSDKError {
        // Remember the old values before calling 'handleUpdatedValueData'
        TiltSensorDirection oldDirection = this.direction;
        TiltSensorAngle oldAngle = this.angle;
        TiltSensorCrash oldCrash = this.crash;

        super.handleUpdatedValueData(valueData);

        if (tiltSensorMode == TiltSensorMode.TILT_SENSOR_MODE_TILT) {
            callbackHelper.performDidUpdateTiltDirection(this, oldDirection, getDirection());
        } else if (tiltSensorMode == TiltSensorMode.TILT_SENSOR_MODE_ANGLE) {
            callbackHelper.performDidUpdateTiltAngle(this, oldAngle, getAngle());
        } else if (tiltSensorMode == TiltSensorMode.TILT_SENSOR_MODE_CRASH) {
            callbackHelper.performDidUpdateTiltCrash(this, oldCrash, getCrash());
        }
    }

    @Override
    public void handleUpdatedInputFormat(InputFormat inputFormat) {
        tiltSensorMode = TiltSensorMode.fromInteger(inputFormat.getMode());
        super.handleUpdatedInputFormat(inputFormat);
    }

    @Override
    public boolean equals(Object o) {
        return super.equals(o);
    }

    @Override
    public int hashCode() {
        return super.hashCode();
    }
}
