package dk.lego.devicesdk.unity;

import android.content.Context;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.DeviceManagerCallbackListener;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.device.LegoDeviceManager;
import dk.lego.devicesdk.device.LegoDeviceManagerImpl;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.services.LegoService;

public class LegoDeviceManagerWrapperImpl implements LegoDeviceManagerWrapper, DeviceManagerCallbackListener {

    private static LegoDeviceManagerWrapperImpl instance = null;

    private LegoDeviceManager deviceManager;
    private LegoDeviceWrapperImpl deviceWrapper;
    private WrapperSerialization wrapperSerialization;
    private UnityInvoker unityInvoker;

    private Context context;

    public static LegoDeviceManagerWrapperImpl getInstance() {
        if (instance == null) {
            instance = new LegoDeviceManagerWrapperImpl();
        }
        return instance;
    }

    private LegoDeviceManagerWrapperImpl() {
        this(LegoDeviceManagerImpl.getInstance(), LegoDeviceWrapperImpl.getInstance(), new WrapperSerialization(), new UnityInvoker());
    }

    LegoDeviceManagerWrapperImpl(LegoDeviceManager deviceManager, LegoDeviceWrapperImpl deviceWrapper, WrapperSerialization wrapperSerialization, UnityInvoker unityInvoker) {
        this.deviceManager = deviceManager;
        this.deviceManager.registerCallbackListener(this);

        this.deviceWrapper = deviceWrapper;
        this.wrapperSerialization = wrapperSerialization;
        this.unityInvoker = unityInvoker;
    }

    @Override
    public Context getContext() {
        return context;
    }

    @Override
    public void setContext(Context context) {
        this.context = context;
    }

    @Override
    protected void finalize() throws Throwable {
        deviceManager.unregisterCallbackListener(this);
        super.finalize();
    }

    @Override
    public void scan() {
        try {
            deviceManager.scan(context);
        } catch (Exception e) {
            LDSDKLogger.e(String.format("Scanning for devices could not be started: %s", e.getMessage()));
        }
    }

    @Override
    public void stopScanning() {
        try {
            deviceManager.stopScanning();
        } catch (Exception e) {
            LDSDKLogger.e(String.format("Scanning for devices could not be stopped: %s", e.getMessage()));
        }
    }

    @Override
    public void connectToDevice(String legoDeviceID) {
        LegoDevice legoDevice = findDevice(legoDeviceID);
        if (legoDevice != null) {
            deviceManager.connectToDevice(context, legoDevice);
        }
    }

    @Override
    public void disconnectDevice(String legoDeviceID) {
        LegoDevice legoDevice = findDevice(legoDeviceID);
        if (legoDevice != null) {
            deviceManager.cancelDeviceConnection(legoDevice);
        }
    }

    @Override
    public String getAllDevices() {
        return wrapperSerialization.mapToJSONString(wrapperSerialization.serializeDevices(deviceManager.allDevices()));
    }

    public LegoDevice findDevice(String legoDeviceID) {
        for (LegoDevice device : deviceManager.allDevices()) {
            if (legoDeviceID.equals(device.getDeviceId())) {
                return device;
            }
        }

        LDSDKLogger.w(String.format("Device not found: %s", legoDeviceID));
        return null;
    }

    public LegoService findService(String legoDeviceID, String connectID) throws LDSDKError {
        LegoDevice device = findDevice(legoDeviceID);
        if (device == null) {
            return null;
        }

        int serviceConnectID;
        try {
            serviceConnectID = Integer.parseInt(connectID);
        } catch (NumberFormatException e) {
            throw LDSDKError.deviceError(LDSDKError.ErrorCode.INTERNAL_ERROR, String.format("Invalid service connect ID: %s", connectID));
        }

        for (LegoService service : device.getServices()) {
            if (serviceConnectID == service.getConnectInfo().getConnectId()) {
                return service;
            }
        }

        LDSDKLogger.w(String.format("Service not found for device %s: %s", legoDeviceID, connectID));
        return null;
    }

    private void invokeUnityMethod(String methodName, LegoDevice device) {
        unityInvoker.invokeMethod(methodName, wrapperSerialization.serializeDevice(device));
    }

    @Override
    public void onDeviceAppeared(LegoDevice device) {
        device.registerCallbackListener(deviceWrapper);
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDeviceDidAppear, device);
    }

    @Override
    public void onDeviceDisappeared(LegoDevice device) {
        device.unregisterCallbackListener(deviceWrapper);
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDeviceDidDisappear, device);
    }

    @Override
    public void onWillStartConnectingToDevice(LegoDevice device) {
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerWillStartConnectingToDevice, device);
    }

    @Override
    public void onDidFailToConnectToDevice(LegoDevice device, boolean autoReconnect, LDSDKError error) {
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDidFailToConnectToDevice, device);
    }

    @Override
    public void onDidDisconnectFromDevice(LegoDevice device, boolean autoReconnect, LDSDKError error) {
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDidDisconnectFromDevice, device);
    }

    @Override
    public void onDidStartInterrogatingDevice(LegoDevice device) {
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDidStartInterrogatingDevice, device);
    }

    @Override
    public void onDidFinishInterrogatingDevice(LegoDevice device) {
        invokeUnityMethod(UnitySDKCallbacks.LEDeviceManagerDidFinishInterrogatingDevice, device);
    }

}
