package dk.lego.devicesdk.unity;

import android.util.Log;

import org.json.simple.parser.ParseException;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.services.CurrentSensor;
import dk.lego.devicesdk.services.CurrentSensorCallbackListener;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.MotionSensor;
import dk.lego.devicesdk.services.MotionSensorCallbackListener;
import dk.lego.devicesdk.services.Motor;
import dk.lego.devicesdk.services.PiezoTonePlayer;
import dk.lego.devicesdk.services.RGBLight;
import dk.lego.devicesdk.services.RGBLightCallbackListener;
import dk.lego.devicesdk.services.ServiceCallbackListener;
import dk.lego.devicesdk.services.TiltSensor;
import dk.lego.devicesdk.services.TiltSensorCallbackListener;
import dk.lego.devicesdk.services.VoltageSensor;
import dk.lego.devicesdk.services.VoltageSensorCallbackListener;

public class LegoServiceWrapperImpl implements ServiceCallbackListener, CurrentSensorCallbackListener, MotionSensorCallbackListener, RGBLightCallbackListener, TiltSensorCallbackListener, VoltageSensorCallbackListener, LegoServiceWrapper {

    private static LegoServiceWrapperImpl instance = null;

    private LegoDeviceManagerWrapperImpl deviceManagerWrapper;
    private WrapperSerialization wrapperSerialization;
    private UnityInvoker unityInvoker;

    public static LegoServiceWrapperImpl getInstance() {
        if (instance == null) {
            instance = new LegoServiceWrapperImpl();
        }
        return instance;
    }

    private LegoServiceWrapperImpl() {
        this(LegoDeviceManagerWrapperImpl.getInstance(), new WrapperSerialization(), new UnityInvoker());
    }

    LegoServiceWrapperImpl(LegoDeviceManagerWrapperImpl deviceManagerWrapper, WrapperSerialization wrapperSerialization, UnityInvoker unityInvoker) {
        this.deviceManagerWrapper = deviceManagerWrapper;
        this.wrapperSerialization = wrapperSerialization;
        this.unityInvoker = unityInvoker;
    }

    // LEService

    @Override
    public void updateServiceData(String legoDeviceID, String connectID) {
        try {
            LegoService service = deviceManagerWrapper.findService(legoDeviceID, connectID);
            if (service == null) {
                return;
            }
            unityInvoker.invokeMethod(UnitySDKCallbacks.LEServiceUpdateServiceData, wrapperSerialization.serializeServiceWithData(service));
        } catch (LDSDKError ldsdkError) {
            LDSDKLogger.e(String.format("Failed to update service data %s for device %s: %s", connectID, legoDeviceID, ldsdkError.getMessage()));
        }
    }

    @Override
    public void updateInputFormat(String legoDeviceID, String connectID, String mode, String unit, String deltaInterval, boolean notificationEnabled) {
        try {
            LegoService service = deviceManagerWrapper.findService(legoDeviceID, connectID);
            if (service == null) {
                return;
            }

            InputFormat inputFormat = InputFormat.inputFormat(
                    Integer.parseInt(connectID),
                    service.getConnectInfo().getType(),
                    Integer.parseInt(mode),
                    Integer.parseInt(deltaInterval),
                    InputFormat.InputFormatUnit.fromInteger(Integer.parseInt(unit)),
                    notificationEnabled);
            service.updateInputFormat(inputFormat);
        } catch (LDSDKError ldsdkError) {
            LDSDKLogger.e(String.format("Failed to update input format of service %s for device %s: %s", connectID, legoDeviceID, ldsdkError.getMessage()));
        } catch (NumberFormatException e) {
            LDSDKLogger.e(e);
        }
    }

    // LEMotor

    @Override
    public void run(String legoDeviceID, String connectID, String motorDirection, String power) {
        Motor motor = (Motor)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (motor != null) {
            motor.run(Motor.MotorDirection.fromInteger(Integer.parseInt(motorDirection)), Integer.parseInt(power));
        }
    }

    @Override
    public void brake(String legoDeviceID, String connectID) {
        Motor motor = (Motor)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (motor != null) {
            motor.brake();
        }
    }

    @Override
    public void drift(String legoDeviceID, String connectID) {
        Motor motor = (Motor)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (motor != null) {
            motor.drift();
        }
    }

    // LERGBLight

    @Override
    public void setRGBLightMode(String legoDeviceID, String connectID, String rgbLightMode) {
        RGBLight light = (RGBLight)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (light != null) {
            light.setRGBMode(RGBLight.RGBLightMode.fromInteger(Integer.parseInt(rgbLightMode)));
        }
    }

    @Override
    public void switchOff(String legoDeviceID, String connectID) {
        RGBLight light = (RGBLight)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (light != null) {
            light.switchOff();
        }
    }

    @Override
    public void changeColor(String legoDeviceID, String connectID, String colorJSON) {
        RGBLight light = (RGBLight)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (light == null) {
            return;
        }

        try {
            light.setColor(wrapperSerialization.deserializeColor(wrapperSerialization.parseJSONObject(colorJSON)));
        } catch (ParseException e) {
            Log.e("LEGODeviceSDK", "Invalid color JSON format: " + colorJSON);
        } catch (IllegalArgumentException e) {
            Log.e("LEGODeviceSDK", "Invalid color: " + colorJSON);
        }
    }

    @Override
    public void changeColorToDefault(String legoDeviceID, String connectID) {
        RGBLight light = (RGBLight)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (light != null) {
            light.switchToDefaultColor();
        }
    }

    @Override
    public void changeColorIndex(String legoDeviceID, String connectID, String colorIndex) {
        RGBLight light = (RGBLight)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (light != null) {
            light.setColorIndex(Integer.parseInt(colorIndex));
        }
    }

    // LEMotionSensor

    @Override
    public void setMotionSensorMode(String legoDeviceID, String connectID, String motionSensorMode) {
        MotionSensor motionSensor = (MotionSensor)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (motionSensor != null) {
            motionSensor.setMotionSensorMode(MotionSensor.MotionSensorMode.fromInteger(Integer.parseInt(motionSensorMode)));
        }
    }

    // LETiltSensor

    @Override
    public void setTiltSensorMode(String legoDeviceID, String connectID, String tiltSensorMode) {
        TiltSensor tiltSensor = (TiltSensor)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (tiltSensor != null) {
            tiltSensor.setTiltSensorMode(TiltSensor.TiltSensorMode.fromInteger(Integer.parseInt(tiltSensorMode)));
        }
    }

    // LEPiezoTonePlayer

    @Override
    public void playFrequency(String legoDeviceID, String connectID, String frequency, String milliseconds) {
        PiezoTonePlayer piezoTonePlayer = (PiezoTonePlayer)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (piezoTonePlayer != null) {
            piezoTonePlayer.playFrequency(Integer.parseInt(frequency), Integer.parseInt(milliseconds));
        }
    }

    @Override
    public void playNote(String legoDeviceID, String connectID, String note, String octave, String milliseconds) {
        PiezoTonePlayer piezoTonePlayer = (PiezoTonePlayer)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (piezoTonePlayer != null) {
            piezoTonePlayer.playNote(PiezoTonePlayer.PiezoTonePlayerNote.fromInteger(Integer.parseInt(note)), Integer.parseInt(octave), Integer.parseInt(milliseconds));
        }
    }

    @Override
    public void stopPlaying(String legoDeviceID, String connectID) {
        PiezoTonePlayer piezoTonePlayer = (PiezoTonePlayer)deviceManagerWrapper.findService(legoDeviceID, connectID);
        if (piezoTonePlayer != null) {
            piezoTonePlayer.stopPlaying();
        }
    }

    // Callbacks

    @Override
    public void didUpdateValueData(LegoService service, byte[] oldValue, byte[] newValue) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LEServiceDidUpdateValueData, wrapperSerialization.serializeServiceValueDataChange(service, oldValue, newValue));
    }

    @Override
    public void didUpdateInputFormat(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LEServiceDidUpdateInputFormat, wrapperSerialization.serializeServiceInputFormatChange(service, oldFormat, newFormat));
    }

    @Override
    public void didUpdateMilliAmps(CurrentSensor sensor, float milliAmps) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LECurrentSensorDidUpdateMilliAmp, wrapperSerialization.serializeServiceCurrentChange(sensor, milliAmps));
    }

    @Override
    public void didUpdateDistance(MotionSensor sensor, float oldDistance, float newDistance) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LEMotionSensorDidUpdateDistance, wrapperSerialization.serializeServiceDistanceChange(sensor, oldDistance, newDistance));
    }

    @Override
    public void didUpdateCount(MotionSensor sensor, int count) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LEMotionSensorDidUpdateCount, wrapperSerialization.serializeServiceCountChange(sensor, count));
    }

    @Override
    public void didReadInitialColor(RGBLight rgbLight) {
    }

    @Override
    public void didUpdateRGBValue(RGBLight rgbLight, Integer oldRGBColor, Integer newRGBColor) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LERGBLightDidUpdateColor, wrapperSerialization.serializeServiceColorChange(rgbLight, oldRGBColor, newRGBColor));
    }

    @Override
    public void didUpdateRGBColorIndex(RGBLight rgbLight, int oldColorIndex, int newColorIndex) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LERGBLightDidUpdateColorIndex, wrapperSerialization.serializeServiceColorIndexChange(rgbLight, oldColorIndex, newColorIndex));
    }

    @Override
    public void didUpdateDirection(TiltSensor sensor, TiltSensor.TiltSensorDirection oldDirection, TiltSensor.TiltSensorDirection newDirection) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LETiltSensorDidUpdateDirection, wrapperSerialization.serializeServiceDirectionChange(sensor, oldDirection, newDirection));
    }

    @Override
    public void didUpdateAngle(TiltSensor sensor, TiltSensor.TiltSensorAngle oldAngle, TiltSensor.TiltSensorAngle newAngle) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LETiltSensorDidUpdateAngle, wrapperSerialization.serializeServiceAngleChange(sensor, oldAngle, newAngle));
    }

    @Override
    public void didUpdateCrash(TiltSensor sensor, TiltSensor.TiltSensorCrash oldCrashValue, TiltSensor.TiltSensorCrash newCrashValue) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LETiltSensorDidUpdateCrash, wrapperSerialization.serializeServiceCrashChange(sensor, oldCrashValue, newCrashValue));
    }

    @Override
    public void didUpdateMilliVolts(VoltageSensor sensor, float milliVolts) {
        unityInvoker.invokeMethod(UnitySDKCallbacks.LEVoltageSensorDidUpdateMilliVolts, wrapperSerialization.serializeServiceVoltageChange(sensor, milliVolts));
    }

}
