package dk.lego.devicesdk.unity;

import android.graphics.Color;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.lego.devicesdk.LDSDKError;
import dk.lego.devicesdk.device.LegoDevice;
import dk.lego.devicesdk.input_output.InputFormat;
import dk.lego.devicesdk.logging.LDSDKLogger;
import dk.lego.devicesdk.services.CurrentSensor;
import dk.lego.devicesdk.services.LegoService;
import dk.lego.devicesdk.services.MotionSensor;
import dk.lego.devicesdk.services.Motor;
import dk.lego.devicesdk.services.PiezoTonePlayer;
import dk.lego.devicesdk.services.RGBLight;
import dk.lego.devicesdk.services.TiltSensor;
import dk.lego.devicesdk.services.VoltageSensor;

public class WrapperSerialization {

    // Device

    public Map<String, Object> serializeDevice(LegoDevice device) {
        Map<String, Object> deviceInfo = new HashMap<>();
        deviceInfo.put("FirmwareRevision", WrapperUtils.objectToString(device.getDeviceInfo().getFirmwareRevision()));
        deviceInfo.put("HardwareRevision", WrapperUtils.objectToString(device.getDeviceInfo().getHardwareRevision()));
        deviceInfo.put("SoftwareRevision", WrapperUtils.objectToString(device.getDeviceInfo().getSoftwareRevision()));
        deviceInfo.put("ManufacturerName", WrapperUtils.objectToString(device.getDeviceInfo().getManufacturerName()));

        Map<String, Object> json = new HashMap<>();
        json.put("DeviceName", WrapperUtils.objectToString(device.getName()));
        json.put("DeviceID", device.getDeviceId());
        json.put("DeviceInfo", deviceInfo);
        json.put("ButtonPressed", device.isButtonPressed());
        json.put("BatteryLevel", device.getBatteryLevel() != null ? device.getBatteryLevel() : 0);
        json.put("ConnectedState", device.getConnectState().ordinal());

        return json;
    }

    public List<Map<String, Object>> serializeDevices(List<? extends LegoDevice> devices) {
        List<Map<String, Object>> serializedDevices = new ArrayList<>();
        for (LegoDevice device : devices) {
            serializedDevices.add(serializeDevice(device));
        }

        return serializedDevices;
    }

    // Device events

    public Map<String, Object> serializeDeviceNameChange(LegoDevice device, String oldName, String newName) {
        Map<String, Object> data = new HashMap<>();
        data.put("DeviceID", device.getDeviceId());
        data.put("OldDeviceName", WrapperUtils.objectToString(oldName));
        data.put("DeviceName", WrapperUtils.objectToString(newName));
        return data;
    }

    public Map<String, Object> serializeDeviceButtonStateChange(LegoDevice device, boolean pressed) {
        Map<String, Object> data = new HashMap<>();
        data.put("DeviceID", device.getDeviceId());
        data.put("ButtonState", pressed);
        return data;
    }

    public Map<String, Object> serializeDeviceBatteryLevel(LegoDevice device, int level) {
        Map<String, Object> data = new HashMap<>();
        data.put("DeviceID", device.getDeviceId());
        data.put("BatteryLevel", level);
        return data;
    }

    public Map<String, Object> serializeDeviceLowVoltage(LegoDevice device, boolean lowVoltage) {
        Map<String, Object> data = new HashMap<>();
        data.put("DeviceID", device.getDeviceId());
        data.put("LowVoltageState", lowVoltage);
        return data;
    }

    public Map<String, Object> serializeDeviceError(LegoDevice device, LDSDKError error) {
        Map<String, Object> data = new HashMap<>();
        data.put("DeviceID", device.getDeviceId());
        data.put("Error", WrapperUtils.objectToString(error.getMessage()));
        return data;
    }

    // Input format

    public Map<String, Object> serializeInputFormat(InputFormat inputFormat) {
        Map<String, Object> json = new HashMap<>();
        if (inputFormat != null) {
            json.put("Revision", inputFormat.getRevision());
            json.put("ConnectID", inputFormat.getConnectId());
            json.put("TypeID", inputFormat.getType().getValue());
            json.put("Mode", inputFormat.getMode());
            json.put("DeltaInterval", inputFormat.getDeltaInterval());
            json.put("Unit", inputFormat.getUnit().getValue());
            json.put("NotificationsEnabled", inputFormat.isNotificationsEnabled());
        }
        return json;
    }

    // Service

    public Map<String, Object> serializeService(LegoService service) {
        return serializeService(service, false);
    }

    public Map<String, Object> serializeService(LegoService service, boolean onlyBasicInfo) {
        Map<String, Object> json = new HashMap<>();

        if (onlyBasicInfo) {
            json.put("DeviceID", service.getDevice().getDeviceId());
            json.put("ConnectID", service.getConnectInfo().getConnectId());
        } else {
            Map<String, Object> connectInfo = new HashMap<>();
            connectInfo.put("ConnectID", service.getConnectInfo().getConnectId());
            connectInfo.put("HubIndex", service.getConnectInfo().getHubIndex());
            connectInfo.put("HardwareRevision", WrapperUtils.objectToString(service.getConnectInfo().getHardwareVersion()));
            connectInfo.put("SoftwareRevision", WrapperUtils.objectToString(service.getConnectInfo().getFirmwareVersion()));
            connectInfo.put("Type", service.getConnectInfo().getType().getValue());

            json.put("ServiceName", WrapperUtils.objectToString(service.getServiceName()));
            json.put("DefaultInputFormat", serializeInputFormat(service.getDefaultInputFormat()));
            json.put("InputFormat", serializeInputFormat(service.getInputFormat()));
            json.put("IsInternalService", service.isInternalService());
            json.put("ConnectInfo", connectInfo);
            json.put("DeviceID", service.getDevice().getDeviceId());
        }

        return json;
    }

    public Map<String, Object> serializeServiceWithData(LegoService service) {
        Map<String, Object> serializedService = serializeService(service);
        Map<String, Object> serviceData = serializeServiceData(service);
        if (serviceData != null) {
            serializedService.put("ServiceData", serviceData);
        }
        return serializedService;
    }

    // Specific services

    public Map<String, Object> serializeServiceData(LegoService service) {
        if (service instanceof Motor) {
            return serializeServiceData((Motor)service);
        } else if (service instanceof RGBLight) {
            return serializeServiceData((RGBLight)service);
        } else if (service instanceof MotionSensor) {
            return serializeServiceData((MotionSensor)service);
        } else if (service instanceof TiltSensor) {
            return serializeServiceData((TiltSensor)service);
        } else if (service instanceof VoltageSensor) {
            return serializeServiceData((VoltageSensor)service);
        } else if (service instanceof CurrentSensor) {
            return serializeServiceData((CurrentSensor)service);
        } else if (service instanceof PiezoTonePlayer) {
            return serializeServiceData((PiezoTonePlayer)service);
        } else {
            // return no service data for unknown services
            return null;
        }
    }

    public Map<String, Object> serializeServiceData(Motor motor) {
        Map<String, Object> data = new HashMap<>();
        data.put("Power", motor.getPower());
        data.put("MotorDirection", motor.getDirection().getValue());
        return data;
    }

    public Map<String, Object> serializeServiceData(RGBLight light) {
        Map<String, Object> data = new HashMap<>();
        data.put("Color", serializeColor(light.getColor()));
        data.put("DefaultColor", serializeColor(light.getDefaultColor()));
        data.put("DefaultColorIndex", serializeColor(light.getDefaultColorIndex()));
        return data;
    }

    public Map<String, Object> serializeServiceData(MotionSensor motionSensor) {
        Map<String, Object> data = new HashMap<>();
        data.put("Count", motionSensor.getCount());
        data.put("Distance", motionSensor.getDistance());
        data.put("MotionSensorMode", motionSensor.getMotionSensorMode().getValue());
        return data;
    }

    public Map<String, Object> serializeServiceData(TiltSensor tiltSensor) {
        Map<String, Object> angle = new HashMap<>();
        angle.put("X", tiltSensor.getAngle().x);
        angle.put("Y", tiltSensor.getAngle().y);

        Map<String, Object> crash = new HashMap<>();
        crash.put("X", tiltSensor.getCrash().x);
        crash.put("Y", tiltSensor.getCrash().y);
        crash.put("Z", tiltSensor.getCrash().z);

        Map<String, Object> data = new HashMap<>();
        data.put("TiltSensorDirection", tiltSensor.getDirection().getValue());
        data.put("TiltSensorMode", tiltSensor.getTiltSensorMode().getValue());
        data.put("Angle", angle);
        data.put("Crash", crash);

        return data;
    }

    public Map<String, Object> serializeServiceData(VoltageSensor voltageSensor) {
        Map<String, Object> data = new HashMap<>();
        data.put("MilliVolts", voltageSensor.getValueAsMilliVolts());
        return data;
    }

    public Map<String, Object> serializeServiceData(CurrentSensor currentSensor) {
        Map<String, Object> data = new HashMap<>();
        data.put("MilliAmp", currentSensor.getValueAsMilliAmps());
        return data;
    }

    public Map<String, Object> serializeServiceData(PiezoTonePlayer piezoTonePlayer) {
        // intentionally return no data for the piezo player
        return null;
    }

    // Service events

    // Note: currently this method broadcasts a value data change event even if there are no valid data formats defined.
    // As a result the data dictionary does not contain any values. Instead there are usually specific events defined
    // for each sensor type, e.g. a color change for the RGB light service.
    public Map<String, Object> serializeServiceValueDataChange(LegoService service, byte[] oldValue, byte[] newValue) {
        // TODO: Take size of data set into account

        Map<String, Object> data = serializeService(service, true);

        InputFormat.InputFormatUnit unit = service.getInputFormat().getUnit();
        if (unit == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_PERCENTAGE ||
            unit == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_RAW) {
            if (oldValue != null) {
                Integer oldNumber = service.getIntegerFromData(oldValue);
                if (oldNumber != null) {
                    data.put("OldValue", oldNumber);
                }
            }
            if (newValue != null) {
                Integer newNumber = service.getIntegerFromData(newValue);
                if (newNumber != null) {
                    data.put("NewValue", newNumber);
                }
            }
        } else if (unit == InputFormat.InputFormatUnit.INPUT_FORMAT_UNIT_SI) {
            if (oldValue != null) {
                Float oldNumber = service.getFloatFromData(oldValue);
                if (oldNumber != null) {
                    data.put("OldValue", oldNumber);
                }
            }
            if (newValue != null) {
                Float newNumber = service.getFloatFromData(newValue);
                if (newNumber != null) {
                    data.put("NewValue", newNumber);
                }
            }
        } else {
            LDSDKLogger.e("Cannot serialize value with unknown input format unit");
        }

        return data;
    }

    public Map<String, Object> serializeServiceInputFormatChange(LegoService service, InputFormat oldFormat, InputFormat newFormat) {
        Map<String, Object> data = serializeService(service, true);
        if (oldFormat != null) {
            data.put("OldInputFormat", serializeInputFormat(oldFormat));
        }
        data.put("NewInputFormat", serializeInputFormat(newFormat));
        return data;
    }

    public Map<String, Object> serializeServiceCurrentChange(CurrentSensor sensor, float milliAmps) {
        Map<String, Object> data = serializeService(sensor, true);
        data.put("MilliAmp", milliAmps);
        return data;
    }

    public Map<String, Object> serializeServiceDistanceChange(MotionSensor sensor, float oldDistance, float newDistance) {
        Map<String, Object> data = serializeService(sensor, true);
        data.put("OldDistance", oldDistance);
        data.put("NewDistance", newDistance);
        return data;
    }

    public Map<String, Object> serializeServiceCountChange(MotionSensor sensor, int count) {
        Map<String, Object> data = serializeService(sensor, true);
        data.put("Count", count);
        return data;
    }

    public Map<String, Object> serializeServiceColorChange(RGBLight rgbLight, Integer oldRGBColor, Integer newRGBColor) {
        Map<String, Object> data = serializeService(rgbLight, true);
        if (oldRGBColor != null) {
            data.put("OldColor", serializeColor(oldRGBColor));
        }
        data.put("NewColor", serializeColor(newRGBColor));
        return data;
    }

    public Map<String, Object> serializeServiceColorIndexChange(RGBLight rgbLight, int oldColorIndex, int newColorIndex) {
        Map<String, Object> data = serializeService(rgbLight, true);
        data.put("OldColorIndex", oldColorIndex);
        data.put("NewColorIndex", newColorIndex);
        return data;
    }

    public Map<String, Object> serializeServiceDirectionChange(TiltSensor sensor, TiltSensor.TiltSensorDirection oldDirection, TiltSensor.TiltSensorDirection newDirection) {
        Map<String, Object> data = serializeService(sensor, true);
        if (oldDirection != null) {
            data.put("OldDirection", oldDirection.getValue());
        }
        data.put("NewDirection", newDirection.getValue());
        return data;
    }

    public Map<String, Object> serializeServiceAngleChange(TiltSensor sensor, TiltSensor.TiltSensorAngle oldAngle, TiltSensor.TiltSensorAngle newAngle) {
        Map<String, Object> data = serializeService(sensor, true);

        if (oldAngle != null) {
            Map<String, Object> oldAngleDict = new HashMap<>();
            oldAngleDict.put("X", oldAngle.x);
            oldAngleDict.put("Y", oldAngle.y);
            data.put("OldAngle", oldAngleDict);
        }

        Map<String, Object> newAngleDict = new HashMap<>();
        newAngleDict.put("X", newAngle.x);
        newAngleDict.put("Y", newAngle.y);
        data.put("NewAngle", newAngleDict);

        return data;
    }

    public Map<String, Object> serializeServiceCrashChange(TiltSensor sensor, TiltSensor.TiltSensorCrash oldCrashValue, TiltSensor.TiltSensorCrash newCrashValue) {
        Map<String, Object> data = serializeService(sensor, true);

        if (oldCrashValue != null) {
            Map<String, Object> oldCrashDict = new HashMap<>();
            oldCrashDict.put("X", oldCrashValue.x);
            oldCrashDict.put("Y", oldCrashValue.y);
            oldCrashDict.put("Z", oldCrashValue.z);
            data.put("OldCrash", oldCrashDict);
        }

        Map<String, Object> newCrashDict = new HashMap<>();
        newCrashDict.put("X", newCrashValue.x);
        newCrashDict.put("Y", newCrashValue.y);
        newCrashDict.put("Z", newCrashValue.z);
        data.put("NewCrash", newCrashDict);

        return data;
    }

    public Map<String, Object> serializeServiceVoltageChange(VoltageSensor sensor, float milliVolts) {
        Map<String, Object> data = serializeService(sensor, true);
        data.put("MilliVolts", milliVolts);
        return data;
    }

    // Helpers

    public Map<String, Object> serializeColor(int color) {
        Map<String, Object> data = new HashMap<>();
        data.put("R", Color.red(color) / 255.0);
        data.put("G", Color.green(color) / 255.0);
        data.put("B", Color.blue(color) / 255.0);
        data.put("A", Color.alpha(color) / 255.0);
        return data;
    }

    public String mapToJSONString(Map<String, Object> map) {
        return JSONObject.toJSONString(map);
    }

    public String mapToJSONString(List<Map<String, Object>> list) {
        return JSONArray.toJSONString(list);
    }

    @SuppressWarnings("unchecked")
    public Map<String, Object> parseJSONObject(String jsonString) throws org.json.simple.parser.ParseException {
        Object jsonObject = JSONValue.parseWithException(jsonString);
        if (!(jsonObject instanceof JSONObject)) {
            throw new IllegalArgumentException("Passed string is not a single JSON object");
        }
        return  (Map<String, Object>)jsonObject;
    }

    public int deserializeColor(Map<String, Object> color) throws IllegalArgumentException {
        if (!(color.get("A") instanceof Number) ||
            !(color.get("R") instanceof Number) ||
            !(color.get("G") instanceof Number) ||
            !(color.get("B") instanceof Number)) {
            throw new IllegalArgumentException("Color object does not contain ARGB number components");
        }

        return Color.argb((int)((double)color.get("A") * 255),
                          (int)((double)color.get("R") * 255),
                          (int)((double)color.get("G") * 255),
                          (int)((double)color.get("B") * 255));
    }

}
